// Chunk array management functions
// Copyright (C) 2000 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

// NOTE: To manage the array data efficiently, and to keep memory
// from fragmenting, variable arrays of structures are kept in "Chunks"
// Each chunk allocates space from the free memory pool
// and the elements of the array are split among them. The chunks are linked
// to each other in order. When an element is inserted, additional chunks are allocated
// as necessary, and linked into the list at the appropriate location.
// Chunks keep track of how many elements are currently used within them.
// Chunks are not allowed to contain 0 elements

static void DisposeArrayChunk(ARRAY_CHUNK_HEADER *chunk)
// deallocate the memory used by chunk
{
	MDisposePtr(chunk->data);
	MDisposePtr(chunk);
}

static ARRAY_CHUNK_HEADER *CreateArrayChunk(UINT32 maxElements,UINT32 elementSize)
// create a chunk header, and its corresponding data
// if there is a problem, set the error, and return NULL
{
	ARRAY_CHUNK_HEADER
		*chunk;

	if((chunk=(ARRAY_CHUNK_HEADER *)MNewPtrClr(sizeof(ARRAY_CHUNK_HEADER))))
	{
		if((chunk->data=(UINT8 *)MNewPtr(maxElements*elementSize)))
		{
			return(chunk);
		}
		MDisposePtr(chunk);
	}
	return((ARRAY_CHUNK_HEADER *)NULL);
}

static void DisposeArrayChunkList(ARRAY_CHUNK_HEADER *startChunk)
// dispose of a linked list of chunks
{
	ARRAY_CHUNK_HEADER
		*nextChunk;

	while(startChunk)
	{
		nextChunk=startChunk->nextHeader;
		DisposeArrayChunk(startChunk);
		startChunk=nextChunk;
	}
}

static bool CreateArrayChunkList(ARRAY_CHUNK_HEADER **startChunk,ARRAY_CHUNK_HEADER **endChunk,UINT32 numChunks,UINT32 maxElements,UINT32 elementSize)
// create a linked list of empty chunks
// if there is a problem, set the error, free anything allocated, and return false
{
	bool
		fail;
	ARRAY_CHUNK_HEADER
		*chunk;

	(*startChunk)=(*endChunk)=NULL;
	fail=false;
	while(numChunks&&!fail)
	{
		if((chunk=CreateArrayChunk(maxElements,elementSize)))
		{
			if(*endChunk)
			{
				(*endChunk)->nextHeader=chunk;
				chunk->previousHeader=(*endChunk);
				(*endChunk)=chunk;
			}
			else
			{
				(*startChunk)=(*endChunk)=chunk;
			}
			numChunks--;
		}
		else
		{
			DisposeArrayChunkList(*startChunk);					// remove the partially created chunk list
			fail=true;
		}
	}
	return(!fail);
}

static void UnlinkArrayChunkList(ARRAY_UNIVERSE *universe,ARRAY_CHUNK_HEADER *startChunk,ARRAY_CHUNK_HEADER *endChunk)
// unlink the chunks from startChunk to endChunk from universe
{
	if(startChunk->previousHeader)
	{
		startChunk->previousHeader->nextHeader=endChunk->nextHeader;
	}
	else
	{
		universe->firstChunkHeader=endChunk->nextHeader;
	}

	if(endChunk->nextHeader)
	{
		endChunk->nextHeader->previousHeader=startChunk->previousHeader;
	}
	else
	{
		universe->lastChunkHeader=startChunk->previousHeader;
	}
	startChunk->previousHeader=endChunk->nextHeader=NULL;			// terminate the list
}

static void DeleteWithinArrayChunk(ARRAY_UNIVERSE *universe,ARRAY_CHUNK_HEADER *chunk,UINT32 offset,UINT32 numElements,ARRAY_CHUNK_HEADER **nextChunk,UINT32 *nextOffset)
// data within one chunk is being deleted
{
	if(numElements<chunk->totalElements)							// see if something will remain after the deletion
	{
		MMoveMem(&(chunk->data[(offset+numElements)*universe->elementSize]),&(chunk->data[offset*universe->elementSize]),(chunk->totalElements-(offset+numElements))*universe->elementSize);	// move back any bytes needed
		chunk->totalElements-=numElements;
		universe->totalElements-=numElements;
		if(offset>=chunk->totalElements)
		{
			(*nextChunk)=chunk->nextHeader;
			(*nextOffset)=0;
		}
		else
		{
			(*nextChunk)=chunk;
			(*nextOffset)=offset;
		}
	}
	else
	{
		(*nextChunk)=chunk->nextHeader;
		(*nextOffset)=0;
		UnlinkArrayChunkList(universe,chunk,chunk);
		universe->totalElements-=chunk->totalElements;
		DisposeArrayChunkList(chunk);
	}
}

void DeleteUniverseArray(ARRAY_UNIVERSE *universe,ARRAY_CHUNK_HEADER *chunk,UINT32 offset,UINT32 numElements,ARRAY_CHUNK_HEADER **nextChunk,UINT32 *nextOffset)
// delete numElements from universe starting at chunk, offset
// return nextChunk/nextOffset as the chunk and offset just past the end of the
// delete, or NULL if none exists past the end
{
	UINT32
		numToDelete;

	while(numElements&&chunk)
	{
		if((chunk->totalElements-offset)<numElements)
		{
			numToDelete=chunk->totalElements-offset;
		}
		else
		{
			numToDelete=numElements;
		}
		DeleteWithinArrayChunk(universe,chunk,offset,numToDelete,&chunk,&offset);
		numElements-=numToDelete;
	}
	(*nextChunk)=chunk;
	(*nextOffset)=offset;
}

static bool InsertWithRoomInChunk(ARRAY_UNIVERSE *universe,ARRAY_CHUNK_HEADER *chunk,UINT32 offset,UINT32 numElements,ARRAY_CHUNK_HEADER **nextChunk,UINT32 *nextOffset)
// the chunk that the elements are going to be inserted into
// contains enough free space to accept the entire inserted list
{
	MMoveMem(&(chunk->data[offset*universe->elementSize]),&(chunk->data[(offset+numElements)*universe->elementSize]),(chunk->totalElements-offset)*universe->elementSize);	// move data within the chunk to make room
	chunk->totalElements+=numElements;					// bump number of elements in this chunk
	universe->totalElements+=numElements;				// adjust universal understanding
	(*nextChunk)=chunk;
	(*nextOffset)=offset;
	return(true);										// this cannot fail
}

static bool InsertWithRoomInTwoChunks(ARRAY_UNIVERSE *universe,ARRAY_CHUNK_HEADER *chunk,UINT32 offset,UINT32 numElements,ARRAY_CHUNK_HEADER **nextChunk,UINT32 *nextOffset)
// the chunk where the element insertion starts, and the next chunk contain enough free space
// to allow the elements to be inserted
{
	ARRAY_CHUNK_HEADER
		*secondChunk;
	UINT32
		freeInSecond,
		amountToMove,
		distanceToMove;

	secondChunk=chunk->nextHeader;
	if(numElements+offset<=universe->maxElements)			// see if the added elements can be contained completely within the first chunk
	{
		freeInSecond=universe->maxElements-secondChunk->totalElements;		// get free space in second chunk
		distanceToMove=chunk->totalElements-offset;			// would like to move as much as possible to second chunk
		if(distanceToMove>freeInSecond)
		{
			distanceToMove=freeInSecond;
		}

		MMoveMem(secondChunk->data,&(secondChunk->data[(distanceToMove)*universe->elementSize]),secondChunk->totalElements*universe->elementSize);	// move second chunk's data out of the way for the add
		MMoveMem(&(chunk->data[(chunk->totalElements-distanceToMove)*universe->elementSize]),secondChunk->data,distanceToMove*universe->elementSize);	// move data from first chunk to second chunk

		chunk->totalElements-=distanceToMove;
		secondChunk->totalElements+=distanceToMove;			// this many more bytes now in use in second chunk

		MMoveMem(&(chunk->data[offset*universe->elementSize]),&(chunk->data[(offset+numElements)*universe->elementSize]),(chunk->totalElements-offset)*universe->elementSize);	// move data within the chunk to make room

		chunk->totalElements+=numElements;					// bump number of bytes in this chunk
		universe->totalElements+=numElements;				// adjust universal understanding
	}
	else
	{
		distanceToMove=chunk->totalElements+numElements-universe->maxElements;		// need to push second chunk data forward this distance
		MMoveMem(secondChunk->data,&(secondChunk->data[distanceToMove*universe->elementSize]),secondChunk->totalElements*universe->elementSize);	// move second chunk's data out of the way for the add
		secondChunk->totalElements+=distanceToMove;			// this many more elements now in use in second chunk
		amountToMove=chunk->totalElements-offset;			// number of elements to move from first to second

		MMoveMem(&(chunk->data[offset*universe->elementSize]),&(secondChunk->data[(offset+numElements-universe->maxElements)*universe->elementSize]),amountToMove*universe->elementSize);	// move data from first chunk to second chunk

		chunk->totalElements=universe->maxElements;
		universe->totalElements+=numElements;				// adjust universal understanding
	}
	(*nextChunk)=chunk;
	(*nextOffset)=offset;
	return(true);											// this cannot fail
}

static bool InsertWithNoRoom(ARRAY_UNIVERSE *universe,ARRAY_CHUNK_HEADER *chunk,UINT32 offset,UINT32 numElements,ARRAY_CHUNK_HEADER **nextChunk,UINT32 *nextOffset)
// the elements to be inserted cannot fit within the chunk passed
// so, chunks will have to be created
// NOTE: this must handle the case where offset is universe->maxElements
{
	UINT32
		numChunks,
		maxToFill,
		numToFill;
	ARRAY_CHUNK_HEADER
		*startChunk,
		*endChunk;

	(*nextChunk)=chunk;
	(*nextOffset)=offset;
	numChunks=(numElements-(universe->maxElements-(chunk->totalElements))+universe->maxElements-1)/universe->maxElements;
	if(CreateArrayChunkList(&startChunk,&endChunk,numChunks,universe->maxElements,universe->elementSize))	// create just enough chunks to hold the elements
	{
		if((endChunk->nextHeader=chunk->nextHeader))		// link new chunks onto the list
		{
			chunk->nextHeader->previousHeader=endChunk;
		}
		else
		{
			universe->lastChunkHeader=endChunk;				// update the universe
		}
		chunk->nextHeader=startChunk;
		startChunk->previousHeader=chunk;
		universe->totalElements+=numElements;				// more elements in the universe now

		if((numElements+offset)/universe->maxElements<numChunks)
		{
			MMoveMem(&(chunk->data[offset*universe->elementSize]),&(endChunk->data[0]),(chunk->totalElements-offset)*universe->elementSize);	// move data from old last chunk to new last chunk
		}
		else
		{
			MMoveMem(&(chunk->data[offset*universe->elementSize]),&(endChunk->data[((numElements+offset)%universe->maxElements)*universe->elementSize]),(chunk->totalElements-offset)*universe->elementSize);	// move data from old last chunk to new last chunk
		}
		endChunk->totalElements=chunk->totalElements-offset;	// this many elements in end chunk at the moment
		chunk->totalElements=offset;							// this many elements in here
		while(numElements&&chunk)
		{
			maxToFill=universe->maxElements-offset;
			if(numElements>maxToFill)
			{
				numToFill=maxToFill;
			}
			else
			{
				numToFill=numElements;
			}
			chunk->totalElements+=numToFill;					// bump number of elements in this chunk
			numElements-=numToFill;								// this many less elements now
			chunk=chunk->nextHeader;							// point to next chunk to fill
			offset=0;
		}
		if((*nextOffset)>=(*nextChunk)->totalElements)			// see if pointing past the end, if so, move to the next
		{
			(*nextChunk)=(*nextChunk)->nextHeader;
			(*nextOffset)=0;
		}
		return(true);											// done
	}
	return(false);
}

static bool InsertWithNoChunks(ARRAY_UNIVERSE *universe,UINT32 numElements,ARRAY_CHUNK_HEADER **nextChunk,UINT32 *nextOffset)
// insertion is happening in an empty universe
{
	UINT32
		numToFill;
	ARRAY_CHUNK_HEADER
		*chunk,
		*startChunk,
		*endChunk;

	if(CreateArrayChunkList(&startChunk,&endChunk,(numElements+universe->maxElements-1)/universe->maxElements,universe->maxElements,universe->elementSize))	// create just enough chunks to hold the data
	{
		(*nextChunk)=startChunk;
		(*nextOffset)=0;
		universe->firstChunkHeader=startChunk;
		universe->lastChunkHeader=endChunk;
		universe->totalElements=numElements;
		chunk=startChunk;
		while(chunk&&numElements)
		{
			if(numElements>universe->maxElements)
			{
				numToFill=universe->maxElements;
			}
			else
			{
				numToFill=numElements;
			}
			chunk->totalElements=numToFill;					// set number of elements in this chunk
			numElements-=numToFill;							// this many less elements now
			chunk=chunk->nextHeader;						// point to next chunk to fill
		}
		return(true);										// done
	}
	return(false);
}

bool InsertUniverseArray(ARRAY_UNIVERSE *universe,ARRAY_CHUNK_HEADER *chunk,UINT32 offset,UINT32 numElements,ARRAY_CHUNK_HEADER **nextChunk,UINT32 *nextOffset)
// make room in the element list for numElements at/before chunk/offset
// return nextChunk and nextOffset as the start of the place where the insertions were made
{
	bool
		fail;

	fail=false;
	(*nextChunk)=NULL;
	(*nextOffset)=0;
	if(numElements)
	{
		if(chunk)
		{
			if((offset==0)&&(chunk->previousHeader)&&(chunk->previousHeader->totalElements<universe->maxElements))	// if inserting at the very start of a chunk, it is more efficient to look one back (if possible)
			{
				chunk=chunk->previousHeader;
				offset=chunk->totalElements;
			}
			if(numElements<=universe->maxElements-(chunk->totalElements))	// see if the elements can be placed in empty space of this chunk
			{
				fail=!InsertWithRoomInChunk(universe,chunk,offset,numElements,nextChunk,nextOffset);
			}
			else
			{
				if(chunk->nextHeader)										// is there another chunk after this one?
				{
					if(numElements<=(2*universe->maxElements)-(chunk->totalElements+chunk->nextHeader->totalElements))	// is there room in the combined free space of this chunk, and the next one?
					{
						fail=!InsertWithRoomInTwoChunks(universe,chunk,offset,numElements,nextChunk,nextOffset);
					}
					else
					{
						fail=!InsertWithNoRoom(universe,chunk,offset,numElements,nextChunk,nextOffset);
					}
				}
				else
				{
					fail=!InsertWithNoRoom(universe,chunk,offset,numElements,nextChunk,nextOffset);
				}
			}
		}
		else
		{
			if((chunk=universe->lastChunkHeader))
			{
				if(numElements<=universe->maxElements-(chunk->totalElements))	// see if the bytes can be placed in empty space of this chunk
				{
					fail=!InsertWithRoomInChunk(universe,chunk,chunk->totalElements,numElements,nextChunk,nextOffset);
				}
				else
				{
					fail=!InsertWithNoRoom(universe,chunk,chunk->totalElements,numElements,nextChunk,nextOffset);
				}
			}
			else
			{
				fail=!InsertWithNoChunks(universe,numElements,nextChunk,nextOffset);
			}
		}
	}
	return(!fail);
}

void UnInitArrayUniverse(ARRAY_UNIVERSE *universe)
// dispose of an array universe
{
	DisposeArrayChunkList(universe->firstChunkHeader);
}

bool InitArrayUniverse(ARRAY_UNIVERSE *universe,UINT32 maxElements,UINT32 elementSize)
// initialize an array universe
// if there is a problem, set the error, and return false
{
	universe->firstChunkHeader=universe->lastChunkHeader=(ARRAY_CHUNK_HEADER *)NULL;
	universe->totalElements=0;
	universe->maxElements=maxElements;
	universe->elementSize=elementSize;
	return(true);
}
