// Cut, copy, and paste functions
// Copyright (C) 2000 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

static EDITOR_BUFFER_HANDLE
	currentClipboardHandle;							// handle to the current clipboard

EDITOR_BUFFER *EditorGetCurrentClipboard()
// return the current clipboard, or NULL if there is none
{
	return(EditorGetBufferFromHandle(&currentClipboardHandle));
}

void EditorSetCurrentClipboard(EDITOR_BUFFER *clipboard)
// set the current clipboard
{
	EditorReleaseBufferHandle(&currentClipboardHandle);				// release our grip on any buffer we held
	if(clipboard)
	{
		EditorGrabBufferHandle(&currentClipboardHandle,clipboard);	// hang onto this one
	}
}

bool EditorCopy(EDITOR_BUFFER *buffer,EDITOR_BUFFER *clipboard)
// copy the current selection (if there is one) from buffer into clipboard
// a "selection" is made in the clipboard text for every selection that is
// copied from the source text. These selections are used when columnar pasting.
// NOTE: it is possible for clipboard to be passed as NULL if no clipboard
// has been selected. This is considered an error.
// Also, if buffer and clipboard are the same, it is not legal to perform this
// operation.
{
	CHUNK_HEADER
		*sourceChunk;
	UINT32
		sourceChunkOffset;
	EDITOR_BUFFER
		*sourceBuffer,
		*destBuffer;
	SELECTION_UNIVERSE
		*sourceSelectionUniverse,
		*destSelectionUniverse;
	UINT32
		sourceStartPosition,
		sourceLength,
		destStartPosition,
		destLength;
	bool
		sawInitialSelection,
		fail;

	fail=false;
	if(clipboard)
	{
		if(clipboard!=buffer)
		{
			destBuffer=clipboard;
			destSelectionUniverse=destBuffer->selectionUniverse;
			sourceBuffer=buffer;
			sourceSelectionUniverse=sourceBuffer->selectionUniverse;

			EditorClearUndo(destBuffer);	// kill any undo information that happened to be lying about in the clipboard (if user opened window to clipboard and started typing)

			EditorStartReplace(destBuffer);

			// kill off what was in the clipboard
			fail=!ReplaceEditorChunks(destBuffer,0,destBuffer->textUniverse->totalBytes,NULL,0,0,false);

			sawInitialSelection=false;
			sourceStartPosition=0;
			while(!fail&&GetSelectionAtOrAfterPosition(sourceSelectionUniverse,sourceStartPosition,&sourceStartPosition,&sourceLength))
			{
				// after every line but the last, place in a newline so that exported columnar selections work as expected
				if(sawInitialSelection)
				{
					fail=!ReplaceEditorText(destBuffer,destBuffer->textUniverse->totalBytes,destBuffer->textUniverse->totalBytes,(UINT8 *)"\n",1,false);
				}
				if(!fail)
				{
					PositionToChunkPosition(sourceBuffer->textUniverse,sourceStartPosition,&sourceChunk,&sourceChunkOffset);	// point to the start of the selection in the source data
					destStartPosition=destBuffer->textUniverse->totalBytes;
					if(ReplaceEditorChunks(destBuffer,destStartPosition,destStartPosition,sourceChunk,sourceChunkOffset,sourceLength,false))	// insert the text at the end of the destination
					{
						if(CopyStyle(sourceBuffer->styleUniverse,sourceStartPosition,destBuffer->styleUniverse,destStartPosition,sourceLength))
						{
							destLength=destBuffer->textUniverse->totalBytes-destStartPosition;
							if(destLength>0)
							{
								fail=!SetSelectionRange(destSelectionUniverse,destStartPosition,destLength);	// create new selection
							}
						}
						else
						{
							fail=true;
						}
					}
					else
					{
						fail=true;
					}
				}
				sawInitialSelection=true;
				sourceStartPosition+=sourceLength;
			}
			EditorEndReplace(destBuffer);
		}
		else
		{
			SetError("Clipboard must not be both source and destination");
			fail=true;
		}
	}
	else
	{
		SetError("No current clipboard");
		fail=true;
	}
	if(!fail)
	{
		if(clipboard==EditorGetCurrentClipboard())		// only export the clipboard if it is the current one, and we did not fail
		{
			fail=!ExportClipboard();
		}
	}
	return(!fail);
}


bool EditorCut(EDITOR_BUFFER *buffer,EDITOR_BUFFER *clipboard)
// Cut the current selection if there is one from buffer into clipboard
// NOTE: it is possible for clipboard to be passed as NULL if no clipboard
// has been selected. This is considered an error.
// Also, if buffer and clipboard are the same, it is not legal to perform this
// operation.
{
	bool
		result;

	if((result=EditorCopy(buffer,clipboard)))
	{
		EditorStartReplace(buffer);
		BeginUndoGroup(buffer);
		result=DeleteAllSelectedText(buffer,buffer->selectionUniverse);
		StrictEndUndoGroup(buffer);
		EditorEndReplace(buffer);
	}
	return(result);
}

bool EditorColumnarPaste(EDITOR_BUFFER *buffer,EDITOR_VIEW *view,EDITOR_BUFFER *clipboard)
// Paste the contents of clipboard into buffer in columnar format
// This means that each "selection" of the clipboard will be placed into buffer at
// the current cursor position, then the position is moved to the next line, and
// the process is repeated until all of the selections have been placed
// NOTE: if there are selections in buffer, they are replaced one at a time
// Any selections remaining after the paste is complete are deleted.
// NOTE: if there are no selections in the clipboard, then it is just pasted as one
// large blob
// Finally, if view is passed as NULL, and the paste requires moving down a line
// at a time, this will just paste the remaining selections at the starts of lines.
// NOTE: it is possible for clipboard to be passed as NULL if no clipboard
// has been selected. This is considered an error.
// Also, if buffer and clipboard are the same, it is not legal to perform this
// operation.
{
	CHUNK_HEADER
		*sourceChunk,
		*destChunk;
	UINT32
		sourceChunkOffset,
		destChunkOffset;
	EDITOR_BUFFER
		*sourceBuffer;
	SELECTION_UNIVERSE
		*sourceSelectionUniverse,
		*destSelectionUniverse;
	UINT32
		sourceStartPosition,
		sourceLength,
		destStartPosition,
		destLength,
		lastReplacePosition,
		finalCursorPosition;
	UINT32
		temp,
		lastNumLines,
		numReplaced,
		newPosition,
		line,
		lineOffset;
	INT32
		desiredX;
	bool
		replacedOne,
		fail;

	fail=false;
	if(clipboard)
	{
		if(clipboard!=buffer)
		{
			sourceBuffer=clipboard;
			if(ImportClipboard())									// give external world a chance to override clipboard
			{
				destSelectionUniverse=buffer->selectionUniverse;
				sourceSelectionUniverse=sourceBuffer->selectionUniverse;
				EditorStartReplace(buffer);
				BeginUndoGroup(buffer);

				finalCursorPosition=lastReplacePosition=GetSelectionCursorPosition(destSelectionUniverse);	// initialize these
				if(!IsSelectionEmpty(sourceSelectionUniverse))		// see if selections in the clipboard
				{
					sourceStartPosition=0;							// get absolute position
					destStartPosition=0;
					numReplaced=0;
					replacedOne=false;								// tells if a replace has been done, so we know not to move the cursor position further

					// first, replace selections in the destination with selections from the source
					while(!fail&&GetSelectionAtOrAfterPosition(sourceSelectionUniverse,sourceStartPosition,&sourceStartPosition,&sourceLength)&&GetSelectionAtOrAfterPosition(destSelectionUniverse,0,&destStartPosition,&destLength))
					{
						PositionToChunkPosition(sourceBuffer->textUniverse,sourceStartPosition,&sourceChunk,&sourceChunkOffset);	// point to the start of the selection in the source data

						lastReplacePosition=destStartPosition;		// update this for later
						if(ReplaceEditorChunks(buffer,destStartPosition,destStartPosition+destLength,sourceChunk,sourceChunkOffset,sourceLength,false))
						{
							if(CopyStyle(sourceBuffer->styleUniverse,sourceStartPosition,buffer->styleUniverse,destStartPosition,sourceLength))
							{
								if(!replacedOne)
								{
									finalCursorPosition=destStartPosition+sourceLength;	// place cursor after start of first replacement when this procedure is finished
									replacedOne=true;
								}
								numReplaced=sourceLength;			// number of bytes that were added at lastReplacePosition
							}
							else
							{
								fail=true;
							}
						}
						else
						{
							fail=true;
						}
						sourceStartPosition+=sourceLength;
					}

					// now, replace by moving down a line at a time
					PositionToLinePosition(buffer->textUniverse,lastReplacePosition,&line,&lineOffset,&destChunk,&destChunkOffset);
					if(view)
					{
						GetEditorViewTextToGraphicPosition(view,lastReplacePosition,&desiredX,false,&temp,&temp);	// find the X offset we will attempt to insert at each time
					}

					if(replacedOne)		// if managed to replace, then move to the line BELOW the last replacement
					{
						PositionToLinePosition(buffer->textUniverse,lastReplacePosition+numReplaced,&line,&lineOffset,&destChunk,&destChunkOffset);
						line++;		// start replacing on this line
					}

					while(!fail&&GetSelectionAtOrAfterPosition(sourceSelectionUniverse,sourceStartPosition,&sourceStartPosition,&sourceLength))
					{
						PositionToChunkPosition(sourceBuffer->textUniverse,sourceStartPosition,&sourceChunk,&sourceChunkOffset);	// point to the start of the selection in the source data

						LineToChunkPosition(buffer->textUniverse,line,&destChunk,&destChunkOffset,&destStartPosition);
						if(view)
						{
							GetEditorViewGraphicToTextPosition(view,destStartPosition,desiredX,&newPosition,&temp);
						}
						else
						{
							newPosition=0;
						}

						destStartPosition+=newPosition;				// this is where the insertion should occur

						if(line>buffer->textUniverse->totalLines)	// attempting to place on non-existent line?
						{
							if(ReplaceEditorText(buffer,destStartPosition,destStartPosition,(UINT8 *)"\n",1,false))	// stick in a newline at the end
							{
								destStartPosition++;				// then move past it
							}
							else
							{
								fail=true;
							}
						}
						lastNumLines=buffer->textUniverse->totalLines;			// remember how many lines there were, so we can see how many were inserted
						if(!fail)
						{
							if(ReplaceEditorChunks(buffer,destStartPosition,destStartPosition,sourceChunk,sourceChunkOffset,sourceLength,false))
							{
								if(CopyStyle(sourceBuffer->styleUniverse,sourceStartPosition,buffer->styleUniverse,destStartPosition,sourceLength))
								{
									if(!replacedOne)
									{
										finalCursorPosition=destStartPosition+sourceLength;	// place cursor after start of first replacement
										replacedOne=true;
									}
									line+=buffer->textUniverse->totalLines-lastNumLines+1;	// skip to next line, plus any inserted
								}
								else
								{
									fail=true;
								}
							}
							else
							{
								fail=true;
							}
						}
						sourceStartPosition+=sourceLength;
					}
				}
				else
				{
					if(!GetSelectionAtOrAfterPosition(destSelectionUniverse,0,&destStartPosition,&destLength))
					{
						destStartPosition=finalCursorPosition;
						destLength=0;
					}
					if(ReplaceEditorChunks(buffer,destStartPosition,destStartPosition+destLength,sourceBuffer->textUniverse->firstChunkHeader,0,sourceBuffer->textUniverse->totalBytes,false))
					{
						fail=!CopyStyle(sourceBuffer->styleUniverse,0,buffer->styleUniverse,destStartPosition,sourceBuffer->textUniverse->totalBytes);
					}
					else
					{
						fail=true;
					}
					finalCursorPosition=destStartPosition+sourceBuffer->textUniverse->totalBytes;		// place cursor after start of replacement
				}

				if(!fail)
				{
					if(DeleteAllSelectedText(buffer,destSelectionUniverse))						// kill any remaining selections
					{
						SetSelectionCursorPosition(destSelectionUniverse,finalCursorPosition);	// drop cursor after start of first replacement
					}
					else
					{
						fail=true;
					}
				}

				StrictEndUndoGroup(buffer);
				EditorEndReplace(buffer);
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			SetError("Clipboard must not be both source and destination");
			fail=true;
		}
	}
	else
	{
		SetError("No current clipboard");
		fail=true;
	}
	return(!fail);
}

EDITOR_BUFFER *EditorStartImportClipboard()
// Call this from the gui to get the editor buffer that is
// the current clipboard, when something external to the editor has modified
// the system clipboard, and it is desired to get that data into the editor
// it is legal for this routine to return NULL
// in which case you should not attempt to place anything into the clipboard
// and you should not call EditorEndImportClipboard
{
	return(EditorGetCurrentClipboard());
}

void EditorEndImportClipboard()
// Call this when you are done modifying the clipboard that was returned
// from EditorStartImportClipboard
{
}

EDITOR_BUFFER *EditorStartExportClipboard()
// Call this from the gui to get the editor buffer that is
// the current clipboard, when you wish to export the current clipboard
// if this returns NULL, you should not call EditorEndExportClipboard
{
	return(EditorGetCurrentClipboard());
}

void EditorEndExportClipboard()
// Call this when you are done exporting the clipboard that was returned
// from EditorStartExportClipboard
{
}

void UnInitClipboard()
// undo whatever InitClipboard did
{
	EditorSetCurrentClipboard(NULL);			// clear the current clipboard, releasing any buffer it held
}

bool InitClipboard()
// initialize clipboard handling
// if there is a problem, SetError, and return false
{
	EditorInitBufferHandle(&currentClipboardHandle);
	return(true);
}
