// Regular expression handling
// Copyright (C) 2000 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

#define BIT_TABLE_LENGTH			(256/8)				// number of bytes in a bit table
#define MAX_STRING_PER_NODE			BIT_TABLE_LENGTH	// maximum number of characters in a string within a node (unioned with bit table)
#define	MAX_RECURSION_DEPTH			20000				// maximum amount of recursion we deem allowable

typedef enum
{
	// NOTE: the order of these IS important, there are tables below that match this order
	PT_STRING=0,				// matching constant string
	PT_LIST,					// matching one character to see if it is in a list
	PT_PREVIOUSMATCH,			// matching against data found in previous match
	PT_SUBEXPRESSIONSTART,		// matching against data found in a sub expression
	PT_SUBEXPRESSIONEND,		// end of sub expression match
	PT_LEFTEDGE,				// left edge must be one of the given characters (or edge of search)
	PT_RIGHTEDGE				// right edge must be one of the given characters (or edge of search)
} PATTERN_TYPE;

typedef struct
{
	CHUNK_HEADER
		*chunk;					// chunk and offset that start this register (if NULL, register not yet matched)
	UINT32
		offset;
	UINT32
		matchIndex;				// index into matched data where this register was matched
	UINT32
		numBytes;				// number of bytes matched by this register
	UINT32
		minimumMatchLength,		// used only during analysis, tells the minimum length of a pattern that could be matched by this register (0 means it can match nothing)
		maximumMatchLength;		// used only during analysis, tells the maximum length of a pattern that could be matched by this register
	bool
		infiniteMatchLength;	// used only during analysis, tells if the maximum length of a pattern could be infinite (in which case maximumMatchLength should be ignored)
} REGISTER;

typedef struct expressionNode
{
	PATTERN_TYPE
		type;					// type of pattern this element represents
	UINT8
		dataLength;				// length of string data if it exists
	union
	{
		UINT8
			data[MAX_STRING_PER_NODE];	// characters of the string to compare against (if string type)
		UINT8
			bits[BIT_TABLE_LENGTH];		// the bits that tell if a character is or is not to be accepted (in a group) (also used in edge searches)
		REGISTER
			*expRegister;		// points to the register which holds the start and end of the last match we are to compare against (for register type)
		struct
		{
			struct expressionListHeader
				*subExpressionList;	// pointer to expression list for this subexpression
			UINT32
				currentCount;	// current number of successful iterations of the subexpression
			REGISTER
				*expRegister;	// points to the register to remember match of subexpression in (NULL if none remembered)
			CHUNK_HEADER
				*lastChunk;		// the chunk pointer that was active last time
			UINT32
				lastOffset;		// the offset that was active last time
			UINT32
				lastNumToSearch;	// number of bytes we were asked to search last time
			bool
				justEntered;	// tells if just entered this subexpression
		} se;
	} pt;
	bool
		matchSpecified;			// user has specified a match limit for this node
	UINT32
		matchMin,				// minimum number of times to match pattern (0 means we dont even have to match this)
		matchMax;				// maximum number of times to match pattern (0 means infinite)
	struct expressionNode
		*nextExpression,		// next expression along the line at this level
		*parentExpression;		// parent expression, or NULL if at top
} EXPRESSION_NODE;

typedef struct expressionListHeader
{
	EXPRESSION_NODE
		*parentExpression;		// parent that got us here, or NULL if at top
	EXPRESSION_NODE
		*firstExpression;		// first expression in the list for this level
	struct expressionListHeader
		*nextAlternate;			// next alternate expression list for this level (NULL for no more)
} EXPRESSION_LIST_HEADER;

typedef struct compiledExpression
{
	EXPRESSION_LIST_HEADER
		*firstList;				// points to the first top-level expression list
	UINT32
		numRegisters;			// incremented every time a register is defined during compile
	bool
		registerDefined[MAX_REGISTERS];	// used during compilation to remember which registers have been defined already
	REGISTER
		expRegisters[MAX_REGISTERS];	// the registers that keep track of various match points
	bool
		charsMatched[256];		// contains list of all characters that can possibly be matched by this expression
	bool
		startCharsMatched[256];	// contains list of all characters that can possibly be matched by this expression at the start
	UINT32
		minimumMatchLength,		// tells the minimum length of a pattern that could be matched by this expression (0 means it can match nothing)
		maximumMatchLength;		// tells the maximum length of a pattern that could be matched by this expression
	bool
		infiniteMatchLength;	// tells if the maximum length of a pattern could be infinite (in which case maximumMatchLength should be ignored)
} COMPILED_EXPRESSION;

typedef struct
{
	bool
		leftEdge,				// tells if search is starting on logical left edge (used for edge match operations < and ^)
		rightEdge;				// tells if search is starting on logical right edge (used for edge match operations > and $)
	ABORT_TEST_FUNCTION
		*abortFunction;			// function to test for abort (or NULL if none)
	UINT32
		initialNumToSearch;		// tells how many bytes we were asked to search through at the start
	CHUNK_HEADER
		*endChunk;				// these are passed in everywhere
	UINT32
		endOffset;
} RECURSE_GLOBALS;

enum
{
	UNBALANCED_PAREN,
	UNBALANCED_SQUARE_BRACKET,
	BAD_RANGE_EXPRESSION,
	BAD_REPETITION,
	INVALID_REGISTER,
	DANGLING_QUOTE,
	INVALID_DIGIT,
	MISSING_DIGIT,
	STACK_OVERFLOW,
	SEARCH_ABORTED
};

static const char *errorDescriptions[]=
{
	"Unbalanced ()",
	"Unbalanced [ or [^",
	"Invalid {} expression",
	"Repetition operator *, +, ?, or {} applied to nothing",
	"Invalid register specification",
	"Incomplete \\ expression",
	"Invalid digit encountered",
	"Digit expected, but not found",
	"Expression stack overflow",
	"Search Aborted"
};

static EXPRESSION_NODE *REAllocateNode()
// allocate, and return an expression node
// if there is a problem, SetError, and return NULL
{
	return((EXPRESSION_NODE *)MNewPtr(sizeof(EXPRESSION_NODE)));
}

static void REFreeNode(EXPRESSION_NODE *expressionNode)
// free the node given by expressionNode
{
	MDisposePtr(expressionNode);
}

static EXPRESSION_LIST_HEADER *REAllocateList()
// allocate, and return an expression list header
// if there is a problem, SetError, and return NULL
{
	return((EXPRESSION_LIST_HEADER *)MNewPtr(sizeof(EXPRESSION_LIST_HEADER)));
}

static void REFreeList(EXPRESSION_LIST_HEADER *expressionList)
// free the list given by expressionList
{
	MDisposePtr(expressionList);
}

static EXPRESSION_NODE *REAddString(EXPRESSION_NODE *parent,EXPRESSION_NODE *currentNode,UINT8 character,const UINT8 *translateTable)
// add character to currentNode. If currentNode is NULL, or full, or not of the correct type
// then create a new node, and make character the first of the string
// NOTE: if currentNode was full, the new node is linked to it's nextExpression
// return the node that character was added to.
// if there is a problem, SetError, and return NULL
{
	EXPRESSION_NODE
		*node;

	if(currentNode&&(currentNode->type==PT_STRING)&&(!currentNode->matchSpecified)&&(currentNode->dataLength<MAX_STRING_PER_NODE))
	{
		node=currentNode;							// there is space in the current node, so add this in
		if(translateTable)
		{
			node->pt.data[node->dataLength++]=translateTable[character];	// drop in the data, bump the amount
		}
		else
		{
			node->pt.data[node->dataLength++]=character;	// drop in the data, bump the amount
		}
	}
	else
	{
		if((node=REAllocateNode()))
		{
			node->type=PT_STRING;					// create a node, and apply the correct type to it
			node->parentExpression=parent;			// set the parent
			node->dataLength=1;						// so far, it contains one byte of string data
			node->pt.data[0]=character;				// drop in the data
			node->matchSpecified=false;				// no match count specified for this node yet
			node->matchMin=1;						// until further notice, this must match exactly once
			node->matchMax=1;
			node->nextExpression=NULL;				// this is now the last expression
			if(currentNode)
			{
				currentNode->nextExpression=node;	// link to previous node if one existed
			}
		}
	}
	return(node);
}

#define	ADD_CHAR_BIT(array,character)	((array)[(character)>>3]|=(1<<((character)&0x07)))		// set bit in field
#define	CLR_CHAR_BIT(array,character)	((array)[(character)>>3]&=~(1<<((character)&0x07)))		// clear bit in field
#define	GET_CHAR_BIT(array,character)	((array)[(character)>>3]&(1<<((character)&0x07)))		// get bit from field

static bool REGetHexDigit(UINT8 *expression,UINT32 *index,UINT32 numBytes,UINT8 *nibble)
// get a hex digit from the input stream, and convert it to a nibble
// if there is a problem, set the error, and return false
{
	bool
		fail;
	UINT8
		character;

	fail=false;
	character='\0';
	if(*index<numBytes)
	{
		character=expression[(*index)++];
		if(character>='a')
		{
			character-='a'-'A';			// make lower case into upper case
		}
		if(character>='A')
		{
			character-='A'-10;			// make letter into number
			if(character>15)
			{
				SetError(errorDescriptions[INVALID_DIGIT]);
				fail=true;
			}
		}
		else
		{
			if(character>='0')
			{
				character-='0';
				if(character>9)
				{
					SetError(errorDescriptions[INVALID_DIGIT]);
					fail=true;
				}
			}
			else
			{
				SetError(errorDescriptions[INVALID_DIGIT]);
				fail=true;
			}
		}
	}
	else
	{
		SetError(errorDescriptions[MISSING_DIGIT]);
		fail=true;
	}
	*nibble=character;
	return(!fail);
}

static bool REConvertQuotedCharacter(UINT8 *expression,UINT32 *index,UINT32 numBytes,UINT8 *character)
// convert character after a \ into what it actually represents
// (for instance \n means newline)
// this can fail if there are no more characters after the \, or
// if the characters do not make sense
{
	bool
		fail;
	UINT8
		highNibble,
		lowNibble;

	fail=false;
	if(*index<numBytes)
	{
		*character=expression[(*index)++];
		switch(*character)
		{
			case 'b':
				*character=0x08;
				break;
			case 'n':
				*character='\n';
				break;
			case 'r':
				*character='\r';
				break;
			case 't':
				*character='\t';
				break;
			case 'x':
				if(REGetHexDigit(expression,index,numBytes,&highNibble))
				{
					if(REGetHexDigit(expression,index,numBytes,&lowNibble))
					{
						*character=highNibble*16+lowNibble;			// create actual character
					}
					else
					{
						fail=true;
					}
				}
				else
				{
					fail=true;
				}
				break;
		}
	}
	else
	{
		SetError(errorDescriptions[DANGLING_QUOTE]);
		fail=true;
	}
	return(!fail);
}

static void REAddRangeToList(EXPRESSION_NODE *node,UINT8 rangeLow,UINT8 rangeHigh,const UINT8 *translateTable)
// add the bytes between rangeLow, and rangeHigh (inclusive) to the list in node
{
	UINT32
		i;

	if(translateTable)
	{
		for(i=rangeLow;i<=(UINT32)rangeHigh;i++)
		{
			ADD_CHAR_BIT(&(node->pt.bits[0]),translateTable[i]);
		}
	}
	else
	{
		for(i=rangeLow;i<=(UINT32)rangeHigh;i++)
		{
			ADD_CHAR_BIT(&(node->pt.bits[0]),i);
		}
	}
}

static EXPRESSION_NODE *REAddList(EXPRESSION_NODE *parent,EXPRESSION_NODE *currentNode,UINT8 *expression,UINT32 *index,UINT32 numBytes,const UINT8 *translateTable)
// list operator has started, parse it, and create a list node in the list
// if there is a problem, SetError, and return NULL
{
	EXPRESSION_NODE
		*node;
	UINT8
		character,
		lastChar;
	bool
		notSearch,
		done,
		fail;
	UINT32
		i;

	if(*index<numBytes)
	{
		if((node=REAllocateNode()))
		{
			node->type=PT_LIST;							// create a node, and apply the correct type to it
			node->parentExpression=parent;				// set the parent
			node->matchSpecified=false;					// no match count specified for this node yet
			node->matchMin=1;							// until further notice, this must match exactly once
			node->matchMax=1;
			node->nextExpression=NULL;					// this is now the last expression
			for(i=0;i<BIT_TABLE_LENGTH;i++)				// clear the array
			{
				node->pt.bits[i]=0;
			}
			done=fail=false;
			notSearch=false;
			i=*index;									// remember where we started
			lastChar=0;									// start range at 0
			while(!done&&!fail&&(*index<numBytes))
			{
				switch(character=expression[(*index)++])
				{
					case '\\':
						if(REConvertQuotedCharacter(expression,index,numBytes,&character))
						{
							if(translateTable)
							{
								ADD_CHAR_BIT(&(node->pt.bits[0]),translateTable[character]);
							}
							else
							{
								ADD_CHAR_BIT(&(node->pt.bits[0]),character);
							}
						}
						else
						{
							fail=true;
						}
						break;
					case '^':
						if(*index-1==i)					// at start??
						{
							notSearch=true;				// set not flag
							character=0;				// keep last char pointed to the start
						}
						else
						{
							if(translateTable)
							{
								ADD_CHAR_BIT(&(node->pt.bits[0]),translateTable[character]);
							}
							else
							{
								ADD_CHAR_BIT(&(node->pt.bits[0]),character);
							}
						}
						break;
					case '-':
						if(*index<numBytes)				// get the next character for the range (if there is one)
						{
							character=expression[(*index)++];
							switch(character)
							{
								case '\\':
									if(REConvertQuotedCharacter(expression,index,numBytes,&character))
									{
										REAddRangeToList(node,lastChar,character,translateTable);
									}
									else
									{
										fail=true;
									}
									break;
								case ']':
									character=255;
									REAddRangeToList(node,lastChar,character,translateTable);		// range pointing into space is assumed max value
									done=true;
									break;
								default:
									REAddRangeToList(node,lastChar,character,translateTable);		// range pointing into space is assumed max value
									break;
							}
						}
						else
						{
							SetError(errorDescriptions[UNBALANCED_SQUARE_BRACKET]);
							fail=true;
						}
						break;
					case ']':
						done=true;						// scanning is done (empty lists are allowed)
						break;
					default:							// otherwise it is normal, and just gets added in
						if(translateTable)
						{
							ADD_CHAR_BIT(&(node->pt.bits[0]),translateTable[character]);
						}
						else
						{
							ADD_CHAR_BIT(&(node->pt.bits[0]),character);
						}
						break;
				}
				lastChar=character;						// remember this for next time around the loop
			}
			if(!done&&!fail)							// ran out of things to scan
			{
				SetError(errorDescriptions[UNBALANCED_SQUARE_BRACKET]);
				fail=true;
			}
			if(!fail)
			{
				if(notSearch)							// see if the table needs to be inverted
				{
					for(i=0;i<BIT_TABLE_LENGTH;i++)
					{
						node->pt.bits[i]^=0xFF;			// invert all the bits in the table if doing "notSearch" search
					}
				}
				if(currentNode)
				{
					currentNode->nextExpression=node;	// link to previous node if one existed
				}
				return(node);
			}
			REFreeNode(node);
		}
	}
	else
	{
		SetError(errorDescriptions[UNBALANCED_SQUARE_BRACKET]);
	}
	return(NULL);
}

static EXPRESSION_NODE *REAddAny(EXPRESSION_NODE *parent,EXPRESSION_NODE *currentNode,UINT8 *expression,UINT32 *index,UINT32 numBytes)
// create a list that matches any character except \n
// if there is a problem, SetError, and return NULL
{
	UINT32
		i;
	EXPRESSION_NODE
		*node;

	if((node=REAllocateNode()))
	{
		node->type=PT_LIST;							// create a node, and apply the correct type to it
		node->parentExpression=parent;				// set the parent
		node->matchSpecified=false;					// no match count specified for this node yet
		node->matchMin=1;							// until further notice, this must match exactly once
		node->matchMax=1;
		node->nextExpression=NULL;					// this is now the last expression
		for(i=0;i<BIT_TABLE_LENGTH;i++)				// set the array
		{
			node->pt.bits[i]=0xFF;					// select ALL characters
		}
		CLR_CHAR_BIT(&(node->pt.bits[0]),'\n');		// remove newline from the list
		if(currentNode)
		{
			currentNode->nextExpression=node;		// link to previous node if one existed
		}
		return(node);
	}
	return(NULL);
}

static EXPRESSION_NODE *REAddPreviousMatch(EXPRESSION_NODE *parent,EXPRESSION_NODE *currentNode,REGISTER *expRegister)
// link in a new node that matches previously matched text
// if there is a problem, SetError, and return NULL
{
	EXPRESSION_NODE
		*node;

	if((node=REAllocateNode()))
	{
		node->type=PT_PREVIOUSMATCH;				// create a node, and apply the correct type to it
		node->parentExpression=parent;				// set the parent
		node->matchSpecified=false;					// no match count specified for this node yet
		node->matchMin=1;							// until further notice, this must match exactly once
		node->matchMax=1;
		node->nextExpression=NULL;					// this is now the last expression
		node->pt.expRegister=expRegister;			// set pointer to the register that needs to be matched
		if(currentNode)
		{
			currentNode->nextExpression=node;		// link to previous node if one existed
		}
		return(node);
	}
	return(NULL);
}

static EXPRESSION_NODE *REAddStartOfWordMatch(EXPRESSION_NODE *parent,EXPRESSION_NODE *currentNode)
// link in a new node that matches the start of a word
// if there is a problem, SetError, and return NULL
{
	UINT32
		i;
	EXPRESSION_NODE
		*node;

	if((node=REAllocateNode()))
	{
		node->type=PT_LEFTEDGE;						// create a node, and apply the correct type to it
		node->parentExpression=parent;				// set the parent
		node->matchSpecified=false;					// no match count specified for this node yet
		node->matchMin=1;							// until further notice, this must match exactly once
		node->matchMax=1;
		node->nextExpression=NULL;					// this is now the last expression

		for(i=0;i<BIT_TABLE_LENGTH;i++)				// clear the array
		{
			node->pt.bits[i]=0;
		}
		for(i=0;i<256;i++)							// set the array to the current word list
		{
			if(wordSpaceTable[i])
			{
				ADD_CHAR_BIT(&(node->pt.bits[0]),i);
			}
		}

		if(currentNode)
		{
			currentNode->nextExpression=node;		// link to previous node if one existed
		}
		return(node);
	}
	return(NULL);
}

static EXPRESSION_NODE *REAddEndOfWordMatch(EXPRESSION_NODE *parent,EXPRESSION_NODE *currentNode)
// link in a new node that matches the end of a word
// if there is a problem, SetError, and return NULL
{
	UINT32
		i;
	EXPRESSION_NODE
		*node;

	if((node=REAllocateNode()))
	{
		node->type=PT_RIGHTEDGE;					// create a node, and apply the correct type to it
		node->parentExpression=parent;				// set the parent
		node->matchSpecified=false;					// no match count specified for this node yet
		node->matchMin=1;							// until further notice, this must match exactly once
		node->matchMax=1;
		node->nextExpression=NULL;					// this is now the last expression

		for(i=0;i<BIT_TABLE_LENGTH;i++)				// clear the array
		{
			node->pt.bits[i]=0;
		}
		for(i=0;i<256;i++)							// set the array to the current word list
		{
			if(wordSpaceTable[i])
			{
				ADD_CHAR_BIT(&(node->pt.bits[0]),i);
			}
		}

		if(currentNode)
		{
			currentNode->nextExpression=node;		// link to previous node if one existed
		}
		return(node);
	}
	return(NULL);
}

static EXPRESSION_NODE *REAddStartOfLineMatch(EXPRESSION_NODE *parent,EXPRESSION_NODE *currentNode)
// link in a new node that matches the start of a line
// if there is a problem, SetError, and return NULL
{
	UINT32
		i;
	EXPRESSION_NODE
		*node;

	if((node=REAllocateNode()))
	{
		node->type=PT_LEFTEDGE;						// create a node, and apply the correct type to it
		node->parentExpression=parent;				// set the parent
		node->matchSpecified=false;					// no match count specified for this node yet
		node->matchMin=1;							// until further notice, this must match exactly once
		node->matchMax=1;
		node->nextExpression=NULL;					// this is now the last expression

		for(i=0;i<BIT_TABLE_LENGTH;i++)				// clear the array
		{
			node->pt.bits[i]=0;
		}
		ADD_CHAR_BIT(&(node->pt.bits[0]),'\n');

		if(currentNode)
		{
			currentNode->nextExpression=node;		// link to previous node if one existed
		}
		return(node);
	}
	return(NULL);
}

static EXPRESSION_NODE *REAddEndOfLineMatch(EXPRESSION_NODE *parent,EXPRESSION_NODE *currentNode)
// link in a new node that matches the end of a line
// if there is a problem, SetError, and return NULL
{
	UINT32
		i;
	EXPRESSION_NODE
		*node;

	if((node=REAllocateNode()))
	{
		node->type=PT_RIGHTEDGE;					// create a node, and apply the correct type to it
		node->parentExpression=parent;				// set the parent
		node->matchSpecified=false;					// no match count specified for this node yet
		node->matchMin=1;							// until further notice, this must match exactly once
		node->matchMax=1;
		node->nextExpression=NULL;					// this is now the last expression

		for(i=0;i<BIT_TABLE_LENGTH;i++)				// clear the array
		{
			node->pt.bits[i]=0;
		}
		ADD_CHAR_BIT(&(node->pt.bits[0]),'\n');

		if(currentNode)
		{
			currentNode->nextExpression=node;		// link to previous node if one existed
		}
		return(node);
	}
	return(NULL);
}

static void REInhaleDigits(UINT8 *string,UINT32 *index,UINT32 numBytes,UINT32 *result)
// inhale as many digits as possible from string, starting at index, and possibly reading up to
// offset numBytes
// return result
// also return with index updated
// NOTE: it is ok for there to be no digits during the read, in this case
// result will return as 0, and index will not be changed
{
	UINT8
		character;

	*result=0;
	while(*index<numBytes&&(character=string[*index])&&character>='0'&&character<='9')
	{
		*result=*result*10+(character-'0');			// add this digit to the result
		(*index)++;
	}
}

static bool REParseRange(UINT8 *expression,UINT32 *index,UINT32 numBytes,UINT32 *rangeLow,UINT32 *rangeHigh)
// repetition range operator has started, read in the ranges, and return them in
// rangeLow, and rangeHigh
// if there is a problem, SetError, and return false
{
	*rangeHigh=0;											// this means infinity
	REInhaleDigits(expression,index,numBytes,rangeLow);	// attempt to read the low number in the range
	if(*index<numBytes)										// see if any more characters (looking for , or })
	{
		switch(expression[(*index)++])
		{
			case ',':
				REInhaleDigits(expression,index,numBytes,rangeHigh);	// attempt to read the high number in the range
				if(*index<numBytes&&expression[(*index)++]=='}')		// see if any more characters (looking for })
				{
					if((*rangeHigh==0)||(*rangeHigh>=*rangeLow))		// make sure low is <= high
					{
						return(true);
					}
				}
				SetError(errorDescriptions[BAD_RANGE_EXPRESSION]);
				break;
			case '}':
				if((*rangeHigh=*rangeLow))					// range with just one number means exactly that many times (0 is invalid)
				{
					return(true);
				}
				else
				{
					SetError(errorDescriptions[BAD_RANGE_EXPRESSION]);
				}
				break;
			default:
				SetError(errorDescriptions[BAD_RANGE_EXPRESSION]);
				break;
		}
	}
	else
	{
		SetError(errorDescriptions[BAD_RANGE_EXPRESSION]);
	}
	return(false);
}

// this is called recursively from below, so we must prototype it
static EXPRESSION_LIST_HEADER *REParseSubExpressionList(COMPILED_EXPRESSION *compiledExpression,EXPRESSION_NODE *parent,UINT8 *expression,UINT32 *index,UINT32 numBytes,UINT32 currentRegister,const UINT8 *translateTable);	// prototype this to keep compiler happy

static EXPRESSION_NODE *REAddSubExpression(COMPILED_EXPRESSION *compiledExpression,EXPRESSION_NODE *parent,EXPRESSION_NODE *currentNode,UINT8 *expression,UINT32 *index,UINT32 numBytes,const UINT8 *translateTable)
// create a subexpression node, and build the subexpression into it
// if there is a problem, SetError, and return NULL
{
	UINT32
		currentRegister;
	EXPRESSION_NODE
		*startNode,
		*endNode;

	if((startNode=REAllocateNode()))
	{
		startNode->type=PT_SUBEXPRESSIONSTART;				// create a start node, and apply the correct type to it
		startNode->parentExpression=parent;
		startNode->matchSpecified=false;					// no match count specified for this node yet (none ever will be)
		startNode->matchMin=1;								// set to some default values (never looked at)
		startNode->matchMax=1;
		if((startNode->nextExpression=endNode=REAllocateNode()))
		{
			endNode->type=PT_SUBEXPRESSIONEND;				// create a node, and apply the correct type to it
			endNode->parentExpression=parent;
			endNode->matchSpecified=false;					// no match count specified for this node yet
			endNode->matchMin=1;							// until further notice, this must match exactly once
			endNode->matchMax=1;
			endNode->nextExpression=NULL;					// this is now the last expression

			currentRegister=compiledExpression->numRegisters;
			if(currentRegister<MAX_REGISTERS)				// see if there is a register
			{
				endNode->pt.se.expRegister=&(compiledExpression->expRegisters[currentRegister]);		// remember the register
				compiledExpression->numRegisters++;			// increment the number of registers in use
			}
			else
			{
				endNode->pt.se.expRegister=NULL;			// no register for this one
			}

			if((endNode->pt.se.subExpressionList=REParseSubExpressionList(compiledExpression,endNode,expression,index,numBytes,currentRegister,translateTable)))		// get subexpression
			{
				if(currentNode)
				{
					currentNode->nextExpression=startNode;	// link to previous node if one existed
				}
				return(startNode);
			}
			REFreeNode(endNode);
		}
		REFreeNode(startNode);
	}
	return(NULL);
}

static void REFreeLists(EXPRESSION_LIST_HEADER *list)
// given a pointer to the top of an expression list, free
// all subnodes, and all alternates below
// then free list
// NOTE: this calls itself recursively for sublists
{
	EXPRESSION_LIST_HEADER
		*currentList,
		*nextAlternate;
	EXPRESSION_NODE
		*currentNode,
		*nextNode;

	currentList=list;
	while(currentList)
	{
		nextAlternate=currentList->nextAlternate;
		currentNode=currentList->firstExpression;
		while(currentNode)
		{
			nextNode=currentNode->nextExpression;
			if(currentNode->type==PT_SUBEXPRESSIONEND)			// free sub-list
			{
				REFreeLists(currentNode->pt.se.subExpressionList);		// free all the nodes of this subexpression
			}
			REFreeNode(currentNode);
			currentNode=nextNode;								// move to the next horizontally
		}
		REFreeList(currentList);
		currentList=nextAlternate;
	}
}

static EXPRESSION_LIST_HEADER *REParseSubExpressionList(COMPILED_EXPRESSION *compiledExpression,EXPRESSION_NODE *parent,UINT8 *expression,UINT32 *index,UINT32 numBytes,UINT32 currentRegister,const UINT8 *translateTable)
// parse expression, and build lists of nodes for compiledExpression
// (if compiledExpression is NULL, then it is assumed that we are compiling a sublist)
// return the head of the created list of expression nodes
// if there is a problem, SetError, and return NULL
{
	EXPRESSION_LIST_HEADER
		*startList,
		*currentList;						// current list in the list of alternates
	EXPRESSION_NODE
		*startNode,
		*currentNode;
	UINT8
		character;
	UINT32
		expRegisterIndex;
	bool
		done,
		fail;

	fail=done=false;
	if((startList=REAllocateList()))
	{
		startList->parentExpression=parent;
		startList->firstExpression=NULL;
		startList->nextAlternate=NULL;
		currentList=startList;								// point to the current list that is being built
		startNode=currentNode=NULL;
		while(!fail&&!done&&((*index)<numBytes))
		{
			switch(character=expression[(*index)++])
			{
				case '(':
					fail=!((currentNode=REAddSubExpression(compiledExpression,parent,currentNode,expression,index,numBytes,translateTable)));
					break;
				case ')':
					if(parent)								// if not at the top, then take this as the end of our group
					{
						if(currentRegister<MAX_REGISTERS)	// define this register now
						{
							compiledExpression->registerDefined[currentRegister]=true;
						}
						done=true;
					}
					else
					{
						SetError(errorDescriptions[UNBALANCED_PAREN]);
						fail=true;
					}
					break;
				case '{':
					if(currentNode&&!currentNode->matchSpecified&&currentNode->type!=PT_LEFTEDGE&&currentNode->type!=PT_RIGHTEDGE)
					{
						currentNode->matchSpecified=true;	// match has been specified
						fail=!REParseRange(expression,index,numBytes,&currentNode->matchMin,&currentNode->matchMax);
					}
					else
					{
						SetError(errorDescriptions[BAD_REPETITION]);
						fail=true;
					}
					break;
				case '*':									// match 0 or more
					if(currentNode&&!currentNode->matchSpecified&&currentNode->type!=PT_LEFTEDGE&&currentNode->type!=PT_RIGHTEDGE)
					{
						currentNode->matchSpecified=true;	// match has been specified
						currentNode->matchMin=0;			// no requirement to match any
						currentNode->matchMax=0;			// match as many as possible
					}
					else
					{
						SetError(errorDescriptions[BAD_REPETITION]);
						fail=true;
					}
					break;
				case '+':									// match 1 or more
					if(currentNode&&!currentNode->matchSpecified&&currentNode->type!=PT_LEFTEDGE&&currentNode->type!=PT_RIGHTEDGE)
					{
						currentNode->matchSpecified=true;	// match has been specified
						currentNode->matchMin=1;			// match at least 1
						currentNode->matchMax=0;			// match as many as possible
					}
					else
					{
						SetError(errorDescriptions[BAD_REPETITION]);
						fail=true;
					}
					break;
				case '?':									// match 0 or 1
					if(currentNode&&!currentNode->matchSpecified&&currentNode->type!=PT_LEFTEDGE&&currentNode->type!=PT_RIGHTEDGE)
					{
						currentNode->matchSpecified=true;	// match has been specified
						currentNode->matchMin=0;			// no requirement to match any
						currentNode->matchMax=1;			// match at most 1
					}
					else
					{
						SetError(errorDescriptions[BAD_REPETITION]);
						fail=true;
					}
					break;
				case '[':
					fail=!(currentNode=REAddList(parent,currentNode,expression,index,numBytes,translateTable));
					break;
				case '|':
					if((currentList->nextAlternate=REAllocateList()))	// create a new list, link it to the old, and make it current
					{
						currentList=currentList->nextAlternate;
						currentList->parentExpression=parent;
						currentList->nextAlternate=NULL;
						currentList->firstExpression=NULL;
						currentNode=NULL;					// no longer in the middle of a node
					}
					else
					{
						fail=true;
					}
					break;
				case '.':
					fail=!(currentNode=REAddAny(parent,currentNode,expression,index,numBytes));
					break;
				case '<':
					fail=!(currentNode=REAddStartOfWordMatch(parent,currentNode));
					break;
				case '>':
					fail=!(currentNode=REAddEndOfWordMatch(parent,currentNode));
					break;
				case '^':
					fail=!(currentNode=REAddStartOfLineMatch(parent,currentNode));
					break;
				case '$':
					fail=!(currentNode=REAddEndOfLineMatch(parent,currentNode));
					break;
				case '\\':
					if(*index<=numBytes)
					{
						character=expression[*index];
						if(character>='0'&&character<='9')								// see if sub pattern match
						{
							(*index)++;
							expRegisterIndex=character-'0';								// make ASCII digit into binary number
							if(expRegisterIndex<MAX_REGISTERS&&compiledExpression->registerDefined[expRegisterIndex])	// make sure it is in range, and that the given register has already been defined (this will guarantee that it is matched before we need it)
							{
								fail=!(currentNode=REAddPreviousMatch(parent,currentNode,&compiledExpression->expRegisters[expRegisterIndex]));
							}
							else
							{
								SetError(errorDescriptions[INVALID_REGISTER]);
								fail=true;
							}
						}
						else
						{
							if(REConvertQuotedCharacter(expression,index,numBytes,&character))
							{
								fail=!(currentNode=REAddString(parent,currentNode,character,translateTable));
							}
							else
							{
								fail=true;
							}
						}
					}
					else
					{
						SetError(errorDescriptions[DANGLING_QUOTE]);
						fail=true;
					}
					break;
				default:
					fail=!(currentNode=REAddString(parent,currentNode,character,translateTable));	// just take the character as part of a string
					break;
			}
			if(currentNode)
			{
				if(!currentList->firstExpression)
				{
					currentList->firstExpression=currentNode;
				}
				if(currentNode->type==PT_SUBEXPRESSIONSTART)							// if just had start of expression, move to the end
				{
					currentNode=currentNode->nextExpression;
				}
			}
		}
		if(!fail&&!done&&parent)														// reached the end, but in middle of group somewhere
		{
			SetError(errorDescriptions[UNBALANCED_PAREN]);
			fail=true;
		}
		if(!fail)
		{
			return(startList);
		}
		REFreeLists(startList);
	}
	return(NULL);
}

void REFree(COMPILED_EXPRESSION *expression)
// free all nodes used by expression
// this must be called for every call to RECompile to deallocate the
// memory used by expression
{
	REFreeLists(expression->firstList);				// free the lists
	MDisposePtr(expression);						// remove the expression header
}

static void MoveOver(EXPRESSION_LIST_HEADER *list,UINT32 depth)
// move over by depth spaces
{
	UINT32
		i;

	for(i=0;i<depth;i++)
	{
		printf(" ");
	}
}

static void PrintNode(EXPRESSION_LIST_HEADER *list,EXPRESSION_NODE *node,UINT32 depth)
// print formatted node information
{
	UINT32
		i;

	switch(node->type)
	{
		case PT_STRING:
			MoveOver(list,depth);
			printf("STRING %d/%d '",(int)node->matchMin,(int)node->matchMax);
			for(i=0;i<node->dataLength;i++)
			{
				printf("%c",node->pt.data[i]);
			}
			printf("'\n");
			break;
		case PT_LIST:
			MoveOver(list,depth);
			printf("LIST %d/%d ",(int)node->matchMin,(int)node->matchMax);
			for(i=0;i<BIT_TABLE_LENGTH;i++)
			{
				printf("%02X ",node->pt.bits[i]);
			}
			printf("\n");
			break;
		case PT_PREVIOUSMATCH:
			MoveOver(list,depth);
			printf("PREVIOUSMATCH %d/%d\n",(int)node->matchMin,(int)node->matchMax);
			break;
		case PT_SUBEXPRESSIONSTART:
			MoveOver(list,depth);
			printf("SUBEXPRESSIONSTART %d/%d\n",(int)node->matchMin,(int)node->matchMax);
			break;
		case PT_SUBEXPRESSIONEND:
			MoveOver(list,depth);
			printf("SUBEXPRESSIONEND %d/%d\n",(int)node->matchMin,(int)node->matchMax);
			break;
		case PT_LEFTEDGE:
			MoveOver(list,depth);
			printf("LEFTEDGE %d/%d\n",(int)node->matchMin,(int)node->matchMax);
			break;
		case PT_RIGHTEDGE:
			MoveOver(list,depth);
			printf("RIGHTEDGE %d/%d\n",(int)node->matchMin,(int)node->matchMax);
			break;
		default:
			MoveOver(list,depth);
			printf("????? (%d)\n",(int)node->type);
			break;
	}
}

static void PrintExpressionList(EXPRESSION_LIST_HEADER *list,UINT32 depth)
// dump debugging output of the compiled expression
// NOTE: this calls itself
{
	EXPRESSION_LIST_HEADER
		*currentList;
	EXPRESSION_NODE
		*currentNode;

	currentList=list;
	while(currentList)
	{
		currentNode=currentList->firstExpression;
		while(currentNode)
		{
			if(currentNode->type==PT_SUBEXPRESSIONEND)
			{
				PrintExpressionList(currentNode->pt.se.subExpressionList,depth+1);		// print the sublist
			}
			PrintNode(currentList,currentNode,depth);
			currentNode=currentNode->nextExpression;
		}
		currentList=currentList->nextAlternate;
	}
}

void REPrintCompiledExpression(COMPILED_EXPRESSION *expression)
// dump debugging output of the compiled expression
{
	UINT32
		i;

	PrintExpressionList(expression->firstList,0);
	if(expression->infiniteMatchLength)
	{
		printf("Max match is infinite\n");
	}
	else
	{
		printf("Max match length=%d\n",expression->maximumMatchLength);
	}
	printf("Min match length=%d\n",expression->minimumMatchLength);
	printf("Matching start chars=");
	for(i=0;i<256;i++)
	{
		printf("%02X ",expression->startCharsMatched[i]);
	}
	printf("\n");
	printf("Matching chars=");
	for(i=0;i<256;i++)
	{
		printf("%02X ",expression->charsMatched[i]);
	}
	printf("\n");
}

static void AnalyzeExpressionList(COMPILED_EXPRESSION *expression,EXPRESSION_LIST_HEADER *list,UINT32 *minLength,UINT32 *maxLength,bool *infiniteLength,bool haveStartChar)
// Analyze the given expression list, recursively calling this routine as needed
{
	EXPRESSION_LIST_HEADER
		*currentList;
	EXPRESSION_NODE
		*currentNode;
	UINT32
		localMinLength,
		localMaxLength,
		alternateMinLength,
		alternateMaxLength,
		subExpressionMinLength,
		subExpressionMaxLength;
	bool
		subExpressionInfiniteLength;
	UINT32
		i;
	bool
		firstAlternate;

	currentList=list;
	firstAlternate=true;

	localMinLength=0;
	localMaxLength=0;

	while(currentList)
	{
		alternateMinLength=0;
		alternateMaxLength=0;
		currentNode=currentList->firstExpression;
		while(currentNode)
		{
			switch(currentNode->type)
			{
				case PT_STRING:
					if((!haveStartChar)&&(alternateMinLength==0))	// if at start character of expression, add character to list of possible start characters
					{
						expression->startCharsMatched[currentNode->pt.data[0]]=true;
					}
					alternateMinLength+=(currentNode->dataLength-1)+currentNode->matchMin;	// last character must occur this many times
					if(currentNode->matchMax)
					{
						alternateMaxLength+=currentNode->dataLength+(currentNode->matchMax-1);	// last character gets repeated
					}
					else
					{
						*infiniteLength=true;
					}
					for(i=0;i<currentNode->dataLength;i++)
					{
						expression->charsMatched[currentNode->pt.data[i]]=true;
					}
					break;
				case PT_LIST:
					if((!haveStartChar)&&(alternateMinLength==0))	// if at start character of expression, add character to list of possible start characters
					{
						for(i=0;i<256;i++)
						{
							if(GET_CHAR_BIT(&currentNode->pt.bits[0],i))
							{
								expression->startCharsMatched[i]=true;
							}
						}
					}
					alternateMinLength+=currentNode->matchMin;
					if(currentNode->matchMax)
					{
						alternateMaxLength+=currentNode->matchMax;
					}
					else
					{
						*infiniteLength=true;
					}
					for(i=0;i<256;i++)
					{
						if(GET_CHAR_BIT(&currentNode->pt.bits[0],i))
						{
							expression->charsMatched[i]=true;
						}
					}
					break;
				case PT_PREVIOUSMATCH:
					alternateMinLength+=currentNode->matchMin*currentNode->pt.expRegister->minimumMatchLength;
					if(currentNode->matchMax)
					{
						*infiniteLength|=currentNode->pt.expRegister->infiniteMatchLength;
						alternateMaxLength+=currentNode->matchMax*currentNode->pt.expRegister->maximumMatchLength;
					}
					else
					{
						*infiniteLength=true;
					}
					break;
				case PT_SUBEXPRESSIONSTART:
					break;
				case PT_SUBEXPRESSIONEND:
					subExpressionMinLength=0;
					subExpressionMaxLength=0;
					subExpressionInfiniteLength=false;
					AnalyzeExpressionList(expression,currentNode->pt.se.subExpressionList,&subExpressionMinLength,&subExpressionMaxLength,&subExpressionInfiniteLength,(haveStartChar||alternateMinLength>0));	// analyze the sublist
					if(currentNode->pt.se.expRegister)
					{
						currentNode->pt.se.expRegister->minimumMatchLength=subExpressionMinLength;	// remember the min length for this sub-expression
						currentNode->pt.se.expRegister->maximumMatchLength=subExpressionMaxLength;	// remember the max length for this sub-expression
						currentNode->pt.se.expRegister->infiniteMatchLength=subExpressionInfiniteLength;	// remember if this could match infinitely
					}
					alternateMinLength+=currentNode->matchMin*subExpressionMinLength;
					if(currentNode->matchMax&&!subExpressionInfiniteLength)
					{
						alternateMaxLength+=currentNode->matchMax*subExpressionMaxLength;
					}
					else
					{
						*infiniteLength=true;
					}
					break;
				case PT_LEFTEDGE:
					break;
				case PT_RIGHTEDGE:
					break;
				default:
					break;
			}
			currentNode=currentNode->nextExpression;
		}

		if(firstAlternate||alternateMinLength<localMinLength)
		{
			localMinLength=alternateMinLength;
		}
		if(alternateMaxLength>localMaxLength)
		{
			localMaxLength=alternateMaxLength;
		}
		firstAlternate=false;

		currentList=currentList->nextAlternate;
	}

	*minLength+=localMinLength;
	*maxLength+=localMaxLength;
}

static void AnalyzeExpression(COMPILED_EXPRESSION *expression)
// Work out which characters can possibly occur in expression.
// Also, work out the maximum length of a match of expression.
// NOTE: if the maximum match length is infinite, set maximumMatchLength to 0.
{
	int
		i;

	for(i=0;i<256;i++)
	{
		expression->charsMatched[i]=false;				// clear the bit table of possible matches
		expression->startCharsMatched[i]=false;			// clear the bit table of possible matches at the start
	}
	expression->minimumMatchLength=0;
	expression->maximumMatchLength=0;
	expression->infiniteMatchLength=false;
	AnalyzeExpressionList(expression,expression->firstList,&expression->minimumMatchLength,&expression->maximumMatchLength,&expression->infiniteMatchLength,false);	// recursively analyze the lists
}

COMPILED_EXPRESSION *RECompile(UINT8 *expression,UINT32 numBytes,const UINT8 *translateTable)
// compile numBytes of expression, and return a pointer to the
// node of the start of the expression.
// if there is a problem, call SetError, and return NULL
{
	COMPILED_EXPRESSION
		*compiledExpression;
	UINT32
		i,
		index;

	if((compiledExpression=(COMPILED_EXPRESSION *)MNewPtr(sizeof(COMPILED_EXPRESSION))))
	{
		compiledExpression->numRegisters=0;						// no registers encountered yet
		for(i=0;i<MAX_REGISTERS;i++)
		{
			compiledExpression->registerDefined[i]=false;		// no registers have been defined yet
		}
		index=0;
		if((compiledExpression->firstList=REParseSubExpressionList(compiledExpression,NULL,expression,&index,numBytes,0,translateTable)))
		{
			AnalyzeExpression(compiledExpression);				// look at the expression, and work out various pieces of information which can be used to determine how far back one needs to go from a text change to locate an expression which includes the change
//			REPrintCompiledExpression(compiledExpression);		// dump out the expression
			return(compiledExpression);
		}
		MDisposePtr(compiledExpression);
	}
	return(NULL);
}

// MATCHING FUNCTIONS -----------------------------------------------------------------------------------------------------------------

static UINT32
	matchRecursionDepth;					// used to keep track of how far we are recursing during match attempts, allows us to fail without running out of stack

static EXPRESSION_NODE *REGetNextNode(EXPRESSION_NODE *node)
// get the next expression node at this level, or move up a level if possible
{
	EXPRESSION_NODE
		*newNode;

	if(!(newNode=node->nextExpression))
	{
		newNode=node->parentExpression;
	}
	return(newNode);
}

// these are called recursively from below, so they must be prototyped
static bool REMatchAlternatives(CHUNK_HEADER *chunk,UINT32 offset,UINT32 numToSearch,EXPRESSION_LIST_HEADER *list,bool *foundMatch,UINT32 *numMatched,RECURSE_GLOBALS *recurseGlobals);
static bool REMatchStringNode(CHUNK_HEADER *chunk,UINT32 offset,UINT32 numToSearch,EXPRESSION_NODE *node,bool *foundMatch,UINT32 *numMatched,RECURSE_GLOBALS *recurseGlobals);
static bool REMatchStringNodeTT(CHUNK_HEADER *chunk,UINT32 offset,UINT32 numToSearch,EXPRESSION_NODE *node,bool *foundMatch,UINT32 *numMatched,RECURSE_GLOBALS *recurseGlobals);
static bool REMatchListNode(CHUNK_HEADER *chunk,UINT32 offset,UINT32 numToSearch,EXPRESSION_NODE *node,bool *foundMatch,UINT32 *numMatched,RECURSE_GLOBALS *recurseGlobals);
static bool REMatchListNodeTT(CHUNK_HEADER *chunk,UINT32 offset,UINT32 numToSearch,EXPRESSION_NODE *node,bool *foundMatch,UINT32 *numMatched,RECURSE_GLOBALS *recurseGlobals);
static bool REMatchPreviousNode(CHUNK_HEADER *chunk,UINT32 offset,UINT32 numToSearch,EXPRESSION_NODE *node,bool *foundMatch,UINT32 *numMatched,RECURSE_GLOBALS *recurseGlobals);
static bool REMatchPreviousNodeTT(CHUNK_HEADER *chunk,UINT32 offset,UINT32 numToSearch,EXPRESSION_NODE *node,bool *foundMatch,UINT32 *numMatched,RECURSE_GLOBALS *recurseGlobals);
static bool REMatchSubStartNode(CHUNK_HEADER *chunk,UINT32 offset,UINT32 numToSearch,EXPRESSION_NODE *node,bool *foundMatch,UINT32 *numMatched,RECURSE_GLOBALS *recurseGlobals);
static bool REMatchSubEndNode(CHUNK_HEADER *chunk,UINT32 offset,UINT32 numToSearch,EXPRESSION_NODE *node,bool *foundMatch,UINT32 *numMatched,RECURSE_GLOBALS *recurseGlobals);
static bool REMatchLeftEdgeNode(CHUNK_HEADER *chunk,UINT32 offset,UINT32 numToSearch,EXPRESSION_NODE *node,bool *foundMatch,UINT32 *numMatched,RECURSE_GLOBALS *recurseGlobals);
static bool REMatchRightEdgeNode(CHUNK_HEADER *chunk,UINT32 offset,UINT32 numToSearch,EXPRESSION_NODE *node,bool *foundMatch,UINT32 *numMatched,RECURSE_GLOBALS *recurseGlobals);

// array of pointers to exact matching routines
static bool
	(*exactRoutines[])(CHUNK_HEADER *,UINT32,UINT32,EXPRESSION_NODE *,bool *,UINT32 *,RECURSE_GLOBALS *)=
	{
		REMatchStringNode,
		REMatchListNode,
		REMatchPreviousNode,
		REMatchSubStartNode,
		REMatchSubEndNode,
		REMatchLeftEdgeNode,
		REMatchRightEdgeNode
	};

// array of pointers to translated matching routines
static bool
	(*translatedRoutines[])(CHUNK_HEADER *,UINT32,UINT32,EXPRESSION_NODE *,bool *,UINT32 *,RECURSE_GLOBALS *)=
	{
		REMatchStringNodeTT,
		REMatchListNodeTT,
		REMatchPreviousNodeTT,
		REMatchSubStartNode,
		REMatchSubEndNode,
		REMatchLeftEdgeNode,
		REMatchRightEdgeNode
	};

// this points to the array that tells what routines to call for each node type
static bool
	(**matchRoutines)(CHUNK_HEADER *,UINT32,UINT32,EXPRESSION_NODE *,bool *,UINT32 *,RECURSE_GLOBALS *);

// points to the translation table to use during translated compares
static const UINT8
	*translationTable;

// macro function that branches to next match routine from current one
#define	RECallNextNode(chunk,offset,numToSearch,node,foundMatch,numMatched,recurseGlobals)					\
{																											\
	fail=!matchRoutines[node->type](chunk,offset,numToSearch,node,foundMatch,numMatched,recurseGlobals);	\
}

static bool REMatchStringNode(CHUNK_HEADER *chunk,UINT32 offset,UINT32 numToSearch,EXPRESSION_NODE *node,bool *foundMatch,UINT32 *numMatched,RECURSE_GLOBALS *recurseGlobals)
// attempt to match node with the data at chunk/offset
// if the pattern could be matched, return true in foundMatch with numMatched as the number of bytes matched
// if no match was located, return false in foundMatch
// NOTE: do not search through more than numToSearch bytes
// endChunk/endOffset will point to the end of the match if one was located
// if there is some sort of hard failure, this routine will SetError, and return false
{
	CHUNK_HEADER
		*lastChunk;												// need to remember this in case we fall off the edge
	UINT32
		i,
		nodeMatched;
	bool
		fail,
		matching;
	UINT8
		character;
	EXPRESSION_NODE
		*nextExpression;

	lastChunk=NULL;
	if(++matchRecursionDepth<MAX_RECURSION_DEPTH)
	{
		fail=false;
		matching=(node->dataLength-1<=(int)numToSearch);		// make sure there are enough bytes to even allow a match
		for(i=0;matching&&((int)i<node->dataLength-1);i++)		// match the constant part of the string (if one exists)
		{
			matching=(chunk->data[offset]==node->pt.data[i]);	// check for match
			if((++offset)>=chunk->totalBytes)					// move forward in the input
			{
				lastChunk=chunk;
				chunk=chunk->nextHeader;
				offset=0;
			}
		}
		if(matching)											// make sure we made it through the first part
		{
			numToSearch-=(node->dataLength-1);					// matched the constant part of the string, now try to get up to matchMax of last character
			character=node->pt.data[node->dataLength-1];
			i=0;
			while(matching&&(i<numToSearch)&&((!node->matchMax)||(i<node->matchMax)))	// read as many as possible until we hit max, or fail
			{
				if((matching=(chunk->data[offset]==character)))	// check for match
				{
					if((++offset)>=chunk->totalBytes)			// move forward in the input iff there was a match
					{
						lastChunk=chunk;
						chunk=chunk->nextHeader;
						offset=0;
					}
					i++;
				}
			}
			if(i>=node->matchMin)								// see if we got the minimum number of matches
			{
				if((nextExpression=REGetNextNode(node)))		// see if there is another node after this one (if so, recurse into it)
				{
					RECallNextNode(chunk,offset,numToSearch-i,nextExpression,&matching,&nodeMatched,recurseGlobals);
					while(!fail&&!matching&&i>node->matchMin)	// see if possible to backup and try again
					{
						i--;
						if(chunk)								// back up 1 character
						{
							if(offset)
							{
								offset--;
							}
							else
							{
								chunk=chunk->previousHeader;
								offset=chunk->totalBytes-1;
							}
						}
						else
						{
							chunk=lastChunk;
							offset=chunk->totalBytes-1;
						}
						RECallNextNode(chunk,offset,numToSearch-i,nextExpression,&matching,&nodeMatched,recurseGlobals);
					}
					if(!fail&&matching)							// match was located!!
					{
						(*numMatched)=node->dataLength-1+i+nodeMatched;		// return number of characters that matched
					}
				}
				else											// no other nodes, this was at the end, so it is done
				{
					(*numMatched)=node->dataLength-1+i;			// return number of characters that matched
					recurseGlobals->endChunk=chunk;
					recurseGlobals->endOffset=offset;
					matching=true;								// if no more nodes, then this list is complete, and it matched
				}
			}
			else
			{
				matching=false;									// did not find minimum number of matches, so this fails here and now
			}
		}
		(*foundMatch)=matching;
		matchRecursionDepth--;
	}
	else
	{
		SetError(errorDescriptions[STACK_OVERFLOW]);
		fail=true;
	}
	return(!fail);
}

static bool REMatchStringNodeTT(CHUNK_HEADER *chunk,UINT32 offset,UINT32 numToSearch,EXPRESSION_NODE *node,bool *foundMatch,UINT32 *numMatched,RECURSE_GLOBALS *recurseGlobals)
// Exactly like REMatchStringNode except that this one uses a translation table during the search
{
	CHUNK_HEADER
		*lastChunk;												// need to remember this in case we fall off the edge
	UINT32
		i,
		nodeMatched;
	bool
		fail,
		matching;
	UINT8
		character;
	EXPRESSION_NODE
		*nextExpression;

	lastChunk=NULL;
	if(++matchRecursionDepth<MAX_RECURSION_DEPTH)
	{
		fail=false;
		matching=(node->dataLength-1<=(int)numToSearch);		// make sure there are enough bytes to even allow a match
		for(i=0;matching&&((int)i<node->dataLength-1);i++)		// match the constant part of the string (if one exists)
		{
			matching=(translationTable[chunk->data[offset]]==translationTable[node->pt.data[i]]);	// check for match
			if((++offset)>=chunk->totalBytes)					// move forward in the input
			{
				lastChunk=chunk;
				chunk=chunk->nextHeader;
				offset=0;
			}
		}
		if(matching)											// make sure we made it through the first part
		{
			numToSearch-=(node->dataLength-1);					// matched the constant part of the string, now try to get up to matchMax of last character
			character=translationTable[node->pt.data[node->dataLength-1]];
			i=0;
			while(matching&&(i<numToSearch)&&((!node->matchMax)||(i<node->matchMax)))	// read as many as possible until we hit max, or fail
			{
				if((matching=(translationTable[chunk->data[offset]]==character)))	// check for match
				{
					if((++offset)>=chunk->totalBytes)			// move forward in the input iff there was a match
					{
						lastChunk=chunk;
						chunk=chunk->nextHeader;
						offset=0;
					}
					i++;
				}
			}
			if(i>=node->matchMin)								// see if we got the minimum number of matches
			{
				if((nextExpression=REGetNextNode(node)))		// see if there is another node after this one (if so, recurse into it)
				{
					RECallNextNode(chunk,offset,numToSearch-i,nextExpression,&matching,&nodeMatched,recurseGlobals);
					while(!fail&&!matching&&i>node->matchMin)	// see if possible to backup and try again
					{
						i--;
						if(chunk)								// back up 1 character
						{
							if(offset)
							{
								offset--;
							}
							else
							{
								chunk=chunk->previousHeader;
								offset=chunk->totalBytes-1;
							}
						}
						else
						{
							chunk=lastChunk;
							offset=chunk->totalBytes-1;
						}
						RECallNextNode(chunk,offset,numToSearch-i,nextExpression,&matching,&nodeMatched,recurseGlobals);
					}
					if(!fail&&matching)							// match was located!!
					{
						(*numMatched)=node->dataLength-1+i+nodeMatched;		// return number of characters that matched
					}
				}
				else											// no other nodes, this was at the end, so it is done
				{
					(*numMatched)=node->dataLength-1+i;			// return number of characters that matched
					recurseGlobals->endChunk=chunk;
					recurseGlobals->endOffset=offset;
					matching=true;								// if no more nodes, then this list is complete, and it matched
				}
			}
			else
			{
				matching=false;									// did not find minimum number of matches, so this fails here and now
			}
		}
		(*foundMatch)=matching;
		matchRecursionDepth--;
	}
	else
	{
		SetError(errorDescriptions[STACK_OVERFLOW]);
		fail=true;
	}
	return(!fail);
}

static bool REMatchListNode(CHUNK_HEADER *chunk,UINT32 offset,UINT32 numToSearch,EXPRESSION_NODE *node,bool *foundMatch,UINT32 *numMatched,RECURSE_GLOBALS *recurseGlobals)
// attempt to match node with the data at chunk/offset
// if the pattern could be matched, return true in foundMatch with numMatched as the number of bytes matched
// if no match was located, return false in foundMatch
// NOTE: do not search through more than numToSearch bytes
// endChunk/endOffset will point to the end of the match if one was located
// if there is some sort of hard failure, this routine will SetError, and return false
{
	CHUNK_HEADER
		*lastChunk;												// need to remember this in case we fall off the edge
	UINT32
		i,
		nodeMatched;
	bool
		fail,
		matching;
	UINT8
		character;
	EXPRESSION_NODE
		*nextExpression;

	lastChunk=NULL;
	if(++matchRecursionDepth<MAX_RECURSION_DEPTH)
	{
		fail=false;
		matching=true;											// start by assuming a match
		i=0;
		while(matching&&(i<numToSearch)&&((!node->matchMax)||(i<node->matchMax)))	// read as many as possible until we hit max, or fail
		{
			character=chunk->data[offset];
			if((matching=(GET_CHAR_BIT(node->pt.bits,character)!=0)))	// check for match
			{
				if((++offset)>=chunk->totalBytes)				// move forward in the input iff there was a match
				{
					lastChunk=chunk;
					chunk=chunk->nextHeader;
					offset=0;
				}
				i++;
			}
		}
		if(i>=node->matchMin)								// see if we got the minimum number of matches
		{
			if((nextExpression=REGetNextNode(node)))		// see if there is another node after this one (if so, recurse into it)
			{
				RECallNextNode(chunk,offset,numToSearch-i,nextExpression,&matching,&nodeMatched,recurseGlobals);
				while(!fail&&!matching&&i>node->matchMin)	// see if possible to backup and try again
				{
					i--;
					if(chunk)								// back up 1 character
					{
						if(offset)
						{
							offset--;
						}
						else
						{
							chunk=chunk->previousHeader;
							offset=chunk->totalBytes-1;
						}
					}
					else
					{
						chunk=lastChunk;
						offset=chunk->totalBytes-1;
					}
					RECallNextNode(chunk,offset,numToSearch-i,nextExpression,&matching,&nodeMatched,recurseGlobals);
				}
				if(!fail&&matching)							// match was located!!
				{
					(*numMatched)=i+nodeMatched;			// return number of characters that matched
				}
			}
			else											// no other nodes, this was at the end, so it is done
			{
				(*numMatched)=i;							// return number of characters that matched
				recurseGlobals->endChunk=chunk;
				recurseGlobals->endOffset=offset;
				matching=true;								// if no more nodes, then this list is complete, and it matched
			}
		}
		else
		{
			matching=false;									// did not find minimum number of matches, so this fails here and now
		}
		(*foundMatch)=matching;
		matchRecursionDepth--;
	}
	else
	{
		SetError(errorDescriptions[STACK_OVERFLOW]);
		fail=true;
	}
	return(!fail);
}

static bool REMatchListNodeTT(CHUNK_HEADER *chunk,UINT32 offset,UINT32 numToSearch,EXPRESSION_NODE *node,bool *foundMatch,UINT32 *numMatched,RECURSE_GLOBALS *recurseGlobals)
// exactly like REMatchListNode but uses a translation table during the compare
{
	CHUNK_HEADER
		*lastChunk;											// need to remember this in case we fall off the edge
	UINT32
		i,
		nodeMatched;
	bool
		fail,
		matching;
	UINT8
		character;
	EXPRESSION_NODE
		*nextExpression;

	lastChunk=NULL;
	if(++matchRecursionDepth<MAX_RECURSION_DEPTH)
	{
		fail=false;
		matching=true;										// start by assuming a match
		i=0;
		while(matching&&(i<numToSearch)&&((!node->matchMax)||(i<node->matchMax)))	// read as many as possible until we hit max, or fail
		{
			character=translationTable[chunk->data[offset]];
			if((matching=(GET_CHAR_BIT(node->pt.bits,character)!=0)))	// check for match
			{
				if((++offset)>=chunk->totalBytes)			// move forward in the input iff there was a match
				{
					lastChunk=chunk;
					chunk=chunk->nextHeader;
					offset=0;
				}
				i++;
			}
		}
		if(i>=node->matchMin)								// see if we got the minimum number of matches
		{
			if((nextExpression=REGetNextNode(node)))		// see if there is another node after this one (if so, recurse into it)
			{
				RECallNextNode(chunk,offset,numToSearch-i,nextExpression,&matching,&nodeMatched,recurseGlobals);
				while(!fail&&!matching&&i>node->matchMin)	// see if possible to backup and try again
				{
					i--;
					if(chunk)								// back up 1 character
					{
						if(offset)
						{
							offset--;
						}
						else
						{
							chunk=chunk->previousHeader;
							offset=chunk->totalBytes-1;
						}
					}
					else
					{
						chunk=lastChunk;
						offset=chunk->totalBytes-1;
					}
					RECallNextNode(chunk,offset,numToSearch-i,nextExpression,&matching,&nodeMatched,recurseGlobals);
				}
				if(!fail&&matching)							// match was located!!
				{
					(*numMatched)=i+nodeMatched;			// return number of characters that matched
				}
			}
			else											// no other nodes, this was at the end, so it is done
			{
				(*numMatched)=i;							// return number of characters that matched
				recurseGlobals->endChunk=chunk;
				recurseGlobals->endOffset=offset;
				matching=true;								// if no more nodes, then this list is complete, and it matched
			}
		}
		else
		{
			matching=false;									// did not find minimum number of matches, so this fails here and now
		}
		(*foundMatch)=matching;
		matchRecursionDepth--;
	}
	else
	{
		SetError(errorDescriptions[STACK_OVERFLOW]);
		fail=true;
	}
	return(!fail);
}

static bool REMatchPreviousNode(CHUNK_HEADER *chunk,UINT32 offset,UINT32 numToSearch,EXPRESSION_NODE *node,bool *foundMatch,UINT32 *numMatched,RECURSE_GLOBALS *recurseGlobals)
// attempt to match node with the data at chunk/offset
// if the pattern could be matched, return true in foundMatch with numMatched as the number of bytes matched
// if no match was located, return false in foundMatch
// NOTE: do not search through more than numToSearch bytes
// endChunk/endOffset will point to the end of the match if one was located
// if there is some sort of hard failure, this routine will SetError, and return false
{
	UINT32
		i,
		totalMatched,
		nodeMatched;
	CHUNK_HEADER
		*lastChunk;
	UINT32
		lastOffset;
	bool
		fail,
		matching;
	EXPRESSION_NODE
		*nextExpression;

	lastOffset=0;
	lastChunk=NULL;
	if(++matchRecursionDepth<MAX_RECURSION_DEPTH)
	{
		fail=false;
		matching=true;										// start by assuming a match
		i=totalMatched=0;
		while(matching&&((!node->matchMax)||(i<node->matchMax)))	// match as many as possible until we hit max, or fail
		{
			if(node->pt.expRegister->chunk&&numToSearch-totalMatched>=node->pt.expRegister->numBytes)	// see if it is defined, and enough bytes to test, if not, matching fails
			{
				if((matching=LiteralMatch(chunk,offset,node->pt.expRegister->chunk,node->pt.expRegister->offset,node->pt.expRegister->numBytes,&(recurseGlobals->endChunk),&(recurseGlobals->endOffset))))
				{
					i++;
					totalMatched+=node->pt.expRegister->numBytes;
					lastChunk=chunk;						// keep these in case we step off the end
					lastOffset=offset;
					chunk=recurseGlobals->endChunk;			// move to new position
					offset=recurseGlobals->endOffset;
				}
			}
			else
			{
				matching=false;
			}
		}
		if(i>=node->matchMin)								// see if we got the minimum number of matches
		{
			if((nextExpression=REGetNextNode(node)))		// see if there is another node after this one (if so, recurse into it)
			{
				RECallNextNode(chunk,offset,numToSearch-totalMatched,nextExpression,&matching,&nodeMatched,recurseGlobals);
				while(!fail&&!matching&&i>node->matchMin)	// see if possible to backup and try again
				{
					i--;
					nodeMatched=node->pt.expRegister->numBytes;
					totalMatched-=nodeMatched;
					if(chunk)								// back up nodeMatched characters
					{
						while(nodeMatched>offset)			// move back through the chunks until we use up all of the bytes matched
						{
							chunk=chunk->previousHeader;	// step back to the previous chunk
							nodeMatched-=offset;
							offset=chunk->totalBytes;
						}
						offset-=nodeMatched;
					}
					else
					{
						chunk=lastChunk;
						offset=lastOffset;
					}
					RECallNextNode(chunk,offset,numToSearch-totalMatched,nextExpression,&matching,&nodeMatched,recurseGlobals);
				}
				if(!fail&&matching)							// match was located!!
				{
					(*numMatched)=totalMatched+nodeMatched;	// return number of characters that matched
				}
			}
			else											// no other nodes, this was at the end, so it is done
			{
				(*numMatched)=totalMatched;					// return number of characters that matched
				recurseGlobals->endChunk=chunk;
				recurseGlobals->endOffset=offset;
				matching=true;								// if no more nodes, then this list is complete, and it matched
			}
		}
		else
		{
			matching=false;									// did not find minimum number of matches, so this fails here and now
		}
		(*foundMatch)=matching;
		matchRecursionDepth--;
	}
	else
	{
		SetError(errorDescriptions[STACK_OVERFLOW]);
		fail=true;
	}
	return(!fail);
}

static bool REMatchPreviousNodeTT(CHUNK_HEADER *chunk,UINT32 offset,UINT32 numToSearch,EXPRESSION_NODE *node,bool *foundMatch,UINT32 *numMatched,RECURSE_GLOBALS *recurseGlobals)
// exactly like REMatchPreviousNode except that this one uses a translation table
{
	UINT32
		i,
		totalMatched,
		nodeMatched;
	CHUNK_HEADER
		*lastChunk;
	UINT32
		lastOffset;
	bool
		fail,
		matching;
	EXPRESSION_NODE
		*nextExpression;

	lastOffset=0;
	lastChunk=NULL;
	if(++matchRecursionDepth<MAX_RECURSION_DEPTH)
	{
		fail=false;
		matching=true;										// start by assuming a match
		i=totalMatched=0;
		while(matching&&((!node->matchMax)||(i<node->matchMax)))	// match as many as possible until we hit max, or fail
		{
			if(node->pt.expRegister->chunk&&numToSearch-totalMatched>=node->pt.expRegister->numBytes)	// see if it is defined, and enough bytes to test, if not, matching fails
			{
				if((matching=LiteralMatchTT(chunk,offset,node->pt.expRegister->chunk,node->pt.expRegister->offset,node->pt.expRegister->numBytes,translationTable,&(recurseGlobals->endChunk),&(recurseGlobals->endOffset))))
				{
					i++;
					totalMatched+=node->pt.expRegister->numBytes;
					lastChunk=chunk;						// keep these in case we step off the end
					lastOffset=offset;
					chunk=recurseGlobals->endChunk;			// move to new position
					offset=recurseGlobals->endOffset;
				}
			}
			else
			{
				matching=false;
			}
		}
		if(i>=node->matchMin)								// see if we got the minimum number of matches
		{
			if((nextExpression=REGetNextNode(node)))		// see if there is another node after this one (if so, recurse into it)
			{
				RECallNextNode(chunk,offset,numToSearch-totalMatched,nextExpression,&matching,&nodeMatched,recurseGlobals);
				while(!fail&&!matching&&i>node->matchMin)	// see if possible to backup and try again
				{
					i--;
					nodeMatched=node->pt.expRegister->numBytes;
					totalMatched-=nodeMatched;
					if(chunk)								// back up nodeMatched characters
					{
						while(nodeMatched>offset)			// move back through the chunks until we use up all of the bytes matched
						{
							chunk=chunk->previousHeader;	// step back to the previous chunk
							nodeMatched-=offset;
							offset=chunk->totalBytes;
						}
						offset-=nodeMatched;
					}
					else
					{
						chunk=lastChunk;
						offset=lastOffset;
					}
					RECallNextNode(chunk,offset,numToSearch-totalMatched,nextExpression,&matching,&nodeMatched,recurseGlobals);
				}
				if(!fail&&matching)							// match was located!!
				{
					(*numMatched)=totalMatched+nodeMatched;	// return number of characters that matched
				}
			}
			else											// no other nodes, this was at the end, so it is done
			{
				(*numMatched)=totalMatched;					// return number of characters that matched
				recurseGlobals->endChunk=chunk;
				recurseGlobals->endOffset=offset;
				matching=true;								// if no more nodes, then this list is complete, and it matched
			}
		}
		else
		{
			matching=false;									// did not find minimum number of matches, so this fails here and now
		}
		(*foundMatch)=matching;
		matchRecursionDepth--;
	}
	else
	{
		SetError(errorDescriptions[STACK_OVERFLOW]);
		fail=true;
	}
	return(!fail);
}

static bool REMatchSubStartNode(CHUNK_HEADER *chunk,UINT32 offset,UINT32 numToSearch,EXPRESSION_NODE *node,bool *foundMatch,UINT32 *numMatched,RECURSE_GLOBALS *recurseGlobals)
// attempt to match node with the data at chunk/offset
// if the pattern could be matched, return true in foundMatch with numMatched as the number of bytes matched
// if no match was located, return false in foundMatch
// NOTE: do not search through more than numToSearch bytes
// endChunk/endOffset will point to the end of the match if one was located
// if there is some sort of hard failure, this routine will SetError, and return false
{
	UINT32
		oldCount;
	bool
		fail;

	if(++matchRecursionDepth<MAX_RECURSION_DEPTH)
	{
		fail=false;
		node->nextExpression->pt.se.justEntered=true;		// set up the end of the subexpression so that it knows we just hit the start
		oldCount=node->nextExpression->pt.se.currentCount;	// keep track of how many times this expression matched last time
		RECallNextNode(chunk,offset,numToSearch,node->nextExpression,foundMatch,numMatched,recurseGlobals);
		node->nextExpression->pt.se.currentCount=oldCount;	// restore match count
		matchRecursionDepth--;
	}
	else
	{
		SetError(errorDescriptions[STACK_OVERFLOW]);
		fail=true;
	}
	return(!fail);
}

static bool REMatchSubEndNode(CHUNK_HEADER *chunk,UINT32 offset,UINT32 numToSearch,EXPRESSION_NODE *node,bool *foundMatch,UINT32 *numMatched,RECURSE_GLOBALS *recurseGlobals)
// attempt to match node with the data at chunk/offset
// if the pattern could be matched, return true in foundMatch with numMatched as the number of bytes matched
// if no match was located, return false in foundMatch
// NOTE: do not search through more than numToSearch bytes
// endChunk/endOffset will point to the end of the match if one was located
// if there is some sort of hard failure, this routine will SetError, and return false
{
	bool
		fail,
		matching;
	EXPRESSION_NODE
		*nextExpression;
	REGISTER
		oldRegister;
	CHUNK_HEADER
		*oldLastChunk;									// used to keep the contents of the old register in case we back up
	UINT32
		oldLastOffset,
		oldLastNumToSearch;

	oldLastOffset=0;
	oldLastNumToSearch=0;
	oldLastChunk=NULL;

	oldRegister.chunk=NULL;								// keep gcc quiet
	oldRegister.offset=0;
	oldRegister.matchIndex=0;
	oldRegister.numBytes=0;

	if(++matchRecursionDepth<MAX_RECURSION_DEPTH)
	{
		fail=matching=false;
		if(node->pt.se.justEntered)
		{
			node->pt.se.currentCount=0;					// start off the counter
			node->pt.se.lastChunk=chunk;				// remember where we were, so the register can be set if needed
			node->pt.se.lastOffset=offset;
			node->pt.se.lastNumToSearch=numToSearch;
			node->pt.se.justEntered=false;				// next time this is seen, we will know it came from a successful completion!
			if(REMatchAlternatives(chunk,offset,numToSearch,node->pt.se.subExpressionList,&matching,numMatched,recurseGlobals))
			{
				if(!matching)							// if could not match, then see if we can move forward from here
				{
					if(!node->matchMin)					// if no matches needed, then we have success
					{
						if((nextExpression=REGetNextNode(node)))		// see if there is another node after this one (if so, recurse into it)
						{
							RECallNextNode(chunk,offset,numToSearch,nextExpression,&matching,numMatched,recurseGlobals);
						}
						else
						{
							matching=true;
							*numMatched=0;				// this node is done here and now
							recurseGlobals->endChunk=chunk;
							recurseGlobals->endOffset=offset;
						}
					}
				}
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			node->pt.se.currentCount++;					// made another successful match
			if(node->pt.se.expRegister)					// see if this is one that is actually remembered
			{
				oldRegister.chunk=node->pt.se.expRegister->chunk;		// copy old register contents (in case we need to back out)
				oldRegister.offset=node->pt.se.expRegister->offset;
				oldRegister.matchIndex=node->pt.se.expRegister->matchIndex;
				oldRegister.numBytes=node->pt.se.expRegister->numBytes;

				oldLastChunk=node->pt.se.lastChunk;
				oldLastOffset=node->pt.se.lastOffset;
				oldLastNumToSearch=node->pt.se.lastNumToSearch;

				node->pt.se.expRegister->chunk=node->pt.se.lastChunk;
				node->pt.se.expRegister->offset=node->pt.se.lastOffset;
				node->pt.se.expRegister->matchIndex=recurseGlobals->initialNumToSearch-node->pt.se.lastNumToSearch;
				node->pt.se.expRegister->numBytes=node->pt.se.lastNumToSearch-numToSearch;

				node->pt.se.lastChunk=chunk;
				node->pt.se.lastOffset=offset;
				node->pt.se.lastNumToSearch=numToSearch;
			}
			if((!node->matchMax)||node->pt.se.currentCount<node->matchMax)
			{
				if(REMatchAlternatives(chunk,offset,numToSearch,node->pt.se.subExpressionList,&matching,numMatched,recurseGlobals))
				{
					if(!matching)							// if could not match, then see if we can move forward from here
					{
						if(node->pt.se.currentCount>=node->matchMin)
						{
							if((nextExpression=REGetNextNode(node)))		// see if there is another node after this one (if so, recurse into it)
							{
								RECallNextNode(chunk,offset,numToSearch,nextExpression,&matching,numMatched,recurseGlobals);
							}
							else
							{
								matching=true;
								*numMatched=0;				// this node is done here and now
								recurseGlobals->endChunk=chunk;
								recurseGlobals->endOffset=offset;
							}
						}
					}
				}
				else
				{
					fail=true;
				}
			}
			else
			{
				if((nextExpression=REGetNextNode(node)))	// see if there is another node after this one (if so, recurse into it)
				{
					RECallNextNode(chunk,offset,numToSearch,nextExpression,&matching,numMatched,recurseGlobals);
				}
				else
				{
					matching=true;
					*numMatched=0;
					recurseGlobals->endChunk=chunk;
					recurseGlobals->endOffset=offset;
				}
			}
			if(!matching)
			{
				node->pt.se.currentCount--;				// this match failed, so back out

				if(node->pt.se.expRegister)				// see if this is one that is actually remembered
				{
					node->pt.se.lastChunk=oldLastChunk;
					node->pt.se.lastOffset=oldLastOffset;
					node->pt.se.lastNumToSearch=oldLastNumToSearch;

					node->pt.se.expRegister->chunk=oldRegister.chunk;
					node->pt.se.expRegister->offset=oldRegister.offset;
					node->pt.se.expRegister->matchIndex=oldRegister.matchIndex;
					node->pt.se.expRegister->numBytes=oldRegister.numBytes;
				}
			}
		}
		(*foundMatch)=matching;
		matchRecursionDepth--;
	}
	else
	{
		SetError(errorDescriptions[STACK_OVERFLOW]);
		fail=true;
	}
	return(!fail);
}


static bool REMatchLeftEdgeNode(CHUNK_HEADER *chunk,UINT32 offset,UINT32 numToSearch,EXPRESSION_NODE *node,bool *foundMatch,UINT32 *numMatched,RECURSE_GLOBALS *recurseGlobals)
// attempt to match node with the data at chunk/offset
// if the pattern could be matched, return true in foundMatch with numMatched as the number of bytes matched
// if no match was located, return false in foundMatch
// NOTE: do not search through more than numToSearch bytes
// endChunk/endOffset will point to the end of the match if one was located
// if there is some sort of hard failure, this routine will SetError, and return false
{
	CHUNK_HEADER
		*lastChunk;												// need to remember this in case we fall off the edge
	UINT32
		lastOffset;
	bool
		fail,
		matching;
	UINT8
		character;
	EXPRESSION_NODE
		*nextExpression;

	if(++matchRecursionDepth<MAX_RECURSION_DEPTH)
	{
		fail=false;
		matching=true;										// start by assuming a match
		(*numMatched)=0;									// return number of characters that matched
		recurseGlobals->endChunk=lastChunk=chunk;			// return position of match
		recurseGlobals->endOffset=lastOffset=offset;
		if(lastOffset)										// back up from current position, and look for start character
		{
			lastOffset--;
		}
		else
		{
			if(lastChunk&&(lastChunk=lastChunk->previousHeader))
			{
				lastOffset=lastChunk->totalBytes-1;
			}
		}
		if(lastChunk)										// if no chunk back, then we were at the start (or end) of text
		{
			character=lastChunk->data[lastOffset];
			if(!GET_CHAR_BIT(node->pt.bits,character))		// see if not left edge character
			{
				if((!recurseGlobals->leftEdge)||(numToSearch<recurseGlobals->initialNumToSearch))	// make sure we have not been told that this is an edge
				{
					matching=false;
				}
			}
		}
		else
		{
			if(!recurseGlobals->leftEdge)					// if this is not an edge, then do not match it
			{
				matching=false;
			}
		}
		if(matching&&(nextExpression=REGetNextNode(node)))	// see if there is another node after this one (if so, recurse into it)
		{
			RECallNextNode(chunk,offset,numToSearch,nextExpression,&matching,numMatched,recurseGlobals);
		}
		(*foundMatch)=matching;
		matchRecursionDepth--;
	}
	else
	{
		SetError(errorDescriptions[STACK_OVERFLOW]);
		fail=true;
	}
	return(!fail);
}

static bool REMatchRightEdgeNode(CHUNK_HEADER *chunk,UINT32 offset,UINT32 numToSearch,EXPRESSION_NODE *node,bool *foundMatch,UINT32 *numMatched,RECURSE_GLOBALS *recurseGlobals)
// attempt to match node with the data at chunk/offset
// if the pattern could be matched, return true in foundMatch with numMatched as the number of bytes matched
// if no match was located, return false in foundMatch
// NOTE: do not search through more than numToSearch bytes
// endChunk/endOffset will point to the end of the match if one was located
// if there is some sort of hard failure, this routine will SetError, and return false
{
	bool
		fail,
		matching;
	UINT8
		character;
	EXPRESSION_NODE
		*nextExpression;

	if(++matchRecursionDepth<MAX_RECURSION_DEPTH)
	{
		fail=false;
		matching=true;										// start by assuming a match
		(*numMatched)=0;									// return number of characters that matched
		recurseGlobals->endChunk=chunk;						// return position of match
		recurseGlobals->endOffset=offset;
		if(chunk)											// if no chunk, then we are at the end of text
		{
			character=chunk->data[offset];
			if(!GET_CHAR_BIT(node->pt.bits,character))		// see if not right edge character
			{
				if((!recurseGlobals->rightEdge)||numToSearch)	// make sure we have not been told that this is an edge
				{
					matching=false;
				}
			}
		}
		else
		{
			if(!recurseGlobals->rightEdge)					// if this is not an edge, then do not match it
			{
				matching=false;
			}
		}
		if(matching&&(nextExpression=REGetNextNode(node)))	// see if there is another node after this one (if so, recurse into it)
		{
			RECallNextNode(chunk,offset,numToSearch,nextExpression,&matching,numMatched,recurseGlobals);
		}
		(*foundMatch)=matching;
		matchRecursionDepth--;
	}
	else
	{
		SetError(errorDescriptions[STACK_OVERFLOW]);
		fail=true;
	}
	return(!fail);
}

static bool REMatchAlternatives(CHUNK_HEADER *chunk,UINT32 offset,UINT32 numToSearch,EXPRESSION_LIST_HEADER *list,bool *foundMatch,UINT32 *numMatched,RECURSE_GLOBALS *recurseGlobals)
// attempt to match alternatives at the given level
// if the pattern could be matched, return true in foundMatch with numMatched as the number of bytes matched
// if no match was located, return false in foundMatch
// NOTE: do not search through more than numToSearch bytes
// endChunk/endOffset will point to the end of the match if one was located
// this tries each alternative in order until a match is located, or no match could be found
// if there is some sort of hard failure, this routine will SetError, and return false
{
	EXPRESSION_LIST_HEADER
		*currentList;
	bool
		fail;

	if(++matchRecursionDepth<MAX_RECURSION_DEPTH)
	{
		if(!recurseGlobals->abortFunction||!recurseGlobals->abortFunction())
		{
			currentList=list;
			(*foundMatch)=false;
			fail=false;
			while(currentList&&!(*foundMatch)&&!fail)
			{
				if(currentList->firstExpression)	// if non-empty list of expressions, go handle it
				{
					RECallNextNode(chunk,offset,numToSearch,currentList->firstExpression,foundMatch,numMatched,recurseGlobals);
				}
				else	// list is empty, so it matches, and we continue with the parent's next node
				{
					if(currentList->parentExpression)				// go back to the parent
					{
						RECallNextNode(chunk,offset,numToSearch,currentList->parentExpression,foundMatch,numMatched,recurseGlobals);
					}
					else											// no other nodes, this was at the end, so it is done
					{
						*numMatched=0;								// return number of characters that matched
						recurseGlobals->endChunk=chunk;
						recurseGlobals->endOffset=offset;
						*foundMatch=true;							// if list is empty, then it matches
					}
				}
				currentList=currentList->nextAlternate;
			}
		}
		else
		{
			SetError(errorDescriptions[SEARCH_ABORTED]);
			fail=true;												// abort happened
		}
		matchRecursionDepth--;
	}
	else
	{
		SetError(errorDescriptions[STACK_OVERFLOW]);
		fail=true;
	}
	return(!fail);
}

static bool REMatch(CHUNK_HEADER *chunk,UINT32 offset,COMPILED_EXPRESSION *compiledExpression,bool leftEdge,bool rightEdge,UINT32 numToSearch,bool *foundMatch,UINT32 *numMatched,CHUNK_HEADER **endChunk,UINT32 *endOffset,ABORT_TEST_FUNCTION *abortFunction)
// this is the regular expression match routine
// it will return true in foundMatch if a match is found
// if there is a hard failure, SetError will be called, and false returned
// NOTE: this can be called when chunk is NULL (for instance in the case
// when there is no text to search). It is even possible for this to find
// a match when chunk is NULL (as in the case when the expression is '$')
// Of course, if chunk is NULL, numToSearch must be 0
{
	RECURSE_GLOBALS
		recurseGlobals;
	UINT32
		i;

	*foundMatch=false;
	for(i=0;i<compiledExpression->numRegisters;i++)
	{
		compiledExpression->expRegisters[i].chunk=NULL;			// no registers have been matched yet
	}
	matchRecursionDepth=0;										// clear recursion depth

	// to save time/stack while calling recursively, this structure is used to hold variables that are global to the recursive search routines, and only a pointer to this is passed
	recurseGlobals.leftEdge=leftEdge;
	recurseGlobals.rightEdge=rightEdge;
	recurseGlobals.abortFunction=abortFunction;
	recurseGlobals.initialNumToSearch=numToSearch;
	recurseGlobals.endChunk=NULL;
	recurseGlobals.endOffset=0;

	if(REMatchAlternatives(chunk,offset,numToSearch,compiledExpression->firstList,foundMatch,numMatched,&recurseGlobals))
	{
		*endChunk=recurseGlobals.endChunk;
		*endOffset=recurseGlobals.endOffset;

		return(true);
	}
	return(false);
}

bool GetRERegisterMatch(COMPILED_EXPRESSION *compiledExpression,UINT32 expRegisterIndex,UINT32 *matchOffset,UINT32 *matchLength)
// After a call to a regular expression search which succeeds, this
// may be called to locate information about the tagged matches
// If the passed tag did not match, set matchOffset to 0, matchLength to 0, and return false
{
	if(expRegisterIndex<compiledExpression->numRegisters)
	{
		if(compiledExpression->expRegisters[expRegisterIndex].chunk)
		{
			*matchOffset=compiledExpression->expRegisters[expRegisterIndex].matchIndex;
			*matchLength=compiledExpression->expRegisters[expRegisterIndex].numBytes;
			return(true);
		}
	}
	*matchOffset=0;
	*matchLength=0;
	return(false);
}

bool GetRERegisterMatchChunkAndOffset(COMPILED_EXPRESSION *compiledExpression,UINT32 expRegisterIndex,CHUNK_HEADER **chunk,UINT32 *offset,UINT32 *matchLength)
// After a call to a regular expression search which succeeds, this
// may be called to locate information about the tagged matches
// If the passed tag did not match, set chunk to NULL, offset to 0, matchLength to 0, and return false
{
	if(expRegisterIndex<compiledExpression->numRegisters)
	{
		if(compiledExpression->expRegisters[expRegisterIndex].chunk)
		{
			*chunk=compiledExpression->expRegisters[expRegisterIndex].chunk;
			*offset=compiledExpression->expRegisters[expRegisterIndex].offset;
			*matchLength=compiledExpression->expRegisters[expRegisterIndex].numBytes;
			return(true);
		}
	}
	*chunk=NULL;
	*offset=0;
	*matchLength=0;
	return(false);
}

bool SearchForwardRE(CHUNK_HEADER *chunk,UINT32 offset,COMPILED_EXPRESSION *compiledExpression,bool leftEdge,bool rightEdge,UINT32 numToSearchStart,UINT32 numToSearch,bool ignoreCase,bool *foundMatch,UINT32 *matchOffset,UINT32 *numMatched,CHUNK_HEADER **startChunk,UINT32 *startOffset,CHUNK_HEADER **endChunk,UINT32 *endOffset,ABORT_TEST_FUNCTION *abortFunction)
// Search for compiledExpression starting at chunk, offset
// Return true in foundMatch if it is found
// leftEdge tells if the starting point is considered a left edge for matching purposes
// rightEdge tells if the ending point is considered a right edge for matching purposes
// numToSearchStart gives the maximum number of bytes to look through trying to find the start of a match
// numToSearch gives the maximum number of bytes to look through
// if ignoreCase is true, search in a case-insensitive way
// matchOffset returns the offset from the start of the data where a match was located
// numMatched returns the number of characters that matched
// startChunk, startOffset give the address in the text which matched
// endChunk, endOffset give the address of the end of the match+1
// If abortFunction is non NULL, it will be called during the search, and an error
// will be returned if it ever returns true.
// if there is a hard error, SetError will be called, and false returned
{
	bool
		fail;
	UINT32
		localOffset;
	bool
		localFound;

	localFound=false;
	localOffset=0;
	fail=false;
	if(ignoreCase)
	{
		matchRoutines=translatedRoutines;
		translationTable=upperCaseTable;							// point at the table to use

		while(!fail&&!localFound&&localOffset<=numToSearchStart)	// '<=' because regular expressions can match 0 bytes
		{
			if(REMatch(chunk,offset,compiledExpression,leftEdge,rightEdge,numToSearch-localOffset,&localFound,numMatched,endChunk,endOffset,abortFunction))
			{
				leftEdge=false;
				if(!localFound)
				{
					localOffset++;
					if(localOffset<=numToSearchStart)
					{
						if((++offset)>=chunk->totalBytes)			// move forward in the input, and try a match there
						{
							chunk=chunk->nextHeader;
							offset=0;
						}
					}
				}
			}
			else
			{
				fail=true;
			}
		}
	}
	else		// this is the case which needs to be the fastest, since it is used most often (especially for syntax highlighting)
	{
		matchRoutines=exactRoutines;

		if(compiledExpression->minimumMatchLength)	// this case can be sped up
		{
			while(!fail&&!localFound&&localOffset<numToSearchStart)
			{
				if(compiledExpression->startCharsMatched[chunk->data[offset]])	// do some quick tests to bail out now if possible
				{
					if(REMatch(chunk,offset,compiledExpression,leftEdge,rightEdge,numToSearch-localOffset,&localFound,numMatched,endChunk,endOffset,abortFunction))
					{
						if(!localFound)
						{
							localOffset++;
							if((++offset)>=chunk->totalBytes)	// move forward in the input, and try a match there
							{
								chunk=chunk->nextHeader;
								offset=0;
							}
						}
					}
					else
					{
						fail=true;
					}
				}
				else
				{
					localOffset++;
					if((++offset)>=chunk->totalBytes)			// move forward in the input, and try a match there
					{
						chunk=chunk->nextHeader;
						offset=0;
					}
				}
				leftEdge=false;
			}
		}
		else
		{
			while(!fail&&!localFound&&localOffset<=numToSearchStart)	// '<=' because regular expressions can match 0 bytes
			{
				if(REMatch(chunk,offset,compiledExpression,leftEdge,rightEdge,numToSearch-localOffset,&localFound,numMatched,endChunk,endOffset,abortFunction))
				{
					leftEdge=false;
					if(!localFound)
					{
						localOffset++;
						if(localOffset<=numToSearchStart)
						{
							if((++offset)>=chunk->totalBytes)	// move forward in the input, and try a match there
							{
								chunk=chunk->nextHeader;
								offset=0;
							}
						}
					}
				}
				else
				{
					fail=true;
				}
			}
		}
	}
	*startChunk=chunk;
	*startOffset=offset;
	*foundMatch=localFound;
	*matchOffset=localOffset;

	return(!fail);
}

bool SearchBackwardRE(CHUNK_HEADER *chunk,UINT32 offset,COMPILED_EXPRESSION *compiledExpression,bool leftEdge,bool rightEdge,UINT32 numToSearchStart,UINT32 numToSearch,bool ignoreCase,bool *foundMatch,UINT32 *matchOffset,UINT32 *numMatched,CHUNK_HEADER **startChunk,UINT32 *startOffset,CHUNK_HEADER **endChunk,UINT32 *endOffset,ABORT_TEST_FUNCTION *abortFunction)
// search backward for the expression return true in foundMatch if it is found
// This repeatedly calls REMatch
// if there is a hard error, SetError will be called, and false returned
{
	bool
		fail;
	UINT32
		localOffset;
	bool
		localFound;

	localFound=false;
	localOffset=numToSearch-numToSearchStart;
	fail=false;
	if(ignoreCase)
	{
		matchRoutines=translatedRoutines;
		translationTable=upperCaseTable;							// point at the table to use

		while(!fail&&!localFound&&localOffset<=numToSearch)			// '<=' because regular expressions can match 0 bytes
		{
			if(REMatch(chunk,offset,compiledExpression,(localOffset==numToSearch)?leftEdge:false,rightEdge,localOffset,&localFound,numMatched,endChunk,endOffset,abortFunction))
			{
				if(!localFound)
				{
					localOffset++;
					if(localOffset<=numToSearch)
					{
						if(!offset)									// move backward in the input
						{
							chunk=chunk->previousHeader;
							offset=chunk->totalBytes;
						}
						offset--;
					}
				}
			}
			else
			{
				fail=true;
			}
		}
	}
	else
	{
		matchRoutines=exactRoutines;

		while(!fail&&!localFound&&localOffset<=numToSearch)			// '<=' because regular expressions can match 0 bytes
		{
			if(!compiledExpression->minimumMatchLength||(localOffset&&compiledExpression->startCharsMatched[chunk->data[offset]]))	// do some quick tests to bail out now if possible
			{
				if(REMatch(chunk,offset,compiledExpression,(localOffset==numToSearch)?leftEdge:false,rightEdge,localOffset,&localFound,numMatched,endChunk,endOffset,abortFunction))
				{
					if(!localFound)
					{
						localOffset++;
						if(localOffset<=numToSearch)
						{
							if(!offset)								// move backward in the input
							{
								chunk=chunk->previousHeader;
								offset=chunk->totalBytes;
							}
							offset--;
						}
					}
				}
				else
				{
					fail=true;
				}
			}
			else
			{
				localOffset++;
				if(localOffset<=numToSearch)
				{
					if(!offset)										// move backward in the input
					{
						chunk=chunk->previousHeader;
						offset=chunk->totalBytes;
					}
					offset--;
				}
			}
		}
	}
	*startChunk=chunk;
	*startOffset=offset;
	*foundMatch=localFound;
	*matchOffset=localOffset;

	return(!fail);
}

void REHuntBackToFarthestPossibleStart(CHUNK_HEADER *chunk,UINT32 offset,COMPILED_EXPRESSION *compiledExpression,UINT32 *numBack,CHUNK_HEADER **endChunk,UINT32 *endOffset)
// Given a chunk and offset into some text, this will move backwards through the text to a point which
// is the farthest possible starting point for a regular expression which matches at the passed point
// numBack is the number of characters backwards that we moved, and endChunk/endOffset point to
// the starting point of a search.
// NOTE: this is too slow and ugly
{
	bool
		done;
	UINT32
		currentBack;

	*numBack=0;
	*endChunk=chunk;
	*endOffset=offset;

	if(chunk)			// if no text, then we cannot move back any further
	{
		done=false;
		currentBack=0;

		while(!done&&(compiledExpression->infiniteMatchLength||(currentBack<compiledExpression->maximumMatchLength)))
		{
			if(compiledExpression->charsMatched[chunk->data[offset]])	// see if this is a matching character
			{
				*endChunk=chunk;
				*endOffset=offset;
				*numBack=currentBack;

				currentBack++;

				if(offset)
				{
					offset--;
				}
				else
				{
					if(chunk->previousHeader)
					{
						chunk=chunk->previousHeader;
						offset=chunk->totalBytes-1;
					}
					else
					{
						done=true;	// hit the start
					}
				}
			}
			else
			{
				done=true;		// non matching, so this is the spot
			}
		}
	}
}
