// Shell interface for e93 (currently using John Ousterhout's Tcl)
// Copyright (C) 2000 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

#define MAX_SCROLL_SPEED	2

static Tcl_Trace
	tclTrace;					// the TCL trace object we created

void InsertBufferTaskData(EDITOR_BUFFER *buffer,UINT8 *data,UINT32 numBytes)
// Insert numBytes of data into buffer, update any
// views of buffer that may need it
// if there is a problem, ignore it!
{
	UINT32
		topLine,
		numLines,
		numPixels;
	INT32
		leftPixel;
	EDITOR_VIEW
		*currentView;
	UINT32
		tempPosition,
		startPosition,
		endPosition;
	UINT32
		startLine;
	CHUNK_HEADER
		*startLineChunk;
	UINT32
		startLineOffset;

// run through the views, see which ones will need to be homed after the insert:
// The ones to home, are those that had the cursor in view at the time of the insert

	GetSelectionEndPositions(buffer->selectionUniverse,&startPosition,&endPosition);
	currentView=buffer->firstView;					// get first view of this buffer (if any)
	while(currentView)								// walk through all views, update each one as needed
	{
		GetEditorViewTextInfo(currentView,&topLine,&numLines,&leftPixel,&numPixels);
		PositionToLinePosition(buffer->textUniverse,startPosition,&startLine,&tempPosition,&startLineChunk,&startLineOffset);
		currentView->wantHome=(startLine>=topLine&&startLine<topLine+numLines);	// see if the cursor is on the view
		currentView=currentView->nextBufferView;	// locate next view of this buffer
	}

	EditorAuxInsert(buffer,data,numBytes);			// drop the text into buffer

	GetSelectionEndPositions(buffer->selectionUniverse,&startPosition,&endPosition);	// get selection position after the insertion
// now home all of those views that need it
	currentView=buffer->firstView;					// get first view of this buffer (if any)
	while(currentView)								// walk through all views, update each one as needed
	{
		if(currentView->wantHome)
		{
			EditorHomeViewToRange(currentView,startPosition,endPosition,false,HT_NONE,HT_LENIENT);
		}
		currentView=currentView->nextBufferView;	// locate next view of this buffer
	}
}

static bool Safe_Tcl_Eval(Tcl_Interp *interp, char *cmd, int *tclResult)
// Some Tcl commands can delete the text of the script which
// is running. If this happens, it is possible for things to get confused.
// For example: a menu command could try to delete itself, which would cause
// the script of the command to be deleted mid-execution.
// To solve this problem, the script is copied into a safe temporary place
// then executed, and finally deleted
// If there is a problem copying the script, this will return false, and not
// execute it. If this returns true, the result from the Tcl_Eval command will
// be placed in tclResult
{
	char
		*scriptCopy;

	if((scriptCopy=(char *)MNewPtr(strlen(cmd)+1)))
	{
		strcpy(scriptCopy,cmd);								// copy the command
		*tclResult=Tcl_Eval(tclInterpreter,scriptCopy);		// execute it
		MDisposePtr(scriptCopy);							// get rid of the copy
		return(true);
	}
	return(false);
}

bool HandleBoundKeyEvent(UINT32 keyCode,UINT32 modifierValue)
// See if there is a bound key that matches the given code, and modifiers
// if there is, perform the script given by the bound key, and return true
// if there is not, return false
{
	EDITOR_KEY_BINDING
		*entry;
	int
		tclResult;
	const char
		*stringResult;

	if((entry=LocateKeyBindingMatch(keyCode,modifierValue)))
	{
		ClearUserAbort();
		if(Safe_Tcl_Eval(tclInterpreter,GetKeyBindingText(entry),&tclResult))
		{
			stringResult=Tcl_GetStringResult(tclInterpreter);
			if((tclResult!=TCL_OK)&&stringResult[0])
			{
				ReportMessage("Bound key execution error:\n%.256s\n",stringResult);
			}
		}
		else
		{
			ReportMessage("Bound key execution error:\nOut of memory\n");
		}
		return(true);
	}
	return(false);
}

void HandleMenuEvent(EDITOR_MENU *menu)
// When a menu event arrives, this is called (by the menu manager of the GUI) to handle the menu
{
	int
		tclResult;
	const char
		*stringResult;

	ClearUserAbort();
	if(Safe_Tcl_Eval(tclInterpreter,GetEditorMenuDataText(menu),&tclResult))
	{
		stringResult=Tcl_GetStringResult(tclInterpreter);
		if((tclResult!=TCL_OK)&&stringResult[0])
		{
			ReportMessage("Menu execution error:\n%.256s\n",stringResult);
		}
	}
	else
	{
		ReportMessage("Menu execution error:\nOut of memory\n");
	}
}

void HandleNonStandardControlEvent(const char *string)
// When the GUI needs to implement some GUI specific control, it can call this to execute shell
// commands for that control
{
	int
		tclResult;
	const char
		*stringResult;

	ClearUserAbort();
	tclResult=Tcl_Eval(tclInterpreter,string);
	stringResult=Tcl_GetStringResult(tclInterpreter);
	if((tclResult!=TCL_OK)&&stringResult[0])
	{
		ReportMessage("Execution error:\n%.256s\n",stringResult);
	}
}

void HandleViewEvent(VIEW_EVENT *event)
// High level view events are handled here
{
	UINT32
		topLine,
		numLines,
		numPixels;
	INT32
		leftPixel;
	UINT8
		key;
	static UINT32
		scrollTimeout;
	static UINT32
		scrollSpeed;
	UINT16
		trackMode;							// tells editor how to place cursor when cursor position keys arrive
	EDITOR_KEY
		*keyEventData;
	VIEW_POS_EVENT_DATA
		*posEventData;
	VIEW_CLICK_EVENT_DATA
		*clickEventData;
	bool
		result;
	bool
		didSomething;

	keyEventData=NULL;						// keep compiler quiet
	result=true;							// success until further notice
	switch(event->eventType)
	{
		case VET_KEYDOWN:
		case VET_KEYREPEAT:
			keyEventData=(EDITOR_KEY *)event->eventData;
			if(keyEventData->isVirtual)
			{
				switch(keyEventData->keyCode)
				{
					case VVK_SCROLLUP:
						if(event->eventType==VET_KEYDOWN)
						{
							scrollSpeed=1;
							scrollTimeout=0;
						}
						EditorVerticalScroll(event->view,-((INT32)scrollSpeed));
						if(scrollTimeout++>3)
						{
							if(scrollSpeed<MAX_SCROLL_SPEED)
							{
								scrollSpeed++;
							}
							scrollTimeout=0;
						}
						break;
					case VVK_SCROLLDOWN:
						if(event->eventType==VET_KEYDOWN)
						{
							scrollSpeed=1;
							scrollTimeout=0;
						}
						EditorVerticalScroll(event->view,scrollSpeed);
						if(scrollTimeout++>3)
						{
							if(scrollSpeed<MAX_SCROLL_SPEED)
							{
								scrollSpeed++;
							}
							scrollTimeout=0;
						}
						break;
					case VVK_DOCUMENTPAGEUP:
						EditorVerticalScrollByPages(event->view,-1);
						break;
					case VVK_DOCUMENTPAGEDOWN:
						EditorVerticalScrollByPages(event->view,1);
						break;
					case VVK_SCROLLLEFT:
						EditorHorizontalScroll(event->view,-HORIZONTAL_SCROLL_THRESHOLD);
						break;
					case VVK_SCROLLRIGHT:
						EditorHorizontalScroll(event->view,HORIZONTAL_SCROLL_THRESHOLD);
						break;
					case VVK_DOCUMENTPAGELEFT:
						EditorHorizontalScrollByPages(event->view,-1);
						break;
					case VVK_DOCUMENTPAGERIGHT:
						EditorHorizontalScrollByPages(event->view,1);
						break;
					case VVK_RETURN:
						EditorHomeViewToSelectionEdge(event->view,false,HT_SEMISTRICT,HT_SEMISTRICT);
						if((result=EditorAutoIndent(event->view->parentBuffer)))
						{
							EditorHomeViewToSelectionEdge(event->view,false,HT_LENIENT,HT_LENIENT);
						}
						break;
					case VVK_TAB:
						key='\t';				// insert a tab
						EditorHomeViewToSelectionEdge(event->view,false,HT_SEMISTRICT,HT_SEMISTRICT);
						if((result=EditorInsert(event->view->parentBuffer,&key,1)))
						{
							EditorHomeViewToSelectionEdge(event->view,false,HT_LENIENT,HT_LENIENT);
						}
						break;
					case VVK_LEFTARROW:
						if(!(keyEventData->modifiers&EEM_CTL))
						{
							if(!(keyEventData->modifiers&EEM_MOD0))
							{
								if(!(keyEventData->modifiers&EEM_SHIFT))
								{
									if((result=EditorMoveCursor(event->view,RPM_mBACKWARD|RPM_CHAR)))
									{
										EditorHomeViewToSelectionEdge(event->view,false,HT_LENIENT,HT_LENIENT);
									}
								}
								else
								{
									if((result=EditorMoveSelection(event->view,RPM_mBACKWARD|RPM_CHAR)))
									{
										EditorHomeViewToSelectionEdge(event->view,!event->view->parentBuffer->currentIsStart,HT_SEMISTRICT,HT_SEMISTRICT);
									}
								}
							}
							else
							{
								if(!(keyEventData->modifiers&EEM_SHIFT))
								{
									if((result=EditorMoveCursor(event->view,RPM_mBACKWARD|RPM_LINEEDGE)))
									{
										EditorHomeViewToSelectionEdge(event->view,false,HT_LENIENT,HT_LENIENT);
									}
								}
								else
								{
									if((result=EditorMoveSelection(event->view,RPM_mBACKWARD|RPM_LINEEDGE)))
									{
										EditorHomeViewToSelectionEdge(event->view,!event->view->parentBuffer->currentIsStart,HT_SEMISTRICT,HT_SEMISTRICT);
									}
								}
							}
						}
						else
						{
							if(!(keyEventData->modifiers&EEM_SHIFT))
							{
								if((result=EditorMoveCursor(event->view,RPM_mBACKWARD|RPM_WORD)))
								{
									EditorHomeViewToSelectionEdge(event->view,false,HT_LENIENT,HT_LENIENT);
								}
							}
							else
							{
								if((result=EditorMoveSelection(event->view,RPM_mBACKWARD|RPM_WORD)))
								{
									EditorHomeViewToSelectionEdge(event->view,!event->view->parentBuffer->currentIsStart,HT_SEMISTRICT,HT_SEMISTRICT);
								}
							}
						}
						break;
					case VVK_RIGHTARROW:
						if(!(keyEventData->modifiers&EEM_CTL))
						{
							if(!(keyEventData->modifiers&EEM_MOD0))
							{
								if(!(keyEventData->modifiers&EEM_SHIFT))
								{
									if((result=EditorMoveCursor(event->view,RPM_CHAR)))
									{
										EditorHomeViewToSelectionEdge(event->view,true,HT_LENIENT,HT_LENIENT);
									}
								}
								else
								{
									if((result=EditorMoveSelection(event->view,RPM_CHAR)))
									{
										EditorHomeViewToSelectionEdge(event->view,!event->view->parentBuffer->currentIsStart,HT_SEMISTRICT,HT_SEMISTRICT);
									}
								}
							}
							else
							{
								if(!(keyEventData->modifiers&EEM_SHIFT))
								{
									if((result=EditorMoveCursor(event->view,RPM_LINEEDGE)))
									{
										EditorHomeViewToSelectionEdge(event->view,true,HT_LENIENT,HT_LENIENT);
									}
								}
								else
								{
									if((result=EditorMoveSelection(event->view,RPM_LINEEDGE)))
									{
										EditorHomeViewToSelectionEdge(event->view,!event->view->parentBuffer->currentIsStart,HT_SEMISTRICT,HT_SEMISTRICT);
									}
								}
							}
						}
						else
						{
							if(!(keyEventData->modifiers&EEM_SHIFT))
							{
								if((result=EditorMoveCursor(event->view,RPM_WORD)))
								{
									EditorHomeViewToSelectionEdge(event->view,true,HT_LENIENT,HT_LENIENT);
								}
							}
							else
							{
								if((result=EditorMoveSelection(event->view,RPM_WORD)))
								{
									EditorHomeViewToSelectionEdge(event->view,!event->view->parentBuffer->currentIsStart,HT_SEMISTRICT,HT_SEMISTRICT);
								}
							}
						}
						break;
					case VVK_UPARROW:
						if(!(keyEventData->modifiers&EEM_CTL))
						{
							if(!(keyEventData->modifiers&EEM_MOD0))
							{
								if(!(keyEventData->modifiers&EEM_SHIFT))
								{
									if((result=EditorMoveCursor(event->view,RPM_mBACKWARD|RPM_LINE)))
									{
										EditorHomeViewToSelectionEdge(event->view,false,HT_LENIENT,HT_LENIENT);
									}
								}
								else
								{
									if((result=EditorMoveSelection(event->view,RPM_mBACKWARD|RPM_LINE)))
									{
										EditorHomeViewToSelectionEdge(event->view,!event->view->parentBuffer->currentIsStart,HT_SEMISTRICT,HT_SEMISTRICT);
									}
								}
							}
							else
							{
								if(!(keyEventData->modifiers&EEM_SHIFT))
								{
									if((result=EditorMoveCursor(event->view,RPM_mBACKWARD|RPM_PAGE)))
									{
										EditorHomeViewToSelectionEdge(event->view,false,HT_LENIENT,HT_LENIENT);
									}
								}
								else
								{
									if((result=EditorMoveSelection(event->view,RPM_mBACKWARD|RPM_PAGE)))
									{
										EditorHomeViewToSelectionEdge(event->view,!event->view->parentBuffer->currentIsStart,HT_LENIENT,HT_LENIENT);
									}
								}
							}
						}
						else
						{
							EditorVerticalScroll(event->view,-1);
						}
						break;
					case VVK_DOWNARROW:
						if(!(keyEventData->modifiers&EEM_CTL))
						{
							if(!(keyEventData->modifiers&EEM_MOD0))
							{
								if(!(keyEventData->modifiers&EEM_SHIFT))
								{
									if((result=EditorMoveCursor(event->view,RPM_LINE)))
									{
										EditorHomeViewToSelectionEdge(event->view,true,HT_LENIENT,HT_LENIENT);
									}
								}
								else
								{
									if((result=EditorMoveSelection(event->view,RPM_LINE)))
									{
										EditorHomeViewToSelectionEdge(event->view,!event->view->parentBuffer->currentIsStart,HT_SEMISTRICT,HT_SEMISTRICT);
									}
								}
							}
							else
							{
								if(!(keyEventData->modifiers&EEM_SHIFT))
								{
									if((result=EditorMoveCursor(event->view,RPM_PAGE)))
									{
										EditorHomeViewToSelectionEdge(event->view,true,HT_LENIENT,HT_LENIENT);
									}
								}
								else
								{
									if((result=EditorMoveSelection(event->view,RPM_PAGE)))
									{
										EditorHomeViewToSelectionEdge(event->view,!event->view->parentBuffer->currentIsStart,HT_LENIENT,HT_LENIENT);
									}
								}
							}
						}
						else
						{
							EditorVerticalScroll(event->view,1);
						}
						break;
					case VVK_BACKSPACE:
						if(!(keyEventData->modifiers&EEM_CTL)&&!(keyEventData->modifiers&EEM_MOD1))
						{
							if(!(keyEventData->modifiers&EEM_MOD0))
							{
								if(!(keyEventData->modifiers&EEM_SHIFT))
								{
									EditorHomeViewToSelectionEdge(event->view,false,HT_SEMISTRICT,HT_SEMISTRICT);
									if((result=EditorDelete(event->view,RPM_mBACKWARD|RPM_CHAR)))
									{
										EditorHomeViewToSelectionEdge(event->view,false,HT_LENIENT,HT_LENIENT);
									}
								}
								else
								{
									EditorHomeViewToSelectionEdge(event->view,false,HT_SEMISTRICT,HT_SEMISTRICT);
									if((result=EditorDelete(event->view,RPM_CHAR)))
									{
										EditorHomeViewToSelectionEdge(event->view,false,HT_LENIENT,HT_LENIENT);
									}
								}
							}
							else
							{
								if(!(keyEventData->modifiers&EEM_SHIFT))
								{
									EditorHomeViewToSelectionEdge(event->view,false,HT_SEMISTRICT,HT_SEMISTRICT);
									if((result=EditorDelete(event->view,RPM_LINEEDGE)))
									{
										EditorHomeViewToSelectionEdge(event->view,false,HT_LENIENT,HT_LENIENT);
									}
								}
								else
								{
									EditorHomeViewToSelectionEdge(event->view,false,HT_SEMISTRICT,HT_SEMISTRICT);
									if((result=EditorDelete(event->view,RPM_mBACKWARD|RPM_LINEEDGE)))
									{
										EditorHomeViewToSelectionEdge(event->view,false,HT_LENIENT,HT_LENIENT);
									}
								}
							}
						}
						else
						{
							if(!(keyEventData->modifiers&EEM_MOD0))
							{
								if(!(keyEventData->modifiers&EEM_SHIFT))
								{
									EditorHomeViewToSelectionEdge(event->view,false,HT_SEMISTRICT,HT_SEMISTRICT);
									if((result=EditorDelete(event->view,RPM_mBACKWARD|RPM_WORD)))
									{
										EditorHomeViewToSelectionEdge(event->view,false,HT_LENIENT,HT_LENIENT);
									}
								}
								else
								{
									EditorHomeViewToSelectionEdge(event->view,false,HT_SEMISTRICT,HT_SEMISTRICT);
									if((result=EditorDelete(event->view,RPM_WORD)))
									{
										EditorHomeViewToSelectionEdge(event->view,false,HT_LENIENT,HT_LENIENT);
									}
								}
							}
							else
							{
								if(!(keyEventData->modifiers&EEM_SHIFT))
								{
									EditorHomeViewToSelectionEdge(event->view,false,HT_SEMISTRICT,HT_SEMISTRICT);
									if((result=EditorDelete(event->view,RPM_DOCEDGE)))
									{
										EditorHomeViewToSelectionEdge(event->view,false,HT_LENIENT,HT_LENIENT);
									}
								}
								else
								{
									EditorHomeViewToSelectionEdge(event->view,false,HT_SEMISTRICT,HT_SEMISTRICT);
									if((result=EditorDelete(event->view,RPM_mBACKWARD|RPM_DOCEDGE)))
									{
										EditorHomeViewToSelectionEdge(event->view,false,HT_LENIENT,HT_LENIENT);
									}
								}
							}
						}
						break;
					case VVK_DELETE:
						if(!(keyEventData->modifiers&EEM_CTL)&&!(keyEventData->modifiers&EEM_MOD1))
						{
							if(!(keyEventData->modifiers&EEM_MOD0))
							{
								if(!(keyEventData->modifiers&EEM_SHIFT))
								{
									EditorHomeViewToSelectionEdge(event->view,false,HT_SEMISTRICT,HT_SEMISTRICT);
									if((result=EditorDelete(event->view,RPM_CHAR)))
									{
										EditorHomeViewToSelectionEdge(event->view,false,HT_LENIENT,HT_LENIENT);
									}
								}
								else
								{
									EditorHomeViewToSelectionEdge(event->view,false,HT_SEMISTRICT,HT_SEMISTRICT);
									if((result=EditorDelete(event->view,RPM_mBACKWARD|RPM_CHAR)))
									{
										EditorHomeViewToSelectionEdge(event->view,false,HT_LENIENT,HT_LENIENT);
									}
								}
							}
							else
							{
								if(!(keyEventData->modifiers&EEM_SHIFT))
								{
									EditorHomeViewToSelectionEdge(event->view,false,HT_SEMISTRICT,HT_SEMISTRICT);
									if((result=EditorDelete(event->view,RPM_mBACKWARD|RPM_LINEEDGE)))
									{
										EditorHomeViewToSelectionEdge(event->view,false,HT_LENIENT,HT_LENIENT);
									}
								}
								else
								{
									EditorHomeViewToSelectionEdge(event->view,false,HT_SEMISTRICT,HT_SEMISTRICT);
									if((result=EditorDelete(event->view,RPM_LINEEDGE)))
									{
										EditorHomeViewToSelectionEdge(event->view,false,HT_LENIENT,HT_LENIENT);
									}
								}
							}
						}
						else
						{
							if(!(keyEventData->modifiers&EEM_MOD0))
							{
								if(!(keyEventData->modifiers&EEM_SHIFT))
								{
									EditorHomeViewToSelectionEdge(event->view,false,HT_SEMISTRICT,HT_SEMISTRICT);
									if((result=EditorDelete(event->view,RPM_WORD)))
									{
										EditorHomeViewToSelectionEdge(event->view,false,HT_LENIENT,HT_LENIENT);
									}
								}
								else
								{
									EditorHomeViewToSelectionEdge(event->view,false,HT_SEMISTRICT,HT_SEMISTRICT);
									if((result=EditorDelete(event->view,RPM_mBACKWARD|RPM_WORD)))
									{
										EditorHomeViewToSelectionEdge(event->view,false,HT_LENIENT,HT_LENIENT);
									}
								}
							}
							else
							{
								if(!(keyEventData->modifiers&EEM_SHIFT))
								{
									EditorHomeViewToSelectionEdge(event->view,false,HT_SEMISTRICT,HT_SEMISTRICT);
									if((result=EditorDelete(event->view,RPM_mBACKWARD|RPM_DOCEDGE)))
									{
										EditorHomeViewToSelectionEdge(event->view,false,HT_LENIENT,HT_LENIENT);
									}
								}
								else
								{
									EditorHomeViewToSelectionEdge(event->view,false,HT_SEMISTRICT,HT_SEMISTRICT);
									if((result=EditorDelete(event->view,RPM_DOCEDGE)))
									{
										EditorHomeViewToSelectionEdge(event->view,false,HT_LENIENT,HT_LENIENT);
									}
								}
							}
						}
						break;
					case VVK_HOME:
						if(!(keyEventData->modifiers&EEM_MOD0))
						{
							if(!(keyEventData->modifiers&EEM_SHIFT))
							{
								if((result=EditorMoveCursor(event->view,RPM_mBACKWARD|RPM_LINEEDGE)))
								{
									EditorHomeViewToSelectionEdge(event->view,false,HT_LENIENT,HT_LENIENT);
								}
							}
							else
							{
								if((result=EditorExpandNormalSelection(event->view,RPM_mBACKWARD|RPM_LINEEDGE)))
								{
									EditorHomeViewToSelectionEdge(event->view,false,HT_LENIENT,HT_LENIENT);
								}
							}
						}
						else
						{
							if(!(keyEventData->modifiers&EEM_SHIFT))
							{
								if((result=EditorMoveCursor(event->view,RPM_mBACKWARD|RPM_DOCEDGE)))
								{
									EditorHomeViewToSelectionEdge(event->view,false,HT_LENIENT,HT_LENIENT);
								}
							}
							else
							{
								if((result=EditorExpandNormalSelection(event->view,RPM_mBACKWARD|RPM_DOCEDGE)))
								{
									EditorHomeViewToSelectionEdge(event->view,false,HT_LENIENT,HT_LENIENT);
								}
							}
						}
						break;
					case VVK_END:
						if(!(keyEventData->modifiers&EEM_MOD0))
						{
							if(!(keyEventData->modifiers&EEM_SHIFT))
							{
								if((result=EditorMoveCursor(event->view,RPM_LINEEDGE)))
								{
									EditorHomeViewToSelectionEdge(event->view,true,HT_LENIENT,HT_LENIENT);
								}
							}
							else
							{
								if((result=EditorExpandNormalSelection(event->view,RPM_LINEEDGE)))
								{
									EditorHomeViewToSelectionEdge(event->view,true,HT_LENIENT,HT_LENIENT);
								}
							}
						}
						else
						{
							if(!(keyEventData->modifiers&EEM_SHIFT))
							{
								if((result=EditorMoveCursor(event->view,RPM_DOCEDGE)))
								{
									EditorHomeViewToSelectionEdge(event->view,true,HT_LENIENT,HT_LENIENT);
								}
							}
							else
							{
								if((result=EditorExpandNormalSelection(event->view,RPM_DOCEDGE)))
								{
									EditorHomeViewToSelectionEdge(event->view,true,HT_LENIENT,HT_LENIENT);
								}
							}
						}
						break;
					case VVK_PAGEUP:
						if(keyEventData->modifiers&EEM_CTL)
						{
							EditorVerticalScrollByPages(event->view,-1);
						}
						else
						{
							if(!(keyEventData->modifiers&EEM_SHIFT))
							{
								if((result=EditorMoveCursor(event->view,RPM_mBACKWARD|RPM_PAGE)))
								{
									EditorHomeViewToSelectionEdge(event->view,false,HT_LENIENT,HT_LENIENT);
								}
							}
							else
							{
								if((result=EditorMoveSelection(event->view,RPM_mBACKWARD|RPM_PAGE)))
								{
									EditorHomeViewToSelectionEdge(event->view,!event->view->parentBuffer->currentIsStart,HT_LENIENT,HT_LENIENT);
								}
							}
						}
						break;
					case VVK_PAGEDOWN:
						if(keyEventData->modifiers&EEM_CTL)
						{
							EditorVerticalScrollByPages(event->view,1);
						}
						else
						{
							if(!(keyEventData->modifiers&EEM_SHIFT))
							{
								if((result=EditorMoveCursor(event->view,RPM_PAGE)))
								{
									EditorHomeViewToSelectionEdge(event->view,true,HT_LENIENT,HT_LENIENT);
								}
							}
							else
							{
								if((result=EditorMoveSelection(event->view,RPM_PAGE)))
								{
									EditorHomeViewToSelectionEdge(event->view,!event->view->parentBuffer->currentIsStart,HT_LENIENT,HT_LENIENT);
								}
							}
						}
						break;
					case VVK_UNDO:
						if((result=EditorUndo(event->view->parentBuffer,&didSomething)))
						{
							if(didSomething)
							{
								EditorHomeViewToSelectionEdge(event->view,false,HT_SEMISTRICT,HT_SEMISTRICT);
							}
							else
							{
								EditorBeep();
							}
						}
						break;
					case VVK_REDO:
						if((result=EditorRedo(event->view->parentBuffer,&didSomething)))
						{
							if(didSomething)
							{
								EditorHomeViewToSelectionEdge(event->view,false,HT_SEMISTRICT,HT_SEMISTRICT);
							}
							else
							{
								EditorBeep();
							}
						}
						break;
					case VVK_UNDOTOGGLE:
						if((result=EditorToggleUndo(event->view->parentBuffer,&didSomething)))
						{
							if(didSomething)
							{
								EditorHomeViewToSelectionEdge(event->view,false,HT_SEMISTRICT,HT_SEMISTRICT);
							}
							else
							{
								EditorBeep();
							}
						}
						break;
					case VVK_CUT:
						if((result=EditorCut(event->view->parentBuffer,EditorGetCurrentClipboard())))
						{
							EditorHomeViewToSelectionEdge(event->view,false,HT_SEMISTRICT,HT_SEMISTRICT);
						}
						break;
					case VVK_COPY:
						result=EditorCopy(event->view->parentBuffer,EditorGetCurrentClipboard());
						break;
					case VVK_PASTE:
						EditorHomeViewToSelectionEdge(event->view,false,HT_SEMISTRICT,HT_SEMISTRICT);
						if((result=EditorColumnarPaste(event->view->parentBuffer,event->view,EditorGetCurrentClipboard())))
						{
							EditorHomeViewToSelectionEdge(event->view,false,HT_SEMISTRICT,HT_SEMISTRICT);
						}
						break;
				}
			}
			else
			{
				key=(UINT8)keyEventData->keyCode;
				if(keyEventData->modifiers&EEM_MOD0)		// check for bound keys
				{
					switch(key)
					{
						case 'j':
							UniverseSanityCheck(event->view->parentBuffer->textUniverse);
							break;
						case 'a':
							result=EditorSelectAll(event->view->parentBuffer);
							break;
						case 'y':
							if((result=EditorRedo(event->view->parentBuffer,&didSomething)))
							{
								if(didSomething)
								{
									EditorHomeViewToSelectionEdge(event->view,false,HT_SEMISTRICT,HT_SEMISTRICT);
								}
								else
								{
									EditorBeep();
								}
							}
							break;
						case 'u':
							if((result=EditorUndo(event->view->parentBuffer,&didSomething)))
							{
								if(didSomething)
								{
									EditorHomeViewToSelectionEdge(event->view,false,HT_SEMISTRICT,HT_SEMISTRICT);
								}
								else
								{
									EditorBeep();
								}
							}
							break;
						case 'z':
							if((result=EditorToggleUndo(event->view->parentBuffer,&didSomething)))
							{
								if(didSomething)
								{
									EditorHomeViewToSelectionEdge(event->view,false,HT_SEMISTRICT,HT_SEMISTRICT);
								}
								else
								{
									EditorBeep();
								}
							}
							break;
						case 'x':
							if((result=EditorCut(event->view->parentBuffer,EditorGetCurrentClipboard())))
							{
								EditorHomeViewToSelectionEdge(event->view,false,HT_SEMISTRICT,HT_SEMISTRICT);
							}
							break;
						case 'c':
							result=EditorCopy(event->view->parentBuffer,EditorGetCurrentClipboard());
							break;
						case 'v':
							EditorHomeViewToSelectionEdge(event->view,false,HT_SEMISTRICT,HT_SEMISTRICT);
							if((result=EditorColumnarPaste(event->view->parentBuffer,event->view,EditorGetCurrentClipboard())))
							{
								EditorHomeViewToSelectionEdge(event->view,false,HT_SEMISTRICT,HT_SEMISTRICT);
							}
							break;
					}
				}
				else
				{
					EditorHomeViewToSelectionEdge(event->view,false,HT_SEMISTRICT,HT_SEMISTRICT);	// if position was off the view, put it well back on
					if((result=EditorInsert(event->view->parentBuffer,&key,1)))
					{
						EditorHomeViewToSelectionEdge(event->view,false,HT_LENIENT,HT_LENIENT);			// now scroll into position
					}
				}
			}
			break;
		case VET_POSITIONVERTICAL:
			posEventData=(VIEW_POS_EVENT_DATA *)event->eventData;
			GetEditorViewTextInfo(event->view,&topLine,&numLines,&leftPixel,&numPixels);
			SetViewTopLeft(event->view,posEventData->position,leftPixel);
			break;
		case VET_POSITIONHORIZONTAL:
			posEventData=(VIEW_POS_EVENT_DATA *)event->eventData;
			GetEditorViewTextInfo(event->view,&topLine,&numLines,&leftPixel,&numPixels);
			SetViewTopLeft(event->view,topLine,posEventData->position);
			break;
		case VET_CLICK:
		case VET_CLICKHOLD:
			clickEventData=(VIEW_CLICK_EVENT_DATA *)event->eventData;
			trackMode=0;								// default place mode
			if(event->eventType==VET_CLICKHOLD)			// see if still tracking
			{
				trackMode|=TM_mREPEAT;
			}
			if(clickEventData->modifiers&EEM_SHIFT)		// attempting to add to current selection?
			{
				trackMode|=TM_mCONTINUE;
			}
			if((clickEventData->modifiers&EEM_MOD0)||(clickEventData->keyCode==2))
			{
				trackMode|=TM_mCOLUMNAR;
			}
			if((clickEventData->modifiers&EEM_MOD1)||(clickEventData->keyCode==1))
			{
				trackMode|=TM_mHAND;
			}
			switch((clickEventData->modifiers&EEM_STATE0)>>EES_STATE0)
			{
				case 0:
					trackMode|=TM_CHAR;
					break;
				case 1:
					trackMode|=TM_WORD;
					break;
				case 2:
					trackMode|=TM_LINE;
					break;
				default:
					trackMode|=TM_ALL;
					break;
			}
			result=EditorTrackViewPointer(event->view,clickEventData->xClick,clickEventData->yClick,trackMode);
			break;
	}
	// @@@ check "result" -- complain if something went wrong
	ResetEditorViewCursorBlink(event->view);	// reset cursor blinking (if we even have a cursor that blinks)
}

bool HandleShellCommand(const char *command,int argc,char *argv[])
// Sometimes things in the gui request that the shell should perform
// certain operations. This is the interface to allow that to happen
// If the operation has some sort of failure (defined by the operation)
// this will report it (if there is a message) and return false
{
	char
		*list;
	int
		tclResult;
	const char
		*stringResult;

	list=Tcl_Merge(argc,argv);
	ClearUserAbort();
	tclResult=Tcl_VarEval(tclInterpreter,"ShellCommand ",command," ",list,(char *)NULL);
	Tcl_Free(list);
	stringResult=Tcl_GetStringResult(tclInterpreter);
	if(tclResult!=TCL_OK)
	{
		if(stringResult[0])
		{
			ReportMessage("Shell command execution error:\n%.256s\n",stringResult);
		}
		return(false);
	}
	return(true);
}

static bool ExecuteStartupScript(Tcl_Interp *interpreter)
// Locate the startup script file, open, and execute it
// if there is any problem, SetError, and return false
{
	bool
		fail;
	char
		scriptPath[MAX_PATH_NAME_LENGTH];

	fail=false;
	if(LocateStartupScript(scriptPath))
	{
		ClearUserAbort();
		Tcl_SetVar(interpreter,"SCRIPTPATH",scriptPath,TCL_LEAVE_ERR_MSG);
		if(Tcl_EvalFile(interpreter,scriptPath)!=TCL_OK)
		{
			SetError("%s:%d:%s",scriptPath,interpreter->errorLine,Tcl_GetStringResult(interpreter));
			fail=true;
		}
	}
	else
	{
		fail=true;
	}
	return(!fail);
}

static int TraceCheckAbortProc(ClientData clientData,Tcl_Interp *interpreter,int level,const char *command,Tcl_Command commandToken,int objc,Tcl_Obj *const objv[])
// This is a small trick on Tcl. We tell it we want to trace, but really, we want
// to check to see if the user is trying to abort the execution of a script.
// So, every time we are called, we check to see if the user is aborting.
{
	if(CheckUserAbort())
	{
		Tcl_AppendResult(interpreter,"Script Aborted",NULL);
		return(TCL_ERROR);
	}
	return(TCL_OK);
}

static void UnSetUpTclAbortHandling(Tcl_Interp *interpreter)
// Undo what SetUpTclAbortHandling did
{
	Tcl_DeleteTrace(interpreter,tclTrace);
}

static bool SetUpTclAbortHandling(Tcl_Interp *interpreter)
// Set up Tcl so that it manages to call CheckUserAbort, and if CheckUserAbort
// returns true, then we force an error in the Tcl control Flow.
{
	if((tclTrace=Tcl_CreateObjTrace(interpreter,0,TCL_ALLOW_INLINE_COMPILATION,TraceCheckAbortProc,(ClientData)NULL,(Tcl_CmdObjTraceDeleteProc *)NULL)))
	{
		return(true);
	}
	return(false);
}

void ShellDoBackground()
// Handle Tcl's event loop
{
	while(Tcl_DoOneEvent(TCL_ALL_EVENTS|TCL_DONT_WAIT))
		;							// tell Tcl to check for any events it may care about
}

static void UnInitTcl()
// Undo what InitTcl did
{
	UnSetUpTclAbortHandling(tclInterpreter);
	UnInitChannels(tclInterpreter);
	Tcl_DeleteInterp(tclInterpreter);	// have TCL clean up
}

static bool InitTcl(char *pathName)
// Initialize Tcl
{
	if((tclInterpreter=Tcl_CreateInterp()))
	{
		if(InitChannels(tclInterpreter))
		{
			if(Tcl_Init(tclInterpreter)==TCL_OK)
			{
			    Tcl_FindExecutable(pathName);

				if(SetUpTclAbortHandling(tclInterpreter))
				{
					return(true);
				}
			}
			else
			{
				SetError("Failed to Tcl_Init(): %s",tclInterpreter->result);
			}
			UnInitChannels(tclInterpreter);
		}
		else
		{
			SetError("Failed to redirect Tcl's stdout and stderr");
		}
		Tcl_DeleteInterp(tclInterpreter);						// have TCL clean up
	}
	else
	{
		SetError("Failed to create TCL interpreter");
	}
	return(false);
}

void ShellLoop(int argc,char *argv[])
// This is where the editor shell gets and handles events
// it actually defers to the gui level to call us back with events
{
	ClearErrorTrace();
	if(InitTcl(argv[0]))
	{
		ClearErrorTrace();
		if(CreateEditorShellCommands(tclInterpreter))
		{
			ClearErrorTrace();
			if(AddSupplementalShellCommands(tclInterpreter))
			{
				ClearErrorTrace();
				if(ExecuteStartupScript(tclInterpreter))			// deal with the start-up script, leave if there is a problem
				{
					EditorEventLoop(argc,argv);						// pass initial parameters
				}
				else
				{
					fprintf(stderr,"Failed to execute startup script: %s\n",GetErrorTrace());
				}
			}
			else
			{
				fprintf(stderr,"Failed to add supplemental shell commands: %s\n",GetErrorTrace());
			}
		}
		else
		{
			fprintf(stderr,"Failed to create editor shell commands: %s\n",GetErrorTrace());
		}
		UnInitTcl();
	}
	else
	{
		fprintf(stderr,"Failed to initialize Tcl: %s\n",GetErrorTrace());
	}
}
