// TCL commands that e93 supports
// Copyright (C) 2000 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

// NOTES about TCL, and its subtle interactions with the editor
//
// First, Tcl should never be allowed to evaluate script while an editor
// window, view, buffer, selectionlist, etc... is in a state of flux that
// would cause it to draw improperly if the script caused the editor to
// update its windows. This is because certain editor operations that
// scripts can call, can force updates of all editor windows.
//
// Second, if a script is running, and decides to operate on a window that
// is in some state of flux, (say in the middle of a replace-all), it must
// be prohibited from doing so, and the operation must fail. NOTE, that by
// rule one, if you are executing script in the middle of replace-all, that
// all buffers must be in a state where they can be drawn.
//
// Finally, if a script is run while some code of the editor is relying on
// a given window, or buffer to be around, the code MUST set the
// buffer Busy, or check that the buffer still exists after
// the script is run.
// If a buffer is set Busy, no script command, or side-effect can allow the
// buffer to be modified in ANY way, including text/selection changes, undo
// changes, new views, deleting views, etc...

#define	TEMP_STRING_BUFFER_LENGTH	1024

typedef bool OptionParseFunction(Tcl_Interp *localInterpreter,char *optionName,int optionValue,void *variable,int objc,Tcl_Obj *const objv[],int *optionIndex);

typedef struct
{
	const char
		*optionString;
	int
		optionValue;
	OptionParseFunction
		*function;
} OPTION_DESCRIPTOR_RECORD;

typedef struct
{
	EDITOR_BUFFER
		*buffer;
	EDITOR_MARK
		*mark;
} PARSE_MARK;

static TOKEN_LIST
	menuRelationshipTokens[]=
	{
		{"nextsibling",MR_NEXTSIBLING},
		{"previoussibling",MR_PREVIOUSSIBLING},
		{"firstchild",MR_FIRSTCHILD},
		{"lastchild",MR_LASTCHILD},
		{"",0}
	};

static TOKEN_LIST
	cursorMovementTokens[]=
	{
		{"leftchar",RPM_mBACKWARD|RPM_CHAR},
		{"rightchar",RPM_CHAR},
		{"leftword",RPM_mBACKWARD|RPM_WORD},
		{"rightword",RPM_WORD},
		{"upline",RPM_mBACKWARD|RPM_LINE},
		{"downline",RPM_LINE},
		{"startline",RPM_mBACKWARD|RPM_LINEEDGE},
		{"endline",RPM_LINEEDGE},
		{"startparagraph",RPM_mBACKWARD|RPM_PARAGRAPHEDGE},
		{"endparagraph",RPM_PARAGRAPHEDGE},
		{"uppage",RPM_mBACKWARD|RPM_PAGE},
		{"downpage",RPM_PAGE},
		{"startdoc",RPM_mBACKWARD|RPM_DOCEDGE},
		{"enddoc",RPM_DOCEDGE},
		{"",0}
	};

typedef enum
{
	SM_EXPRESSION=0,
	SM_MAPPING,
	SM_AT,
	SM_ROOT
} SYNTAX_MAP_COMMANDS;

static TOKEN_LIST
	syntaxMapCommands[]=
	{
		{"exp",SM_EXPRESSION},
		{"map",SM_MAPPING},
		{"at",SM_AT},
		{"root",SM_ROOT},
		{"",0}
	};

static UINT32 ForcePositionIntoRange(EDITOR_BUFFER *buffer,UINT32 position)
// take position, and make sure it is in range for buffer
{
	if(position>buffer->textUniverse->totalBytes)
	{
		position=buffer->textUniverse->totalBytes;
	}
	return(position);
}

static void ArrangePositions(UINT32 *startPosition,UINT32 *endPosition)
// Make sure startPosition is <= endPosition
// if not, flip them around
{
	UINT32
		temp;

	if(*startPosition>*endPosition)
	{
		temp=*startPosition;
		*startPosition=*endPosition;
		*endPosition=temp;
	}
}

static EDITOR_WINDOW *GetEditorWindow(Tcl_Interp *localInterpreter,Tcl_Obj *name)
// see if string is a window, if so, return it
// if not, return NULL, and set the Tcl result
{
	EDITOR_WINDOW
		*window;
	char
		*string;

	string=Tcl_GetStringFromObj(name,NULL);
	if((window=LocateEditorDocumentWindow(string)))
	{
		return(window);
	}
	Tcl_AppendResult(localInterpreter,"Failed to locate window '",string,"'",NULL);
	return(NULL);
}

static EDITOR_BUFFER *GetEditorBuffer(Tcl_Interp *localInterpreter,Tcl_Obj *name)
// see if string is a buffer, if so, return it
// if not, return NULL, and set the Tcl result
{
	EDITOR_BUFFER
		*buffer;
	char
		*string;

	string=Tcl_GetStringFromObj(name,NULL);
	if((buffer=LocateBuffer(string)))
	{
		return(buffer);
	}
	Tcl_AppendResult(localInterpreter,"Failed to locate buffer '",string,"'",NULL);
	return(NULL);
}

static bool GetINT32(Tcl_Interp *localInterpreter,Tcl_Obj *object,INT32 *number)
// convert object to a INT32
// return false if the conversion failed, and fill the Tcl result
{
	long
		longValue;

	if(Tcl_ExprLongObj(localInterpreter,object,&longValue)==TCL_OK)
	{
		*number=(INT32)longValue;
		return(true);
	}
	return(false);
}

static bool GetUINT32(Tcl_Interp *localInterpreter,Tcl_Obj *object,UINT32 *number)
// convert object to a UINT32
// return false if the conversion failed, and fill the Tcl result
{
	long
		longValue;

	if(Tcl_ExprLongObj(localInterpreter,object,&longValue)==TCL_OK)
	{
		*number=(UINT32)longValue;					// ### this should really be fixed to be a true unsigned
		return(true);
	}
	return(false);
}

static bool GetUINT32String(Tcl_Interp *localInterpreter,const char *string,UINT32 *number)
// convert string to a UINT32
// return false if the conversion failed, and fill the Tcl result
{
	long
		longValue;

	if(Tcl_ExprLong(localInterpreter,string,&longValue)==TCL_OK)
	{
		*number=(UINT32)longValue;					// ### this should really be fixed to be a true unsigned
		return(true);
	}
	return(false);
}

static bool GetRectangle(Tcl_Interp *localInterpreter,Tcl_Obj *const *objs,EDITOR_RECT *rect)
// convert objs (of which there must be at least 4) to an EDITOR_RECT
// return false if the conversion failed, and fill the Tcl result
{
	if(GetINT32(localInterpreter,objs[0],&rect->x))
	{
		if(GetINT32(localInterpreter,objs[1],&rect->y))
		{
			if(GetUINT32(localInterpreter,objs[2],&rect->w))
			{
				if(GetUINT32(localInterpreter,objs[3],&rect->h))
				{
					return(true);
				}
			}
		}
	}
	return(false);
}

static bool GetNibble(char character,UINT32 *nibble)
// interpret character as a hex nibble, and return it
// if there is a problem return false
{
	if(character>='0'&&character<='9')
	{
		*nibble=character-'0';
		return(true);
	}
	else
	{
		if(character>='A'&&character<='F')
		{
			*nibble=10+character-'A';
			return(true);
		}
		else
		{
			if(character>='a'&&character<='f')
			{
				*nibble=10+character-'a';
				return(true);
			}
		}
	}
	return(false);
}

static bool GetColor(Tcl_Interp *localInterpreter,Tcl_Obj *object,EDITOR_COLOR *color)
// convert object to a color value
// return false if the conversion failed, and fill the Tcl result
{
	UINT32
		index,
		nibble;
	char
		character;
	bool
		fail;
	char
		*string;

	fail=false;
	string=Tcl_GetStringFromObj(object,NULL);
	if(!EditorColorNameToColor(string,color))
	{
		*color=0;
		index=0;
		while(!fail&&(character=string[index]))
		{
			if(GetNibble(character,&nibble))
			{
				(*color)<<=4;
				(*color)|=nibble;
				index++;
			}
			else
			{
				fail=true;
			}
		}
		(*color)&=0xFFFFFF;			// if number was large, ignore preceding digits
		if(!index)					// no characters in the input string
		{
			fail=true;
		}
	}
	if(fail)
	{
		Tcl_AppendResult(localInterpreter,"Invalid color:",string,NULL);
	}
	return(!fail);
}

static bool GetBoolean(Tcl_Interp *localInterpreter,Tcl_Obj *object,bool *value)
// convert object to a boolean value
// return false if the conversion failed, and fill the Tcl result
{
	int
		intValue;

	if(Tcl_GetBoolean(localInterpreter,Tcl_GetStringFromObj(object,NULL),&intValue)==TCL_OK)
	{
		*value=intValue?true:false;
		return(true);
	}
	return(false);
}

static bool OptionBoolean(Tcl_Interp *localInterpreter,char *optionName,int optionValue,void *variable,int objc,Tcl_Obj *const objv[],int *optionIndex)
// Handle parsing of a boolean option
{
	*((bool *)variable)=(optionValue!=0);
	return(true);
}

static bool OptionEnum(Tcl_Interp *localInterpreter,char *optionName,int optionValue,void *variable,int objc,Tcl_Obj *const objv[],int *optionIndex)
// Handle parsing of a enumeration option
{
	*((int *)variable)=optionValue;
	return(true);
}

/* unused code
static bool OptionUINT32(Tcl_Interp *localInterpreter,char *optionName,int optionValue,void *variable,int objc,Tcl_Obj *const objv[],int *optionIndex)
// Return an unsigned integer
{
	if((*optionIndex)<objc)
	{
		if(GetUINT32(localInterpreter,objv[*optionIndex],(UINT32 *)variable))
		{
			(*optionIndex)++;
			return(true);
		}
	}
	else
	{
		Tcl_AppendResult(localInterpreter,optionName,": Missing arguments",NULL);
	}
	return(false);
}
*/

static bool OptionMark(Tcl_Interp *localInterpreter,char *optionName,int optionValue,void *variable,int objc,Tcl_Obj *const objv[],int *optionIndex)
// Attempt to locate a currently existing mark within an editor buffer
{
	char
		*markName;
	PARSE_MARK
		*parseMark;

	if((*optionIndex)<objc)
	{
		parseMark=(PARSE_MARK *)variable;
		markName=Tcl_GetStringFromObj(objv[*optionIndex],NULL);
		(*optionIndex)++;
		if((parseMark->mark=LocateEditorMark(parseMark->buffer,markName)))
		{
			return(true);
		}
		else
		{
			Tcl_AppendResult(localInterpreter,"Failed to locate mark '",markName,"'",NULL);
		}
	}
	else
	{
		Tcl_AppendResult(localInterpreter,optionName,": Missing arguments",NULL);
	}
	return(false);
}

static bool LocateOptionMatch(OPTION_DESCRIPTOR_RECORD *options,char *option,int *matchIndex)
// attempt to locate an option that matches option
// if one was located, return true with the index in matchIndex
// else return false
{
	int
		i;
	bool
		found,reachedEnd;

	i=0;
	found=reachedEnd=false;
	while(!found&&!reachedEnd)
	{
		if(options[i].optionString)
		{
			if(strcmp(option,options[i].optionString)==0)
			{
				*matchIndex=i;
				found=true;
			}
			i++;
		}
		else
		{
			reachedEnd=true;
		}
	}
	return(found);
}

static void AddOptionsToResult(Tcl_Interp *localInterpreter,OPTION_DESCRIPTOR_RECORD *options)
// used when reporting bad options
{
	int
		i;

	Tcl_AppendResult(localInterpreter,options[0].optionString,NULL);
	i=1;
	while(options[i].optionString)
	{
		if(options[i+1].optionString)
		{
			Tcl_AppendResult(localInterpreter,", ",NULL);
		}
		else
		{
			Tcl_AppendResult(localInterpreter,", or ",NULL);
		}
		Tcl_AppendResult(localInterpreter,options[i].optionString,NULL);
		i++;
	}
}

static bool ParseOptions(Tcl_Interp *localInterpreter,OPTION_DESCRIPTOR_RECORD *options,void **variables,int objc,Tcl_Obj *const objv[],int *optionIndex,bool unrecognizedIsError)
// parse command options given a description of the options, and a list of parameters for them
// to modify
// if there is a problem, set the error string in localInterpreter, and return false
// NOTE: if unrecognizedIsError is false, this will stop parsing at the first
// unrecognized token, and return true
// if unrecognizedIsError is true, this will only return true if the list of
// options can be parsed correctly until the end.
{
	bool
		fail,
		mismatch;
	char
		*optionName;
	int
		matchIndex;

	fail=mismatch=false;
	while(!fail&&!mismatch&&((*optionIndex)<objc))
	{
		optionName=Tcl_GetStringFromObj(objv[*optionIndex],NULL);
		(*optionIndex)++;
		matchIndex=0;			// keep GCC quiet
		if(LocateOptionMatch(options,optionName,&matchIndex))
		{
			fail=!options[matchIndex].function(localInterpreter,optionName,options[matchIndex].optionValue,variables[matchIndex],objc,objv,optionIndex);
		}
		else
		{
			mismatch=true;
			if(unrecognizedIsError)
			{
				Tcl_AppendResult(localInterpreter,"bad option \"",optionName,"\": must be ",NULL);
				AddOptionsToResult(localInterpreter,options);
				fail=true;
			}
		}
	}
	return(!fail);
}

static UINT32
	modifierFlagTable[]=
	{
		EEM_CAPSLOCK,
		EEM_SHIFT,
		EEM_CTL,
		EEM_MOD0,
		EEM_MOD1,
		EEM_MOD2,
		EEM_MOD3,
		EEM_MOD4,
		EEM_MOD5,
		EEM_MOD6,
		EEM_MOD7
	};

static bool GetModifierMaskAndValue(Tcl_Interp *localInterpreter,Tcl_Obj *object,UINT32 *modifierMask,UINT32 *modifierValue)
// convert object of modifiers into a mask, and value
// if there is a problem, set the Tcl result, and return false
{
	int
		i,
		length;
	char
		*string;
	bool
		fail;

	*modifierMask=0;
	*modifierValue=0;
	string=Tcl_GetStringFromObj(object,NULL);
	length=strlen(string);
	fail=false;
	if(length==11)				// must be an entry for each flag
	{
		for(i=0;(!fail)&&(i<length);i++)
		{
			switch(string[i])
			{
				case 'x':
				case 'X':
					break;
				case '0':
					*modifierMask|=modifierFlagTable[i];
					break;
				case '1':
					*modifierMask|=modifierFlagTable[i];
					*modifierValue|=modifierFlagTable[i];
					break;
				default:
					Tcl_AppendResult(localInterpreter,"Invalid modifiers",NULL);
					fail=true;
					break;
			}
		}
	}
	else
	{
		Tcl_AppendResult(localInterpreter,"Invalid modifiers (incorrect length)",NULL);
		fail=true;
	}
	return(!fail);
}

static void ModifierMaskAndValueToString(UINT32 modifierMask,UINT32 modifierValue,char *string)
// given a modifier mask, and value, make a modifier string
// NOTE: string must be long enough to hold the result (which is never longer than 12 characters
// including the terminator)
{
	int
		i;

	for(i=0;i<11;i++)
	{
		if(modifierMask&modifierFlagTable[i])
		{
			if(modifierValue&modifierFlagTable[i])
			{
				string[i]='1';
			}
			else
			{
				string[i]='0';
			}
		}
		else
		{
			string[i]='X';
		}
	}
	string[i]='\0';
}

static bool GetVarFlagList(Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[],bool *flagList)
// write values from the given variable array to the given boolean array
// if there is a problem, set the TCL error, and return false
{
	Tcl_Obj
		*varValue;
	int
		currentCount;
	bool
		fail;

	fail=false;
	for(currentCount=0;currentCount<objc&&!fail;currentCount++)
	{
		if((varValue=Tcl_ObjGetVar2(localInterpreter,objv[currentCount],NULL,TCL_LEAVE_ERR_MSG)))
		{
			fail=!GetBoolean(localInterpreter,varValue,&(flagList[currentCount]));
		}
		else
		{
			fail=true;
		}
	}
	return(!fail);
}

static bool PutVarFlagList(Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[],bool *flagList)
// write the boolean variables back into Tcl variables
// if there is a problem, set the TCL error, and return false
{
	int
		currentCount;
	bool
		fail;

	fail=false;
	for(currentCount=0;currentCount<objc&&!fail;currentCount++)
	{
		fail=!Tcl_ObjSetVar2(localInterpreter,objv[currentCount],NULL,Tcl_NewStringObj(flagList[currentCount]?"1":"0",-1),TCL_LEAVE_ERR_MSG);
	}
	return(!fail);
}

static bool ShellBufferNotBusy(Tcl_Interp *localInterpreter,EDITOR_BUFFER *buffer)
// check the given buffer's busy state
// if it is true, set the Tcl result, and return false
{
	if(!BufferBusy(buffer))
	{
		return(true);
	}
	else
	{
		Tcl_AppendResult(localInterpreter,"Buffer '",buffer->contentName,"' is busy",NULL);
	}
	return(false);
}

static bool CurrentClipboardNotBusy(Tcl_Interp *localInterpreter)
// make sure the current clipboard is not busy, if it is, set the Tcl result
// and return false
{
	EDITOR_BUFFER
		*clipboardBuffer;

	if((clipboardBuffer=EditorGetCurrentClipboard()))
	{
		return(ShellBufferNotBusy(localInterpreter,clipboardBuffer));
	}
	return(true);
}

// ACTUAL TCL COMMAND IMPLEMENTATIONS

static int Cmd_DeleteMenu(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// delete a menu from the menu list
// NOTE: if this deletes an element, it will also delete the element's children
{
	EDITOR_MENU
		*menu;
	UINT32
		index;
	int
		pathArgc;
	const char
		**pathArgv;
	bool
		fail;

	fail=false;
	if(objc>1)
	{
		for(index=1;!fail&&(int)index<objc;index++)
		{
			if(Tcl_SplitList(localInterpreter,Tcl_GetStringFromObj(objv[index],NULL),&pathArgc,&pathArgv)==TCL_OK)
			{
				menu=NULL;										// start at the root
				if(GetEditorMenu(pathArgc,pathArgv,&menu))
				{
					DisposeEditorMenu(menu);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,"Invalid menu specification",NULL);
					fail=true;
				}
				Tcl_Free((char *)pathArgv);
			}
			else
			{
				fail=true;
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"menuPathList ?menuPathList ...?");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_AddMenu(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// add a menu to the menu list
// NOTE: if a menu already exists, this will delete it before attempting
// a creation
// NOTE ALSO: if this deletes an element, it will also delete the element's children
{
	int
		relationship;
	bool
		active;
	int
		pathArgc;
	const char
		**pathArgv;
	bool
		succeed;

	succeed=false;
	if(objc==7)
	{
		if(MatchToken(Tcl_GetStringFromObj(objv[2],NULL),menuRelationshipTokens,&relationship))
		{
			if(GetBoolean(localInterpreter,objv[3],&active))
			{
				if(Tcl_SplitList(localInterpreter,Tcl_GetStringFromObj(objv[1],NULL),&pathArgc,&pathArgv)==TCL_OK)
				{
					ClearErrorTrace();
					if(CreateEditorMenu(pathArgc,pathArgv,relationship,Tcl_GetStringFromObj(objv[4],NULL),Tcl_GetStringFromObj(objv[5],NULL),Tcl_GetStringFromObj(objv[6],NULL),active))
					{
						succeed=true;
					}
					else
					{
						Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
					}
					Tcl_Free((char *)pathArgv);
				}
			}
		}
		else
		{
			Tcl_AppendResult(localInterpreter,"Syntax error in menu relationship field",NULL);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"menuPathList relationship active menuName menuModifiers menuFunction");
	}
	return(succeed?TCL_OK:TCL_ERROR);
}

static int Cmd_BindKey(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// add a key binding to the bindings list
// NOTE: if a binding already exists, this will delete it before attempting
// a creation
{
	UINT32
		keyCode,
		modifierMask,
		modifierValue;

	if(objc==4)
	{
		if(EditorKeyNameToKeyCode(Tcl_GetStringFromObj(objv[1],NULL),&keyCode))						// find the code for this key
		{
			if(GetModifierMaskAndValue(localInterpreter,objv[2],&modifierMask,&modifierValue))
			{
				ClearErrorTrace();
				if(CreateEditorKeyBinding(keyCode,modifierMask,modifierValue,Tcl_GetStringFromObj(objv[3],NULL)))
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
				}
			}
		}
		else
		{
			Tcl_AppendResult(localInterpreter,"Invalid key code",NULL);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"keyName modifiers keyFunction");
	}
	return(TCL_ERROR);
}

static int Cmd_UnBindKey(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// delete a key binding from the bindings list
{
	UINT32
		keyCode,
		modifierMask,
		modifierValue;

	if(objc==3)
	{
		if(EditorKeyNameToKeyCode(Tcl_GetStringFromObj(objv[1],NULL),&keyCode))						// find the code for this key
		{
			if(GetModifierMaskAndValue(localInterpreter,objv[2],&modifierMask,&modifierValue))
			{
				if(DeleteEditorKeyBinding(keyCode,modifierMask,modifierValue))
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,"Key binding does not exist",NULL);
				}
			}
		}
		else
		{
			Tcl_AppendResult(localInterpreter,"Invalid key code",NULL);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"keyName modifiers");
	}
	return(TCL_ERROR);
}

static int Cmd_KeyBindings(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// dump all key bindings
{
	UINT32
		keyCode,
		modifierMask,
		modifierValue;
	EDITOR_KEY_BINDING
		*currentBinding;
	UINT32
		currentCode,
		currentMask,
		currentValue;
	char
		*keyName,
		modifierString[TEMP_STRING_BUFFER_LENGTH];
	bool
		haveName,
		haveModifiers;
	bool
		fail;
	bool
		didOne;

	if(objc>=1&&objc<=3)
	{
		fail=false;
		haveName=haveModifiers=false;
		if(objc>=2)
		{
			haveName=true;
			if(!EditorKeyNameToKeyCode(Tcl_GetStringFromObj(objv[1],NULL),&keyCode))						// find the code for this key
			{
				Tcl_AppendResult(localInterpreter,"Invalid key code",NULL);
				fail=true;
			}
		}
		if(!fail&&objc>=3)
		{
			haveModifiers=true;
			fail=!GetModifierMaskAndValue(localInterpreter,objv[2],&modifierMask,&modifierValue);
		}
		if(!fail)
		{
			currentBinding=keyBindingListHead;
			didOne=false;
			while(currentBinding)
			{
				GetKeyBindingCodeAndModifiers(currentBinding,&currentCode,&currentMask,&currentValue);

				if(!haveName||(currentCode==keyCode))
				{
					if(!haveModifiers||((currentMask==modifierMask)&&(currentValue==modifierValue)))
					{
						if((keyName=EditorKeyCodeToKeyName(currentCode)))
						{
							ModifierMaskAndValueToString(currentMask,currentValue,&(modifierString[0]));
							if(didOne)				// do not put return on last line
							{
								Tcl_AppendResult(localInterpreter,"\n",NULL);
							}
							Tcl_AppendResult(localInterpreter,"bindkey",NULL);
							Tcl_AppendElement(localInterpreter,keyName);
							Tcl_AppendElement(localInterpreter,&(modifierString[0]));
							Tcl_AppendElement(localInterpreter,GetKeyBindingText(currentBinding));
							didOne=true;
						}
					}
				}
				currentBinding=LocateNextKeyBinding(currentBinding);
			}
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"?keyName? ?modifiers?");
	}
	return(TCL_ERROR);
}

static int Cmd_WaitKey(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// wait for a keyboard press, return the string for the code, and modifiers
{
	UINT32
		keyCode,
		modifierValue;
	char
		*keyName,
		modifierString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==1)
	{
		EditorGetKeyPress(&keyCode,&modifierValue,true,false);
		if((keyName=EditorKeyCodeToKeyName(keyCode)))
		{
			ModifierMaskAndValueToString(0xFFFFFFFF,modifierValue,&(modifierString[0]));
			Tcl_AppendElement(localInterpreter,keyName);
			Tcl_AppendElement(localInterpreter,&(modifierString[0]));
			return(TCL_OK);
		}
		else
		{
			Tcl_AppendResult(localInterpreter,"Unnamed key",NULL);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_GetKey(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// check to see if there is a keyboard event in the queue, if so, get it, if not, return an error
{
	UINT32
		keyCode,
		modifierValue;
	char
		*keyName,
		modifierString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==1)
	{
		if(EditorGetKeyPress(&keyCode,&modifierValue,false,false))
		{
			if((keyName=EditorKeyCodeToKeyName(keyCode)))
			{
				ModifierMaskAndValueToString(0xFFFFFFFF,modifierValue,&(modifierString[0]));
				Tcl_AppendElement(localInterpreter,keyName);
				Tcl_AppendElement(localInterpreter,&(modifierString[0]));
				return(TCL_OK);
			}
			else
			{
				Tcl_AppendResult(localInterpreter,"Unnamed key",NULL);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_OkDialog(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// create a dialog with some text, and an OK button
{
	bool
		fail;

	fail=false;
	if(objc==2)
	{
		ClearErrorTrace();
		if(OkDialog(Tcl_GetStringFromObj(objv[1],NULL)))
		{
			ClearUserAbort();				// if user managed to send break sequence during the dialog, we would rather not abort now!
		}
		else
		{
			Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"dialogText");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_OkCancelDialog(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// create a dialog with some text, and an OK button, and a cancel button
{
	bool
		cancel,
		fail;

	fail=false;
	if(objc==2)
	{
		ClearErrorTrace();
		if(OkCancelDialog(Tcl_GetStringFromObj(objv[1],NULL),&cancel))
		{
			if(cancel)
			{
				fail=true;
			}
			ClearUserAbort();				// if user managed to send break sequence during the dialog, we would rather not abort now!
		}
		else
		{
			Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"dialogText");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_YesNoDialog(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// create a dialog with some text, and yes, no, and cancel buttons
{
	bool
		yes,
		cancel,
		fail;

	fail=false;
	if(objc==2)
	{
		ClearErrorTrace();
		if(YesNoCancelDialog(Tcl_GetStringFromObj(objv[1],NULL),&yes,&cancel))
		{
			if(cancel)
			{
				fail=true;
			}
			else
			{
				Tcl_AppendResult(localInterpreter,yes?"1":"0",NULL);
			}
			ClearUserAbort();				// if user managed to send break sequence during the dialog, we would rather not abort now!
		}
		else
		{
			Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"dialogText");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_TextDialog(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// create a get text dialog
// if InitialText is not specified, then no initial text will be present in the dialog
// NOTE: since this can indirectly modify the current clipboard, the clipboard must not be busy
{
	bool
		cancel,
		fail;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	fail=false;
	if(objc==2||objc==3)
	{
		if(objc==3)
		{
			strncpy(tempString,Tcl_GetStringFromObj(objv[2],NULL),TEMP_STRING_BUFFER_LENGTH-1);
			tempString[TEMP_STRING_BUFFER_LENGTH-1]='\0';
		}
		else
		{
			tempString[0]='\0';
		}
		if(CurrentClipboardNotBusy(localInterpreter))
		{
			ClearErrorTrace();
			if(GetSimpleTextDialog(Tcl_GetStringFromObj(objv[1],NULL),&(tempString[0]),TEMP_STRING_BUFFER_LENGTH,&cancel))
			{
				if(cancel)
				{
					fail=true;
				}
				else
				{
					Tcl_AppendResult(localInterpreter,&(tempString[0]),NULL);
				}
				ClearUserAbort();				// if user managed to send break sequence during the dialog, we would rather not abort now!
			}
			else
			{
				Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"dialogTitle ?initialText?");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_SearchDialog(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// create a search dialog
// findBuffer, and replaceBuffer must exist, or an error will be returned
// NOTE: currently DialogTitle is ignored, later it should be implemented
// NOTE: since this can indirectly modify the current clipboard, the clipboard must not be busy
{
	bool
		flagList[6];
	UINT16
		searchMode;
	EDITOR_BUFFER
		*findBuffer,
		*replaceBuffer;
	bool
		cancel,
		result;

	result=false;
	if(objc==10)
	{
		if((findBuffer=GetEditorBuffer(localInterpreter,objv[2])))
		{
			if(ShellBufferNotBusy(localInterpreter,findBuffer))
			{
				if((replaceBuffer=GetEditorBuffer(localInterpreter,objv[3])))
				{
					if(ShellBufferNotBusy(localInterpreter,replaceBuffer))
					{
						if(GetVarFlagList(localInterpreter,6,&(objv[4]),&(flagList[0])))
						{
							if(CurrentClipboardNotBusy(localInterpreter))
							{
								ClearErrorTrace();
								if(SearchReplaceDialog(findBuffer,replaceBuffer,&(flagList[0]),&(flagList[1]),&(flagList[2]),&(flagList[3]),&(flagList[4]),&(flagList[5]),&searchMode,&cancel))
								{
									if(!cancel)
									{
										result=true;
									}
									ClearUserAbort();				// if user managed to send break sequence during the dialog, we would rather not abort now!
									PutVarFlagList(localInterpreter,6,&(objv[4]),&(flagList[0]));	// put these back even if user canceled
								}
								else
								{
									Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
								}
							}
							if(result)
							{
								switch(searchMode)
								{
									case ST_FIND:
										Tcl_AppendResult(localInterpreter,"find",NULL);
										break;
									case ST_FINDALL:
										Tcl_AppendResult(localInterpreter,"findall",NULL);
										break;
									case ST_REPLACE:
										Tcl_AppendResult(localInterpreter,"replace",NULL);
										break;
									case ST_REPLACEALL:
										Tcl_AppendResult(localInterpreter,"replaceall",NULL);
										break;
								}
								return(TCL_OK);
							}
						}
					}
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"dialogTitle findBuffer replaceBuffer backwardsVar wrapAroundVar selectionExprVar ignoreCaseVar limitScopeVar replaceProcVar");
	}
	return(TCL_ERROR);
}

static int Cmd_ListDialog(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// create a list dialog, handle it, and if all is well, return a new list
// that contains the chosen elements of the old list
{
	UINT32
		index;
	int
		listArgc;
	const char
		**listArgv;
	bool
		*selectedElements;
	bool
		cancel,
		fail;

	fail=false;
	if(objc==3)
	{
		if(Tcl_SplitList(localInterpreter,Tcl_GetStringFromObj(objv[2],NULL),&listArgc,&listArgv)==TCL_OK)
		{
			if((selectedElements=(bool *)MNewPtrClr(sizeof(bool)*listArgc)))	// get array of "falses"
			{
				if(listArgc)
				{
					selectedElements[0]=true;	// select the first element if there is one
				}
				ClearErrorTrace();
				if(SimpleListBoxDialog(Tcl_GetStringFromObj(objv[1],NULL),listArgc,listArgv,selectedElements,&cancel))	// let user choose elements
				{
					if(!cancel)
					{
						for(index=0;(int)index<listArgc;index++)
						{
							if(selectedElements[index])
							{
								Tcl_AppendElement(localInterpreter,listArgv[index]);	// make a result that contains the selected elements
							}
						}
					}
					else
					{
						fail=true;
					}
					ClearUserAbort();				// if user managed to send break sequence during the dialog, we would rather not abort now!
				}
				else
				{
					Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
					fail=true;
				}
				MDisposePtr(selectedElements);
			}
			else
			{
				Tcl_AppendResult(localInterpreter,"Failed to allocate memory for list result",NULL);
				fail=true;
			}
			Tcl_Free((char *)listArgv);
		}
		else
		{
			// SplitList will leave a result if it fails
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"dialogTitle inputList");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_OpenDialog(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// create a get file(s) dialog
// if InitialText is not specified, then no initial text will be present in the dialog
// NOTE: since this can indirectly modify the current clipboard, the clipboard must not be busy
{
	bool
		cancel,
		fail;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];
	UINT32
		index;
	char
		**paths;

	fail=false;
	if(objc==2||objc==3)
	{
		if(objc==3)
		{
			strncpy(tempString,Tcl_GetStringFromObj(objv[2],NULL),TEMP_STRING_BUFFER_LENGTH-1);
			tempString[TEMP_STRING_BUFFER_LENGTH-1]='\0';
		}
		else
		{
			tempString[0]='\0';
		}
		if(CurrentClipboardNotBusy(localInterpreter))
		{
			ClearErrorTrace();
			if(OpenFileDialog(Tcl_GetStringFromObj(objv[1],NULL),&(tempString[0]),TEMP_STRING_BUFFER_LENGTH,&paths,&cancel))
			{
				if(!cancel)
				{
					index=0;
					while(paths[index])
					{
						Tcl_AppendElement(localInterpreter,paths[index++]);
					}
					FreeOpenFileDialogPaths(paths);
				}
				else
				{
					fail=true;
				}
				ClearUserAbort();				// if user managed to send break sequence during the dialog, we would rather not abort now!
			}
			else
			{
				Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"dialogTitle ?initialText?");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_SaveDialog(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// create a save file dialog
// if InitialText is not specified, then no initial text will be present in the dialog
// NOTE: since this can indirectly modify the current clipboard, the clipboard must not be busy
{
	bool
		cancel,
		fail;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	fail=false;
	if(objc==2||objc==3)
	{
		if(objc==3)
		{
			strncpy(tempString,Tcl_GetStringFromObj(objv[2],NULL),TEMP_STRING_BUFFER_LENGTH-1);
			tempString[TEMP_STRING_BUFFER_LENGTH-1]='\0';
		}
		else
		{
			tempString[0]='\0';
		}
		if(CurrentClipboardNotBusy(localInterpreter))
		{
			ClearErrorTrace();
			if(SaveFileDialog(Tcl_GetStringFromObj(objv[1],NULL),&(tempString[0]),TEMP_STRING_BUFFER_LENGTH,&cancel))
			{
				if(!cancel)
				{
					Tcl_AppendResult(localInterpreter,&(tempString[0]),NULL);
				}
				else
				{
					fail=true;
				}
				ClearUserAbort();				// if user managed to send break sequence during the dialog, we would rather not abort now!
			}
			else
			{
				Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"dialogTitle ?initialText?");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_PathDialog(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// create a path choose dialog
// if InitialText is not specified, then no initial text will be present in the dialog
// NOTE: since this can indirectly modify the current clipboard, the clipboard must not be busy
{
	bool
		cancel,
		fail;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	fail=false;
	if(objc==2||objc==3)
	{
		if(objc==3)
		{
			strncpy(tempString,Tcl_GetStringFromObj(objv[2],NULL),TEMP_STRING_BUFFER_LENGTH-1);
			tempString[TEMP_STRING_BUFFER_LENGTH-1]='\0';
		}
		else
		{
			tempString[0]='\0';
		}
		if(CurrentClipboardNotBusy(localInterpreter))
		{
			ClearErrorTrace();
			if(ChoosePathDialog(Tcl_GetStringFromObj(objv[1],NULL),&(tempString[0]),TEMP_STRING_BUFFER_LENGTH,&cancel))
			{
				if(!cancel)
				{
					Tcl_AppendResult(localInterpreter,&(tempString[0]),NULL);
				}
				else
				{
					fail=true;
				}
				ClearUserAbort();				// if user managed to send break sequence during the dialog, we would rather not abort now!
			}
			else
			{
				Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"dialogTitle ?initialText?");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_FontDialog(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// create a font selection dialog
// if InitialFont is not specified, then no initial text will be present in the dialog
// NOTE: since this can indirectly modify the current clipboard, the clipboard must not be busy
{
	bool
		cancel,
		fail;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	fail=false;
	if(objc==2||objc==3)
	{
		if(objc==3)
		{
			strncpy(tempString,Tcl_GetStringFromObj(objv[2],NULL),TEMP_STRING_BUFFER_LENGTH-1);
			tempString[TEMP_STRING_BUFFER_LENGTH-1]='\0';
		}
		else
		{
			tempString[0]='\0';
		}
		if(CurrentClipboardNotBusy(localInterpreter))
		{
			ClearErrorTrace();
			if(ChooseFontDialog(Tcl_GetStringFromObj(objv[1],NULL),&(tempString[0]),TEMP_STRING_BUFFER_LENGTH,&cancel))
			{
				if(!cancel)
				{
					Tcl_AppendResult(localInterpreter,&(tempString[0]),NULL);
				}
				else
				{
					fail=true;
				}
				ClearUserAbort();				// if user managed to send break sequence during the dialog, we would rather not abort now!
			}
			else
			{
				Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"dialogTitle ?initialFont?");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_CloseBuffer(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// close the passed buffer as long as it is not in use, or being modified
{
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				EditorCloseBuffer(buffer);
				return(TCL_OK);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_OpenBuffer(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// open the passed file into a buffer, return the buffer name
{
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		ClearErrorTrace();
		if((buffer=EditorOpenBuffer(Tcl_GetStringFromObj(objv[1],NULL))))
		{
			Tcl_AppendResult(localInterpreter,buffer->contentName,NULL);
			return(TCL_OK);
		}
		else
		{
			Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"pathName");
	}
	return(TCL_ERROR);
}

static int Cmd_NewBuffer(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// create a new buffer with the passed name
// if a buffer already exists with the passed name fail
{
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		ClearErrorTrace();
		if((buffer=EditorNewBuffer(Tcl_GetStringFromObj(objv[1],NULL))))				// create a new buffer
		{
			Tcl_AppendResult(localInterpreter,buffer->contentName,NULL);
			return(TCL_OK);
		}
		else
		{
			Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_BufferList(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// return the current list of document buffers
{
	bool
		fail;
	EDITOR_BUFFER
		*testBuffer;

	fail=false;
	if(objc==1)
	{
		testBuffer=EditorGetFirstBuffer();
		while(testBuffer)
		{
			Tcl_AppendElement(localInterpreter,testBuffer->contentName);
			testBuffer=EditorGetNextBuffer(testBuffer);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_SaveBuffer(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// save the passed buffer
{
	EDITOR_BUFFER
		*buffer;
	bool
		result;

	if(objc==2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				ClearErrorTrace();
				result=EditorSaveBuffer(buffer);
				if(result)
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_SaveBufferAs(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// save the passed buffer to newPath
// rename the buffer to the new name, set clean
// NOTE: this returns the new name of buffer after it is saved
{
	EDITOR_BUFFER
		*buffer;
	bool
		result;

	if(objc==3)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				ClearErrorTrace();
				result=EditorSaveBufferAs(buffer,Tcl_GetStringFromObj(objv[2],NULL));
				if(result)
				{
					Tcl_AppendResult(localInterpreter,buffer->contentName,NULL);
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName newPath");
	}
	return(TCL_ERROR);
}

static int Cmd_SaveBufferTo(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// save the passed buffer to a file, do not modify the buffer's "clean" status
{
	EDITOR_BUFFER
		*buffer;
	bool
		result;

	if(objc==3)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				ClearErrorTrace();
				result=EditorSaveBufferTo(buffer,Tcl_GetStringFromObj(objv[2],NULL));
				if(result)
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName pathName");
	}
	return(TCL_ERROR);
}

static int Cmd_RevertBuffer(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// revert the passed buffer
{
	EDITOR_BUFFER
		*buffer;
	bool
		result;

	if(objc==2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				ClearErrorTrace();
				result=EditorRevertBuffer(buffer);
				if(result)
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_BufferVariables(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// show all the variables linked to the passed buffer
{
	EDITOR_BUFFER
		*buffer;
	VARIABLE_BINDING
		*currentBinding;
	bool
		didOne;

	if(objc==2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			currentBinding=buffer->variableTable.variableBindingListHead;
			didOne=false;
			while(currentBinding)
			{
				if(didOne)				// do not put return on last line
				{
					Tcl_AppendResult(localInterpreter,"\n",NULL);
				}
				Tcl_AppendElement(localInterpreter,GetVariableBindingName(currentBinding));
				Tcl_AppendElement(localInterpreter,GetVariableBindingText(currentBinding));
				didOne=true;
				currentBinding=LocateNextVariableBinding(currentBinding);
			}
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_SetBufferVariable(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// set/remove a buffer relative variable
{
	EDITOR_BUFFER
		*buffer;
	char
		*variableName;

	if(objc==3||objc==4)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			variableName=Tcl_GetStringFromObj(objv[2],NULL);
			if(objc==3)
			{
				if(DeleteVariableBinding(&(buffer->variableTable),Tcl_GetStringFromObj(objv[2],NULL)))
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,"Failed to locate variable '",Tcl_GetStringFromObj(objv[2],NULL),"'",NULL);
				}
			}
			else
			{
				ClearErrorTrace();
				if(CreateVariableBinding(&(buffer->variableTable),variableName,Tcl_GetStringFromObj(objv[3],NULL)))
				{
					Tcl_AppendResult(localInterpreter,Tcl_GetStringFromObj(objv[3],NULL),NULL);
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName variableName ?variableText?");
	}
	return(TCL_ERROR);
}

static int Cmd_GetBufferVariable(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// get a variable hung off of a buffer
// these variables can be used by Tcl scripts to mark buffers, and associate
// arbitrary data with them
{
	EDITOR_BUFFER
		*buffer;
	VARIABLE_BINDING
		*bufferVariable;

	if(objc==3)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if((bufferVariable=LocateVariableBinding(&(buffer->variableTable),Tcl_GetStringFromObj(objv[2],NULL))))
			{
				Tcl_AppendResult(localInterpreter,GetVariableBindingText(bufferVariable),NULL);
				return(TCL_OK);
			}
			else
			{
				Tcl_AppendResult(localInterpreter,"Failed to locate variable '",Tcl_GetStringFromObj(objv[2],NULL),"'",NULL);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName variableName");
	}
	return(TCL_ERROR);
}

static int Cmd_BufferDirty(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// return 1 if the passed buffer is dirty, 0 if not
// if the buffer cannot be located, fail
{
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(AtUndoCleanPoint(buffer))
			{
				Tcl_AppendResult(localInterpreter,"0",NULL);
			}
			else
			{
				Tcl_AppendResult(localInterpreter,"1",NULL);
			}
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_FromFile(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// return 1 if the passed buffer is linked to a file, 0 if not
// if the buffer cannot be located, fail
{
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(buffer->fromFile)
			{
				Tcl_AppendResult(localInterpreter,"1",NULL);
			}
			else
			{
				Tcl_AppendResult(localInterpreter,"0",NULL);
			}
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_HasWindow(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// return 1 if the passed buffer has a window, 0 if not
// if the buffer cannot be located, fail
{
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(buffer->window)
			{
				Tcl_AppendResult(localInterpreter,"1",NULL);
			}
			else
			{
				Tcl_AppendResult(localInterpreter,"0",NULL);
			}
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_ClearDirty(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// clear the dirty flag for the passed buffer
// if the buffer cannot be located, fail
{
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				EditorStartTextChange(buffer);			// start text change, so that status bars can update if needed
				SetUndoCleanPoint(buffer);				// make it clean again
				EditorEndTextChange(buffer);			// change is complete
				return(TCL_OK);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_TextInfo(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// return the number of lines, and number of characters of text in the given buffer
{
	EDITOR_BUFFER
		*buffer;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				sprintf(tempString,"%u %u",buffer->textUniverse->totalLines,buffer->textUniverse->totalBytes);
				Tcl_AppendResult(localInterpreter,tempString,NULL);
				return(TCL_OK);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_PositionToLineOffset(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// given a position in a buffer, return a line number, and offset in characters into the line
// that matches the position
// NOTE: to be consistent with line numbering starting at 1, we fudge the number in here
{
	EDITOR_BUFFER
		*buffer;
	UINT32
		position,
		line,
		lineOffset;
	CHUNK_HEADER
		*chunk;
	UINT32
		offset;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==3)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(GetUINT32(localInterpreter,objv[2],&position))
			{
				if(ShellBufferNotBusy(localInterpreter,buffer))
				{
					PositionToLinePosition(buffer->textUniverse,position,&line,&lineOffset,&chunk,&offset);
					sprintf(tempString,"%u %u",line+1,lineOffset);
					Tcl_AppendResult(localInterpreter,tempString,NULL);
					return(TCL_OK);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName position");
	}
	return(TCL_ERROR);
}

static int Cmd_LineOffsetToPosition(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// given a line, and offset in a buffer, return a position in the buffer that matches
// NOTE: to be consistent with line numbering starting at 1, we fudge the number in here
{
	EDITOR_BUFFER
		*buffer;
	UINT32
		position,
		line,
		lineOffset,
		distanceMoved;
	CHUNK_HEADER
		*chunk;
	UINT32
		offset;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==4)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(GetUINT32(localInterpreter,objv[2],&line))
			{
				if(GetUINT32(localInterpreter,objv[3],&lineOffset))
				{
					if(ShellBufferNotBusy(localInterpreter,buffer))
					{
						if(line==0)
						{
							line=1;
						}
						LineToChunkPosition(buffer->textUniverse,line-1,&chunk,&offset,&position);	// get position that is the start of the given line
						ChunkPositionToNextLine(buffer->textUniverse,chunk,offset,&chunk,&offset,&distanceMoved);
						if(lineOffset>distanceMoved)		// see if offset is past end of line, if so, pin it
						{
							lineOffset=distanceMoved;
						}
						sprintf(tempString,"%u",position+lineOffset);
						Tcl_AppendResult(localInterpreter,tempString,NULL);
						return(TCL_OK);
					}
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName lineNumber offset");
	}
	return(TCL_ERROR);
}

static int Cmd_CheckBuffer(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// check the text universe on the given buffer, report the results
// NOTE: this is mainly for debugging, and draws its own dialog boxes
{
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				ClearErrorTrace();
				if(UniverseSanityCheck(buffer->textUniverse))
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_CloseWindow(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// close a given window
{
	EDITOR_WINDOW
		*window;
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		if((window=GetEditorWindow(localInterpreter,objv[1])))
		{
			buffer=window->buffer;								// hang onto this, we will need it after the window closes
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				EditorCloseDocumentWindow(window);
				return(TCL_OK);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName");
	}
	return(TCL_ERROR);
}

static int Cmd_OpenWindow(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// open a window onto the given buffer
// if there is already a window open on the buffer, just bring it to the top
{
	EDITOR_BUFFER
		*buffer;
	EDITOR_WINDOW
		*window;
	EDITOR_RECT
		rect;
	UINT32
		tabSize;
	EDITOR_COLOR
		foreground,
		background;

	if(objc==10)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(GetRectangle(localInterpreter,&(objv[2]),&rect))
			{
				if(GetUINT32(localInterpreter,objv[7],&tabSize))
				{
					if(GetColor(localInterpreter,objv[8],&foreground))
					{
						if(GetColor(localInterpreter,objv[9],&background))
						{
							if(ShellBufferNotBusy(localInterpreter,buffer))
							{
								ClearErrorTrace();
								window=EditorOpenDocumentWindow(buffer,&rect,Tcl_GetStringFromObj(objv[6],NULL),tabSize,foreground,background);
								if(window)
								{
									return(TCL_OK);
								}
								else
								{
									Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
								}
							}
						}
					}
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName x y w h fontName tabSize foregroundColor backgroundColor");
	}
	return(TCL_ERROR);
}

static int Cmd_WindowList(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// return the current list of document windows
{
	bool
		fail;
	UINT32
		windowIndex,
		numWindows;
	EDITOR_WINDOW
		**windowList;

	fail=false;
	if(objc==1)
	{
		if(GetSortedDocumentWindowList(&numWindows,&windowList))
		{
			for(windowIndex=0;windowIndex<numWindows;windowIndex++)
			{
				Tcl_AppendElement(localInterpreter,windowList[windowIndex]->buffer->contentName);
			}
			MDisposePtr(windowList);
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_GetTopLeft(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// get the top line, and lefthand coordinate of the active view in the given window
// NOTE: to be consistent with line numbering starting at 1, we fudge the number in here
{
	EDITOR_WINDOW
		*window;
	EDITOR_VIEW
		*view;
	UINT32
		topLine,
		numLines,
		numPixels;
	INT32
		leftPixel;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==2)
	{
		if((window=GetEditorWindow(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,window->buffer))
			{
				view=GetDocumentWindowCurrentView(window);
				GetEditorViewTextInfo(view,&topLine,&numLines,&leftPixel,&numPixels);
				sprintf(tempString,"%u %d",topLine+1,leftPixel);
				Tcl_AppendResult(localInterpreter,tempString,NULL);
				return(TCL_OK);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName");
	}
	return(TCL_ERROR);
}

static int Cmd_SetTopLeft(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// set the top line, and lefthand coordinate of the active view in the given window
// NOTE: to be consistent with line numbering starting at 1, we fudge the number in here
{
	EDITOR_WINDOW
		*window;
	EDITOR_VIEW
		*view;
	UINT32
		topLine;
	INT32
		leftPixel;

	if(objc==4)
	{
		if((window=GetEditorWindow(localInterpreter,objv[1])))
		{
			if(GetUINT32(localInterpreter,objv[2],&topLine))
			{
				if(GetINT32(localInterpreter,objv[3],&leftPixel))
				{
					if(ShellBufferNotBusy(localInterpreter,window->buffer))
					{
						view=GetDocumentWindowCurrentView(window);
						if(topLine==0)				// if line 0 was given, just make it line 1
						{
							topLine=1;
						}
						SetViewTopLeft(view,topLine-1,leftPixel);
						return(TCL_OK);
					}
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName topLine leftPixel");
	}
	return(TCL_ERROR);
}

static OPTION_DESCRIPTOR_RECORD
	homeOptions[]=
	{
		{"-lenient",		HT_LENIENT,		OptionEnum},
		{"-semistrict",		HT_SEMISTRICT,	OptionEnum},
		{"-strict",			HT_STRICT,		OptionEnum},
		{"-start",			false,			OptionBoolean},
		{"-end",			true,			OptionBoolean},
		{NULL,0,NULL}
	};

static int Cmd_HomeWindow(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// home a document view to a given position, in the given way
// by default, the mode is HT_SEMISTRICT, and useEnd is false
{
	EDITOR_WINDOW
		*window;
	EDITOR_VIEW
		*view;
	UINT32
		startPosition,
		endPosition;
	bool
		fail;
	int
		mode;
	bool
		useEnd;
	void
		*parsedVars[]=
		{
			(void *)&mode,
			(void *)&mode,
			(void *)&mode,
			(void *)&useEnd,
			(void *)&useEnd,
		};
	int
		optionIndex;

	fail=false;
	if(objc>=4)
	{
		if((window=GetEditorWindow(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,window->buffer))
			{
				if(GetUINT32(localInterpreter,objv[2],&startPosition))
				{
					if(GetUINT32(localInterpreter,objv[3],&endPosition))
					{
						startPosition=ForcePositionIntoRange(window->buffer,startPosition);
						endPosition=ForcePositionIntoRange(window->buffer,endPosition);
						ArrangePositions(&startPosition,&endPosition);
						optionIndex=4;

						mode=HT_SEMISTRICT;
						useEnd=false;

						if(ParseOptions(localInterpreter,homeOptions,parsedVars,objc,objv,&optionIndex,true))
						{
							view=GetDocumentWindowCurrentView(window);
							EditorHomeViewToRange(view,startPosition,endPosition,useEnd,mode,mode);
							ResetEditorViewCursorBlink(view);				// reset cursor blinking when homing view
						}
						else
						{
							fail=true;
						}

					}
					else
					{
						fail=true;
					}
				}
				else
				{
					fail=true;
				}
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName startPosition endPosition ?flags?");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_SetRect(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set the rectangle of a window (relative to the primary screen).
{
	EDITOR_WINDOW
		*window;
	EDITOR_RECT
		rect;
	bool
		result;

	if(objc==6)
	{
		if((window=GetEditorWindow(localInterpreter,objv[1])))
		{
			if(GetRectangle(localInterpreter,&(objv[2]),&rect))
			{
				if(ShellBufferNotBusy(localInterpreter,window->buffer))
				{
					ClearErrorTrace();
					result=SetEditorDocumentWindowRect(window,&rect);
					if(result)
					{
						return(TCL_OK);
					}
					else
					{
						Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
					}
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName x y w h");
	}
	return(TCL_ERROR);
}

static int Cmd_GetRect(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// get the rectangle of a window
{
	EDITOR_WINDOW
		*window;
	EDITOR_RECT
		rect;
	bool
		result;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==2)
	{
		if((window=GetEditorWindow(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,window->buffer))
			{
				ClearErrorTrace();
				result=GetEditorDocumentWindowRect(window,&rect);
				if(result)
				{
					sprintf(tempString,"%d %d %d %d",(int)rect.x,(int)rect.y,(int)rect.w,(int)rect.h);
					Tcl_AppendResult(localInterpreter,tempString,NULL);
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName");
	}
	return(TCL_ERROR);
}

static int Cmd_SetFont(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// set the font of a window
{
	EDITOR_WINDOW
		*window;
	EDITOR_VIEW
		*view;
	UINT32
		style;
	bool
		haveFont;
	bool
		result;
	char
		*fontName;

	if(objc==3||objc==4)
	{
		if((window=GetEditorWindow(localInterpreter,objv[1])))
		{
			fontName=Tcl_GetStringFromObj(objv[2],NULL);
			haveFont=fontName[0]!='\0';
			style=0;
			if((objc==3)||GetUINT32(localInterpreter,objv[3],&style))
			{
				if(ShellBufferNotBusy(localInterpreter,window->buffer))
				{
					result=true;
					view=GetDocumentWindowCurrentView(window);
					ClearErrorTrace();
					if(haveFont||style==0)
					{
						result=SetEditorViewStyleFont(view,style,fontName);
					}
					else
					{
						ClearEditorViewStyleFont(view,style);
					}
					if(result)
					{
						return(TCL_OK);
					}
					else
					{
						Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
					}
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName fontName ?styleNum?");
	}
	return(TCL_ERROR);
}

static int Cmd_GetFont(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// get the font of a window/style
{
	EDITOR_WINDOW
		*window;
	EDITOR_VIEW
		*view;
	UINT32
		style;
	bool
		result;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==2||objc==3)
	{
		if((window=GetEditorWindow(localInterpreter,objv[1])))
		{
			style=0;
			if((objc==2)||GetUINT32(localInterpreter,objv[2],&style))
			{
				if(ShellBufferNotBusy(localInterpreter,window->buffer))
				{
					view=GetDocumentWindowCurrentView(window);
					result=GetEditorViewStyleFont(view,style,&tempString[0],TEMP_STRING_BUFFER_LENGTH);
					if(result)
					{
						Tcl_AppendResult(localInterpreter,tempString,NULL);
						return(TCL_OK);
					}
					else
					{
						Tcl_AppendResult(localInterpreter,"Failed to get font",NULL);
					}
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName ?styleNum?");
	}
	return(TCL_ERROR);
}

static int Cmd_SetTabSize(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// set the tab size of a window
{
	EDITOR_WINDOW
		*window;
	EDITOR_VIEW
		*view;
	UINT32
		tabSize;
	bool
		result;

	if(objc==3)
	{
		if((window=GetEditorWindow(localInterpreter,objv[1])))
		{
			if(GetUINT32(localInterpreter,objv[2],&tabSize))
			{
				if(ShellBufferNotBusy(localInterpreter,window->buffer))
				{
					view=GetDocumentWindowCurrentView(window);
					ClearErrorTrace();
					result=SetEditorViewTabSize(view,tabSize);
					if(result)
					{
						return(TCL_OK);
					}
					else
					{
						Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
					}
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName tabSize");
	}
	return(TCL_ERROR);
}

static int Cmd_GetTabSize(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// get the tab size of a window
{
	EDITOR_WINDOW
		*window;
	EDITOR_VIEW
		*view;
	UINT32
		tabSize;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==2)
	{
		if((window=GetEditorWindow(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,window->buffer))
			{
				view=GetDocumentWindowCurrentView(window);
				tabSize=GetEditorViewTabSize(view);
				sprintf(tempString,"%d",(int)tabSize);
				Tcl_AppendResult(localInterpreter,tempString,NULL);
				return(TCL_OK);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName");
	}
	return(TCL_ERROR);
}

static int Cmd_SetColors(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// set the colors of a window
{
	EDITOR_WINDOW
		*window;
	EDITOR_VIEW
		*view;
	bool
		haveForeground,
		haveBackground;
	EDITOR_COLOR
		foregroundColor,
		backgroundColor;
	UINT32
		style;
	bool
		result;
	char
		*foregroundName,
		*backgroundName;

	if(objc==4||objc==5)
	{
		foregroundName=Tcl_GetStringFromObj(objv[2],NULL);
		backgroundName=Tcl_GetStringFromObj(objv[3],NULL);

		haveForeground=foregroundName[0]!='\0';	// if these strings are 0 length, they indicate that the color should be cleared
		haveBackground=backgroundName[0]!='\0';

		if((window=GetEditorWindow(localInterpreter,objv[1])))
		{
			style=0;
			if((objc==4)||GetUINT32(localInterpreter,objv[4],&style))
			{
				if(ShellBufferNotBusy(localInterpreter,window->buffer))
				{
					view=GetDocumentWindowCurrentView(window);
					result=true;
					if(haveForeground||style==0)
					{
						if(GetColor(localInterpreter,objv[2],&foregroundColor))
						{
							ClearErrorTrace();
							result=SetEditorViewStyleForegroundColor(view,style,foregroundColor);
							if(result&&(style==0))
							{
								result=SetEditorDocumentWindowForegroundColor(window,foregroundColor);
							}
							if(!result)
							{
								Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
							}
						}
						else
						{
							result=false;
						}
					}
					else
					{
						ClearEditorViewStyleForegroundColor(view,style);
					}
					if(result)
					{
						if(haveBackground||style==0)
						{
							if(GetColor(localInterpreter,objv[3],&backgroundColor))
							{
								ClearErrorTrace();
								result=SetEditorViewStyleBackgroundColor(view,style,backgroundColor);
								if(result&&(style==0))
								{
									result=SetEditorDocumentWindowBackgroundColor(window,backgroundColor);
								}
								if(!result)
								{
									Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
								}
							}
							else
							{
								result=false;
							}
						}
						else
						{
							ClearEditorViewStyleBackgroundColor(view,style);
						}
					}
					if(result)
					{
						return(TCL_OK);
					}
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName foregroundColor backgroundColor ?styleNum?");
	}
	return(TCL_ERROR);
}

static int Cmd_GetColors(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// get the colors of a window
{
	EDITOR_WINDOW
		*window;
	EDITOR_VIEW
		*view;
	bool
		haveForeground,
		haveBackground;
	EDITOR_COLOR
		foregroundColor,
		backgroundColor;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];
	UINT32
		style;

	if(objc==2||objc==3)
	{
		if((window=GetEditorWindow(localInterpreter,objv[1])))
		{
			style=0;
			if((objc==2)||GetUINT32(localInterpreter,objv[2],&style))
			{
				if(ShellBufferNotBusy(localInterpreter,window->buffer))
				{
					view=GetDocumentWindowCurrentView(window);
					haveForeground=GetEditorViewStyleForegroundColor(view,style,&foregroundColor);
					haveBackground=GetEditorViewStyleBackgroundColor(view,style,&backgroundColor);
					if(haveForeground)
					{
						sprintf(tempString,"%06X ",(int)foregroundColor);
						Tcl_AppendResult(localInterpreter,tempString,NULL);
					}
					else
					{
						Tcl_AppendResult(localInterpreter,"\"\" ",NULL);
					}
					if(haveBackground)
					{
						sprintf(tempString,"%06X",(int)backgroundColor);
						Tcl_AppendResult(localInterpreter,tempString,NULL);
					}
					else
					{
						Tcl_AppendResult(localInterpreter,"\"\"",NULL);
					}
					return(TCL_OK);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName ?styleNum?");
	}
	return(TCL_ERROR);
}

static int Cmd_SetTopWindow(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// make the passed window the top/active window
{
	EDITOR_WINDOW
		*window;

	if(objc==2)
	{
		if((window=GetEditorWindow(localInterpreter,objv[1])))
		{
			SetTopDocumentWindow(window);
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName");
	}
	return(TCL_ERROR);
}

static int Cmd_MinimizeWindow(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// attempt to minimize (in whatever way the current OS supports) the given document window
{
	EDITOR_WINDOW
		*window;

	if(objc==2)
	{
		if((window=GetEditorWindow(localInterpreter,objv[1])))
		{
			MinimizeDocumentWindow(window);
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName");
	}
	return(TCL_ERROR);
}

static int Cmd_UnminimizeWindow(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// attempt to un-minimize (in whatever way the current OS supports) the given document window
{
	EDITOR_WINDOW
		*window;

	if(objc==2)
	{
		if((window=GetEditorWindow(localInterpreter,objv[1])))
		{
			UnminimizeDocumentWindow(window);
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName");
	}
	return(TCL_ERROR);
}

static int Cmd_GetActiveWindow(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// return the active window name, or error if none
{
	EDITOR_WINDOW
		*window;

	if(objc==1)
	{
		if((window=GetActiveDocumentWindow()))
		{
			Tcl_AppendResult(localInterpreter,window->buffer->contentName,NULL);
			return(TCL_OK);
		}
		else
		{
			Tcl_AppendResult(localInterpreter,"No active window",NULL);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_UpdateWindows(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// update any window that has invalid areas
// on some implementations, this may do nothing
{
	if(objc==1)
	{
		UpdateEditorWindows();
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_Clear(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// perform a selection clear in the given buffer
{
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				ClearErrorTrace();
				if(EditorDeleteSelection(buffer,buffer->selectionUniverse))
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_Extract(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Returns the bytes given in a range of positions
// for a buffer.
{
	EDITOR_BUFFER
		*buffer;
	UINT32
		startPosition,
		endPosition;
	CHUNK_HEADER
		*chunk;
	UINT32
		offset;
	UINT8
		*tempBuffer;

	if(objc==4)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(GetUINT32(localInterpreter,objv[2],&startPosition))
			{
				if(GetUINT32(localInterpreter,objv[3],&endPosition))
				{
					startPosition=ForcePositionIntoRange(buffer,startPosition);
					endPosition=ForcePositionIntoRange(buffer,endPosition);
					ArrangePositions(&startPosition,&endPosition);

					if((tempBuffer=(UINT8 *)MNewPtr(endPosition-startPosition+1)))
					{
		 				PositionToChunkPosition(buffer->textUniverse,startPosition,&chunk,&offset);
		 				ExtractUniverseText(buffer->textUniverse,chunk,offset,tempBuffer,endPosition-startPosition,&chunk,&offset);
		 				tempBuffer[endPosition-startPosition]='\0';
						Tcl_AppendElement(localInterpreter,(char *)tempBuffer);
						MDisposePtr(tempBuffer);
						return(TCL_OK);
					}
					else
					{
						Tcl_AppendResult(localInterpreter,"Failed to allocate memory for result",NULL);
					}
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName startPosition endPosition");
	}
	return(TCL_ERROR);
}

static int Cmd_Insert(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// insert the given text into the current selection
// of the passed buffer
{
	EDITOR_BUFFER
		*buffer;
	char
		*dataToAdd;

	if(objc==3)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				dataToAdd=Tcl_GetStringFromObj(objv[2],NULL);
				ClearErrorTrace();
				if(EditorInsert(buffer,(UINT8 *)dataToAdd,strlen(dataToAdd)))
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName textToInsert");
	}
	return(TCL_ERROR);
}

static int Cmd_InsertFile(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// insert the given file into the current selection
// of the passed buffer
{
	EDITOR_BUFFER
		*buffer;
	bool
		result;

	if(objc==3)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				ClearErrorTrace();
				result=EditorInsertFile(buffer,Tcl_GetStringFromObj(objv[2],NULL));
				if(result)
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName pathName");
	}
	return(TCL_ERROR);
}

static int Cmd_AutoIndent(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// perform an autoindent on the passed buffer
{
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				ClearErrorTrace();
				if(EditorAutoIndent(buffer))
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_MoveCursor(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// move the cursor in the given window by the relative mode
{
	EDITOR_WINDOW
		*window;
	EDITOR_VIEW
		*view;
	int
		relativeMode;

	if(objc==3)
	{
		if((window=GetEditorWindow(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,window->buffer))
			{
				if(MatchToken(Tcl_GetStringFromObj(objv[2],NULL),cursorMovementTokens,&relativeMode))
				{
					view=GetDocumentWindowCurrentView(window);
					ClearErrorTrace();
					if(EditorMoveCursor(view,(UINT16)relativeMode))
					{
						ResetEditorViewCursorBlink(view);				// reset cursor blinking when done moving
						return(TCL_OK);
					}
					else
					{
						Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
					}
				}
				else
				{
					Tcl_AppendResult(localInterpreter,"Invalid relative mode",NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName relativeMode");
	}
	return(TCL_ERROR);
}

static int Cmd_Delete(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// delete the characters in the window between the current cursor position, and
// the given relative mode away, if there is a selection, it will be deleted instead
{
	EDITOR_WINDOW
		*window;
	EDITOR_VIEW
		*view;
	int
		relativeMode;

	if(objc==3)
	{
		if((window=GetEditorWindow(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,window->buffer))
			{
				if(MatchToken(Tcl_GetStringFromObj(objv[2],NULL),cursorMovementTokens,&relativeMode))
				{
					view=GetDocumentWindowCurrentView(window);
					ClearErrorTrace();
					if(EditorDelete(view,(UINT16)relativeMode))
					{
						return(TCL_OK);
					}
					else
					{
						Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
					}
				}
				else
				{
					Tcl_AppendResult(localInterpreter,"Invalid relative mode",NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName relativeMode");
	}
	return(TCL_ERROR);
}

static int Cmd_ExpandSelection(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// expand the selection in the given window in the given relative direction
{
	EDITOR_WINDOW
		*window;
	EDITOR_VIEW
		*view;
	int
		relativeMode;

	if(objc==3)
	{
		if((window=GetEditorWindow(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,window->buffer))
			{
				if(MatchToken(Tcl_GetStringFromObj(objv[2],NULL),cursorMovementTokens,&relativeMode))
				{
					view=GetDocumentWindowCurrentView(window);
					ClearErrorTrace();
					if(EditorExpandNormalSelection(view,(UINT16)relativeMode))
					{
						return(TCL_OK);
					}
					else
					{
						Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
					}
				}
				else
				{
					Tcl_AppendResult(localInterpreter,"Invalid relative mode",NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName relativeMode");
	}
	return(TCL_ERROR);
}

static int Cmd_ReduceSelection(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// reduce the selection in the given window in the given relative direction
{
	EDITOR_WINDOW
		*window;
	EDITOR_VIEW
		*view;
	int
		relativeMode;

	if(objc==3)
	{
		if((window=GetEditorWindow(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,window->buffer))
			{
				if(MatchToken(Tcl_GetStringFromObj(objv[2],NULL),cursorMovementTokens,&relativeMode))
				{
					view=GetDocumentWindowCurrentView(window);
					ClearErrorTrace();
					if(EditorReduceNormalSelection(view,(UINT16)relativeMode))
					{
						return(TCL_OK);
					}
					else
					{
						Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
					}
				}
				else
				{
					Tcl_AppendResult(localInterpreter,"Invalid relative mode",NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName relativeMode");
	}
	return(TCL_ERROR);
}

static int Cmd_SetSelectionColors(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// set the selection colors of a window
{
	EDITOR_WINDOW
		*window;
	EDITOR_VIEW
		*view;
	bool
		haveForeground,
		haveBackground;
	EDITOR_COLOR
		foregroundColor,
		backgroundColor;
	bool
		result;
	char
		*foregroundName,
		*backgroundName;

	if(objc==4)
	{
		foregroundName=Tcl_GetStringFromObj(objv[2],NULL);
		backgroundName=Tcl_GetStringFromObj(objv[3],NULL);

		haveForeground=foregroundName[0]!='\0';	// if these strings are 0 length, they indicate that the color should be cleared
		haveBackground=backgroundName[0]!='\0';

		if((window=GetEditorWindow(localInterpreter,objv[1])))
		{
			if(!haveForeground||GetColor(localInterpreter,objv[2],&foregroundColor))
			{
				if(!haveBackground||GetColor(localInterpreter,objv[3],&backgroundColor))
				{
					if(ShellBufferNotBusy(localInterpreter,window->buffer))
					{
						ClearErrorTrace();
						view=GetDocumentWindowCurrentView(window);
						result=true;
						if(haveForeground)
						{
							result=SetEditorViewSelectionForegroundColor(view,foregroundColor);
						}
						else
						{
							ClearEditorViewSelectionForegroundColor(view);
						}
						if(result)
						{
							if(haveBackground)
							{
								result=SetEditorViewSelectionBackgroundColor(view,backgroundColor);
							}
							else
							{
								ClearEditorViewSelectionBackgroundColor(view);
							}
						}
						if(result)
						{
							return(TCL_OK);
						}
						else
						{
							Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
						}
					}
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName foregroundColor backgroundColor");
	}
	return(TCL_ERROR);
}

static int Cmd_GetSelectionColors(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// get the selection colors of a window
{
	EDITOR_WINDOW
		*window;
	EDITOR_VIEW
		*view;
	bool
		haveForeground,
		haveBackground;
	EDITOR_COLOR
		foregroundColor,
		backgroundColor;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==2)
	{
		if((window=GetEditorWindow(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,window->buffer))
			{
				view=GetDocumentWindowCurrentView(window);
				haveForeground=GetEditorViewSelectionForegroundColor(view,&foregroundColor);
				haveBackground=GetEditorViewSelectionBackgroundColor(view,&backgroundColor);
				if(haveForeground)
				{
					sprintf(tempString,"%06X ",(int)foregroundColor);
					Tcl_AppendResult(localInterpreter,tempString,NULL);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,"\"\" ",NULL);
				}
				if(haveBackground)
				{
					sprintf(tempString,"%06X",(int)backgroundColor);
					Tcl_AppendResult(localInterpreter,tempString,NULL);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,"\"\"",NULL);
				}
				return(TCL_OK);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName");
	}
	return(TCL_ERROR);
}

static int Cmd_SelectionInfo(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// return information about the selection:
// startPosition,endPosition,startLine,endLine,startLinePosition,endLinePosition,totalSegments,totalSpan
// NOTE: to be consistent with line numbering starting at 1, we fudge the numbers in here
// ### obsolete -- build out of Tcl
{
	UINT32
		startPosition,						// selection information
		endPosition,
		startLine,
		endLine,
		startLinePosition,
		endLinePosition,
		totalSegments,
		totalSpan;
	EDITOR_BUFFER
		*buffer;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				EditorGetSelectionInfo(buffer,buffer->selectionUniverse,&startPosition,&endPosition,&startLine,&endLine,&startLinePosition,&endLinePosition,&totalSegments,&totalSpan);
				sprintf(tempString,"%u %u %u %u %u %u %u %u",startPosition,endPosition,startLine+1,endLine+1,startLinePosition,endLinePosition,totalSegments,totalSpan);
				Tcl_AppendResult(localInterpreter,tempString,NULL);
				return(TCL_OK);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_SelectAll(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// perform a select all in the given buffer
// ### obsolete -- build out of Tcl
{
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				ClearErrorTrace();
				if(EditorSelectAll(buffer))
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_SelectedTextList(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// return the currently selected text as a list -- one element per selection
{
	EDITOR_BUFFER
		*buffer;
	UINT32
		currentPosition;
	UINT32
		actualLength;
	UINT8
		*selectionBuffer;
	bool
		done;
	bool
		fail;

	fail=false;
	if(objc==2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				currentPosition=0;
				done=false;
				while(!done&&!fail)
				{
					ClearErrorTrace();
					if((selectionBuffer=EditorNextSelectionToBuffer(buffer,buffer->selectionUniverse,&currentPosition,1,&actualLength,&done)))
					{
						selectionBuffer[actualLength-1]='\0';					// terminate the string
						Tcl_AppendElement(localInterpreter,(char *)selectionBuffer);
						MDisposePtr(selectionBuffer);
					}
					else
					{
						if(!done)
						{
							Tcl_ResetResult(localInterpreter);
							Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
							fail=true;
						}
					}
				}
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_SelectLine(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// locate the given line, and select it
// if lineNumber is out of range, go to the nearest one
// ### obsolete -- build out of Tcl
{
	EDITOR_BUFFER
		*buffer;
	UINT32
		line;

	if(objc==3)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				if(GetUINT32(localInterpreter,objv[2],&line))
				{
					if(line==0)		// if user enters 0, just take it as 1
					{
						line=1;
					}
					ClearErrorTrace();
					if(EditorLocateLine(buffer,line-1))
					{
						return(TCL_OK);
					}
					else
					{
						Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
					}
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName lineNumber");
	}
	return(TCL_ERROR);
}

static int Cmd_GetSelectionEnds(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// return the start and end of the current selection
// of the passed buffer
// if there is no selection, the cursor position is returned in both values
{
	EDITOR_BUFFER
		*buffer;
	UINT32
		startPosition,
		endPosition;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				GetSelectionEndPositions(buffer->selectionUniverse,&startPosition,&endPosition);
				sprintf(tempString,"%u %u",startPosition,endPosition);
				Tcl_AppendResult(localInterpreter,tempString,NULL);
				return(TCL_OK);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_SetSelectionEnds(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// set the selection of the passed buffer
// if the given end position is < the start position,
// the positions will be reversed
// if any position is out of range, it will be forced in range
{
	EDITOR_BUFFER
		*buffer;
	UINT32
		startPosition,
		endPosition;

	if(objc==4)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				if(GetUINT32(localInterpreter,objv[2],&startPosition))
				{
					if(GetUINT32(localInterpreter,objv[3],&endPosition))
					{
						startPosition=ForcePositionIntoRange(buffer,startPosition);
						endPosition=ForcePositionIntoRange(buffer,endPosition);
						ArrangePositions(&startPosition,&endPosition);
						ClearErrorTrace();
						if(EditorSetNormalSelection(buffer,buffer->selectionUniverse,startPosition,endPosition))
						{
							return(TCL_OK);
						}
						else
						{
							Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
						}
					}
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName startPosition endPosition");
	}
	return(TCL_ERROR);
}

static int Cmd_GetSelectionEndList(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// return offsets of the ends of each selection segment
// if there is no selection, an empty list is returned
{
	EDITOR_BUFFER
		*buffer;
	UINT32
		currentPosition,
		currentLength;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				currentPosition=0;
				while(GetSelectionAtOrAfterPosition(buffer->selectionUniverse,currentPosition,&currentPosition,&currentLength))
				{
					sprintf(tempString,"%u %u",currentPosition,currentPosition+currentLength);
					Tcl_AppendElement(localInterpreter,tempString);
					currentPosition+=currentLength;
				}
				return(TCL_OK);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_AddSelectionEndList(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// add list of ends given to the selection
// if any position is out of range, it will be forced in range
// if there is any overlap, the new segment boundaries will remain
{
	EDITOR_BUFFER
		*buffer;
	int
		listArgc;
	const char
		**listArgv;
	int
		elementArgc;
	const char
		**elementArgv;
	UINT32
		startPosition,
		endPosition;
	int
		i;
	bool
		fail;

	if(objc==3)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				if(Tcl_SplitList(localInterpreter,Tcl_GetStringFromObj(objv[2],NULL),&listArgc,&listArgv)==TCL_OK)
				{
					EditorStartSelectionChange(buffer);
					fail=false;
					for(i=0;!fail&&i<listArgc;i++)
					{
						if(Tcl_SplitList(localInterpreter,listArgv[i],&elementArgc,&elementArgv)==TCL_OK)
						{
							if(elementArgc==2)
							{
								if(GetUINT32String(localInterpreter,elementArgv[0],&startPosition))
								{
									if(GetUINT32String(localInterpreter,elementArgv[1],&endPosition))
									{
										startPosition=ForcePositionIntoRange(buffer,startPosition);
										endPosition=ForcePositionIntoRange(buffer,endPosition);
										ArrangePositions(&startPosition,&endPosition);
										if(endPosition>startPosition)	// make sure there is at least something to select
										{
											if(!SetSelectionRange(buffer->selectionUniverse,startPosition,endPosition-startPosition))
											{
												Tcl_AppendResult(localInterpreter,"Failed to set selection",NULL);
												fail=true;
											}
										}
									}
									else
									{
										fail=true;
									}
								}
								else
								{
									fail=true;
								}
							}
							else
							{
								Tcl_AppendResult(localInterpreter,"Exactly two elements are allowed per list entry",NULL);
								fail=true;
							}
							Tcl_Free((char *)elementArgv);
						}
						else
						{
							fail=true;
						}
					}
					EditorEndSelectionChange(buffer);
					Tcl_Free((char *)listArgv);
					if(!fail)
					{
						return(TCL_OK);
					}
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName selectionList");
	}
	return(TCL_ERROR);
}

static int Cmd_GetSelectionAtPosition(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// return information about the selection at the given character position in the text
// The result is a value (0 if no selection, 1 if selection) and the position <= the given
// one where the current selection begins, and the position just after the end of the
// selection
{
	EDITOR_BUFFER
		*buffer;
	UINT32
		position;
	UINT32
		startPosition,
		numElements;
	bool
		isSelected;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==3)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(GetUINT32(localInterpreter,objv[2],&position))
			{
				position=ForcePositionIntoRange(buffer,position);
				if(!GetSelectionRange(buffer->selectionUniverse,position,&startPosition,&numElements,&isSelected))
				{
					numElements=buffer->textUniverse->totalBytes-startPosition;
				}
				sprintf(tempString,"%u %u %u",isSelected?1:0,startPosition,startPosition+numElements);
				Tcl_AppendElement(localInterpreter,tempString);
				return(TCL_OK);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName position");
	}
	return(TCL_ERROR);
}

static OPTION_DESCRIPTOR_RECORD
	markOptions[]=
	{
		{"-sourcemark",		0,		OptionMark},
		{"-destmark",		0,		OptionMark},
		{NULL,0,NULL}
	};

static int Cmd_CopySelection(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Copy selection data from one selection to another
{
	EDITOR_BUFFER
		*buffer;
	PARSE_MARK
		sourceMark,
		destMark;
	SELECTION_UNIVERSE
		*sourceSelection,
		*destSelection;
	void
		*parsedVars[]=
		{
			(void *)&sourceMark,
			(void *)&destMark,
		};
	int
		optionIndex;
	bool
		fail;

	fail=false;
	if(objc>=2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				optionIndex=2;

				sourceMark.buffer=buffer;
				sourceMark.mark=NULL;

				destMark.buffer=buffer;
				destMark.mark=NULL;

				if(ParseOptions(localInterpreter,markOptions,parsedVars,objc,objv,&optionIndex,true))
				{
					sourceSelection=sourceMark.mark?sourceMark.mark->selectionUniverse:buffer->selectionUniverse;
					destSelection=destMark.mark?destMark.mark->selectionUniverse:buffer->selectionUniverse;

					if(destMark.mark)			// work out what's changing
					{
						ViewsStartMarkChange(destMark.mark);
					}
					else
					{
						EditorStartSelectionChange(buffer);
					}
					CopySelectionUniverse(sourceSelection,destSelection);	// perform the operation
					if(destMark.mark)			// work out what's changed
					{
						ViewsEndMarkChange(destMark.mark);
					}
					else
					{
						EditorEndSelectionChange(buffer);
					}
				}
				else
				{
					fail=true;
				}
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"buffer ?flags?");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_AndSelection(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Logical AND selection data from one selection to another
{
	EDITOR_BUFFER
		*buffer;
	PARSE_MARK
		sourceMark,
		destMark;
	SELECTION_UNIVERSE
		*sourceSelection,
		*destSelection;
	void
		*parsedVars[]=
		{
			(void *)&sourceMark,
			(void *)&destMark,
		};
	int
		optionIndex;
	bool
		fail;

	fail=false;
	if(objc>=2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				optionIndex=2;

				sourceMark.buffer=buffer;
				sourceMark.mark=NULL;

				destMark.buffer=buffer;
				destMark.mark=NULL;

				if(ParseOptions(localInterpreter,markOptions,parsedVars,objc,objv,&optionIndex,true))
				{
					sourceSelection=sourceMark.mark?sourceMark.mark->selectionUniverse:buffer->selectionUniverse;
					destSelection=destMark.mark?destMark.mark->selectionUniverse:buffer->selectionUniverse;

					if(destMark.mark)			// work out what's changing
					{
						ViewsStartMarkChange(destMark.mark);
					}
					else
					{
						EditorStartSelectionChange(buffer);
					}
					AndSelectionUniverse(sourceSelection,destSelection);	// perform the operation
					if(destMark.mark)			// work out what's changed
					{
						ViewsEndMarkChange(destMark.mark);
					}
					else
					{
						EditorEndSelectionChange(buffer);
					}
				}
				else
				{
					fail=true;
				}
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"buffer ?flags?");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_OrSelection(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Logical OR selection data from one selection to another
{
	EDITOR_BUFFER
		*buffer;
	PARSE_MARK
		sourceMark,
		destMark;
	SELECTION_UNIVERSE
		*sourceSelection,
		*destSelection;
	void
		*parsedVars[]=
		{
			(void *)&sourceMark,
			(void *)&destMark,
		};
	int
		optionIndex;
	bool
		fail;

	fail=false;
	if(objc>=2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				optionIndex=2;

				sourceMark.buffer=buffer;
				sourceMark.mark=NULL;

				destMark.buffer=buffer;
				destMark.mark=NULL;

				if(ParseOptions(localInterpreter,markOptions,parsedVars,objc,objv,&optionIndex,true))
				{
					sourceSelection=sourceMark.mark?sourceMark.mark->selectionUniverse:buffer->selectionUniverse;
					destSelection=destMark.mark?destMark.mark->selectionUniverse:buffer->selectionUniverse;

					if(destMark.mark)			// work out what's changing
					{
						ViewsStartMarkChange(destMark.mark);
					}
					else
					{
						EditorStartSelectionChange(buffer);
					}
					OrSelectionUniverse(sourceSelection,destSelection);	// perform the operation
					if(destMark.mark)			// work out what's changed
					{
						ViewsEndMarkChange(destMark.mark);
					}
					else
					{
						EditorEndSelectionChange(buffer);
					}
				}
				else
				{
					fail=true;
				}
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"buffer ?flags?");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_XorSelection(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Logical OR selection data from one selection to another
{
	EDITOR_BUFFER
		*buffer;
	PARSE_MARK
		sourceMark,
		destMark;
	SELECTION_UNIVERSE
		*sourceSelection,
		*destSelection;
	void
		*parsedVars[]=
		{
			(void *)&sourceMark,
			(void *)&destMark,
		};
	int
		optionIndex;
	bool
		fail;

	fail=false;
	if(objc>=2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				optionIndex=2;

				sourceMark.buffer=buffer;
				sourceMark.mark=NULL;

				destMark.buffer=buffer;
				destMark.mark=NULL;

				if(ParseOptions(localInterpreter,markOptions,parsedVars,objc,objv,&optionIndex,true))
				{
					sourceSelection=sourceMark.mark?sourceMark.mark->selectionUniverse:buffer->selectionUniverse;
					destSelection=destMark.mark?destMark.mark->selectionUniverse:buffer->selectionUniverse;

					if(destMark.mark)			// work out what's changing
					{
						ViewsStartMarkChange(destMark.mark);
					}
					else
					{
						EditorStartSelectionChange(buffer);
					}
					XorSelectionUniverse(sourceSelection,destSelection);	// perform the operation
					if(destMark.mark)			// work out what's changed
					{
						ViewsEndMarkChange(destMark.mark);
					}
					else
					{
						EditorEndSelectionChange(buffer);
					}
				}
				else
				{
					fail=true;
				}
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"buffer ?flags?");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_MarkInvisible(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Make a mark invisible within a window
{
	EDITOR_WINDOW
		*window;
	EDITOR_MARK
		*mark;
	EDITOR_VIEW
		*view;

	if(objc==3)
	{
		if((window=GetEditorWindow(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,window->buffer))
			{
				if((mark=LocateEditorMark(window->buffer,Tcl_GetStringFromObj(objv[2],NULL))))
				{
					view=GetDocumentWindowCurrentView(window);
					ClearEditorMarkVisible(mark,view);
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,"Failed to locate mark '",Tcl_GetStringFromObj(objv[2],NULL),"'",NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName markName");
	}
	return(TCL_ERROR);
}

static int Cmd_MarkVisible(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Make a mark visible in a window
{
	EDITOR_WINDOW
		*window;
	EDITOR_MARK
		*mark;
	EDITOR_VIEW
		*view;
	EDITOR_COLOR
		background;

	if(objc==4)
	{
		if((window=GetEditorWindow(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,window->buffer))
			{
				if((mark=LocateEditorMark(window->buffer,Tcl_GetStringFromObj(objv[2],NULL))))
				{
					if(GetColor(localInterpreter,objv[3],&background))
					{
						view=GetDocumentWindowCurrentView(window);
						SetEditorMarkVisible(mark,view,background);
						return(TCL_OK);
					}
				}
				else
				{
					Tcl_AppendResult(localInterpreter,"Failed to locate mark '",Tcl_GetStringFromObj(objv[2],NULL),"'",NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName markName backgroundColor");
	}
	return(TCL_ERROR);
}

static int Cmd_MarkList(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// return a list of marks for the given buffer
{
	EDITOR_BUFFER
		*buffer;
	EDITOR_MARK
		*mark;

	if(objc==2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			mark=buffer->marks;
			while(mark)
			{
				if(mark->markName)
				{
					Tcl_AppendElement(localInterpreter,mark->markName);
				}
				mark=mark->nextMark;
			}
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_CloseMark(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Remove a mark from a buffer
{
	EDITOR_BUFFER
		*buffer;
	EDITOR_MARK
		*mark;

	if(objc==3)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				if((mark=LocateEditorMark(buffer,Tcl_GetStringFromObj(objv[2],NULL))))
				{
					CloseEditorMark(buffer,mark);
				}
				if(mark)
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,"Failed to locate mark '",Tcl_GetStringFromObj(objv[2],NULL),"'",NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName markName");
	}
	return(TCL_ERROR);
}

static int Cmd_NewMark(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// create a new empty mark in the given buffer
{
	EDITOR_BUFFER
		*buffer;
	EDITOR_MARK
		*mark;

	if(objc==3)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				ClearErrorTrace();
				mark=NewEditorMark(buffer,Tcl_GetStringFromObj(objv[2],NULL));
				if(mark)
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName markName");
	}
	return(TCL_ERROR);
}

static int Cmd_GetStyleAtPosition(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// return information about the style at the given character position in the text
// The result is a value (style index), the position <= the given
// one where the style begins, and the position just after the end of the
// style
{
	EDITOR_BUFFER
		*buffer;
	UINT32
		position;
	UINT32
		startPosition,
		numElements;
	UINT32
		style;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==3)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(GetUINT32(localInterpreter,objv[2],&position))
			{
				position=ForcePositionIntoRange(buffer,position);
				if(!GetStyleRange(buffer->styleUniverse,position,&startPosition,&numElements,&style))
				{
					numElements=buffer->textUniverse->totalBytes-startPosition;
				}
				sprintf(tempString,"%u %u %u",style,startPosition,startPosition+numElements);
				Tcl_AppendElement(localInterpreter,tempString);
				return(TCL_OK);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName position");
	}
	return(TCL_ERROR);
}

static int Cmd_SetStyle(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// set the style of text of the passed buffer
// if the given end position is < the start position,
// the positions will be reversed
// if any position is out of range, it will be forced in range
{
	EDITOR_BUFFER
		*buffer;
	UINT32
		startPosition,
		endPosition;
	UINT32
		style;

	if(objc==5)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				if(GetUINT32(localInterpreter,objv[2],&startPosition))
				{
					if(GetUINT32(localInterpreter,objv[3],&endPosition))
					{
						if(GetUINT32(localInterpreter,objv[4],&style))
						{
							startPosition=ForcePositionIntoRange(buffer,startPosition);
							endPosition=ForcePositionIntoRange(buffer,endPosition);
							ArrangePositions(&startPosition,&endPosition);
							ViewsStartStyleChange(buffer);
							SetStyleRange(buffer->styleUniverse,startPosition,endPosition-startPosition,style);
							ViewsEndStyleChange(buffer);
							return(TCL_OK);
						}
					}
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName startPosition endPosition styleNum");
	}
	return(TCL_ERROR);
}

static int Cmd_SelectionToStyle(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set the style of a buffer to a given value anywhere text is selected
{
	EDITOR_BUFFER
		*buffer;
	UINT32
		style;
	bool
		fail;
	UINT32
		startPosition,
		numElements;

	if(objc==3)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				if(GetUINT32(localInterpreter,objv[2],&style))
				{
					ViewsStartStyleChange(buffer);
					fail=false;

					startPosition=0;
					while(!fail&&GetSelectionAtOrAfterPosition(buffer->selectionUniverse,startPosition,&startPosition,&numElements))
					{
						fail=!SetStyleRange(buffer->styleUniverse,startPosition,numElements,style);
						startPosition+=numElements;
					}

					ViewsEndStyleChange(buffer);
					return(TCL_OK);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName styleNum");
	}
	return(TCL_ERROR);
}

static int Cmd_StyleToSelection(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set the selection in a buffer everywhere a given style is in the buffer
// NOTE: this does not clear the old selection, it will just add to it
{
	EDITOR_BUFFER
		*buffer;
	UINT32
		style;
	bool
		fail;
	UINT32
		startPosition,
		numElements;
	UINT32
		sawStyle;

	if(objc==3)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				if(GetUINT32(localInterpreter,objv[2],&style))
				{
					EditorStartSelectionChange(buffer);

					fail=false;
					startPosition=0;
					while(!fail&&GetStyleRange(buffer->styleUniverse,startPosition,&startPosition,&numElements,&sawStyle))
					{
						if(sawStyle==style)
						{
							fail=!SetSelectionRange(buffer->selectionUniverse,startPosition,numElements);
						}
						startPosition+=numElements;
					}
					if(!fail)	// handle the last element if it happens to be in the default style
					{
						if(sawStyle==style)
						{
							fail=!SetSelectionRange(buffer->selectionUniverse,startPosition,buffer->textUniverse->totalBytes-startPosition);
						}
					}

					EditorEndSelectionChange(buffer);
					return(TCL_OK);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName styleNum");
	}
	return(TCL_ERROR);
}

static void DisposeSplitExpressionName(char *name)
// get rid of a split expression name created by SplitExpressionName
{
	MDisposePtr(name);
}

static char *SplitExpressionName(const char *name,bool *useRegisterMatch,UINT32 *registerIndex)
// Split name into a name, and information about any register match.
// Expression names may be specified as just a name, or a name followed by a ':'
// and then a register number.
// NOTE: if this fails, it will SetError, and return NULL
// NOTE: if this succeeds, the returned name should be disposed of by calling
// DisposeSplitExpressionName
{
	int
		length;
	char
		*newName;

	*useRegisterMatch=false;

	length=strlen(name);
	if((newName=(char *)MNewPtr(length+1)))		// +1 to get space for termination
	{
		strcpy(newName,name);					// copy it over
		if(length>=2)
		{
			if(newName[length-2]==':'&&newName[length-1]>='0'&&newName[length-1]<='9')	// see if indication of register specification
			{
				*useRegisterMatch=true;
				*registerIndex=newName[length-1]-'0';
				newName[length-2]='\0';	// terminate
			}
		}
	}
	return(newName);
}

static EXPRESSION_PIECE *GetExpressionPiece(Tcl_Interp *localInterpreter,SYNTAX_MAP *map,const char *name)
// See if an expression piece of the given name can be added to the list
// If there is a problem, add it to the Tcl result, and return NULL
{
	char
		*newName;
	bool
		useRegisterMatch;
	UINT32
		registerIndex;
	SYNTAX_EXPRESSION
		*expression;
	EXPRESSION_PIECE
		*piece;

	piece=NULL;
	registerIndex=0;		// keep GCC quiet
	if((newName=SplitExpressionName(name,&useRegisterMatch,&registerIndex)))
	{
		if((expression=LocateSyntaxMapExpression(map,newName)))
		{
			ClearErrorTrace();
			if(!(piece=AddPieceToExpression(expression,useRegisterMatch,registerIndex)))
			{
				Tcl_AppendResult(localInterpreter,"Failed to add piece to expression '",newName,"': ",GetErrorTrace(),NULL);
			}
		}
		else
		{
			Tcl_AppendResult(localInterpreter,"Failed to locate expression '",newName,"'",NULL);
		}
		DisposeSplitExpressionName(newName);
	}
	else
	{
		Tcl_AppendResult(localInterpreter,"Failed to split name '",name,"'",NULL);
	}
	return(piece);
}

static bool ParseEndExpressions(Tcl_Interp *localInterpreter,SYNTAX_MAP *map,SYNTAX_STYLE_MAPPING *mapping,const char *endExpressionList)
// parse up the ending expressions, and add them to mapping
// If there is a problem, add it to the Tcl result, and return false
{
	EXPRESSION_PIECE
		*expressionPiece;
	int
		argc;
	const char
		**argv;
	int
		i;
	bool
		fail;

	fail=false;
	if(Tcl_SplitList(localInterpreter,endExpressionList,&argc,&argv)==TCL_OK)	// get list of end expressions
	{
		for(i=0;!fail&&i<argc;i++)
		{
			if((expressionPiece=GetExpressionPiece(localInterpreter,map,argv[i])))
			{
				ClearErrorTrace();
				if(AddMappingEndExpressionPiece(mapping,expressionPiece))
				{
				}
				else
				{
					Tcl_AppendResult(localInterpreter,"Failed to end expression '",argv[i],"' to mapping: ",GetErrorTrace(),NULL);
					fail=true;
				}
			}
			else
			{
				fail=true;
			}
		}
		Tcl_Free((char *)argv);
	}
	else
	{
		fail=true;				// SplitList will leave a result if it fails
	}
	return(!fail);
}

static bool ParseMapping(Tcl_Interp *localInterpreter,SYNTAX_MAP *map,int argc,const char *argv[])
// Use the arguments to build a mapping onto map
// If there is a problem, add it to the Tcl result, and return false
{
	SYNTAX_STYLE_MAPPING
		*mapping;
	EXPRESSION_PIECE
		*expressionPiece;
	UINT32
		startStyle,
		betweenStyle,
		endStyle;

	ClearErrorTrace();
	if((mapping=AddMappingToSyntaxMap(map,argv[0])))
	{
		if((expressionPiece=GetExpressionPiece(localInterpreter,map,argv[1])))
		{
			SetMappingStartExpressionPiece(mapping,expressionPiece);
			if(ParseEndExpressions(localInterpreter,map,mapping,argv[2]))
			{
				if(GetUINT32String(localInterpreter,argv[3],&startStyle))
				{
					if(GetUINT32String(localInterpreter,argv[4],&betweenStyle))
					{
						if(GetUINT32String(localInterpreter,argv[5],&endStyle))
						{
							SetMappingStyles(mapping,startStyle,betweenStyle,endStyle);
							return(true);
						}
					}
				}
			}
		}
		RemoveMappingFromSyntaxMap(map,mapping);
	}
	else
	{
		Tcl_AppendResult(localInterpreter,"Failed to add mapping '",argv[0],"' to map: ",GetErrorTrace(),NULL);
	}
	return(false);
}

static bool ParseSyntaxMap(Tcl_Interp *localInterpreter,SYNTAX_MAP *map,char *mapContents)
// Parse the mapContents, and use it to build map
// If there is a problem, add it to the Tcl result, and return false
{
	bool
		fail;
	int
		listArgc;
	const char
		**listArgv;
	int
		argc;
	const char
		**argv;
	int
		mappinglistArgc;
	const char
		**mappinglistArgv;
	int
		syntaxMapToken;
	SYNTAX_STYLE_MAPPING
		*mapping,
		*otherMapping;
	int
		i,j;

	fail=false;
	if(Tcl_SplitList(localInterpreter,mapContents,&listArgc,&listArgv)==TCL_OK)	// make list of command/parameter
	{
		for(i=0;(i<listArgc)&&!fail;i+=2)					// step over each command, parameter list pair
		{
			if(MatchToken(listArgv[i],syntaxMapCommands,&syntaxMapToken))
			{
				if(Tcl_SplitList(localInterpreter,listArgv[i+1],&argc,&argv)==TCL_OK)	// make parameters into arguments
				{
					switch(syntaxMapToken)
					{
						case SM_EXPRESSION:
							if(argc==2)
							{
								ClearErrorTrace();
								if(AddExpressionToSyntaxMap(map,argv[0],argv[1]))
								{
								}
								else
								{
									Tcl_AppendResult(localInterpreter,"Failed to add expression '",argv[0],"' to map: ",GetErrorTrace(),NULL);
									fail=true;
								}
							}
							else
							{
								Tcl_AppendResult(localInterpreter,"wrong # args: should be \"",listArgv[i]," expressionName expression\"",NULL);
								fail=true;
							}
							break;
						case SM_MAPPING:
							if(argc==6)
							{
								fail=!ParseMapping(localInterpreter,map,argc,argv);
							}
							else
							{
								Tcl_AppendResult(localInterpreter,"wrong # args: should be \"",listArgv[i]," mapName startExpressionName endExpressionList startStyle betweenStyle endStyle\"",NULL);
								fail=true;
							}
							break;
						case SM_AT:
							if(argc==2)
							{
								if((mapping=LocateSyntaxStyleMapping(map,argv[0])))
								{
									if(Tcl_SplitList(localInterpreter,argv[1],&mappinglistArgc,&mappinglistArgv)==TCL_OK)	// make parameters into arguments
									{
										for(j=0;j<mappinglistArgc&&!fail;j++)
										{
											if((otherMapping=LocateSyntaxStyleMapping(map,mappinglistArgv[j])))
											{
												ClearErrorTrace();
												if(AddBetweenListElementToMapping(mapping,otherMapping))
												{
												}
												else
												{
													Tcl_AppendResult(localInterpreter,"Failed to add mapping '",mappinglistArgv[j],"': ",GetErrorTrace(),NULL);
													fail=true;
												}
											}
											else
											{
												Tcl_AppendResult(localInterpreter,"Failed to locate mapping '",mappinglistArgv[j],"'",NULL);
												fail=true;
											}
										}
										Tcl_Free((char *)mappinglistArgv);
									}
									else
									{
										fail=true;
									}
								}
								else
								{
									Tcl_AppendResult(localInterpreter,"Failed to locate mapping '",argv[0],"'",NULL);
									fail=true;
								}
							}
							else
							{
								Tcl_AppendResult(localInterpreter,"wrong # args: should be \"",listArgv[i]," mapName betweenList\"",NULL);
								fail=true;
							}
							break;
						case SM_ROOT:
							for(j=0;j<argc&&!fail;j++)
							{
								if((otherMapping=LocateSyntaxStyleMapping(map,argv[j])))
								{
									ClearErrorTrace();
									if(AddRootListElementToMap(map,otherMapping))
									{
									}
									else
									{
										Tcl_AppendResult(localInterpreter,"Failed to add root mapping '",argv[j],"': ",GetErrorTrace(),NULL);
										fail=true;
									}
								}
								else
								{
									Tcl_AppendResult(localInterpreter,"Failed to locate mapping '",argv[j],"'",NULL);
									fail=true;
								}
							}
							break;
					}
					Tcl_Free((char *)argv);
				}
				else
				{
					fail=true;		// SplitList will leave a result if it fails
				}
			}
			else
			{
				Tcl_AppendResult(localInterpreter,"Invalid syntax map command '",listArgv[i],"'",NULL);
				fail=true;
			}
		}
		Tcl_Free((char *)listArgv);
	}
	else
	{
		fail=true;		// SplitList will leave a result if it fails
	}
	return(!fail);
}

static int Cmd_AddSyntaxMap(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// create a syntax highlighting map from a list of expressions and a list of mappings from expressions to styles
// NOTE: this can be used to replace an existing syntax map, but
// if the existing map is in use by any buffer which is busy, this will fail
// before it does anything
{
	EDITOR_BUFFER
		*buffer,
		*nextBuffer;
	SYNTAX_MAP
		*oldMap,
		*map;
	bool
		fail;

	if(objc==3)
	{
		ClearErrorTrace();
		if((map=OpenSyntaxMap(Tcl_GetStringFromObj(objv[1],NULL))))
		{
			if(ParseSyntaxMap(localInterpreter,map,Tcl_GetStringFromObj(objv[2],NULL)))
			{
				fail=false;
				if((oldMap=LocateSyntaxMap(Tcl_GetStringFromObj(objv[1],NULL))))	// see if there is an old map with this name that is about to be replaced
				{
					buffer=NULL;
					while(!fail&&(buffer=LocateNextEditorBufferOnMap(oldMap,buffer)))	// verify all buffers are not busy
					{
						fail=!ShellBufferNotBusy(localInterpreter,buffer);
					}
					if(!fail)							// move all buffers from old map to new map
					{
						nextBuffer=LocateNextEditorBufferOnMap(oldMap,NULL);
						while(nextBuffer)				// re-assign the maps
						{
							buffer=nextBuffer;			// point at the current buffer
							nextBuffer=LocateNextEditorBufferOnMap(oldMap,nextBuffer);	// move to the next before we do any modification to the current
							AssignSyntaxMap(buffer,map);	// if this fails, there's not much we can do anyway, so ignore it
						}
						UnlinkSyntaxMap(oldMap);
						CloseSyntaxMap(oldMap);
					}
				}
				if(!fail)
				{
					ClearErrorTrace();
					if(LinkSyntaxMap(map))				// link this map onto the global list
					{
						return(TCL_OK);
					}
					else
					{
						Tcl_AppendResult(localInterpreter,"Error linking syntax map '",Tcl_GetStringFromObj(objv[1],NULL),"': ",GetErrorTrace(),NULL);

						// failed to link new map to list, so untie any buffers that got linked to it
						buffer=NULL;
						while((buffer=LocateNextEditorBufferOnMap(map,buffer)))
						{
							AssignSyntaxMap(buffer,NULL);
						}
					}
				}
			}
			CloseSyntaxMap(map);
		}
		else
		{
			Tcl_AppendResult(localInterpreter,"Error creating syntax map '",Tcl_GetStringFromObj(objv[1],NULL),"': ",GetErrorTrace(),NULL);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"syntaxMapName syntaxMap");
	}
	return(TCL_ERROR);
}

static int Cmd_DeleteSyntaxMap(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// delete a syntax map that was created by "syntaxmap"
// Remove the map from all buffers which are using it
// NOTE: if any buffer which is using the map is currently busy, this will
// fail before it does anything.
{
	EDITOR_BUFFER
		*buffer,
		*nextBuffer;
	SYNTAX_MAP
		*map;
	bool
		fail;

	if(objc==2)
	{
		if((map=LocateSyntaxMap(Tcl_GetStringFromObj(objv[1],NULL))))	// find map to delete
		{
			fail=false;
			buffer=NULL;
			while(!fail&&(buffer=LocateNextEditorBufferOnMap(map,buffer)))	// verify all buffers are not busy
			{
				fail=!ShellBufferNotBusy(localInterpreter,buffer);
			}
			if(!fail)							// remove all buffers from the map, and kill the map
			{
				nextBuffer=LocateNextEditorBufferOnMap(map,NULL);
				while(nextBuffer)				// re-assign the maps
				{
					buffer=nextBuffer;			// point at the current buffer
					nextBuffer=LocateNextEditorBufferOnMap(map,nextBuffer);	// move to the next before we do any modification to the current
					AssignSyntaxMap(buffer,NULL);	// if this fails, there's not much we can do anyway, so ignore it
				}
				UnlinkSyntaxMap(map);
				CloseSyntaxMap(map);
				return(TCL_OK);
			}
		}
		else
		{
			Tcl_AppendResult(localInterpreter,"Could not locate syntax map '",Tcl_GetStringFromObj(objv[1],NULL),"'",NULL);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"syntaxMap");
	}
	return(TCL_ERROR);
}

static int Cmd_SyntaxMaps(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// return the list of currently defined syntax maps
{
	bool
		fail;
	SYNTAX_MAP
		*currentMap;

	fail=false;
	if(objc==1)
	{
		currentMap=NULL;
		while((currentMap=LocateNextSyntaxMap(currentMap)))
		{
			Tcl_AppendElement(localInterpreter,GetSyntaxMapName(currentMap));
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_SetSyntaxMap(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set the syntax map in use for the given buffer
// If an empty map string is specified, then remove any map from the current buffer
{
	EDITOR_BUFFER
		*buffer;
	SYNTAX_MAP
		*map;
	char
		*mapName;

	if(objc==3)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				map=NULL;
				mapName=Tcl_GetStringFromObj(objv[2],NULL);
				if(mapName[0]=='\0'||(map=LocateSyntaxMap(mapName)))
				{
					ClearErrorTrace();
					if(AssignSyntaxMap(buffer,map))
					{
						return(TCL_OK);
					}
					else
					{
						Tcl_AppendResult(localInterpreter,"Failed to assign syntax map: ",GetErrorTrace(),NULL);
					}
				}
				else
				{
					Tcl_AppendResult(localInterpreter,"Could not locate syntax map '",mapName,"'",NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName syntaxMapName");
	}
	return(TCL_ERROR);
}

static int Cmd_GetSyntaxMap(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// read back the syntax map in use for the given buffer
{
	EDITOR_BUFFER
		*buffer;
	SYNTAX_MAP
		*map;

	if(objc==2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if((map=GetAssignedSyntaxMap(buffer)))
			{
				Tcl_AppendElement(localInterpreter,GetSyntaxMapName(map));
			}
			else
			{
				Tcl_AppendElement(localInterpreter,"");
			}
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_SyntaxMapBuffers(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Return the list of buffer names which are currently using the given syntax map
{
	EDITOR_BUFFER
		*buffer;
	SYNTAX_MAP
		*map;

	if(objc==2)
	{
		if((map=LocateSyntaxMap(Tcl_GetStringFromObj(objv[1],NULL))))
		{
			buffer=NULL;
			while((buffer=LocateNextEditorBufferOnMap(map,buffer)))
			{
				Tcl_AppendElement(localInterpreter,buffer->contentName);
			}
			return(TCL_OK);
		}
		else
		{
			Tcl_AppendResult(localInterpreter,"Could not locate syntax map '",Tcl_GetStringFromObj(objv[1],NULL),"'",NULL);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"syntaxMapName");
	}
	return(TCL_ERROR);
}

static void CreateSearchResult(char *string,bool foundMatch,SELECTION_UNIVERSE *selection)
// When a find, findall, replace, or replaceall has finished executing, this
// is called to create the Tcl result for the search
// If the search was successful, this will return the range of the first
// match/replacement
// If the search was not successful, this will return an empty string
{
	UINT32
		matchPosition;
	UINT32
		startPosition,
		numElements;
	bool
		isActive;

	if(foundMatch)
	{
		matchPosition=GetSelectionCursorPosition(selection);	// get the position of the match/first match
		GetSelectionRange(selection,matchPosition,&startPosition,&numElements,&isActive);
		if(!isActive)
		{
			startPosition=matchPosition;
			numElements=0;
		}
		sprintf(string,"%u %u",startPosition,startPosition+numElements);
	}
	else
	{
		string[0]='\0';					// just return an empty string
	}
}

static OPTION_DESCRIPTOR_RECORD
	searchOptions[]=
	{
		{"-backward",		true,	OptionBoolean},
		{"-forward",		false,	OptionBoolean},
		{"-wrap",			true,	OptionBoolean},
		{"-nowrap",			false,	OptionBoolean},
		{"-regex",			true,	OptionBoolean},
		{"-literal",		false,	OptionBoolean},
		{"-ignorecase",		true,	OptionBoolean},
		{"-case",			false,	OptionBoolean},
		{"-limitscope",		true,	OptionBoolean},
		{"-globalscope",	false,	OptionBoolean},
		{"-replacescript",	true,	OptionBoolean},
		{"-replaceliteral",	false,	OptionBoolean},
		{"-sourcemark",		0,		OptionMark},
		{"-destmark",		0,		OptionMark},
		{NULL,0,NULL}
	};

static int Cmd_Find(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// find within the given buffer
// findBuffer must exist, or an error will be returned
// if there is no error, then if there as a match, the first match span is returned, if no
// matches, a null string is returned
{
	EDITOR_BUFFER
		*buffer;
	EDITOR_BUFFER
		*findBuffer;
	bool
		backwards,
		wrap,
		regex,
		ignoreCase,
		limitScope,
		replaceScript;
	PARSE_MARK
		sourceMark,
		destMark;
	SELECTION_UNIVERSE
		*sourceSelection,
		*resultSelection;
	void
		*parsedVars[]=
		{
			(void *)&backwards,
			(void *)&backwards,
			(void *)&wrap,
			(void *)&wrap,
			(void *)&regex,
			(void *)&regex,
			(void *)&ignoreCase,
			(void *)&ignoreCase,
			(void *)&limitScope,
			(void *)&limitScope,
			(void *)&replaceScript,
			(void *)&replaceScript,
			(void *)&sourceMark,
			(void *)&destMark,
		};
	int
		optionIndex;
	bool
		foundMatch;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];
	bool
		fail;

	fail=false;
	if(objc>=3)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				SetBufferBusy(buffer);
				if((findBuffer=GetEditorBuffer(localInterpreter,objv[2])))
				{
					if(ShellBufferNotBusy(localInterpreter,findBuffer))
					{
						SetBufferBusy(findBuffer);
						optionIndex=3;

						backwards=false;
						wrap=false;
						regex=false;
						ignoreCase=false;
						limitScope=false;
						replaceScript=false;

						sourceMark.buffer=buffer;
						sourceMark.mark=NULL;

						destMark.buffer=buffer;
						destMark.mark=NULL;

						if(ParseOptions(localInterpreter,searchOptions,parsedVars,objc,objv,&optionIndex,true))
						{
							sourceSelection=sourceMark.mark?sourceMark.mark->selectionUniverse:buffer->selectionUniverse;

							ClearErrorTrace();
							if((resultSelection=OpenSelectionUniverse()))	// get a selection for the result
							{
								ClearErrorTrace();
								if(EditorFind(buffer,sourceSelection,findBuffer->textUniverse,backwards,wrap,regex,ignoreCase,&foundMatch,resultSelection,CheckUserAbort))
								{
									CreateSearchResult(tempString,foundMatch,resultSelection);
									Tcl_AppendResult(localInterpreter,tempString,NULL);
									if(foundMatch)
									{
										if(destMark.mark)
										{
											ViewsStartMarkChange(destMark.mark);
											CopySelectionUniverse(resultSelection,destMark.mark->selectionUniverse);
											ViewsEndMarkChange(destMark.mark);
										}
										else
										{
											EditorStartSelectionChange(buffer);
											CopySelectionUniverse(resultSelection,buffer->selectionUniverse);
											EditorEndSelectionChange(buffer);
										}
									}
								}
								else
								{
									Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
									fail=true;
								}
								CloseSelectionUniverse(resultSelection);
							}
							else
							{
								Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
								fail=true;
							}
						}
						else
						{
							fail=true;
						}
						ClearBufferBusy(findBuffer);
					}
					else
					{
						fail=true;
					}
				}
				else
				{
					fail=true;
				}
				ClearBufferBusy(buffer);
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"searchInBuffer findBuffer ?flags?");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_FindAll(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// find all within the given buffer
// findBuffer must exist, or an error will be returned
// if there is no error, then if there as a match, the first match span is returned, if no
// matches, a null string is returned
{
	EDITOR_BUFFER
		*buffer;
	EDITOR_BUFFER
		*findBuffer;
	bool
		backwards,
		wrap,
		regex,
		ignoreCase,
		limitScope,
		replaceScript;
	PARSE_MARK
		sourceMark,
		destMark;
	SELECTION_UNIVERSE
		*sourceSelection,
		*resultSelection;
	void
		*parsedVars[]=
		{
			(void *)&backwards,
			(void *)&backwards,
			(void *)&wrap,
			(void *)&wrap,
			(void *)&regex,
			(void *)&regex,
			(void *)&ignoreCase,
			(void *)&ignoreCase,
			(void *)&limitScope,
			(void *)&limitScope,
			(void *)&replaceScript,
			(void *)&replaceScript,
			(void *)&sourceMark,
			(void *)&destMark,
		};
	int
		optionIndex;
	bool
		foundMatch;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];
	bool
		fail;

	fail=false;
	if(objc>=3)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				SetBufferBusy(buffer);
				if((findBuffer=GetEditorBuffer(localInterpreter,objv[2])))
				{
					if(ShellBufferNotBusy(localInterpreter,findBuffer))
					{
						SetBufferBusy(findBuffer);
						optionIndex=3;

						backwards=false;
						wrap=false;
						regex=false;
						ignoreCase=false;
						limitScope=false;
						replaceScript=false;

						sourceMark.buffer=buffer;
						sourceMark.mark=NULL;

						destMark.buffer=buffer;
						destMark.mark=NULL;

						if(ParseOptions(localInterpreter,searchOptions,parsedVars,objc,objv,&optionIndex,true))
						{
							sourceSelection=sourceMark.mark?sourceMark.mark->selectionUniverse:buffer->selectionUniverse;

							ClearErrorTrace();
							if((resultSelection=OpenSelectionUniverse()))	// get a selection for the result
							{
								ClearErrorTrace();
								if(EditorFindAll(buffer,sourceSelection,findBuffer->textUniverse,backwards,wrap,regex,ignoreCase,limitScope,&foundMatch,resultSelection,CheckUserAbort))
								{
									CreateSearchResult(tempString,foundMatch,resultSelection);
									Tcl_AppendResult(localInterpreter,tempString,NULL);
									if(foundMatch)
									{
										if(destMark.mark)
										{
											ViewsStartMarkChange(destMark.mark);
											CopySelectionUniverse(resultSelection,destMark.mark->selectionUniverse);
											ViewsEndMarkChange(destMark.mark);
										}
										else
										{
											EditorStartSelectionChange(buffer);
											CopySelectionUniverse(resultSelection,buffer->selectionUniverse);
											EditorEndSelectionChange(buffer);
										}
									}
								}
								else
								{
									Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
									fail=true;
								}
								CloseSelectionUniverse(resultSelection);
							}
							else
							{
								Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
								fail=true;
							}
						}
						else
						{
							fail=true;
						}
						ClearBufferBusy(findBuffer);
					}
					else
					{
						fail=true;
					}
				}
				else
				{
					fail=true;
				}
				ClearBufferBusy(buffer);
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"searchInBuffer findBuffer ?flags?");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_Replace(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// replace within the given buffer
// findBuffer and replaceBuffer must exist, or an error will be returned
// if there is no error, then if there as a match, the first match span is returned, if no
// matches, a null string is returned
{
	EDITOR_BUFFER
		*buffer;
	EDITOR_BUFFER
		*findBuffer,
		*replaceBuffer;
	bool
		backwards,
		wrap,
		regex,
		ignoreCase,
		limitScope,
		replaceScript;
	PARSE_MARK
		sourceMark,
		destMark;
	SELECTION_UNIVERSE
		*sourceSelection,
		*resultSelection;
	void
		*parsedVars[]=
		{
			(void *)&backwards,
			(void *)&backwards,
			(void *)&wrap,
			(void *)&wrap,
			(void *)&regex,
			(void *)&regex,
			(void *)&ignoreCase,
			(void *)&ignoreCase,
			(void *)&limitScope,
			(void *)&limitScope,
			(void *)&replaceScript,
			(void *)&replaceScript,
			(void *)&sourceMark,
			(void *)&destMark,
		};
	int
		optionIndex;
	bool
		foundMatch;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];
	bool
		fail;

	fail=false;
	if(objc>=3)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				SetBufferBusy(buffer);
				if((findBuffer=GetEditorBuffer(localInterpreter,objv[2])))
				{
					if(ShellBufferNotBusy(localInterpreter,findBuffer))
					{
						SetBufferBusy(findBuffer);
						if((replaceBuffer=GetEditorBuffer(localInterpreter,objv[3])))
						{
							if(ShellBufferNotBusy(localInterpreter,replaceBuffer))
							{
								SetBufferBusy(replaceBuffer);
								optionIndex=4;

								backwards=false;
								wrap=false;
								regex=false;
								ignoreCase=false;
								limitScope=false;
								replaceScript=false;

								sourceMark.buffer=buffer;
								sourceMark.mark=NULL;

								destMark.buffer=buffer;
								destMark.mark=NULL;

								if(ParseOptions(localInterpreter,searchOptions,parsedVars,objc,objv,&optionIndex,true))
								{
									sourceSelection=sourceMark.mark?sourceMark.mark->selectionUniverse:buffer->selectionUniverse;

									ClearErrorTrace();
									if((resultSelection=OpenSelectionUniverse()))	// get a selection for the result
									{
										ClearErrorTrace();
										if(EditorReplace(buffer,sourceSelection,findBuffer->textUniverse,replaceBuffer->textUniverse,backwards,wrap,regex,ignoreCase,replaceScript,&foundMatch,resultSelection,CheckUserAbort))
										{
											CreateSearchResult(tempString,foundMatch,resultSelection);
											Tcl_AppendResult(localInterpreter,tempString,NULL);
											if(foundMatch)
											{
												if(destMark.mark)
												{
													ViewsStartMarkChange(destMark.mark);
													CopySelectionUniverse(resultSelection,destMark.mark->selectionUniverse);
													ViewsEndMarkChange(destMark.mark);
												}
												else
												{
													EditorStartSelectionChange(buffer);
													CopySelectionUniverse(resultSelection,buffer->selectionUniverse);
													EditorEndSelectionChange(buffer);
												}
											}
										}
										else
										{
											Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
											fail=true;
										}
										CloseSelectionUniverse(resultSelection);
									}
									else
									{
										Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
										fail=true;
									}
								}
								else
								{
									fail=true;
								}
								ClearBufferBusy(replaceBuffer);
							}
							else
							{
								fail=true;
							}
						}
						else
						{
							fail=true;
						}
						ClearBufferBusy(findBuffer);
					}
				}
				else
				{
					fail=true;
				}
				ClearBufferBusy(buffer);
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"searchInBuffer findBuffer replaceBuffer ?flags?");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_ReplaceAll(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// replace all within the given buffer
// findBuffer and replaceBuffer must exist, or an error will be returned
// if there is no error, then if there as a match, the first match span is returned, if no
// matches, a null string is returned
{
	EDITOR_BUFFER
		*buffer;
	EDITOR_BUFFER
		*findBuffer,
		*replaceBuffer;
	bool
		backwards,
		wrap,
		regex,
		ignoreCase,
		limitScope,
		replaceScript;
	PARSE_MARK
		sourceMark,
		destMark;
	SELECTION_UNIVERSE
		*sourceSelection,
		*resultSelection;
	void
		*parsedVars[]=
		{
			(void *)&backwards,
			(void *)&backwards,
			(void *)&wrap,
			(void *)&wrap,
			(void *)&regex,
			(void *)&regex,
			(void *)&ignoreCase,
			(void *)&ignoreCase,
			(void *)&limitScope,
			(void *)&limitScope,
			(void *)&replaceScript,
			(void *)&replaceScript,
			(void *)&sourceMark,
			(void *)&destMark,
		};
	int
		optionIndex;
	bool
		foundMatch;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];
	bool
		fail;

	fail=false;
	if(objc>=3)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				SetBufferBusy(buffer);
				if((findBuffer=GetEditorBuffer(localInterpreter,objv[2])))
				{
					if(ShellBufferNotBusy(localInterpreter,findBuffer))
					{
						SetBufferBusy(findBuffer);
						if((replaceBuffer=GetEditorBuffer(localInterpreter,objv[3])))
						{
							if(ShellBufferNotBusy(localInterpreter,replaceBuffer))
							{
								SetBufferBusy(replaceBuffer);
								optionIndex=4;

								backwards=false;
								wrap=false;
								regex=false;
								ignoreCase=false;
								limitScope=false;
								replaceScript=false;

								sourceMark.buffer=buffer;
								sourceMark.mark=NULL;

								destMark.buffer=buffer;
								destMark.mark=NULL;

								if(ParseOptions(localInterpreter,searchOptions,parsedVars,objc,objv,&optionIndex,true))
								{
									sourceSelection=sourceMark.mark?sourceMark.mark->selectionUniverse:buffer->selectionUniverse;

									ClearErrorTrace();
									if((resultSelection=OpenSelectionUniverse()))	// get a selection for the result
									{
										ClearErrorTrace();
										if(EditorReplaceAll(buffer,sourceSelection,findBuffer->textUniverse,replaceBuffer->textUniverse,backwards,wrap,regex,ignoreCase,limitScope,replaceScript,&foundMatch,resultSelection,CheckUserAbort))
										{
											CreateSearchResult(tempString,foundMatch,resultSelection);
											Tcl_AppendResult(localInterpreter,tempString,NULL);
											if(foundMatch)
											{
												if(destMark.mark)
												{
													ViewsStartMarkChange(destMark.mark);
													CopySelectionUniverse(resultSelection,destMark.mark->selectionUniverse);
													ViewsEndMarkChange(destMark.mark);
												}
												else
												{
													EditorStartSelectionChange(buffer);
													CopySelectionUniverse(resultSelection,buffer->selectionUniverse);
													EditorEndSelectionChange(buffer);
												}
											}
										}
										else
										{
											Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
											fail=true;
										}
										CloseSelectionUniverse(resultSelection);
									}
									else
									{
										Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
										fail=true;
									}
								}
								else
								{
									fail=true;
								}
								ClearBufferBusy(replaceBuffer);
							}
							else
							{
								fail=true;
							}
						}
						else
						{
							fail=true;
						}
						ClearBufferBusy(findBuffer);
					}
				}
				else
				{
					fail=true;
				}
				ClearBufferBusy(buffer);
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"searchInBuffer findBuffer replaceBuffer ?flags?");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_UndoToggle(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// perform an undo toggle in the given buffer
{
	EDITOR_BUFFER
		*buffer;
	bool
		didSomething;

	if(objc==2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				ClearErrorTrace();
				if(EditorToggleUndo(buffer,&didSomething))
				{
					Tcl_AppendResult(localInterpreter,didSomething?"1":"0",NULL);
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_Undo(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// perform an undo in the given buffer
{
	EDITOR_BUFFER
		*buffer;
	bool
		didSomething;

	if(objc==2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				ClearErrorTrace();
				if(EditorUndo(buffer,&didSomething))
				{
					Tcl_AppendResult(localInterpreter,didSomething?"1":"0",NULL);
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_Redo(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// perform a redo in the given buffer
{
	EDITOR_BUFFER
		*buffer;
	bool
		didSomething;

	if(objc==2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				ClearErrorTrace();
				if(EditorRedo(buffer,&didSomething))
				{
					Tcl_AppendResult(localInterpreter,didSomething?"1":"0",NULL);
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_BreakUndo(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// cause a break in the undo flow for the passed buffer
{
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				BeginUndoGroup(buffer);
				StrictEndUndoGroup(buffer);
				return(TCL_OK);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_FlushUndos(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// perform a clear of the undo/redo buffers in the given buffer
{
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				EditorClearUndo(buffer);
				return(TCL_OK);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_Cut(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// perform a cut in the given buffer
{
	EDITOR_BUFFER
		*buffer;
	EDITOR_BUFFER
		*clipboard;

	if(objc==2||objc==3)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				if(objc==3)
				{
					clipboard=LocateBuffer(Tcl_GetStringFromObj(objv[2],NULL));
				}
				else
				{
					clipboard=EditorGetCurrentClipboard();
				}
				if(clipboard)
				{
					if(ShellBufferNotBusy(localInterpreter,clipboard))
					{
						ClearErrorTrace();
						if(EditorCut(buffer,clipboard))
						{
							return(TCL_OK);
						}
						else
						{
							Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
						}
					}
				}
				else
				{
					Tcl_AppendResult(localInterpreter,"No clipboard",NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName ?clipboardBuffer?");
	}
	return(TCL_ERROR);
}

static int Cmd_Copy(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// perform a copy in the given buffer
{
	EDITOR_BUFFER
		*buffer;
	EDITOR_BUFFER
		*clipboard;

	if(objc==2||objc==3)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				if(objc==3)
				{
					clipboard=LocateBuffer(Tcl_GetStringFromObj(objv[2],NULL));
				}
				else
				{
					clipboard=EditorGetCurrentClipboard();
				}
				if(clipboard)
				{
					if(ShellBufferNotBusy(localInterpreter,clipboard))
					{
						ClearErrorTrace();
						if(EditorCopy(buffer,clipboard))
						{
							return(TCL_OK);
						}
						else
						{
							Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
						}
					}
				}
				else
				{
					Tcl_AppendResult(localInterpreter,"No clipboard",NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName ?clipboardBuffer?");
	}
	return(TCL_ERROR);
}

static int Cmd_Paste(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// perform a columnar paste in the given window/buffer
{
	EDITOR_BUFFER
		*buffer;
	EDITOR_BUFFER
		*clipboard;
	EDITOR_VIEW
		*view;

	if(objc==2||objc==3)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				if(buffer->window)
				{
					view=GetDocumentWindowCurrentView(buffer->window);
				}
				else
				{
					view=NULL;
				}
				if(objc==3)
				{
					clipboard=LocateBuffer(Tcl_GetStringFromObj(objv[2],NULL));
				}
				else
				{
					clipboard=EditorGetCurrentClipboard();
				}
				if(clipboard)
				{
					if(ShellBufferNotBusy(localInterpreter,clipboard))
					{
						ClearErrorTrace();
						if(EditorColumnarPaste(buffer,view,clipboard))
						{
							return(TCL_OK);
						}
						else
						{
							Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
						}
					}
				}
				else
				{
					Tcl_AppendResult(localInterpreter,"No clipboard",NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName ?clipboardBuffer?");
	}
	return(TCL_ERROR);
}

static int Cmd_SetClipboard(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// set the current clipboard to the given buffer (or NULL)
{
	EDITOR_BUFFER
		*clipboard;

	if(objc==1)
	{
		EditorSetCurrentClipboard(NULL);
		return(TCL_OK);
	}
	else if(objc==2)
	{
		if((clipboard=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,clipboard))
			{
				EditorSetCurrentClipboard(clipboard);
				return(TCL_OK);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"?clipboardBuffer?");
	}
	return(TCL_ERROR);
}

static int Cmd_GetClipboard(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// return the current clipboard, fail if there is none
{
	EDITOR_BUFFER
		*clipboard;

	if(objc==1)
	{
		if((clipboard=EditorGetCurrentClipboard()))
		{
			Tcl_AppendResult(localInterpreter,clipboard->contentName,NULL);
			return(TCL_OK);
		}
		else
		{
			Tcl_AppendResult(localInterpreter,"No clipboard",NULL);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_ScreenSize(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// return the dimensions of the screen that the editor is running on
{
	INT32
		xOffset,
		yOffset;
	UINT32
		width,
		height;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==1)
	{
		GetEditorScreenOffsetAndDimensions(&xOffset,&yOffset,&width,&height);
		sprintf(tempString,"%u %u",width,height);
		Tcl_AppendResult(localInterpreter,tempString,NULL);
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_SetWordChars(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// set the wordSpaceTable so that all of the characters in the passed argument
// are not considered space
{
	int
		i;
	char
		*chars;

	if(objc==2)
	{
		for(i=0;i<256;i++)
		{
			wordSpaceTable[i]=true;
		}
		chars=Tcl_GetStringFromObj(objv[1],NULL);

		for(i=0;chars[i];i++)
		{
			wordSpaceTable[(unsigned int)chars[i]]=false;
		}
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"wordCharacters");
	}
	return(TCL_ERROR);
}

static int Cmd_GetWordChars(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// return the list of characters that are considered parts of words
{
	char
		tempString[256];					// hold the list of characters that are part of a word
	int
		i,j;

	if(objc==1)
	{
		j=0;
		for(i=1;i<256;i++)					// never look at character 0, because it would terminate the string, and confuse things
		{
			if(!wordSpaceTable[i])
			{
				tempString[j++]=(char)i;	// set this one into the string
			}
		}
		tempString[j++]='\0';				// terminate the string
		Tcl_AppendElement(localInterpreter,&(tempString[0]));
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_Execute(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// execute the given tcl script, place the results in the given buffer
// if it still exists after the script is finished, otherwise throw them away.
{
	EDITOR_BUFFER_HANDLE
		handle;
	EDITOR_BUFFER
		*buffer;
	UINT32
		startPosition,
		endPosition;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];
	int
		tclResult;
	const char
		*stringResult;

	if(objc==3)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				// NOTE: do not set busy here, we will allow executed script to mess with the window, but we must be careful:
				// If the buffer goes away as the result of executing the script, we have to make sure it is not referenced again

				EditorInitBufferHandle(&handle);
				EditorGrabBufferHandle(&handle,buffer);					// hold on to this buffer

				GetSelectionEndPositions(buffer->selectionUniverse,&startPosition,&endPosition);
				SetSelectionCursorPosition(buffer->auxSelectionUniverse,endPosition);	// move the aux cursor to the end of any selection/cursor, and start output there

				// do not call ClearUserAbort here, it will have already been called when the script that got us here was started
				tclResult=Tcl_EvalObj(localInterpreter,objv[2]);
				stringResult=Tcl_GetStringResult(localInterpreter);
				if(*stringResult)										// if no result, do not try to insert into buffer
				{
					if((buffer=EditorGetBufferFromHandle(&handle)))		// see if this buffer is still alive
					{
						if(tclResult!=TCL_OK)
						{
							sprintf(tempString,"Error in line %d: ",localInterpreter->errorLine);
							EditorAuxInsert(buffer,(UINT8 *)tempString,(UINT32)strlen(tempString));
						}
						EditorAuxInsert(buffer,(UINT8 *)stringResult,(UINT32)strlen(stringResult));
						EditorAuxInsert(buffer,(UINT8 *)"\n",1);
					}
				}
				EditorReleaseBufferHandle(&handle);						// do not need this any more
				return(TCL_OK);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName script");
	}
	return(TCL_ERROR);
}

static int Cmd_Task(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// start a task running in the given buffer, or send data to a task in a buffer
{
	EDITOR_BUFFER
		*buffer;
	UINT32
		startPosition,
		endPosition;
	char
		*taskData;

	if(objc==3)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				GetSelectionEndPositions(buffer->selectionUniverse,&startPosition,&endPosition);
				SetSelectionCursorPosition(buffer->auxSelectionUniverse,endPosition);	// move the aux cursor to the end of any selection/cursor, and start output there
				taskData=Tcl_GetStringFromObj(objv[2],NULL);
				ClearErrorTrace();
				if(WriteBufferTaskData(buffer,(UINT8 *)taskData,(UINT32)strlen(taskData)))
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName taskData");
	}
	return(TCL_ERROR);
}

static int Cmd_UpdateTask(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// if a task is running in the given buffer, this can be called to
// update the status of the task
// If the buffer is busy, this will fail
// NOTE: while scripts are not running, the tasks on all buffers are
// continuously updated, but while scripts are running, this must
// be called explicitly to get a buffer to update
{
	EDITOR_BUFFER
		*buffer;
	bool
		didUpdate;

	if(objc==2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				ClearErrorTrace();
				if(UpdateBufferTask(buffer,&didUpdate))
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_EOFTask(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// send an EOF to the task on the passed buffer, if there is a problem, complain
{
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				ClearErrorTrace();
				if(SendEOFToBufferTask(buffer))
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_KillTask(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// send a signal to the task on the passed buffer, if there is a problem, complain
{
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				ClearErrorTrace();
				if(KillBufferTask(buffer))
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,GetErrorTrace(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_TaskBytes(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// return the number of characters the last/current task has dumped into the given buffer
// NOTE: this count is only reset when as task is launched
{
	EDITOR_BUFFER
		*buffer;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			sprintf(tempString,"%u",buffer->taskBytes);
			Tcl_AppendResult(localInterpreter,tempString,NULL);
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_HasTask(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// see if a task is running in the given buffer, if so, return 1, else 0
{
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		if((buffer=GetEditorBuffer(localInterpreter,objv[1])))
		{
			Tcl_AppendResult(localInterpreter,buffer->task?"1":"0",NULL);
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_Beep(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// make a noise
{
	if(objc==1)
	{
		EditorBeep();
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_Version(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// get version numbers
{
	const char
		*localVersion;
	bool
		fail;

	fail=false;
	if(objc==1)
	{
		Tcl_AppendResult(localInterpreter,programName," ",VERSION,EDITION,NULL);
		if((localVersion=GetEditorLocalVersion()))
		{
			Tcl_AppendResult(localInterpreter," (",localVersion,")",NULL);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_ShellCommand(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// This exists to provide a stub for the shell commands which e93 will call internally
// for various functions. Normally, when the initial script is executed, it will define
// this function, but if the script does not define it, or fails before defining it,
// this acts as a safety net
{
	return(TCL_OK);
}

static int Cmd_ForceQuit(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// execute editor quit function (the editor WILL quit, no questions asked if you call this)
{
	bool
		fail;

	fail=false;
	if(objc==1)
	{
		EditorQuit();
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

/* @@@ removed until I get new channels.c from Michael
static int Cmd_GetChannel(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// return the buffer being used for Tcl's channel named in objv 1
// if the buffer does not exist, fail
{
	EDITOR_BUFFER
		*buffer;
	char
		*channelId;

	if(objc==2)
	{
		channelId=Tcl_GetStringFromObj(objv[1],NULL);
		if (Tcl_GetChannel(localInterpreter, channelId, NULL) != NULL)	// see if there is such a channel
		{
			if((buffer=GetChannelBuffer(localInterpreter, channelId)))
			{
				Tcl_AppendResult(localInterpreter,buffer->contentName,NULL);
				return(TCL_OK);
			}
			else
			{
				Tcl_AppendResult(localInterpreter,"No buffer",NULL);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"channelId");
	}
	return(TCL_ERROR);
}

static int Cmd_SetChannel(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// set the buffer to be used for Tcl's channel
// if the buffer does not exist, complain, but if it is deleted while it
// is being used as the output, that's ok -- the output is just discarded
{
	EDITOR_BUFFER
		*buffer=NULL;
	char
		*channelId;

	if(objc == 2 || objc == 3)
	{
		if(objc == 3)
		{
			if((buffer=GetEditorBuffer(localInterpreter,objv[2]))==NULL)
			{
				return(TCL_ERROR);
			}
		}

		channelId=Tcl_GetStringFromObj(objv[1],NULL);
		SetChannelBuffer(localInterpreter,channelId, buffer);
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"channelId ?bufferName?");
	}
	return(TCL_ERROR);
}
*/

// Create commands so TCL can find them

bool CreateEditorShellCommands(Tcl_Interp *interpreter)
// create the shell commands that are implemented by the editor
// if there is a problem, SetError, return false
{
// menus and key bindings
	Tcl_CreateObjCommand(interpreter,"deletemenu",Cmd_DeleteMenu,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"addmenu",Cmd_AddMenu,NULL,NULL);

	Tcl_CreateObjCommand(interpreter,"bindkey",Cmd_BindKey,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"unbindkey",Cmd_UnBindKey,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"keybindings",Cmd_KeyBindings,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"waitkey",Cmd_WaitKey,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"getkey",Cmd_GetKey,NULL,NULL);

// dialogs
	Tcl_CreateObjCommand(interpreter,"okdialog",Cmd_OkDialog,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"okcanceldialog",Cmd_OkCancelDialog,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"yesnodialog",Cmd_YesNoDialog,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"textdialog",Cmd_TextDialog,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"searchdialog",Cmd_SearchDialog,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"listdialog",Cmd_ListDialog,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"opendialog",Cmd_OpenDialog,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"savedialog",Cmd_SaveDialog,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"pathdialog",Cmd_PathDialog,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"fontdialog",Cmd_FontDialog,NULL,NULL);

// buffers, windows, and views
	Tcl_CreateObjCommand(interpreter,"closebuffer",Cmd_CloseBuffer,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"openbuffer",Cmd_OpenBuffer,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"newbuffer",Cmd_NewBuffer,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"bufferlist",Cmd_BufferList,NULL,NULL);

	Tcl_CreateObjCommand(interpreter,"savebuffer",Cmd_SaveBuffer,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"savebufferas",Cmd_SaveBufferAs,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"savebufferto",Cmd_SaveBufferTo,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"revertbuffer",Cmd_RevertBuffer,NULL,NULL);

	Tcl_CreateObjCommand(interpreter,"buffervariables",Cmd_BufferVariables,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"setbuffervariable",Cmd_SetBufferVariable,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"getbuffervariable",Cmd_GetBufferVariable,NULL,NULL);

	Tcl_CreateObjCommand(interpreter,"isdirty",Cmd_BufferDirty,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"fromfile",Cmd_FromFile,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"haswindow",Cmd_HasWindow,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"cleardirty",Cmd_ClearDirty,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"textinfo",Cmd_TextInfo,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"positiontolineoffset",Cmd_PositionToLineOffset,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"lineoffsettoposition",Cmd_LineOffsetToPosition,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"checkbuffer",Cmd_CheckBuffer,NULL,NULL);

	Tcl_CreateObjCommand(interpreter,"closewindow",Cmd_CloseWindow,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"openwindow",Cmd_OpenWindow,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"windowlist",Cmd_WindowList,NULL,NULL);

	Tcl_CreateObjCommand(interpreter,"gettopleft",Cmd_GetTopLeft,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"settopleft",Cmd_SetTopLeft,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"homewindow",Cmd_HomeWindow,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"setrect",Cmd_SetRect,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"getrect",Cmd_GetRect,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"setfont",Cmd_SetFont,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"getfont",Cmd_GetFont,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"settabsize",Cmd_SetTabSize,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"gettabsize",Cmd_GetTabSize,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"setcolors",Cmd_SetColors,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"getcolors",Cmd_GetColors,NULL,NULL);

	Tcl_CreateObjCommand(interpreter,"settopwindow",Cmd_SetTopWindow,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"minimizewindow",Cmd_MinimizeWindow,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"unminimizewindow",Cmd_UnminimizeWindow,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"activewindow",Cmd_GetActiveWindow,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"updatewindows",Cmd_UpdateWindows,NULL,NULL);

// editing, cursor, and selection functions
	Tcl_CreateObjCommand(interpreter,"clear",Cmd_Clear,NULL,NULL);							// -destmark
	Tcl_CreateObjCommand(interpreter,"extract",Cmd_Extract,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"insert",Cmd_Insert,NULL,NULL);						// -destmark
	Tcl_CreateObjCommand(interpreter,"insertfile",Cmd_InsertFile,NULL,NULL);				// -destmark
	Tcl_CreateObjCommand(interpreter,"autoindent",Cmd_AutoIndent,NULL,NULL);				// -destmark

	Tcl_CreateObjCommand(interpreter,"movecursor",Cmd_MoveCursor,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"delete",Cmd_Delete,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"expandselection",Cmd_ExpandSelection,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"reduceselection",Cmd_ReduceSelection,NULL,NULL);

// selection/marks
	Tcl_CreateObjCommand(interpreter,"setselectioncolors",Cmd_SetSelectionColors,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"getselectioncolors",Cmd_GetSelectionColors,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"selectioninfo",Cmd_SelectionInfo,NULL,NULL);			// -sourcemark only
	Tcl_CreateObjCommand(interpreter,"selectall",Cmd_SelectAll,NULL,NULL);					// -destmark only
	Tcl_CreateObjCommand(interpreter,"selectedtextlist",Cmd_SelectedTextList,NULL,NULL);	// -sourcemark only
	Tcl_CreateObjCommand(interpreter,"selectline",Cmd_SelectLine,NULL,NULL);				// -destmark only
	Tcl_CreateObjCommand(interpreter,"getselectionends",Cmd_GetSelectionEnds,NULL,NULL);	// -sourcemark only
	Tcl_CreateObjCommand(interpreter,"setselectionends",Cmd_SetSelectionEnds,NULL,NULL);	// -destmark only
	Tcl_CreateObjCommand(interpreter,"getselectionendlist",Cmd_GetSelectionEndList,NULL,NULL);	// -sourcemark only
	Tcl_CreateObjCommand(interpreter,"addselectionendlist",Cmd_AddSelectionEndList,NULL,NULL);	// -destmark only
	Tcl_CreateObjCommand(interpreter,"getselectionatposition",Cmd_GetSelectionAtPosition,NULL,NULL);	// -sourcemark only

	Tcl_CreateObjCommand(interpreter,"copyselection",Cmd_CopySelection,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"andselection",Cmd_AndSelection,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"orselection",Cmd_OrSelection,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"xorselection",Cmd_XorSelection,NULL,NULL);

	Tcl_CreateObjCommand(interpreter,"markinvisible",Cmd_MarkInvisible,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"markvisible",Cmd_MarkVisible,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"marklist",Cmd_MarkList,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"closemark",Cmd_CloseMark,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"newmark",Cmd_NewMark,NULL,NULL);

// styles and syntax highlighting
	Tcl_CreateObjCommand(interpreter,"getstyleatposition",Cmd_GetStyleAtPosition,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"setstyle",Cmd_SetStyle,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"selectiontostyle",Cmd_SelectionToStyle,NULL,NULL);	// -sourcemark only
	Tcl_CreateObjCommand(interpreter,"styletoselection",Cmd_StyleToSelection,NULL,NULL);	// -destmark only
	Tcl_CreateObjCommand(interpreter,"addsyntaxmap",Cmd_AddSyntaxMap,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"deletesyntaxmap",Cmd_DeleteSyntaxMap,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"syntaxmaps",Cmd_SyntaxMaps,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"setsyntaxmap",Cmd_SetSyntaxMap,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"getsyntaxmap",Cmd_GetSyntaxMap,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"syntaxmapbuffers",Cmd_SyntaxMapBuffers,NULL,NULL);

// search/replace
	Tcl_CreateObjCommand(interpreter,"find",Cmd_Find,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"findall",Cmd_FindAll,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"replace",Cmd_Replace,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"replaceall",Cmd_ReplaceAll,NULL,NULL);

// undo
	Tcl_CreateObjCommand(interpreter,"undotoggle",Cmd_UndoToggle,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"undo",Cmd_Undo,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"redo",Cmd_Redo,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"breakundo",Cmd_BreakUndo,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"flushundos",Cmd_FlushUndos,NULL,NULL);

// clipboard
	Tcl_CreateObjCommand(interpreter,"cut",Cmd_Cut,NULL,NULL);							// -sourcemark
	Tcl_CreateObjCommand(interpreter,"copy",Cmd_Copy,NULL,NULL);						// -sourcemark
	Tcl_CreateObjCommand(interpreter,"paste",Cmd_Paste,NULL,NULL);						// -destmark
	Tcl_CreateObjCommand(interpreter,"setclipboard",Cmd_SetClipboard,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"getclipboard",Cmd_GetClipboard,NULL,NULL);

// misc functions
	Tcl_CreateObjCommand(interpreter,"screensize",Cmd_ScreenSize,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"setwordchars",Cmd_SetWordChars,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"getwordchars",Cmd_GetWordChars,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"execute",Cmd_Execute,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"task",Cmd_Task,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"updatetask",Cmd_UpdateTask,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"eoftask",Cmd_EOFTask,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"killtask",Cmd_KillTask,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"taskbytes",Cmd_TaskBytes,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"hastask",Cmd_HasTask,NULL,NULL);

	Tcl_CreateObjCommand(interpreter,"beep",Cmd_Beep,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"version",Cmd_Version,NULL,NULL);

	Tcl_CreateObjCommand(interpreter,"ShellCommand",Cmd_ShellCommand,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"forceQUIT",Cmd_ForceQuit,NULL,NULL);

// Built-in Tcl commands we DONT want
	Tcl_DeleteCommand(interpreter,"exit");

//	Tcl_CreateObjCommand(interpreter,"setchannelbuffer",Cmd_SetChannel,NULL,NULL);
//	Tcl_CreateObjCommand(interpreter,"getchannelbuffer",Cmd_GetChannel,NULL,NULL);

	return(true);
}
