// Sparse array management functions
// Copyright (C) 2000 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

// Sparse arrays are arrays of arbitrary items of which most
// elements are usually unoccupied.
// These are handy for describing selections, styles, and other
// things within e93.

// To create a sparse array, define a structure of which the first
// element is a SPARSE_ARRAY_HEADER.

#include	"includes.h"

#define	HEADER_OFFSET(sparseArray,chunkVal,offsetVal) ((SPARSE_ARRAY_HEADER *)(&chunkVal->data[(offsetVal)*sparseArray->elementSize]))->offset
#define	HEADER_LENGTH(sparseArray,chunkVal,offsetVal) ((SPARSE_ARRAY_HEADER *)(&chunkVal->data[(offsetVal)*sparseArray->elementSize]))->length

/* This should not be needed anymore ---------------------------------------------------------
static void SanityTest(SPARSE_ARRAY *sparseArray)
// Look at the cached information about sparseArray, and verify that
// it is correct
// If there is a problem, send it to the console
{
	ARRAY_CHUNK_HEADER
		*chunk;
	UINT32
		offset;
	UINT32
		sumOffset;
	bool
		sawCacheChunk;


	chunk=sparseArray->elementChunks.firstChunkHeader;
	sumOffset=0;
	offset=0;
	sawCacheChunk=false;
	while(chunk)
	{
		if(chunk==sparseArray->cacheChunkHeader&&offset==sparseArray->cacheOffset)
		{
			sawCacheChunk=true;
			if(sumOffset!=sparseArray->cacheCurrentElement)
			{
				fprintf(stderr,"BAD current element: CE:%d ACE:%d%c\n",sparseArray->cacheCurrentElement,sumOffset,7);
			}
		}
		if(HEADER_OFFSET(sparseArray,chunk,offset)>2000000)
		{
			fprintf(stderr,"Offset looks a little large:%d%c\n",HEADER_OFFSET(sparseArray,chunk,offset),7);
		}
		if(HEADER_LENGTH(sparseArray,chunk,offset)>2000000)
		{
			fprintf(stderr,"Length looks a little large:%d%c\n",HEADER_OFFSET(sparseArray,chunk,offset),7);
		}
		sumOffset+=HEADER_OFFSET(sparseArray,chunk,offset);
		sumOffset+=HEADER_LENGTH(sparseArray,chunk,offset);
		offset++;
		if(offset>=chunk->totalElements)	// move to the next chunk
		{
			chunk=chunk->nextHeader;
			offset=0;
		}
	}
	if(sumOffset!=sparseArray->totalElements)
	{
		fprintf(stderr,"BAD total elements: TE:%d ATE:%d%c\n",sparseArray->totalElements,sumOffset,7);
	}
	if(sparseArray->cacheChunkHeader&&!sawCacheChunk)
	{
		fprintf(stderr,"BAD cache chunk/offset %lX/%d (not seen in array!)%c\n",sparseArray->cacheChunkHeader,sparseArray->cacheOffset,7);
	}
}

void DumpSATable(SPARSE_ARRAY *sparseArray)
// spew out the sparse array table passed (for debugging)
{
	ARRAY_CHUNK_HEADER
		*chunk;
	UINT32
		offset;
	UINT32
		sumOffset;

	fprintf(stderr,"offset=%d cachecount=%d, totalE=%d\n",sparseArray->cacheOffset,sparseArray->cacheCurrentElement,sparseArray->totalElements);

	fprintf(stderr,"\n");
	chunk=sparseArray->elementChunks.firstChunkHeader;
	sumOffset=0;
	offset=0;
	while(chunk)
	{
		sumOffset+=HEADER_OFFSET(sparseArray,chunk,offset);
		fprintf(stderr,"offset:%d (%d) length:%d\n",HEADER_OFFSET(sparseArray,chunk,offset),sumOffset,HEADER_LENGTH(sparseArray,chunk,offset));
		sumOffset+=HEADER_LENGTH(sparseArray,chunk,offset);
		offset++;
		if(offset>=chunk->totalElements)	// move to the next chunk
		{
			chunk=chunk->nextHeader;
			offset=0;
		}
	}
}
*/

static void LocateSARecord(SPARSE_ARRAY *sparseArray,UINT32 position,ARRAY_CHUNK_HEADER **outChunk,UINT32 *outOffset,UINT32 *startPosition)
// Use the cache to quickly locate the record in sparseArray which contains position
// return an offset to the chunk of the array universe which contains the
// record in question, and the offset to the record
// also return the actual starting position of the record (the position where its offset begins)
// NOTE: if the passed position is after the last entry in the array, return
// outChunk as NULL, outOffset at 0, and startPosition as the total
// number of elements in the array
// NOTE: except for the cases when this returns NULL, the caller can always
// guarantee that the cache is left pointing to the values returned
{
	ARRAY_CHUNK_HEADER
		*chunk;
	UINT32
		offset;
	UINT32
		currentIndex;
	UINT32
		recordLength;

	if(sparseArray->elementChunks.firstChunkHeader&&(position<sparseArray->totalElements))
	{
		if(!sparseArray->cacheChunkHeader)	// if no current cache header, see if position is closer to start or end
		{
			if(position<=(sparseArray->totalElements/2))	// see if search should start from beginning or end
			{
				sparseArray->cacheChunkHeader=sparseArray->elementChunks.firstChunkHeader;
				sparseArray->cacheOffset=0;
				sparseArray->cacheCurrentElement=0;
			}
			else
			{
				sparseArray->cacheChunkHeader=sparseArray->elementChunks.lastChunkHeader;
				sparseArray->cacheOffset=sparseArray->elementChunks.lastChunkHeader->totalElements;
				sparseArray->cacheCurrentElement=sparseArray->totalElements;
			}
		}

		chunk=sparseArray->cacheChunkHeader;
		offset=sparseArray->cacheOffset;
		currentIndex=sparseArray->cacheCurrentElement;

		if(position>=currentIndex)							// hunt forward for the element
		{
			while(position>=currentIndex+(recordLength=HEADER_OFFSET(sparseArray,chunk,offset)+HEADER_LENGTH(sparseArray,chunk,offset)))	// this should never run off the end of the list because of the test above to make sure position is not too large
			{
				currentIndex+=recordLength;
				offset++;
				if(offset>=chunk->totalElements)
				{
					chunk=chunk->nextHeader;
					offset=0;
				}
			}
			*outChunk=sparseArray->cacheChunkHeader=chunk;	// write back to the cache
			*outOffset=sparseArray->cacheOffset=offset;
			*startPosition=sparseArray->cacheCurrentElement=currentIndex;
		}
		else	// hunting backwards
		{
			if(offset)									// step back one
			{
				offset--;
			}
			else
			{
				chunk=chunk->previousHeader;			// we know this will succeed because of previous tests
				offset=chunk->totalElements-1;
			}
			currentIndex-=(HEADER_OFFSET(sparseArray,chunk,offset)+HEADER_LENGTH(sparseArray,chunk,offset));
			while(position<currentIndex)
			{
				if(offset)								// step back one
				{
					offset--;
				}
				else
				{
					chunk=chunk->previousHeader;		// we know this will succeed because of previous tests
					offset=chunk->totalElements-1;
				}
				currentIndex-=(HEADER_OFFSET(sparseArray,chunk,offset)+HEADER_LENGTH(sparseArray,chunk,offset));
			}
			*outChunk=sparseArray->cacheChunkHeader=chunk;		// write back to the cache
			*outOffset=sparseArray->cacheOffset=offset;
			*startPosition=sparseArray->cacheCurrentElement=currentIndex;
		}
	}
	else	// position beyond end of array, or no array
	{
		*outChunk=NULL;
		*outOffset=0;
		*startPosition=sparseArray->totalElements;
	}
}

void DeleteSARange(SPARSE_ARRAY *sparseArray,UINT32 startPosition,UINT32 numElements)
// Remove sparse array elements from startPosition for numElements
// NOTE: this does not mind being called for positions outside of the current array
// NOTE: This was a complete pain to write and is hard to think about.
{
	ARRAY_CHUNK_HEADER
		*chunk,
		*firstChunk;
	UINT32
		offset,
		firstOffset;
	UINT32
		actualStart;
	UINT32
		numRecords;
	UINT32
		startRecordIndex,
		endRecordIndex,
		elementsLeftInRecord;
	UINT32
		index,
		startIndex,
		length,
		startLength;

	if(numElements&&(startPosition<sparseArray->totalElements))	// make sure there's work to be done
	{
		if(numElements>=(sparseArray->totalElements-startPosition))
		{
			numElements=sparseArray->totalElements-startPosition;	// truncate down the number of elements to the number we have
		}

		LocateSARecord(sparseArray,startPosition,&chunk,&offset,&actualStart);

		sparseArray->totalElements-=numElements;			// removing at least this many elements from the array

		startIndex=index=HEADER_OFFSET(sparseArray,chunk,offset);
		startLength=length=HEADER_LENGTH(sparseArray,chunk,offset);

		startRecordIndex=startPosition-actualStart;			// find out how far into the starting record we are

		elementsLeftInRecord=index+length-startRecordIndex;

		firstChunk=chunk;									// hold onto these
		firstOffset=offset;

		numRecords=0;										// keep track of the number of records stepped over here

		while(numElements>elementsLeftInRecord)
		{
			numElements-=elementsLeftInRecord;
			numRecords++;
			offset++;
			if(offset>=chunk->totalElements)				// move to the next chunk
			{
				chunk=chunk->nextHeader;
				offset=0;
			}
			index=HEADER_OFFSET(sparseArray,chunk,offset);
			length=HEADER_LENGTH(sparseArray,chunk,offset);
			elementsLeftInRecord=index+length;
		}
		elementsLeftInRecord-=numElements;

		endRecordIndex=index+length-elementsLeftInRecord;

		if(numRecords)										// see if we stepped to a different record than we started in
		{
			if(startRecordIndex>startIndex)					// see if first record needs truncation or deletion
			{
				HEADER_LENGTH(sparseArray,firstChunk,firstOffset)=startRecordIndex-startIndex;
				firstOffset++;
				if(firstOffset>=firstChunk->totalElements)	// move to the next chunk to start deletion
				{
					firstChunk=firstChunk->nextHeader;
					firstOffset=0;
				}
				numRecords--;
			}
			if(endRecordIndex==index+length)				// does last record need deletion?
			{
				numRecords++;
			}
			if(numRecords)									// anything needs deletion?
			{
				DeleteUniverseArray(&sparseArray->elementChunks,firstChunk,firstOffset,numRecords,&chunk,&offset);
			}
			if(chunk)										// anything remaining after deletion?
			{
				if(endRecordIndex<index+length)				// see if last record was completely deleted, or needs to be cropped
				{
					if(endRecordIndex<=index)
					{
						if(startRecordIndex<=startIndex)	// see if the start record was deleted, and there is some offset which needs to be propagated to this record
						{
							HEADER_OFFSET(sparseArray,chunk,offset)+=startRecordIndex-endRecordIndex;	// add in the offset left over from the start
						}
						else
						{
							HEADER_OFFSET(sparseArray,chunk,offset)-=endRecordIndex;
							sparseArray->cacheCurrentElement=startPosition;	// start record was not deleted, so new record begins at start position
						}
					}
					else
					{
						if(startRecordIndex<=startIndex)	// see if the start record was deleted, and there is some offset which needs to be propagated to this record
						{
							HEADER_OFFSET(sparseArray,chunk,offset)=startRecordIndex;
						}
						else
						{
							HEADER_OFFSET(sparseArray,chunk,offset)=0;
							sparseArray->cacheCurrentElement=startPosition;	// start record was not deleted, so new record begins at start position
						}
						HEADER_LENGTH(sparseArray,chunk,offset)=index+length-endRecordIndex;
					}
				}
				else										// last record was deleted
				{
					if(startRecordIndex<=startIndex)		// see if the start record was deleted, and there is some offset which needs to be propagated to this record
					{
						HEADER_OFFSET(sparseArray,chunk,offset)+=startRecordIndex;	// add in the offset left over from the start
					}
					else
					{
						sparseArray->cacheCurrentElement=startPosition;	// start record was not deleted, so new record begins at start position
					}
				}
			}
			else
			{
				if(startRecordIndex<=startIndex)			// see if first record needed truncation or deletion
				{
					sparseArray->totalElements-=startRecordIndex;		// get rid of offset at end of array
				}
			}
			sparseArray->cacheChunkHeader=chunk;
			sparseArray->cacheOffset=offset;
		}
		else												// deletion began an ended in one record
		{
			if(endRecordIndex<index+length)					// ended within the record
			{
				if(endRecordIndex>index)					// ended within the length
				{
					if(startRecordIndex>=index)				// started within the length, so just reduce the length
					{
						HEADER_LENGTH(sparseArray,chunk,offset)-=endRecordIndex-startRecordIndex;
					}
					else									// started in the offset, ended in the length
					{
						HEADER_OFFSET(sparseArray,chunk,offset)-=index-startRecordIndex;
						HEADER_LENGTH(sparseArray,chunk,offset)-=endRecordIndex-index;
					}
				}
				else										// ended within the offset (so it started within the offset, so just reduce the offset)
				{
					HEADER_OFFSET(sparseArray,chunk,offset)-=endRecordIndex-startRecordIndex;
				}
			}
			else	// ended at the end of the record
			{
				if(startRecordIndex>index)					// ended at the end, but started within the length, so just truncate
				{
					HEADER_LENGTH(sparseArray,chunk,offset)=startRecordIndex-index;
				}
				else				// ended at the end, and started before the length, so kill this record, and apply remaining offset to next (if there is one)
				{
					DeleteUniverseArray(&sparseArray->elementChunks,chunk,offset,1,&chunk,&offset);
					if(chunk)
					{
						HEADER_OFFSET(sparseArray,chunk,offset)+=startRecordIndex;
					}
					else
					{
						sparseArray->totalElements-=startRecordIndex;		// get rid of offset at end of array
					}
					sparseArray->cacheChunkHeader=chunk;
					sparseArray->cacheOffset=offset;
				}
			}
		}
	}
}

bool InsertSARange(SPARSE_ARRAY *sparseArray,UINT32 startPosition,UINT32 numElements)
// insert numElements into the array at startPosition
// This will expand the array as needed, pushing out either the offset or length
// of a record (or doing nothing if the insertion is past the end)
{
	ARRAY_CHUNK_HEADER
		*chunk;
	UINT32
		offset;
	UINT32
		actualStart;
	bool
		fail;

	fail=false;
	if(numElements)			// if no elements to insert, then do nothing
	{
		LocateSARecord(sparseArray,startPosition,&chunk,&offset,&actualStart);
		if(chunk)		// did something already exist in the array at this point? If not, then do nothing
		{
			sparseArray->totalElements+=numElements;		// this many elements going in
			if((startPosition-actualStart)>HEADER_OFFSET(sparseArray,chunk,offset))
			{
				HEADER_LENGTH(sparseArray,chunk,offset)+=numElements;
			}
			else
			{
				HEADER_OFFSET(sparseArray,chunk,offset)+=numElements;
			}
		}
	}
	return(!fail);
}

bool GetSARecordForPosition(SPARSE_ARRAY *sparseArray,UINT32 position,UINT32 *startPosition,UINT32 *numElements,void **record)
// Return the pointer to the record of sparseArray which occupies position.
// If the passed position is past the end of the array, return false.
// If the passed position is in an empty part of the array, return record as NULL, with
// startPosition and numElements describing the empty part of the array.
// NOTE: this uses a cache so it is efficient to call
// repeatedly for positions which are nearby each other
// NOTE: the pointer is only valid until the next call to any sparse array routine
{
	ARRAY_CHUNK_HEADER
		*chunk;
	UINT32
		offset;
	UINT32
		headerOffset;

	LocateSARecord(sparseArray,position,&chunk,&offset,startPosition);
	if(chunk)
	{
		headerOffset=HEADER_OFFSET(sparseArray,chunk,offset);
		if(position>=(*startPosition+headerOffset))
		{
			*startPosition+=headerOffset;
			*numElements=HEADER_LENGTH(sparseArray,chunk,offset);
			*record=&chunk->data[offset*sparseArray->elementSize];
		}
		else
		{
			*numElements=headerOffset;
			*record=NULL;
		}
		return(true);
	}
	return(false);
}

bool GetSARecordAtOrAfterPosition(SPARSE_ARRAY *sparseArray,UINT32 position,UINT32 *startPosition,UINT32 *numElements,void **record)
// Return the pointer to the record of sparseArray which occupies position,
// or the first one after the position if the position is not occupied.
// If the passed position is past the end of the array, return false
// NOTE: this uses a cache so it is efficient to call
// repeatedly for positions which are nearby each other
// NOTE: the pointer is only valid until the next call to any sparse array routine
{
	ARRAY_CHUNK_HEADER
		*chunk;
	UINT32
		offset;

	LocateSARecord(sparseArray,position,&chunk,&offset,startPosition);
	if(chunk)
	{
		*startPosition+=HEADER_OFFSET(sparseArray,chunk,offset);
		*numElements=HEADER_LENGTH(sparseArray,chunk,offset);
		*record=&chunk->data[offset*sparseArray->elementSize];
		return(true);
	}
	return(false);
}

bool GetSARecordAtOrBeforePosition(SPARSE_ARRAY *sparseArray,UINT32 position,UINT32 *startPosition,UINT32 *numElements,void **record)
// Return the pointer to the record of sparseArray which occupies position,
// or the first one before the position if the position is not occupied.
// If the passed position is before the first element of the array, return false.
// NOTE: this uses a cache so it is efficient to call
// repeatedly for positions which are nearby each other
// NOTE: the pointer is only valid until the next call to any sparse array routine
{
	ARRAY_CHUNK_HEADER
		*chunk;
	UINT32
		offset;
	UINT32
		headerOffset;

	LocateSARecord(sparseArray,position,&chunk,&offset,startPosition);
	if(chunk)
	{
		headerOffset=HEADER_OFFSET(sparseArray,chunk,offset);
		if(position>=*startPosition+headerOffset)
		{
			*startPosition+=headerOffset;
			*numElements=HEADER_LENGTH(sparseArray,chunk,offset);
			*record=&chunk->data[offset*sparseArray->elementSize];
			return(true);
		}
		else	// get element that precedes this one
		{
			if(offset)							// step back one
			{
				offset--;
			}
			else
			{
				if((chunk=chunk->previousHeader))
				{
					offset=chunk->totalElements-1;
				}
			}
			if(chunk)
			{
				*startPosition-=HEADER_LENGTH(sparseArray,chunk,offset);
				*numElements=HEADER_LENGTH(sparseArray,chunk,offset);
				*record=&chunk->data[offset*sparseArray->elementSize];
				return(true);
			}
		}
	}
	else
	{
		if((chunk=sparseArray->elementChunks.lastChunkHeader))
		{
			offset=sparseArray->elementChunks.lastChunkHeader->totalElements;
			*startPosition=sparseArray->totalElements-HEADER_LENGTH(sparseArray,chunk,offset);
			*numElements=HEADER_LENGTH(sparseArray,chunk,offset);
			*record=&chunk->data[offset*sparseArray->elementSize];
			return(true);
		}
	}
	return(false);
}

bool ClearSARange(SPARSE_ARRAY *sparseArray,UINT32 startPosition,UINT32 numElements)
// Clear out elements of sparseArray from startPosition for numElements
// if there is a problem, set the error and return false
// this will overwrite anything currently in the array at
// the given positions, and will cause items to be split
// if it bisects them.
{
	ARRAY_CHUNK_HEADER
		*chunk,
		*splitChunk;
	UINT32
		offset,
		splitOffset;
	UINT32
		actualStart;
	bool
		fail;

	fail=false;
	if(numElements)
	{
		DeleteSARange(sparseArray,startPosition,numElements);	// get rid of any elements that were in that spot
		LocateSARecord(sparseArray,startPosition,&chunk,&offset,&actualStart);	// find the record containing the start
		if(chunk)												// if there is a record there, see how we need to clear
		{
			if(actualStart+HEADER_OFFSET(sparseArray,chunk,offset)<startPosition)	// need to split?
			{
				if(InsertUniverseArray(&sparseArray->elementChunks,chunk,offset,1,&chunk,&offset))	// add 1 element for the split we need
				{
					sparseArray->cacheChunkHeader=chunk;		// adjust cache for new chunk just added
					sparseArray->cacheOffset=offset;

					splitOffset=offset;
					splitChunk=chunk;
					splitOffset++;
					if(splitOffset>=splitChunk->totalElements)	// move to the next chunk (which is the original we need to split)
					{
						splitChunk=splitChunk->nextHeader;
						splitOffset=0;
					}

					memcpy(&chunk->data[offset*sparseArray->elementSize],&splitChunk->data[splitOffset*sparseArray->elementSize],sparseArray->elementSize);	// blast in split value

					// finally, adjust offsets and lengths
					HEADER_LENGTH(sparseArray,chunk,offset)=startPosition-actualStart-HEADER_OFFSET(sparseArray,chunk,offset);	// knock off some length on the first item
					HEADER_OFFSET(sparseArray,splitChunk,splitOffset)=numElements;
					HEADER_LENGTH(sparseArray,splitChunk,splitOffset)-=startPosition-actualStart-HEADER_OFFSET(sparseArray,chunk,offset);

					sparseArray->totalElements+=numElements;
				}
				else
				{
					fail=true;
				}
			}
			else	// no splitting, just push other record over by the amount to clear
			{
				HEADER_OFFSET(sparseArray,chunk,offset)+=numElements;
				sparseArray->totalElements+=numElements;
			}
		}
	}
	return(!fail);
}

bool SetSARange(SPARSE_ARRAY *sparseArray,UINT32 startPosition,UINT32 numElements,void *value)
// Set elements of sparseArray from startPosition for numElements
// to value
// if there is a problem, set the error and return false
// the value will overwrite anything currently in the array at
// the given positions, and will cause items to be split
// if it bisects them.
// NOTE: if the range goes beyond the current end of the array,
// the array will be extended.
{
	ARRAY_CHUNK_HEADER
		*chunk,
		*newChunk,
		*splitChunk;
	UINT32
		offset,
		newOffset,
		splitOffset;
	UINT32
		actualStart;
	bool
		fail;

	fail=false;
	if(numElements)
	{
		DeleteSARange(sparseArray,startPosition,numElements);	// get rid of any elements that were in that spot
		LocateSARecord(sparseArray,startPosition,&chunk,&offset,&actualStart);	// find the record containing the start
		if(chunk)												// if there is a record there, see how we need to add
		{
			if(actualStart+HEADER_OFFSET(sparseArray,chunk,offset)<startPosition)	// need to split?
			{
				if(InsertUniverseArray(&sparseArray->elementChunks,chunk,offset,2,&chunk,&offset))	// add 2 elements, one for the new stuff, and one for the half of the split we need
				{
					sparseArray->cacheChunkHeader=chunk;		// adjust cache for new chunks just added
					sparseArray->cacheOffset=offset;

					newOffset=offset;
					newChunk=chunk;
					newOffset++;
					if(newOffset>=newChunk->totalElements)		// move to the next chunk
					{
						newChunk=newChunk->nextHeader;
						newOffset=0;
					}

					splitOffset=newOffset;
					splitChunk=newChunk;
					splitOffset++;
					if(splitOffset>=splitChunk->totalElements)	// move to the next chunk (which is the original we need to split)
					{
						splitChunk=splitChunk->nextHeader;
						splitOffset=0;
					}

					memcpy(&newChunk->data[newOffset*sparseArray->elementSize],value,sparseArray->elementSize);	// blast in new value
					memcpy(&chunk->data[offset*sparseArray->elementSize],&splitChunk->data[splitOffset*sparseArray->elementSize],sparseArray->elementSize);	// blast in split value

					// finally, adjust offsets and lengths
					HEADER_LENGTH(sparseArray,chunk,offset)=startPosition-actualStart-HEADER_OFFSET(sparseArray,chunk,offset);	// knock off some length on the first item
					HEADER_OFFSET(sparseArray,newChunk,newOffset)=0;
					HEADER_LENGTH(sparseArray,newChunk,newOffset)=numElements;
					HEADER_OFFSET(sparseArray,splitChunk,splitOffset)=0;
					HEADER_LENGTH(sparseArray,splitChunk,splitOffset)-=startPosition-actualStart-HEADER_OFFSET(sparseArray,chunk,offset);

					sparseArray->totalElements+=numElements;
				}
				else
				{
					fail=true;
				}
			}
			else	// no splitting, just push other record over, and add this one
			{
				HEADER_OFFSET(sparseArray,chunk,offset)-=startPosition-actualStart;	// reduce the offset of this element
				if(InsertUniverseArray(&sparseArray->elementChunks,chunk,offset,1,&chunk,&offset))	// add an element
				{
					memcpy(&chunk->data[offset*sparseArray->elementSize],value,sparseArray->elementSize);	// blast in new value
					HEADER_OFFSET(sparseArray,chunk,offset)=startPosition-actualStart;
					HEADER_LENGTH(sparseArray,chunk,offset)=numElements;
					sparseArray->cacheCurrentElement=actualStart;
					sparseArray->cacheChunkHeader=chunk;
					sparseArray->cacheOffset=offset;

					sparseArray->totalElements+=numElements;
				}
				else
				{
					fail=true;	// things will not be pretty if this fails, as it leaves the array in an inconsistent state
				}
			}
		}
		else	// no record, so must be at the end, so add a new record there, and set the needed offset at the start (leave cache set to null)
		{
			if(InsertUniverseArray(&sparseArray->elementChunks,NULL,0,1,&chunk,&offset))	// add an element
			{
				memcpy(&chunk->data[offset*sparseArray->elementSize],value,sparseArray->elementSize);	// blast in new value
				HEADER_OFFSET(sparseArray,chunk,offset)=startPosition-sparseArray->totalElements;
				HEADER_LENGTH(sparseArray,chunk,offset)=numElements;

				sparseArray->totalElements=startPosition+numElements;
			}
			else
			{
				fail=true;
			}
		}
	}
	return(!fail);
}

bool IsSAEmpty(SPARSE_ARRAY *sparseArray)
// return true if the given sparse array contains no elements, false otherwise
{
	if(sparseArray->totalElements)
	{
		return(false);
	}
	return(true);
}

bool GetSAElementEnds(SPARSE_ARRAY *sparseArray,UINT32 *startElementOffset,UINT32 *endElementOffset)
// if the given sparse array contains elements, fill in startElementOffset as the
// index to the first used element in the array, and endElementOffset as the offset
// of the last used element in the array (+1), then return true
// otherwise, return false.
{
	if(sparseArray->totalElements)
	{
		(*startElementOffset)=HEADER_OFFSET(sparseArray,sparseArray->elementChunks.firstChunkHeader,0);
		(*endElementOffset)=sparseArray->totalElements;
		return(true);
	}
	return(false);
}

void UnInitSA(SPARSE_ARRAY *sparseArray)
// uninitialize a sparse array
{
	UnInitArrayUniverse(&(sparseArray->elementChunks));
}

bool InitSA(SPARSE_ARRAY *sparseArray,UINT32 elementSize)
// initialize a sparse array
// if there is a problem, set the error, and return false
{
	sparseArray->elementSize=elementSize;	// remember this as we need it when moving through the array
	sparseArray->totalElements=0;

	sparseArray->cacheChunkHeader=NULL;		// initialize the cache handling
	sparseArray->cacheOffset=0;
	sparseArray->cacheCurrentElement=0;

	if(InitArrayUniverse(&(sparseArray->elementChunks),256,elementSize))
	{
		return(true);
	}
	return(false);
}
