// Style list management functions
// Copyright (C) 2000 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

// This code handles keeping track of style in the text
// it uses the sparse-array concept to hold the style information
// NOTE: styles are coalesced in the array to make them work
// in a more intuitive manner

#include	"includes.h"

typedef struct styleElement
{
	SPARSE_ARRAY_HEADER
		saHeader;				// header needed for sparse array code
	UINT32
		style;					// style index
} STYLE_ELEMENT;

typedef struct styleUniverse
{
	SPARSE_ARRAY
		styleArray;
} STYLE_UNIVERSE;

static void CoalescePosition(STYLE_UNIVERSE *universe,UINT32 position)
// look at the style at position, and position-1
// if they are the same, coalesce the array
{
	STYLE_ELEMENT
		*leftElement,
		*rightElement;
	STYLE_ELEMENT
		newElement;
	UINT32
		leftStartPosition,
		leftElements,
		rightStartPosition,
		rightElements;

	if(position>0)
	{
		if(GetSARecordForPosition(&universe->styleArray,position-1,&leftStartPosition,&leftElements,(void **)&leftElement))
		{
			if(leftElement&&(leftStartPosition+leftElements==position))	// make sure there's something to the left, and it ends just before position
			{
				if(GetSARecordForPosition(&universe->styleArray,position,&rightStartPosition,&rightElements,(void **)&rightElement))
				{
					if(rightElement&&(leftElement->style==rightElement->style))	// make sure there's something to the right, and it matches
					{
						newElement.style=leftElement->style;
						SetSARange(&universe->styleArray,leftStartPosition,leftElements+rightElements,&newElement);
						// @@@ this might fail, but the routine should never fail
						// Look into changing this to call "MergeSARange"
					}
				}
			}
		}
	}
}

void DeleteStyleRange(STYLE_UNIVERSE *universe,UINT32 startPosition,UINT32 numElements)
// Remove style array elements from startPosition for numElements
// NOTE: this does not mind being called for positions outside of the current array
{
	DeleteSARange(&universe->styleArray,startPosition,numElements);
	CoalescePosition(universe,startPosition);	// if two of the same style ended up next to each other, merge them
}

bool InsertStyleRange(STYLE_UNIVERSE *universe,UINT32 startPosition,UINT32 numElements)
// insert numElements into the style array at startPosition
{
	return(InsertSARange(&universe->styleArray,startPosition,numElements));
}

bool GetStyleRange(STYLE_UNIVERSE *universe,UINT32 position,UINT32 *startPosition,UINT32 *numElements,UINT32 *style)
// Return the range of positions in the array which include position
// and have the same style as position
// NOTE: this uses a cache so it is efficient to call
// repeatedly for positions which are nearby each other
// NOTE: if position is off the end of the array, style will be returned as 0, and
// this will return false
{
	STYLE_ELEMENT
		*element;

	*style=0;
	if(GetSARecordForPosition(&universe->styleArray,position,startPosition,numElements,(void **)&element))
	{
		if(element)
		{
			*style=element->style;
		}
		return(true);
	}
	return(false);
}

UINT32 GetStyleAtPosition(STYLE_UNIVERSE *universe,UINT32 position)
// call this to return the style at the given position of
// the given style universe.
// NOTE: if position is off the end of the array, this will return style 0.
// NOTE: this uses a cache so it is efficient to call
// repeatedly for positions which are nearby each other
{
	UINT32
		startPosition,
		numElements;
	UINT32
		style;

	if(!GetStyleRange(universe,position,&startPosition,&numElements,&style))
	{
		style=0;			// default to style 0
	}
	return(style);
}

bool SetStyleRange(STYLE_UNIVERSE *universe,UINT32 startPosition,UINT32 numElements,UINT32 style)
// set style at startPosition for numElements of universe
// if there is a problem, set the error and return false
// the style will overwrite anything currently in the array at
// the given positions
// NOTE: if the range goes beyond the current end of the array,
// the array will be extended by setting style 0 from the end of
// the array until startPosition, and then placing numElements of
// the given style into the array at startPosition
{
	STYLE_ELEMENT
		element;

	if(style)
	{
		element.style=style;
		if(SetSARange(&universe->styleArray,startPosition,numElements,&element))
		{
			CoalescePosition(universe,startPosition);
			CoalescePosition(universe,startPosition+numElements);
			return(true);
		}
		return(false);
	}
	else
	{
		return(ClearSARange(&universe->styleArray,startPosition,numElements));
	}
}

bool SetStyleAtPosition(STYLE_UNIVERSE *universe,UINT32 position,UINT32 style)
// set style at position of universe
// if there is a problem, set the error and return false
{
	return(SetStyleRange(universe,position,1,style));
}

void AdjustStylesForChange(STYLE_UNIVERSE *universe,UINT32 startOffset,UINT32 oldEndOffset,UINT32 newEndOffset)
// given a change at startOffset, run through the given style list, and fix it up
{
	DeleteStyleRange(universe,startOffset,oldEndOffset-startOffset);
	InsertStyleRange(universe,startOffset,newEndOffset-startOffset);
}

bool CopyStyle(STYLE_UNIVERSE *sourceUniverse,UINT32 sourceOffset,STYLE_UNIVERSE *destUniverse,UINT32 destOffset,UINT32 numElements)
// copy numElements of style information from sourceUniverse starting at sourceOffset
// to destUniverse at destOffset
// If there is a problem, set the error, and return false
{
	bool
		fail;
	UINT32
		startPosition;
	UINT32
		elementCount;
	UINT32
		style;
	UINT32
		elementsToSet;

	fail=false;
	while(!fail&&numElements)
	{
		if(GetStyleRange(sourceUniverse,sourceOffset,&startPosition,&elementCount,&style))
		{
			elementsToSet=elementCount-(sourceOffset-startPosition);
			if(elementsToSet>numElements)
			{
				elementsToSet=numElements;
			}
			if(SetStyleRange(destUniverse,destOffset,elementsToSet,style))
			{
				sourceOffset+=elementsToSet;
				destOffset+=elementsToSet;
				numElements-=elementsToSet;
			}
			else
			{
				fail=true;
			}
		}
		else	// ran off the end of the source universe, so treat the rest as style 0
		{
			fail=!SetStyleRange(destUniverse,destOffset,numElements,0);
			numElements=0;	// terminate the loop
		}
	}
	return(!fail);
}

void CloseStyleUniverse(STYLE_UNIVERSE *universe)
// dispose of a style universe
{
	UnInitSA(&(universe->styleArray));
	MDisposePtr(universe);
}

STYLE_UNIVERSE *OpenStyleUniverse()
// open a style universe (with no elements)
// if there is a problem, set the error, and return NULL
{
	STYLE_UNIVERSE
		*universe;

	if((universe=(STYLE_UNIVERSE *)MNewPtr(sizeof(STYLE_UNIVERSE))))
	{
		if(InitSA(&(universe->styleArray),sizeof(STYLE_ELEMENT)))
		{
			return(universe);
		}
		MDisposePtr(universe);
	}
	return(NULL);
}
