// Undo/redo functions
// Copyright (C) 2000 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

static void DoClearUndo(UNDO_GROUP *group)
// Clear the text, style, and undo arrays from group
{
	ARRAY_CHUNK_HEADER
		*undoChunk;
	UINT32
		undoOffset;

	DeleteUniverseArray(&(group->undoArray),group->undoArray.firstChunkHeader,0,group->undoArray.totalElements,&undoChunk,&undoOffset);
	DeleteStyleRange(group->undoStyles,0,group->undoText->totalBytes);
	DeleteUniverseText(group->undoText,0,group->undoText->totalBytes);
}

void BeginUndoGroup(EDITOR_BUFFER *buffer)
// All of the things that are done between now, and the matching EndUndoGroup
// will be remembered as a group in the undo buffer
{
	if(!buffer->undoUniverse->groupingDepth++)			// increment the depth, if it was 0, then start a new group
	{
		buffer->undoUniverse->currentFrame++;			// bump the frame number
	}
}

void EndUndoGroup(EDITOR_BUFFER *buffer)
// Call this to balance a call to BeginUndoGroup
// It ends the undo group, but will allow subsequent undos to be joined
{
	if(buffer->undoUniverse->groupingDepth)
	{
		--buffer->undoUniverse->groupingDepth;			// decrement the group nesting count
	}
}

void StrictEndUndoGroup(EDITOR_BUFFER *buffer)
// Call this to balance a call to BeginUndoGroup
// It ends the undo group, and will NOT allow subsequent undos to be joined
{
	if(buffer->undoUniverse->groupingDepth)
	{
		if(!(--buffer->undoUniverse->groupingDepth))	// decrement the group nesting count, if it went to 0, force new group
		{
			buffer->undoUniverse->currentFrame++;		// bump the frame number
		}
	}
}

static bool CreateUndoDeleteElement(EDITOR_BUFFER *buffer,UINT32 startPosition,UINT32 numBytes,UNDO_GROUP *group)
// Create a new element at the end of the undo array
// If there is a problem, set the error, and return false
{
	CHUNK_HEADER
		*chunk;
	UINT32
		offset;
	ARRAY_CHUNK_HEADER
		*undoChunk;
	UINT32
		undoOffset;
	UNDO_ELEMENT
		*undoElement;

	if(!buffer->undoUniverse->groupingDepth)							// if not currently in a group, then make this a new frame
	{
		buffer->undoUniverse->currentFrame++;							// bump the frame number
	}
	if(InsertUniverseArray(&(group->undoArray),NULL,0,1,&undoChunk,&undoOffset))		// create the element
	{
		undoElement=&(((UNDO_ELEMENT *)undoChunk->data)[undoOffset]);
		undoElement->frameNumber=buffer->undoUniverse->currentFrame;	// set the frame number to the current undo frame
		undoElement->textStartPosition=startPosition;					// when the text is blasted back into place, this is where it goes
		undoElement->textLength=0;										// number of bytes to overwrite (none, we are deleting)
		undoElement->undoStartPosition=group->undoText->totalBytes;		// start at what is now the current end of the undo text
		undoElement->undoLength=numBytes;								// number of bytes to put back
		PositionToChunkPosition(buffer->textUniverse,startPosition,&chunk,&offset);	// point to bytes about to be deleted
		if(CopyStyle(buffer->styleUniverse,startPosition,group->undoStyles,group->undoText->totalBytes,numBytes))
		{
			if(InsertUniverseChunks(group->undoText,group->undoText->totalBytes,chunk,offset,numBytes))	// copy into undo holding area
			{
				return(true);											// if we got them, then we are done
			}
			DeleteStyleRange(group->undoStyles,group->undoText->totalBytes,numBytes);	// kill the styles we copied
		}
		DeleteUniverseArray(&(group->undoArray),undoChunk,undoOffset,1,&undoChunk,&undoOffset);	// we failed, so get rid of the last element
	}
	return(false);
}

static bool MergeUndoDeleteElement(EDITOR_BUFFER *buffer,UNDO_ELEMENT *lastElement,UINT32 startPosition,UINT32 numBytes,UNDO_GROUP *group)
// A delete is happening, and the last element of the undo list is in the current
// frame. See if the delete can be combined with the last element of the undo list,
// or if another entirely new element needs to be created
// If there is a problem, set the error, and return false
{
	CHUNK_HEADER
		*chunk;
	UINT32
		offset;
	ARRAY_CHUNK_HEADER
		*undoChunk;
	UINT32
		undoOffset;
	UINT32
		bytesToEdge;

	if((lastElement->textStartPosition==startPosition)||(lastElement->textStartPosition+lastElement->textLength==startPosition+numBytes))	// for a delete to be grouped, it must start exactly in line with last element, or end exactly in line with it
	{
		if(startPosition>=lastElement->textStartPosition)
		{
			bytesToEdge=lastElement->textStartPosition+lastElement->textLength-startPosition;
			if(bytesToEdge>=numBytes)
			{
				lastElement->textLength-=numBytes;
				if(!lastElement->textLength&&!lastElement->undoLength)	// if this element has become empty, then get rid of it
				{
					DeleteUniverseArray(&(group->undoArray),group->undoArray.lastChunkHeader,group->undoArray.lastChunkHeader->totalElements-1,1,&undoChunk,&undoOffset);
				}
				return(true);											// all that was needed was to remove bytes
			}
			else
			{
				startPosition+=bytesToEdge;
				numBytes-=bytesToEdge;
				PositionToChunkPosition(buffer->textUniverse,startPosition,&chunk,&offset);	// point to bytes about to be deleted

				if(CopyStyle(buffer->styleUniverse,startPosition,group->undoStyles,group->undoText->totalBytes,numBytes))
				{
					if(InsertUniverseChunks(group->undoText,group->undoText->totalBytes,chunk,offset,numBytes))		// copy into undo holding area
					{
						lastElement->textLength-=bytesToEdge;
						lastElement->undoLength+=numBytes;
						return(true);																// if we got them, then we are done
					}
					DeleteStyleRange(group->undoStyles,group->undoText->totalBytes,numBytes);		// kill the styles we copied
				}
				return(false);
			}
		}
		else
		{
			if((startPosition+numBytes)<=(lastElement->textStartPosition+lastElement->textLength))	// make sure it does not span entire element
			{
				bytesToEdge=startPosition+numBytes-lastElement->textStartPosition;
				numBytes-=bytesToEdge;
				PositionToChunkPosition(buffer->textUniverse,startPosition,&chunk,&offset);			// point to bytes about to be deleted
				if(InsertStyleRange(group->undoStyles,lastElement->undoStartPosition,numBytes))		// move old styles over to account for new ones
				{
					if(CopyStyle(buffer->styleUniverse,startPosition,group->undoStyles,lastElement->undoStartPosition,numBytes))
					{
						if(InsertUniverseChunks(group->undoText,lastElement->undoStartPosition,chunk,offset,numBytes))	// copy into undo holding area
						{
							lastElement->textStartPosition=startPosition;
							lastElement->textLength-=bytesToEdge;
							lastElement->undoLength+=numBytes;
							return(true);															// if we got them, then we are done
						}
					}
					DeleteStyleRange(group->undoStyles,lastElement->undoStartPosition,numBytes);	// kill the styles we inserted
				}
				return(false);
			}
			else
			{
				return(CreateUndoDeleteElement(buffer,startPosition,numBytes,group));				// this delete spans entire previous element, so make new one
			}
		}
	}
// if we cannot merge, then create a new element
	return(CreateUndoDeleteElement(buffer,startPosition,numBytes,group));	// delete in a different place
}

static bool RegisterDelete(EDITOR_BUFFER *buffer,UINT32 startPosition,UINT32 numBytes,UNDO_GROUP *group,bool allowMerge)
// Register the delete into the given text buffer, and chunk array
{
	ARRAY_CHUNK_HEADER
		*lastUndoChunk;
	UNDO_ELEMENT
		*lastElement;

	if((lastUndoChunk=group->undoArray.lastChunkHeader))									// is there a current item in the array?
	{
		lastElement=&(((UNDO_ELEMENT *)lastUndoChunk->data)[lastUndoChunk->totalElements-1]);	// point to last element of undo array
		if(allowMerge&&(buffer->undoUniverse->currentFrame==lastElement->frameNumber))		// is it in the current frame?
		{
			return(MergeUndoDeleteElement(buffer,lastElement,startPosition,numBytes,group));
		}
		else
		{
			return(CreateUndoDeleteElement(buffer,startPosition,numBytes,group));
		}
	}
	else
	{
		return(CreateUndoDeleteElement(buffer,startPosition,numBytes,group));
	}
}

bool RegisterUndoDelete(EDITOR_BUFFER *buffer,UINT32 startPosition,UINT32 numBytes,bool allowMerge)
// Call this BEFORE deleting numBytes of text from startPosition of buffer
// If there is a problem, return false
// NOTE: for speed, it is the caller's responsibility to determine chunk and offset
// which point to startPosition of buffer
{
	UNDO_UNIVERSE
		*undoUniverse;

	if(numBytes)
	{
		undoUniverse=buffer->undoUniverse;

		if(!undoUniverse->undoActive)
		{
			if(undoUniverse->cleanElements>undoUniverse->undoGroup.undoArray.totalElements)	// if position which is considered "clean" is beyond where we are now, it becomes meaningless
			{
				undoUniverse->cleanKnown=false;
			}
			DoClearUndo(&(undoUniverse->redoGroup));	// get rid of redo information
		}
		if(undoUniverse->undoing)						// if undoing, then place undo info into redo buffers
		{
			return(RegisterDelete(buffer,startPosition,numBytes,&(undoUniverse->redoGroup),allowMerge));
		}
		else
		{
			return(RegisterDelete(buffer,startPosition,numBytes,&(undoUniverse->undoGroup),allowMerge));
		}
	}
	else
	{
		return(true);
	}
}

static bool CreateUndoInsertElement(EDITOR_BUFFER *buffer,UINT32 startPosition,UINT32 numBytes,UNDO_GROUP *group)
// Create a new element at the end of the undo array
// If there is a problem, set the error, and return false
{
	ARRAY_CHUNK_HEADER
		*undoChunk;
	UINT32
		undoOffset;
	UNDO_ELEMENT
		*undoElement;

	if(!buffer->undoUniverse->groupingDepth)							// if not currently in a group, then make this a new frame
	{
		buffer->undoUniverse->currentFrame++;							// bump the frame number
	}
	if(InsertUniverseArray(&(group->undoArray),NULL,0,1,&undoChunk,&undoOffset))	// create the element
	{
		undoElement=&(((UNDO_ELEMENT *)undoChunk->data)[undoOffset]);
		undoElement->frameNumber=buffer->undoUniverse->currentFrame;	// set the frame number to the current undo frame
		undoElement->textStartPosition=startPosition;					// when the text is blasted back into place, this is where it goes
		undoElement->textLength=numBytes;								// number of bytes to overwrite (all that is being inserted)
		undoElement->undoStartPosition=group->undoText->totalBytes;		// start at what is now the current end of the undo text
		undoElement->undoLength=0;										// number of bytes to put back (none, we are inserting)
		return(true);
	}
	return(false);
}

static bool MergeUndoInsertElement(EDITOR_BUFFER *buffer,UNDO_ELEMENT *lastElement,UINT32 startPosition,UINT32 numBytes,UNDO_GROUP *group)
// An insert is happening, and the last element of the undo list is in the current
// frame. See if the insert can be combined with the last element of the undo list,
// or if another entirely new element needs to be created
// If there is a problem, set the error, and return false
{
	if(startPosition==lastElement->textStartPosition+lastElement->textLength)	// see if insert just after end of current start
	{
		lastElement->textLength+=numBytes;								// add these bytes to what should be deleted during undo
		return(true);
	}
// if we cannot merge, then create a new element
	return(CreateUndoInsertElement(buffer,startPosition,numBytes,group));	// insert in a different place
}

static bool RegisterInsert(EDITOR_BUFFER *buffer,UINT32 startPosition,UINT32 numBytes,UNDO_GROUP *group,bool allowMerge)
// Register the insert into the given text buffer, and chunk array
// If there is a problem, set the error, and return false
{
	ARRAY_CHUNK_HEADER
		*lastUndoChunk;
	UNDO_ELEMENT
		*lastElement;

	if((lastUndoChunk=group->undoArray.lastChunkHeader))							// is there a current item in the array?
	{
		lastElement=&(((UNDO_ELEMENT *)lastUndoChunk->data)[lastUndoChunk->totalElements-1]);	// point to last element of undo array
		if(allowMerge&&(buffer->undoUniverse->currentFrame==lastElement->frameNumber))			// is it in the current frame?
		{
			return(MergeUndoInsertElement(buffer,lastElement,startPosition,numBytes,group));
		}
		else
		{
			return(CreateUndoInsertElement(buffer,startPosition,numBytes,group));
		}
	}
	else
	{
		return(CreateUndoInsertElement(buffer,startPosition,numBytes,group));
	}
}

bool RegisterUndoInsert(EDITOR_BUFFER *buffer,UINT32 startPosition,UINT32 numBytes,bool allowMerge)
// Call this AFTER inserting numBytes of text into buffer at startPosition
// If there is a problem, set the error, return false
{
	UNDO_UNIVERSE
		*undoUniverse;

	if(numBytes)
	{
		undoUniverse=buffer->undoUniverse;

		if(!undoUniverse->undoActive)
		{
			if(undoUniverse->cleanElements>undoUniverse->undoGroup.undoArray.totalElements)	// if position which is considered "clean" is beyond where we are now, it becomes meaningless
			{
				undoUniverse->cleanKnown=false;
			}
			DoClearUndo(&(undoUniverse->redoGroup));	// get rid of redo information
		}
		if(undoUniverse->undoing)						// if undoing, then place undo info into redo buffers
		{
			return(RegisterInsert(buffer,startPosition,numBytes,&(undoUniverse->redoGroup),allowMerge));
		}
		else
		{
			return(RegisterInsert(buffer,startPosition,numBytes,&(undoUniverse->undoGroup),allowMerge));
		}
	}
	else
	{
		return(true);
	}
}

static bool UndoSelectionChange(SELECTION_UNIVERSE *selectionUniverse,UINT32 position,UINT32 numBytesOut,UINT32 numBytesIn)
// Handle updating the selection, an creating new selections during undo
// NOTE: if something goes wrong, attempt to leave the selection universe in a sane state
{
	AdjustSelectionsForChange(selectionUniverse,position,position+numBytesOut,position+numBytesIn);
	if(SetSelectionRange(selectionUniverse,position,numBytesIn))
	{
		return(true);
	}
	return(false);
}

static bool DoUndo(EDITOR_BUFFER *buffer,UNDO_GROUP *group,bool *didSomething)
// Undo from the given text and array into buffer
// NOTE: if there is nothing to undo, return false in *didSomething
// else return true.
// If something bad happens while trying to undo, set the error, and return false
{
	SELECTION_UNIVERSE
		*selectionUniverse,
		*newSelectionUniverse;
	ARRAY_CHUNK_HEADER
		*lastUndoChunk;
	UINT32
		tempOffset;
	UNDO_ELEMENT
		*lastElement;
	UINT32
		frameToUse;
	bool
		haveMinimumPosition;
	UINT32
		minimumPosition;													// used to set the cursor position after the change
	CHUNK_HEADER
		*chunk;
	UINT32
		offset;
	bool
		fail;

	fail=*didSomething=false;
	haveMinimumPosition=false;
	minimumPosition=0;
	if((lastUndoChunk=group->undoArray.lastChunkHeader))					// is there a current item in the array?
	{
		lastElement=&(((UNDO_ELEMENT *)lastUndoChunk->data)[lastUndoChunk->totalElements-1]);	// point to last element of undo array
		frameToUse=lastElement->frameNumber;								// undo all elements of this frame

		if((newSelectionUniverse=OpenSelectionUniverse()))					// create a new selection universe where the output selection will be created
		{
			selectionUniverse=buffer->selectionUniverse;					// point to current selection universe
			EditorStartSelectionChange(buffer);
			DeleteAllSelections(selectionUniverse);							// remove all old selections
			EditorEndSelectionChange(buffer);

			EditorStartReplace(buffer);
			BeginUndoGroup(buffer);											// group the redo of this undo
			while(!fail&&lastElement&&lastElement->frameNumber==frameToUse)
			{
				PositionToChunkPosition(group->undoText,lastElement->undoStartPosition,&chunk,&offset);	// point to bytes about to be placed into the text
				if(ReplaceEditorChunks(buffer,lastElement->textStartPosition,lastElement->textStartPosition+lastElement->textLength,chunk,offset,lastElement->undoLength,false))
				{
					if(CopyStyle(group->undoStyles,lastElement->undoStartPosition,buffer->styleUniverse,lastElement->textStartPosition,lastElement->undoLength))
					{
						if(UndoSelectionChange(newSelectionUniverse,lastElement->textStartPosition,lastElement->textLength,lastElement->undoLength))
						{
							if(haveMinimumPosition)
							{
								if(lastElement->textStartPosition<minimumPosition)
								{
									minimumPosition=lastElement->textStartPosition;
								}
							}
							else
							{
								minimumPosition=lastElement->textStartPosition;
								haveMinimumPosition=true;
							}

							DeleteStyleRange(group->undoStyles,lastElement->undoStartPosition,lastElement->undoLength);
							DeleteUniverseText(group->undoText,lastElement->undoStartPosition,lastElement->undoLength);
							DeleteUniverseArray(&(group->undoArray),lastUndoChunk,lastUndoChunk->totalElements-1,1,&lastUndoChunk,&tempOffset);
							if((lastUndoChunk=group->undoArray.lastChunkHeader))		// fetch the last element again
							{
								lastElement=&(((UNDO_ELEMENT *)lastUndoChunk->data)[lastUndoChunk->totalElements-1]);	// point to last element of undo array
							}
							else
							{
								lastElement=NULL;
							}
						}
						else
						{
							fail=true;
						}
					}
					else
					{
						fail=true;
					}
				}
				else
				{
					fail=true;
				}
			}

			EditorEndReplace(buffer);
			StrictEndUndoGroup(buffer);										// end this group, do not let anything combine with it

			// place new selection into the universe

			EditorStartSelectionChange(buffer);
			SetSelectionCursorPosition(newSelectionUniverse,minimumPosition);	// place cursor at the start
			buffer->selectionUniverse=newSelectionUniverse;
			CloseSelectionUniverse(selectionUniverse);						// close old universe
			EditorEndSelectionChange(buffer);

			*didSomething=true;
		}
		else
		{
			fail=true;
		}
	}
	return(!fail);
}

bool EditorUndo(EDITOR_BUFFER *buffer,bool *didSomething)
// Undo whatever was done last (if there is nothing in the undo buffer
// then do not do anything)
// If there is no undo information, return false
// otherwise return true.
// If there is a problem, set the error, return false
{
	bool
		result;

	*didSomething=false;
	ShowBusy();
	buffer->undoUniverse->undoActive=true;							// changes are coming from undo/redo, not user
	buffer->undoUniverse->undoing=true;								// remember that we are undoing so if register calls come in, we can place them in redo buffer
	result=DoUndo(buffer,&(buffer->undoUniverse->undoGroup),didSomething);
	buffer->undoUniverse->undoing=false;
	buffer->undoUniverse->undoActive=false;
	ShowNotBusy();
	return(result);
}

bool EditorRedo(EDITOR_BUFFER *buffer,bool *didSomething)
// Redo whatever was undone last
// If there is no undo information, return false in *didSomething
// otherwise return true.
// If there is a problem, set the error, return false
{
	bool
		result;

	*didSomething=false;
	ShowBusy();
	buffer->undoUniverse->undoActive=true;							// changes are coming from undo/redo, not user
	result=DoUndo(buffer,&(buffer->undoUniverse->redoGroup),didSomething);
	buffer->undoUniverse->undoActive=false;
	ShowNotBusy();
	return(result);
}

bool EditorToggleUndo(EDITOR_BUFFER *buffer,bool *didSomething)
// If there is a redo that can be done, then do it, otherwise, do an undo
// If neither can be done, return didSomething=false
// If there is a problem, set the error, return false
{
	bool
		result;

	if((result=EditorRedo(buffer,didSomething)))	// try to redo
	{
		if(!(*didSomething))						// if nothing to redo, then try to undo
		{
			return(EditorUndo(buffer,didSomething));
		}
	}
	return(result);
}

void EditorClearUndo(EDITOR_BUFFER *buffer)
// When, undo/redo must be cleared, call this to do it
{
	if(buffer->undoUniverse->cleanKnown)	// if clean position is known, then make sure we are there, otherwise we don't know it anymore
	{
		if(buffer->undoUniverse->cleanElements!=buffer->undoUniverse->undoGroup.undoArray.totalElements)
		{
			buffer->undoUniverse->cleanKnown=false;
		}
	}
	buffer->undoUniverse->cleanElements=0;				// clean position moves to 0

	DoClearUndo(&(buffer->undoUniverse->undoGroup));	// get rid of undo information
	DoClearUndo(&(buffer->undoUniverse->redoGroup));	// get rid of redo information
	buffer->undoUniverse->currentFrame=0;				// reset the frame number
}

void SetUndoCleanPoint(EDITOR_BUFFER *buffer)
// Mark the current position in the undo data as the buffer "clean" position
// NOTE: this will force a break in the undo flow if needed
{
	buffer->undoUniverse->currentFrame++;				// bump the frame number to force a break in the undo flow
	buffer->undoUniverse->cleanKnown=true;
	buffer->undoUniverse->cleanElements=buffer->undoUniverse->undoGroup.undoArray.totalElements;
}

bool AtUndoCleanPoint(EDITOR_BUFFER *buffer)
// Tells if the buffer is at the "clean" point in the undo flow
{
	if(buffer->undoUniverse->cleanKnown&&buffer->undoUniverse->cleanElements==buffer->undoUniverse->undoGroup.undoArray.totalElements)
	{
		return(true);
	}
	return(false);
}

static bool InitializeUndoGroup(UNDO_GROUP *group)
// Initialize an undo group
{
	if((group->undoText=OpenTextUniverse()))
	{
		if((group->undoStyles=OpenStyleUniverse()))
		{
			if(InitArrayUniverse(&(group->undoArray),UNDO_CHUNK_ELEMENTS,sizeof(UNDO_ELEMENT)))
			{
				return(true);
			}
			CloseStyleUniverse(group->undoStyles);
		}
		CloseTextUniverse(group->undoText);
	}
	return(false);
}

static void UnInitializeUndoGroup(UNDO_GROUP *group)
// Undo what InitializeUndoGroup did
{
	UnInitArrayUniverse(&(group->undoArray));
	CloseStyleUniverse(group->undoStyles);
	CloseTextUniverse(group->undoText);
}

UNDO_UNIVERSE *OpenUndoUniverse()
// Open an undo universe
// If there is a problem, set the error, and return NULL
{
	UNDO_UNIVERSE
		*universe;

	if((universe=(UNDO_UNIVERSE *)MNewPtr(sizeof(UNDO_UNIVERSE))))
	{
		universe->currentFrame=0;
		universe->groupingDepth=0;
		universe->cleanElements=0;
		universe->cleanKnown=true;
		universe->undoActive=false;
		universe->undoing=false;
		if(InitializeUndoGroup(&(universe->undoGroup)))
		{
			if(InitializeUndoGroup(&(universe->redoGroup)))
			{
				return(universe);
			}
			UnInitializeUndoGroup(&(universe->undoGroup));
		}
		MDisposePtr(universe);
	}
	return(NULL);
}

void CloseUndoUniverse(UNDO_UNIVERSE *universe)
// Dispose of an undo universe
{
	UnInitializeUndoGroup(&(universe->redoGroup));
	UnInitializeUndoGroup(&(universe->undoGroup));
	MDisposePtr(universe);
}
