// Dialog push button handler
// Copyright (C) 2000 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

#define	SHADOW_COLOR_UL			0x91AFBE	// shadow colors (upper left)
#define	SHADOW_COLOR_LR			0x314F5E	// (lower right)

#define	MAX_BUTTON_NAME		256									// maximum length of button name (including terminator)

typedef struct button
{
	EDITOR_WINDOW
		*parentWindow;											// the window that contains this button
	buttonResizeFunction
		*resizeFunction;										// function which is used to resize for this window
	EDITOR_RECT
		rect;
	EDITOR_FONT
		*font;													// info about currently selected font
	char
		buttonName[MAX_BUTTON_NAME];
	bool
		highlight;												// used during tracking to remember if highlighted
	bool
		hasKey;													// true if this button has a key code associated with it
	KeySym
		keySym;													// equivalent key code for this button
	bool
		*pressedState;											// boolean set to true if button pressed
	void
		(*pressedProc)(void *parameters);						// routine called when the button has been pressed
	void
		*pressedProcParameters;									// data passed to pressedProc routine
} BUTTON;


static bool PointInButton(BUTTON *button,INT32 x,INT32 y)
// return true if parent window relative x/y are in button
// false if not
{
	return(PointInRECT(x,y,&button->rect));
}

static void InvalidateButton(BUTTON *button)
// invalidate the button within its parent window
{
	InvalidateWindowRect(button->parentWindow,&button->rect);
}

static void SetButtonRectangle(BUTTON *button)
// Work out the rectangle given the current window
{
	if(button->resizeFunction)	// see if non-null resize function specified
	{
		button->resizeFunction(button->parentWindow,&button->rect);	// get new rectangle
		InvalidateButton(button);									// invalidate it
	}
}

static void DrawButton(BUTTON *button,Pixmap imageBuffer,Region updateRegion)
// redraw button in its parent window, obey the invalid area of the parent's window
{
	int
		stringLength,
		stringWidth;
	int
		stringX,
		stringY;
	EDITOR_COLOR
		background,
		upLeftHighlight,
		downRightHighlight;

	XSetRegion(xDisplay,xGraphicsContext,updateRegion);			// set clipping to what's invalid

	if(button->rect.w&&button->rect.h)
	{
		if(button->highlight)
		{
			background=GRAY_2;
			upLeftHighlight=SHADOW_COLOR_LR;
			downRightHighlight=SHADOW_COLOR_UL;
		}
		else
		{
			background=GRAY_3;
			upLeftHighlight=SHADOW_COLOR_UL;
			downRightHighlight=SHADOW_COLOR_LR;
		}
		XSetFillStyle(xDisplay,xGraphicsContext,FillSolid);
		XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(background));
		XFillRectangle(xDisplay,imageBuffer,xGraphicsContext,button->rect.x,button->rect.y,button->rect.w,button->rect.h);

		OutlineShadowRectangle(imageBuffer,&button->rect,upLeftHighlight,downRightHighlight,2);

		stringLength=strlen(&(button->buttonName[0]));
		stringWidth=XTextWidth(button->font->xFont,&(button->buttonName[0]),stringLength);
		stringX=button->rect.x+(button->rect.w/2)-(stringWidth/2);
		stringY=button->rect.y+(button->rect.h/2)-(button->font->ascent+button->font->descent)/2+button->font->ascent;

		if(button->highlight)
		{
			stringX--;
			stringY--;
		}

		XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(BLACK));
		XSetFont(xDisplay,xGraphicsContext,button->font->xFont->fid);		// point to the current font for this button
		XDrawString(xDisplay,imageBuffer,xGraphicsContext,stringX,stringY,&(button->buttonName[0]),stringLength);
	}
	XSetClipMask(xDisplay,xGraphicsContext,None);							// get rid of clip mask
}

static void PressButton(BUTTON *button)
// call this and button will behave as if it has been pressed
{
	button->highlight=true;
	InvalidateButton(button);
	UpdateEditorWindows();
	if(button->pressedState)
	{
		(*button->pressedState)=true;
	}
	if(button->pressedProc)
	{
		button->pressedProc(button->pressedProcParameters);
	}
	button->highlight=false;
	InvalidateButton(button);
}

static bool TrackButton(BUTTON *button,XEvent *event)
// see if the point given by event falls over button,
// if so, track it, and return true
// if not, return false
{
	bool
		onButton,
		wasOnButton;
	Window
		root,
		child;
	int
		rootX,
		rootY,
		windowX,
		windowY;
	unsigned int
		state;

	if((event->xbutton.button==Button1)&&PointInButton(button,event->xbutton.x,event->xbutton.y))
	{
		button->highlight=true;
		InvalidateButton(button);
		UpdateEditorWindows();												// redraw anything that is needed
		wasOnButton=onButton=true;
		while(StillDown(Button1,1))
		{
			if(XQueryPointer(xDisplay,((WINDOW_LIST_ELEMENT *)button->parentWindow->userData)->xWindow,&root,&child,&rootX,&rootY,&windowX,&windowY,&state))
			{
				onButton=PointInButton(button,windowX,windowY);
			}
			if(wasOnButton!=onButton)
			{
				button->highlight=onButton;
				InvalidateButton(button);
				UpdateEditorWindows();										// redraw anything that is needed
				wasOnButton=onButton;
			}
		}
		if(onButton)
		{
			PressButton(button);
		}
		button->highlight=false;
		InvalidateButton(button);
		return(true);
	}
	return(false);
}

static bool ButtonKeyOverride(BUTTON *button,XEvent *event)
// this is called for buttons that may wish to grab keys before they go
// into the default focus of a dialog
// true is returned if the button wanted the key
{
	KeySym
		keySym;
	XKeyEvent
		localKeyEvent;														// make a local copy so we can strip annoying bits

	if(button->hasKey)
	{
		localKeyEvent=*(XKeyEvent *)event;									// copy the event
		localKeyEvent.state&=(~(Mod1Mask|LockMask));						// remove these from consideration
		ComposeXLookupString(&localKeyEvent,NULL,0,&keySym);
		if(keySym==button->keySym)
		{
			PressButton(button);
			return(true);
		}
	}
	return(false);
}

static BUTTON *CreateButton(EDITOR_WINDOW *window,BUTTON_DESCRIPTOR *descriptor)
// create a button in window, invalidate the area of window
// where the button is to be placed
// return a pointer to it if successful
// SetError, and return NULL if not
{
	BUTTON
		*button;

	if((button=(BUTTON *)MNewPtr(sizeof(BUTTON))))		// create button data structure
	{
		button->parentWindow=window;
		button->resizeFunction=descriptor->resizeFunction;
		strncpy(&(button->buttonName[0]),descriptor->buttonName,MAX_BUTTON_NAME);
		button->buttonName[MAX_BUTTON_NAME-1]='\0';		// terminate the string as needed
		button->highlight=false;						// not highlighted
		button->hasKey=descriptor->hasKey;				// true if this button has a key code associated with it
		button->keySym=descriptor->keySym;				// equivalent key code for this button
		button->pressedState=descriptor->pressedState;	// points to a bool (or to NULL) that is set to true if this button has been pressed
		button->pressedProc=descriptor->pressedProc;	// copy the procedure to be called when the button is pressed
		button->pressedProcParameters=descriptor->pressedProcParameters;	// copy parameters for pressed Proc
		if((button->font=LoadFont(descriptor->fontName)))
		{
			SetButtonRectangle(button);
			return(button);
		}
		MDisposePtr(button);
	}
	return(NULL);
}

static void DisposeButton(BUTTON *button)
// unlink button from its parent window, and delete it
// invalidate the area of the window where button was
{
	InvalidateButton(button);					// invalidate where it is now
	FreeFont(button->font);
	MDisposePtr(button);
}

// dialog interface routines

static void DrawButtonItem(DIALOG_ITEM *item,Pixmap imageBuffer,Region updateRegion)
// draw the button pointed to by item
{
	DrawButton((BUTTON *)item->itemStruct,imageBuffer,updateRegion);
}

static void ResizeButtonItem(DIALOG_ITEM *item)
// When the dialog window size changes, call this to update the item itself
// This will invalidate the old and new item areas
{
	BUTTON
		*button;

	button=(BUTTON *)item->itemStruct;
	InvalidateButton(button);				// invalidate where we're at currently
	SetButtonRectangle(button);
}

void PressButtonItem(DIALOG_ITEM *item)
// Do whatever would be done if the user pressed item
{
	PressButton((BUTTON *)item->itemStruct);
}

static bool TrackButtonItem(DIALOG_ITEM *item,XEvent *event)
// track the button pointed to by item
{
	return(TrackButton((BUTTON *)item->itemStruct,event));
}

static bool EarlyKeyButtonItem(DIALOG_ITEM *item,XEvent *event)
// see if this button wants the key given by event, if so, return true
{
	return(ButtonKeyOverride((BUTTON *)item->itemStruct,event));
}

static void DisposeButtonItem(DIALOG_ITEM *item)
// dispose of the local button structure pointed to by item
{
	DisposeButton((BUTTON *)item->itemStruct);
}

bool CreateButtonItem(DIALOG_ITEM *item,void *descriptor)
// create the button, and fill in item
// if there is a problem, SetError, return false
{
	if((item->itemStruct=(void *)CreateButton(item->dialog->parentWindow,(BUTTON_DESCRIPTOR *)descriptor)))
	{
		item->disposeProc=DisposeButtonItem;
		item->drawProc=DrawButtonItem;
		item->resizeProc=ResizeButtonItem;
		item->trackProc=TrackButtonItem;
		item->earlyKeyProc=EarlyKeyButtonItem;
		item->wantFocus=false;
		item->focusChangeProc=NULL;
		item->focusKeyProc=NULL;
		item->focusPeriodicProc=NULL;
		return(true);
	}
	return(false);
}
