// Dialog check box handler
// Copyright (C) 2000 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

#define	MAX_CHECK_BOX_NAME		256								// maximum length of check box name (including terminator)

typedef struct checkBox
{
	EDITOR_WINDOW
		*parentWindow;											// the window that contains this check box
	checkBoxResizeFunction
		*resizeFunction;										// function which is used to resize for this window
	EDITOR_RECT
		rect;
	EDITOR_FONT
		*font;													// info about currently selected font
	char
		checkBoxName[MAX_CHECK_BOX_NAME];
	bool
		highlight;												// used during tracking to remember if highlighted
	bool
		hasKey;													// true if this check box has a key code associated with it
	KeySym
		keySym;													// equivalent key code for this check box
	bool
		*pressedState;											// boolean set to true if check box pressed
	void
		(*pressedProc)(struct checkBox *checkBox,void *parameters);	// routine called when the check box has been pressed
	void
		*pressedProcParameters;									// data passed to pressedProc routine
} CHECK_BOX;

static bool PointInCheckBox(CHECK_BOX *checkBox,INT32 x,INT32 y)
// return true if parent window relative x/y are in checkBox
// false if not
{
	return(PointInRECT(x,y,&checkBox->rect));
}

static void InvalidateCheckBox(CHECK_BOX *checkBox)
// invalidate the check box within its parent window
{
	InvalidateWindowRect(checkBox->parentWindow,&checkBox->rect);
}

static void SetCheckBoxRectangle(CHECK_BOX *checkBox)
// Work out the rectangle given the current window
{
	if(checkBox->resizeFunction)	// see if non-null resize function specified
	{
		checkBox->resizeFunction(checkBox->parentWindow,&checkBox->rect);	// get new rectangle
		InvalidateCheckBox(checkBox);							// invalidate it
	}
}

static void DrawCheckBox(CHECK_BOX *checkBox,Pixmap imageBuffer,Region updateRegion)
// redraw checkBox in its parent window, obey the invalid area of the parent's window
{
	XRectangle
		regionBox;
	Region
		invalidRegion;
	EDITOR_COLOR
		upLeftHighlight,
		downRightHighlight;
	INT32
		stringX,
		stringY;
	INT32
		xLeft,
		xRight,
		yTop,
		yBottom;

	invalidRegion=XCreateRegion();

	regionBox.x=checkBox->rect.x;
	regionBox.y=checkBox->rect.y;
	regionBox.width=checkBox->rect.w;
	regionBox.height=checkBox->rect.h;

	XUnionRectWithRegion(&regionBox,invalidRegion,invalidRegion);				// create a region which is the rectangle of the text box
	XIntersectRegion(invalidRegion,updateRegion,invalidRegion);					// intersect with what is invalid in this window

	if(!XEmptyRegion(invalidRegion))											// see if it has an invalid area
	{
		XSetRegion(xDisplay,xGraphicsContext,invalidRegion);					// set clipping to what's invalid

		XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(GRAY_3));
		XFillRectangle(xDisplay,imageBuffer,xGraphicsContext,checkBox->rect.x,checkBox->rect.y,checkBox->rect.w,checkBox->rect.h);

		upLeftHighlight=BLACK;
		downRightHighlight=WHITE;
		xLeft=checkBox->rect.x;
		xRight=xLeft+checkBox->rect.w-1;
		yTop=checkBox->rect.y;
		yBottom=yTop+checkBox->rect.h-1;

		XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(upLeftHighlight));
		XDrawLine(xDisplay,imageBuffer,xGraphicsContext,xLeft,yTop,xRight,yTop);
		XDrawLine(xDisplay,imageBuffer,xGraphicsContext,xLeft,yTop,xLeft,yBottom);
		XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(downRightHighlight));
		XDrawLine(xDisplay,imageBuffer,xGraphicsContext,xRight,yTop,xRight,yBottom);
		XDrawLine(xDisplay,imageBuffer,xGraphicsContext,xLeft,yBottom,xRight,yBottom);

		XSetBackground(xDisplay,xGraphicsContext,EditorColorToXPixel(GRAY_3));
		if(checkBox->highlight)
		{
			XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(WHITE));
		}
		else
		{
			XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(BLACK));
		}

		if(checkBox->pressedState&&(*checkBox->pressedState))
		{
			XCopyPlane(xDisplay,checkedBoxPixmap,imageBuffer,xGraphicsContext,0,0,CHECK_BOX_WIDTH,CHECK_BOX_HEIGHT,checkBox->rect.x+2,checkBox->rect.y+(checkBox->rect.h-CHECK_BOX_HEIGHT)/2,1);
		}
		else
		{
			XCopyPlane(xDisplay,checkBoxPixmap,imageBuffer,xGraphicsContext,0,0,CHECK_BOX_WIDTH,CHECK_BOX_HEIGHT,checkBox->rect.x+2,checkBox->rect.y+(checkBox->rect.h-CHECK_BOX_HEIGHT)/2,1);
		}

		stringX=checkBox->rect.x+CHECK_BOX_WIDTH+4;
		stringY=checkBox->rect.y+(checkBox->rect.h/2)-(checkBox->font->ascent+checkBox->font->descent)/2+checkBox->font->ascent;

		XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(BLACK));
		XSetFont(xDisplay,xGraphicsContext,checkBox->font->xFont->fid);				// point to the current font for this check box
		XDrawString(xDisplay,imageBuffer,xGraphicsContext,stringX,stringY,&(checkBox->checkBoxName[0]),strlen(&(checkBox->checkBoxName[0])));
		XSetClipMask(xDisplay,xGraphicsContext,None);								// get rid of clip mask
	}

	XDestroyRegion(invalidRegion);													// get rid of invalid region
}

static bool TrackCheckBox(CHECK_BOX *checkBox,XEvent *event)
// see if the point given by event falls over checkBox,
// if so, track it, and return true
// if not, return false
{
	bool
		onCheckBox,
		wasOnCheckBox;
	Window
		root,
		child;
	int
		rootX,
		rootY,
		windowX,
		windowY;
	unsigned int
		state;

	if((event->xbutton.button==Button1)&&PointInCheckBox(checkBox,event->xbutton.x,event->xbutton.y))
	{
		checkBox->highlight=true;
		InvalidateCheckBox(checkBox);
		UpdateEditorWindows();												// redraw anything that is needed
		wasOnCheckBox=onCheckBox=true;
		while(StillDown(Button1,1))
		{
			if(XQueryPointer(xDisplay,((WINDOW_LIST_ELEMENT *)checkBox->parentWindow->userData)->xWindow,&root,&child,&rootX,&rootY,&windowX,&windowY,&state))
			{
				onCheckBox=PointInCheckBox(checkBox,windowX,windowY);
			}
			if(wasOnCheckBox!=onCheckBox)
			{
				checkBox->highlight=onCheckBox;
				InvalidateCheckBox(checkBox);
				UpdateEditorWindows();										// redraw anything that is needed
				wasOnCheckBox=onCheckBox;
			}
		}
		if(onCheckBox)
		{
			if(checkBox->pressedState)
			{
				if((*checkBox->pressedState))								// toggle the state
				{
					(*checkBox->pressedState)=false;
				}
				else
				{
					(*checkBox->pressedState)=true;
				}
			}
			if(checkBox->pressedProc)
			{
				checkBox->pressedProc(checkBox,checkBox->pressedProcParameters);
			}
		}
		checkBox->highlight=false;
		InvalidateCheckBox(checkBox);
		return(true);
	}
	return(false);
}

static bool CheckBoxKeyOverride(CHECK_BOX *checkBox,XEvent *event)
// this is called for check boxes that may wish to grab keys before they go
// into the default focus of a dialog
// true is returned if the check box wanted the key
{
	KeySym
		keySym;

	if(checkBox->hasKey)
	{
		ComposeXLookupString((XKeyEvent *)event,NULL,0,&keySym);
		if(keySym==checkBox->keySym)
		{
			checkBox->highlight=true;
			InvalidateCheckBox(checkBox);
			if(checkBox->pressedState)
			{
				(*checkBox->pressedState)=true;
			}
			if(checkBox->pressedProc)
			{
				checkBox->pressedProc(checkBox,checkBox->pressedProcParameters);
			}
			checkBox->highlight=false;
			InvalidateCheckBox(checkBox);
			return(true);
		}
	}
	return(false);
}

static CHECK_BOX *CreateCheckBox(EDITOR_WINDOW *window,CHECK_BOX_DESCRIPTOR *descriptor)
// create a check box in window, invalidate the area of window
// where the check box is to be placed
// return a pointer to it if successful
// SetError, and return NULL if not
{
	CHECK_BOX
		*checkBox;

	if((checkBox=(CHECK_BOX *)MNewPtr(sizeof(CHECK_BOX))))	// create check box data structure
	{
		checkBox->parentWindow=window;
		checkBox->resizeFunction=descriptor->resizeFunction;
		strncpy(&(checkBox->checkBoxName[0]),descriptor->checkBoxName,MAX_CHECK_BOX_NAME);
		checkBox->checkBoxName[MAX_CHECK_BOX_NAME-1]='\0';	// terminate the string as needed
		checkBox->highlight=false;							// not highlighted
		checkBox->hasKey=descriptor->hasKey;				// true if this check box has a key code associated with it
		checkBox->keySym=descriptor->keySym;				// equivalent key code for this check box
		checkBox->pressedState=descriptor->pressedState;	// points to a bool (or to NULL) that is set to true if this check box has been pressed
		checkBox->pressedProc=descriptor->pressedProc;		// copy the procedure to be called when the check box is pressed
		checkBox->pressedProcParameters=descriptor->pressedProcParameters;	// copy parameters for pressed Proc
		if((checkBox->font=LoadFont(descriptor->fontName)))
		{
			SetCheckBoxRectangle(checkBox);
			return(checkBox);
		}
		MDisposePtr(checkBox);
	}
	return(NULL);
}

static void DisposeCheckBox(CHECK_BOX *checkBox)
// unlink checkBox from its parent window, and delete it
// invalidate the area of the window where checkBox was
{
	InvalidateCheckBox(checkBox);					// invalidate where it is now
	FreeFont(checkBox->font);
	MDisposePtr(checkBox);
}

// dialog interface routines

static void DrawCheckBoxItem(DIALOG_ITEM *item,Pixmap imageBuffer,Region updateRegion)
// draw the check box pointed to by item
{
	DrawCheckBox((CHECK_BOX *)item->itemStruct,imageBuffer,updateRegion);
}

static void ResizeCheckBoxItem(DIALOG_ITEM *item)
// When the dialog window size changes, call this to update the item itself
// This will invalidate the old and new item areas
{
	CHECK_BOX
		*checkBox;

	checkBox=(CHECK_BOX *)item->itemStruct;
	InvalidateCheckBox(checkBox);					// invalidate where we're at currently
	SetCheckBoxRectangle(checkBox);
}

static bool TrackCheckBoxItem(DIALOG_ITEM *item,XEvent *event)
// track the check box pointed to by item
{
	return(TrackCheckBox((CHECK_BOX *)item->itemStruct,event));
}

static bool EarlyKeyCheckBoxItem(DIALOG_ITEM *item,XEvent *event)
// see if this check box wants the key given by event, if so, return true
{
	return(CheckBoxKeyOverride((CHECK_BOX *)item->itemStruct,event));
}

static void DisposeCheckBoxItem(DIALOG_ITEM *item)
// dispose of the local check box structure pointed to by item
{
	DisposeCheckBox((CHECK_BOX *)item->itemStruct);
}

bool CreateCheckBoxItem(DIALOG_ITEM *item,void *descriptor)
// create the check box, and fill in item
// if there is a problem, SetError, return false
{
	if((item->itemStruct=(void *)CreateCheckBox(item->dialog->parentWindow,(CHECK_BOX_DESCRIPTOR *)descriptor)))
	{
		item->disposeProc=DisposeCheckBoxItem;
		item->drawProc=DrawCheckBoxItem;
		item->resizeProc=ResizeCheckBoxItem;
		item->trackProc=TrackCheckBoxItem;
		item->earlyKeyProc=EarlyKeyCheckBoxItem;
		item->wantFocus=false;
		item->focusChangeProc=NULL;
		item->focusKeyProc=NULL;
		item->focusPeriodicProc=NULL;
		return(true);
	}
	return(false);
}
