// Dialog window handling
// Copyright (C) 2000 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

#define	BACKGROUND_COLOR		0x718F9E		// color to use for dialog backgrounds
#define	SHADOW_COLOR_UL			0x91AFBE		// shadow colors (upper left)
#define	SHADOW_COLOR_LR			0x314F5E		// (lower right)

static void ChangeFocus(DIALOG *dialog,DIALOG_ITEM *newItem)
// Change the focus of dialog to newItem
// If the focus is already at new item, do nothing
// NOTE: this is DIALOG ITEM focus, not X window focus
{
	DIALOG_ITEM
		*oldItem;

	oldItem=dialog->focusItem;
	if(oldItem!=newItem)							// make sure they are not the same, or both NULL
	{
		dialog->focusItem=newItem;
		if(oldItem)
		{
			if(oldItem->focusChangeProc)
			{
				oldItem->focusChangeProc(oldItem);	// tell him about it
			}
		}
		if(newItem)
		{
			if(newItem->focusChangeProc)
			{
				newItem->focusChangeProc(newItem);	// tell him about it
			}
		}
	}
}

static DIALOG_ITEM *LocateForwardFocusItem(DIALOG *dialog)
// Look forward from the dialog item which currently has the focus
// and return the next item which should get focus
// If no items require focus, return NULL
{
	DIALOG_ITEM
		*currentItem,
		*nextFocusItem;

	if((currentItem=dialog->focusItem))
	{
		currentItem=currentItem->nextItem;			// begin looking one beyond the old focus item (if there was an old one)
	}
	nextFocusItem=NULL;
	while(currentItem&&!nextFocusItem)
	{
		if(currentItem->wantFocus)
		{
			nextFocusItem=currentItem;				// this is the next one that wants the focus
		}
		currentItem=currentItem->nextItem;
	}
	currentItem=dialog->firstItem;					// start over at the beginning
	while(currentItem&&!nextFocusItem)
	{
		if(currentItem->wantFocus)
		{
			nextFocusItem=currentItem;				// this is the next one that wants the focus
		}
		currentItem=currentItem->nextItem;
	}
	return(nextFocusItem);
}

static void ForwardFocus(DIALOG *dialog)
// Push the focus forward one notch, and alert the involved parties
// NOTE: this is DIALOG ITEM focus, not X window focus
{
	DIALOG_ITEM
		*nextFocusItem;

	nextFocusItem=LocateForwardFocusItem(dialog);	// this can return NULL, which will be passed to ChangeFocus to select nothing
	ChangeFocus(dialog,nextFocusItem);
}

static void LinkItemToDialog(DIALOG *dialog,DIALOG_ITEM *newItem)
// Link newItem to the end of the linked list of dialog items in dialog
{
	newItem->dialog=dialog;							// point back at the dialog
	newItem->nextItem=NULL;
	if((newItem->previousItem=dialog->lastItem))
	{
		dialog->lastItem->nextItem=newItem;
		dialog->lastItem=newItem;
	}
	else
	{
		dialog->firstItem=dialog->lastItem=newItem;
	}
}

static void UnlinkItemFromDialog(DIALOG_ITEM *dialogItem)
// Unlink dialogItem from the dialog list
{
	if(dialogItem->previousItem)
	{
		dialogItem->previousItem->nextItem=dialogItem->nextItem;
	}
	else
	{
		dialogItem->dialog->firstItem=dialogItem->nextItem;
	}
	if(dialogItem->nextItem)
	{
		dialogItem->nextItem->previousItem=dialogItem->previousItem;
	}
	else
	{
		dialogItem->dialog->lastItem=dialogItem->previousItem;
	}
}

DIALOG_ITEM *NewDialogItem(DIALOG *dialog,bool (*createProc)(DIALOG_ITEM *item,void *itemDescriptor),void *itemDescriptor)
// Create a new dialog item, link it to dialog, call the item's creation procedure
// createProc must not be NULL and must return true for this to succeed
// If there is a problem, SetError, return NULL
{
	DIALOG_ITEM
		*dialogItem;

	if((dialogItem=(DIALOG_ITEM *)MNewPtrClr(sizeof(DIALOG_ITEM))))	// make new item, clear any procedure pointers to NULL
	{
		LinkItemToDialog(dialog,dialogItem);
		if(createProc(dialogItem,itemDescriptor))
		{
			return(dialogItem);
		}
		UnlinkItemFromDialog(dialogItem);
		MDisposePtr(dialogItem);
	}
	return(NULL);
}

void DisposeDialogItem(DIALOG_ITEM *dialogItem)
// Unlink, and dispose of a dialog item, created by NewDialogItem
{
	if(dialogItem->disposeProc)
	{
		dialogItem->disposeProc(dialogItem);
	}
	UnlinkItemFromDialog(dialogItem);
	MDisposePtr(dialogItem);
}

static DIALOG *CreateDialog(EDITOR_RECT *rect,const char *title,EDITOR_COLOR background)
// Open a new empty dialog of the given size, with the given title
// It should not be updated until all the items have been added to it
// If there is a problem, SetError, return NULL
{
	DIALOG
		*dialog;
	EDITOR_WINDOW
		*window;
	WINDOW_LIST_ELEMENT
		*windowElement;
	UINT32
		valueMask;
	Atom
		protocols[2];														// number of protocols
	XWMHints
		hints;																// hints to tell window manager how to treat these windows
	XSizeHints
		sizeHints;															// window size info (used to REALLY convince X that I know where I want my own window)
	XEvent
		xEvent;
	INT32
		xOffset,
		yOffset;
	UINT32
		screenWidth,
		screenHeight;

	if((dialog=(DIALOG *)MNewPtr(sizeof(DIALOG))))
	{
		dialog->backgroundColor=background;
		dialog->dialogComplete=false;										// not done with this yet
		dialog->firstItem=dialog->lastItem=dialog->focusItem=NULL;
		if((window=(EDITOR_WINDOW *)MNewPtr(sizeof(EDITOR_WINDOW))))		// create editor window
		{
			if((windowElement=(WINDOW_LIST_ELEMENT *)MNewPtr(sizeof(WINDOW_LIST_ELEMENT))))	// create a window list element
			{
				LinkWindowElement(windowElement);							// link it to the global window list that we maintain
				windowElement->editorWindow=window;							// point to our high level concept of window
				window->windowType=EWT_DIALOG;								// set the editor window type to "dialog"
				window->userData=windowElement;								// point back to the window element

				valueMask=0;
				// Due to X's nasty error handling, we assume this create does not fail!!!

				GetEditorScreenOffsetAndDimensions(&xOffset,&yOffset,&screenWidth,&screenHeight);
				windowElement->realTopWindow=windowElement->xWindow=XCreateWindow(xDisplay,RootWindow(xDisplay,xScreenNum),rect->x+xOffset,rect->y+yOffset,rect->w,rect->h,2,CopyFromParent,InputOutput,DefaultVisual(xDisplay,xScreenNum),valueMask,NULL);

				sizeHints.flags=USSize|USPosition|PMinSize;					// ### LIE to X, tell it the user requested the size and position so it will honor them
				sizeHints.x=rect->x;
				sizeHints.y=rect->y;
				sizeHints.width=rect->w;
				sizeHints.height=rect->h;
				sizeHints.min_width=rect->w;
				sizeHints.min_height=rect->h;
				sizeHints.max_width=rect->w;
				sizeHints.max_height=rect->h;
				XSetWMNormalHints(xDisplay,windowElement->xWindow,&sizeHints);

				XStoreName(xDisplay,windowElement->xWindow,title);
				XSelectInput(xDisplay,windowElement->xWindow,ExposureMask|ButtonPressMask|ButtonReleaseMask|FocusChangeMask|KeyPressMask|StructureNotifyMask);	// want to see these events

				protocols[0]=takeFocusAtom;
				protocols[1]=deleteWindowAtom;
				XSetWMProtocols(xDisplay,windowElement->xWindow,&(protocols[0]),2);	// tell X we would like to participate in these protocols

				hints.flags=InputHint|StateHint|WindowGroupHint;
				hints.input=True;
				hints.initial_state=NormalState;
				hints.window_group=((WINDOW_LIST_ELEMENT *)rootMenuWindow->userData)->xWindow;	// all dialogs belong to the group headed by the main menu window
				XSetWMHints(xDisplay,windowElement->xWindow,&hints);

// NOTE: when setting transient below, It would be nice if the window manager would allow it to be transient for the GROUP, so that
// it would stay on top of all of the editor windows, but so far, I have not been able to convince it to do that :{
// Note that it is being set transient to itself, this provides behavior which is better than anything else I could get the WM
// to do!
//
// NOTE 10/13/2000 the new Gnome does not like this, so it was removed
//				XSetTransientForHint(xDisplay,windowElement->xWindow,windowElement->xWindow);	// tell WM this is transient (would like it to stay on top, but window manager will not oblige)

				windowElement->invalidRegion=XCreateRegion();				// create new empty invalid region

				windowElement->currentIcon=(Pixmap)NULL;					// no icon associated with this window
				dialog->parentWindow=window;								// link the dialog to the window just created
				window->windowInfo=dialog;									// link the window to the dialog just created
				XMapRaised(xDisplay,windowElement->xWindow);				// map window to the top of the display
				WaitUntilMapped(window);									// wait until this window is actually mapped before continuing (olwm likes to mess with the temporal ordering of our windows, this attempts to prevent its inane meddling

				// This code asks the WM to please keep our dialog window on top
				xEvent.xclient.type=ClientMessage;
				xEvent.xclient.serial=0;
				xEvent.xclient.send_event=True;
				xEvent.xclient.display=xDisplay;
				xEvent.xclient.window=windowElement->xWindow;
				xEvent.xclient.message_type=XInternAtom(xDisplay,"_NET_WM_STATE",True);
				xEvent.xclient.format=32;
				xEvent.xclient.data.l[0]=1;					// 1 _NET_WM_STATE_ADD
				xEvent.xclient.data.l[1]=XInternAtom(xDisplay,"_NET_WM_STATE_ABOVE",True);
				xEvent.xclient.data.l[2]=XInternAtom(xDisplay,"_NET_WM_STATE_MODAL",True);
				xEvent.xclient.data.l[3]=1;					// source indication (1 means application)
				xEvent.xclient.data.l[4]=0;
				XSendEvent(xDisplay,XDefaultRootWindow(xDisplay),False,SubstructureRedirectMask|SubstructureNotifyMask,&xEvent);

				return(dialog);
			}
			MDisposePtr(window);
		}
		MDisposePtr(dialog);
	}
	return(NULL);
}

static void DisposeDialog(DIALOG *dialog)
// Close a dialog window, kill any items it may have linked to it
{
	WINDOW_LIST_ELEMENT
		*windowElement;

	while(dialog->firstItem)
	{
		DisposeDialogItem(dialog->firstItem);								// remove any items
	}
	FocusAwayFrom(dialog->parentWindow);									// trick X by moving focus so window manager will not try to when this window dies
	windowElement=(WINDOW_LIST_ELEMENT *)dialog->parentWindow->userData;	// get the x window associated with this document window
	XDestroyRegion(windowElement->invalidRegion);							// destroy invalid region
	XDestroyWindow(xDisplay,windowElement->xWindow);						// tell x to make the window go away
	WaitUntilDestroyed(dialog->parentWindow);
	UnlinkWindowElement(windowElement);										// unlink it from our list
	MDisposePtr(windowElement);
	MDisposePtr(dialog->parentWindow);
	MDisposePtr(dialog);
	UpdateEditorWindows();													// update the display
}

static DIALOG *CreateCenteredDialog(UINT32 width,UINT32 height,const char *title,EDITOR_COLOR background)
// Open a new dialog window of width, height centered
// nicely on the screen
// If there is a problem, SetError, and return NULL
{
	EDITOR_RECT
		dialogRect;
	INT32
		xOffset,
		yOffset;
	UINT32
		screenWidth,
		screenHeight;

	GetEditorScreenOffsetAndDimensions(&xOffset,&yOffset,&screenWidth,&screenHeight);
	dialogRect.x=screenWidth/2-width/2;						// half the way across
	dialogRect.y=screenHeight/3-height/2;					// 1/3 the way down
	dialogRect.w=width;
	dialogRect.h=height;
	return(CreateDialog(&dialogRect,title,background));
}

static void DrawDialog(DIALOG *dialog)
// Draw the dialog, and all items that request it
// Drawing focus is handled from here
{
	DIALOG_ITEM
		*currentItem;
	WINDOW_LIST_ELEMENT
		*windowElement;
	Window
		xWindow;
	XWindowAttributes
		attributes;
	EDITOR_RECT
		rect;
	Pixmap
		imageBuffer;
	bool
		bufferedUpdate;		// tells if update is buffered

	windowElement=(WINDOW_LIST_ELEMENT *)dialog->parentWindow->userData;		// get the x window information associated with this window
	xWindow=windowElement->xWindow;												// get x window for this window

	XGetWindowAttributes(xDisplay,xWindow,&attributes);
	rect.x=0;
	rect.y=0;
	rect.w=attributes.width;
	rect.h=attributes.height;

	imageBuffer=GetWindowUpdatePixmap(attributes.width,attributes.height);		// go get a pixmap
	bufferedUpdate=true;
	if(imageBuffer==None)		// if we failed to get an offscreen buffer, just update directly
	{
		imageBuffer=xWindow;
		bufferedUpdate=false;
	}

	XSetRegion(xDisplay,xGraphicsContext,windowElement->invalidRegion);			// set clipping to what's invalid
	XSetFillStyle(xDisplay,xGraphicsContext,FillSolid);
	XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(dialog->backgroundColor));
	XFillRectangle(xDisplay,imageBuffer,xGraphicsContext,0,0,attributes.width,attributes.height);
	OutlineShadowRectangle(imageBuffer,&rect,SHADOW_COLOR_UL,SHADOW_COLOR_LR,2);	// drop in border

	currentItem=dialog->firstItem;
	while(currentItem)
	{
		if(currentItem->drawProc)
		{
			currentItem->drawProc(currentItem,imageBuffer,windowElement->invalidRegion);
		}
		currentItem=currentItem->nextItem;
	}

	if(bufferedUpdate)			// if the update was buffered, the move the data to the screen
	{
		XSetRegion(xDisplay,xGraphicsContext,windowElement->invalidRegion);		// only copy what's needed
		XCopyArea(xDisplay,imageBuffer,xWindow,xGraphicsContext,0,0,attributes.width,attributes.height,0,0);	// copy the updates
		XSetClipMask(xDisplay,xGraphicsContext,None);							// get rid of clip mask
	}
}

void UpdateDialogWindow(EDITOR_WINDOW *window)
// Make sure window is up to date
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	DIALOG
		*dialog;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;				// get the X window information associated with this document window
	if(!XEmptyRegion(windowElement->invalidRegion))						// see if it has an invalid area
	{
		dialog=(DIALOG *)window->windowInfo;
		DrawDialog(dialog);
		XDestroyRegion(windowElement->invalidRegion);					// clear out the invalid region now that the window has been drawn
		windowElement->invalidRegion=XCreateRegion();
	}
}

void AdjustDialogWindowForNewSize(EDITOR_WINDOW *window)
// If the window has changed size, then the
// controls will be needing update, so adjust them for the new
// size
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	DIALOG
		*dialog;
	DIALOG_ITEM
		*currentItem;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	dialog=(DIALOG *)window->windowInfo;

	currentItem=dialog->firstItem;
	while(currentItem)							// run through all the items and ask them to resize
	{
		if(currentItem->resizeProc)
		{
			currentItem->resizeProc(currentItem);
		}
		currentItem=currentItem->nextItem;
	}
}

static void DialogButtonPress(DIALOG *dialog,XEvent *event)
// Handle button press events for a dialog (see what the user is clicking
// on, and attempt to deal with it)
// This runs through the items backwards so that it finds the top-most
// item first.
{
	DIALOG_ITEM
		*currentItem;
	bool
		handled;

	handled=false;
	currentItem=dialog->lastItem;
	while(currentItem&&!handled)
	{
		if(currentItem->trackProc)
		{
			handled=currentItem->trackProc(currentItem,event);
		}
		currentItem=currentItem->previousItem;
	}
}

static void DialogKeyPress(DIALOG *dialog,XEvent *event)
// A keyboard event has arrived for window
// Decide who it should go to, and handle it
// If a focus change key comes in, and it is not grabbed early, it will
// be used to change the focus, and not sent to the focus item
{
	DIALOG_ITEM
		*currentItem;
	bool
		handled;
	KeySym
		keySym;

	handled=false;
	currentItem=dialog->firstItem;
	while(currentItem&&!handled)
	{
		if(currentItem->earlyKeyProc)
		{
			handled=currentItem->earlyKeyProc(currentItem,event);
		}
		currentItem=currentItem->nextItem;
	}
	if(!handled)
	{
		ComposeXLookupString((XKeyEvent *)event,NULL,0,&keySym);
		if(keySym!=XK_Tab)
		{
			if(dialog->focusItem)
			{
				if(dialog->focusItem->focusKeyProc)
				{
					dialog->focusItem->focusKeyProc(dialog->focusItem,event);
				}
			}
		}
		else
		{
			ForwardFocus(dialog);
		}
	}
}

void DialogPeriodicProc(EDITOR_WINDOW *window)
// This is called roughly twice a second to flash cursors, or do whatever in the dialog
// when the dialog has focus
{
	DIALOG
		*dialog;

	dialog=(DIALOG *)window->windowInfo;
	if(dialog->focusItem)
	{
		if(dialog->focusItem->focusPeriodicProc)
		{
			dialog->focusItem->focusPeriodicProc(dialog->focusItem);
		}
	}
}

void DialogTakeFocus(DIALOG_ITEM *item)
// An item in the dialog window has decided that it wants the focus, so change the focus to it
{
	ChangeFocus(item->dialog,item);
}

void DialogWindowEvent(EDITOR_WINDOW *window,XEvent *event)
// Handle an event arriving for a dialog window
{
	WINDOW_LIST_ELEMENT
		*windowElement;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	switch(event->type)
	{
		case ButtonPress:
			XRaiseWindow(xDisplay,windowElement->xWindow);							// make sure dialog is on top
			XSetInputFocus(xDisplay,windowElement->xWindow,RevertToPointerRoot,event->xbutton.time);	// give me the focus
			DialogButtonPress((DIALOG *)window->windowInfo,event);					// go handle the button press
			break;
		case KeyPress:
			DialogKeyPress((DIALOG *)window->windowInfo,event);						// go handle the key press
			break;
		case ClientMessage:
			break;
		default:
			break;
	}
}

static void HandleDialog(DIALOG *dialog)
// Handle events for dialog
{
	ForwardFocus(dialog);									// place focus on first item that wants it
	DialogEventLoop(dialog->parentWindow,&(dialog->dialogComplete));
}

static void ExitPress(void *parameters)
// button was pressed, say that dialog is complete
{
	((DIALOG *)parameters)->dialogComplete=true;			// set completion flag to true
}

// ----------------------------------------------------------------------------
static void OkDialogStatTextResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the static text area
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->x=10;
	rect->y=10;
	if(attributes.width>20)
	{
		rect->w=attributes.width-20;
	}
	else
	{
		rect->w=0;
	}
	if(attributes.height>50)
	{
		rect->h=attributes.height-50;
	}
	else
	{
		rect->h=0;
	}
}

static void OkDialogSeparatorResize(EDITOR_WINDOW *window,INT32 *x,INT32 *y,UINT32 *length)
// Resize the separator
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	*x=10;

	if(attributes.height>45)
	{
		*y=attributes.height-45;
	}
	else
	{
		*y=0;
	}

	if(attributes.width>20)
	{
		*length=attributes.width-20;
	}
	else
	{
		*length=0;
	}
}

static void OkDialogButtonOkResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the ok button
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->w=50;
	rect->h=25;

	rect->x=attributes.width/2-rect->w/2;

	if(attributes.height>35)
	{
		rect->y=attributes.height-35;
	}
	else
	{
		rect->y=0;
	}
}

bool OkDialog(const char *text)
// Bring up an ok dialog, with text
// Wait for the user to click on ok, then return
// If there is a problem, SetError, and return false
{
	DIALOG
		*dialog;
	STATIC_TEXT_DESCRIPTOR
		textDescriptor;
	SEPARATOR_DESCRIPTOR
		separatorDescriptor;
	BUTTON_DESCRIPTOR
		buttonDescriptor;
	bool
		result;

	result=false;
	if((dialog=CreateCenteredDialog(500,250," ",BACKGROUND_COLOR)))			// make new empty dialog
	{
		textDescriptor.resizeFunction=OkDialogStatTextResize;
		textDescriptor.backgroundColor=BACKGROUND_COLOR;
		textDescriptor.foregroundColor=BLACK;
		textDescriptor.string=text;
		textDescriptor.fontName=defaultDialogTextFont;

		if(NewDialogItem(dialog,CreateStaticTextItem,&textDescriptor))
		{
			separatorDescriptor.resizeFunction=OkDialogSeparatorResize;
			separatorDescriptor.horizontal=true;

			if(NewDialogItem(dialog,CreateSeparatorItem,&separatorDescriptor))
			{
				buttonDescriptor.resizeFunction=OkDialogButtonOkResize;
				buttonDescriptor.buttonName="Ok";
				buttonDescriptor.fontName=defaultDialogButtonFont;
				buttonDescriptor.hasKey=true;
				buttonDescriptor.keySym=XK_Return;
				buttonDescriptor.pressedProc=ExitPress;
				buttonDescriptor.pressedProcParameters=dialog;	// pass pointer to the dialog
				buttonDescriptor.pressedState=NULL;

				if(NewDialogItem(dialog,CreateButtonItem,&buttonDescriptor))
				{
					HandleDialog(dialog);						// deal with this dialog
					result=true;
				}
			}
		}
		DisposeDialog(dialog);
	}
	return(result);
}

// ----------------------------------------------------------------------------
static void OkCancelDialogStatTextResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the static text area
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->x=10;
	rect->y=10;
	if(attributes.width>20)
	{
		rect->w=attributes.width-20;
	}
	else
	{
		rect->w=0;
	}
	if(attributes.height>50)
	{
		rect->h=attributes.height-50;
	}
	else
	{
		rect->h=0;
	}
}

static void OkCancelDialogSeparatorResize(EDITOR_WINDOW *window,INT32 *x,INT32 *y,UINT32 *length)
// Resize the separator
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	*x=10;

	if(attributes.height>45)
	{
		*y=attributes.height-45;
	}
	else
	{
		*y=0;
	}

	if(attributes.width>20)
	{
		*length=attributes.width-20;
	}
	else
	{
		*length=0;
	}
}

static void OkCancelDialogButtonOkResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the ok button
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->w=50;
	rect->h=25;

	rect->x=attributes.width/5-rect->w/2;

	if(attributes.height>35)
	{
		rect->y=attributes.height-35;
	}
	else
	{
		rect->y=0;
	}
}

static void OkCancelDialogButtonCancelResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the cancel button
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->w=70;
	rect->h=25;

	rect->x=4*attributes.width/5-rect->w/2;

	if(attributes.height>35)
	{
		rect->y=attributes.height-35;
	}
	else
	{
		rect->y=0;
	}
}

bool OkCancelDialog(const char *text,bool *cancel)
// Bring up an ok/cancel dialog, with text
// If there is a problem, SetError, return false
// Otherwise, return true in cancel if the user
// canceled, false if not
{
	DIALOG
		*dialog;
	STATIC_TEXT_DESCRIPTOR
		textDescriptor;
	SEPARATOR_DESCRIPTOR
		separatorDescriptor;
	BUTTON_DESCRIPTOR
		buttonDescriptor;
	bool
		result;

	result=false;
	*cancel=false;
	if((dialog=CreateCenteredDialog(500,250," ",BACKGROUND_COLOR)))			// make new empty dialog
	{
		textDescriptor.resizeFunction=OkCancelDialogStatTextResize;
		textDescriptor.backgroundColor=BACKGROUND_COLOR;
		textDescriptor.foregroundColor=BLACK;
		textDescriptor.string=text;
		textDescriptor.fontName=defaultDialogTextFont;

		if(NewDialogItem(dialog,CreateStaticTextItem,&textDescriptor))
		{
			separatorDescriptor.resizeFunction=OkCancelDialogSeparatorResize;
			separatorDescriptor.horizontal=true;

			if(NewDialogItem(dialog,CreateSeparatorItem,&separatorDescriptor))
			{
				buttonDescriptor.resizeFunction=OkCancelDialogButtonOkResize;
				buttonDescriptor.buttonName="Ok";
				buttonDescriptor.fontName=defaultDialogButtonFont;
				buttonDescriptor.hasKey=true;
				buttonDescriptor.keySym=XK_Return;
				buttonDescriptor.pressedProc=ExitPress;
				buttonDescriptor.pressedProcParameters=dialog;
				buttonDescriptor.pressedState=NULL;

				if(NewDialogItem(dialog,CreateButtonItem,&buttonDescriptor))
				{
					buttonDescriptor.resizeFunction=OkCancelDialogButtonCancelResize;
					buttonDescriptor.buttonName="Cancel";
					buttonDescriptor.fontName=defaultDialogButtonFont;
					buttonDescriptor.hasKey=true;
					buttonDescriptor.keySym=XK_Escape;
					buttonDescriptor.pressedProc=ExitPress;
					buttonDescriptor.pressedProcParameters=dialog;
					buttonDescriptor.pressedState=cancel;

					if(NewDialogItem(dialog,CreateButtonItem,&buttonDescriptor))
					{
						HandleDialog(dialog);					// deal with this dialog
						result=true;
					}
				}
			}
		}
		DisposeDialog(dialog);
	}
	return(result);
}

// ----------------------------------------------------------------------------
static void YesNoCancelDialogStatTextResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the static text area
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->x=10;
	rect->y=10;
	if(attributes.width>20)
	{
		rect->w=attributes.width-20;
	}
	else
	{
		rect->w=0;
	}
	if(attributes.height>50)
	{
		rect->h=attributes.height-50;
	}
	else
	{
		rect->h=0;
	}
}

static void YesNoCancelDialogSeparatorResize(EDITOR_WINDOW *window,INT32 *x,INT32 *y,UINT32 *length)
// Resize the separator
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	*x=10;

	if(attributes.height>45)
	{
		*y=attributes.height-45;
	}
	else
	{
		*y=0;
	}

	if(attributes.width>20)
	{
		*length=attributes.width-20;
	}
	else
	{
		*length=0;
	}
}

static void YesNoCancelDialogButtonYesResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the yes button
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->w=50;
	rect->h=25;

	rect->x=attributes.width/5-rect->w/2;

	if(attributes.height>35)
	{
		rect->y=attributes.height-35;
	}
	else
	{
		rect->y=0;
	}
}

static void YesNoCancelDialogButtonNoResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the no button
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->w=50;
	rect->h=25;

	rect->x=2*attributes.width/5-rect->w/2;

	if(attributes.height>35)
	{
		rect->y=attributes.height-35;
	}
	else
	{
		rect->y=0;
	}
}

static void YesNoCancelDialogButtonCancelResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the cancel button
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->w=70;
	rect->h=25;

	rect->x=4*attributes.width/5-rect->w/2;

	if(attributes.height>35)
	{
		rect->y=attributes.height-35;
	}
	else
	{
		rect->y=0;
	}
}

bool YesNoCancelDialog(const char *text,bool *yes,bool *cancel)
// Bring up a yes, no, or cancel dialog, with text
// If there is a problem, SetError, return false
// Otherwise, if cancel is true the user canceled,
// otherwise, if yes is true, yes was hit,
// otherwise no was hit.
{
	DIALOG
		*dialog;
	STATIC_TEXT_DESCRIPTOR
		textDescriptor;
	SEPARATOR_DESCRIPTOR
		separatorDescriptor;
	BUTTON_DESCRIPTOR
		buttonDescriptor;
	bool
		result;

	result=false;
	*yes=*cancel=false;
	if((dialog=CreateCenteredDialog(500,250," ",BACKGROUND_COLOR)))			// make new empty dialog
	{
		textDescriptor.resizeFunction=YesNoCancelDialogStatTextResize;
		textDescriptor.backgroundColor=BACKGROUND_COLOR;
		textDescriptor.foregroundColor=BLACK;
		textDescriptor.string=text;
		textDescriptor.fontName=defaultDialogTextFont;

		if(NewDialogItem(dialog,CreateStaticTextItem,&textDescriptor))
		{
			separatorDescriptor.resizeFunction=YesNoCancelDialogSeparatorResize;
			separatorDescriptor.horizontal=true;

			if(NewDialogItem(dialog,CreateSeparatorItem,&separatorDescriptor))
			{
				buttonDescriptor.resizeFunction=YesNoCancelDialogButtonYesResize;
				buttonDescriptor.buttonName="Yes";
				buttonDescriptor.fontName=defaultDialogButtonFont;
				buttonDescriptor.hasKey=true;
				buttonDescriptor.keySym=XK_y;
				buttonDescriptor.pressedProc=ExitPress;
				buttonDescriptor.pressedProcParameters=dialog;
				buttonDescriptor.pressedState=yes;

				if(NewDialogItem(dialog,CreateButtonItem,&buttonDescriptor))
				{
					buttonDescriptor.resizeFunction=YesNoCancelDialogButtonNoResize;
					buttonDescriptor.buttonName="No";
					buttonDescriptor.fontName=defaultDialogButtonFont;
					buttonDescriptor.hasKey=true;
					buttonDescriptor.keySym=XK_n;
					buttonDescriptor.pressedProc=ExitPress;
					buttonDescriptor.pressedProcParameters=dialog;
					buttonDescriptor.pressedState=NULL;

					if(NewDialogItem(dialog,CreateButtonItem,&buttonDescriptor))
					{
						buttonDescriptor.resizeFunction=YesNoCancelDialogButtonCancelResize;
						buttonDescriptor.buttonName="Cancel";
						buttonDescriptor.fontName=defaultDialogButtonFont;
						buttonDescriptor.hasKey=true;
						buttonDescriptor.keySym=XK_Escape;
						buttonDescriptor.pressedProc=ExitPress;
						buttonDescriptor.pressedProcParameters=dialog;
						buttonDescriptor.pressedState=cancel;

						if(NewDialogItem(dialog,CreateButtonItem,&buttonDescriptor))
						{
							HandleDialog(dialog);					// deal with this dialog
							result=true;
						}
					}
				}
			}
		}
		DisposeDialog(dialog);
	}
	return(result);
}

// ----------------------------------------------------------------------------
static void GetSimpleTextDialogStatTextResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the static text area
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->x=10;
	rect->y=10;
	if(attributes.width>20)
	{
		rect->w=attributes.width-20;
	}
	else
	{
		rect->w=0;
	}
	rect->h=20;
}

static void GetSimpleTextDialogButtonOkResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the ok button
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->w=50;
	rect->h=25;

	rect->x=attributes.width/5-rect->w/2;

	if(attributes.height>35)
	{
		rect->y=attributes.height-35;
	}
	else
	{
		rect->y=0;
	}
}

static void GetSimpleTextDialogTextBoxResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the text box
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->x=10;
	rect->y=30;
	if(attributes.width>20)
	{
		rect->w=attributes.width-20;
	}
	else
	{
		rect->w=0;
	}
	if(attributes.height>80)
	{
		rect->h=attributes.height-80;
	}
	else
	{
		rect->h=0;
	}
}

static void GetSimpleTextDialogButtonCancelResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the cancel button
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->w=70;
	rect->h=25;

	rect->x=4*attributes.width/5-rect->w/2;

	if(attributes.height>35)
	{
		rect->y=attributes.height-35;
	}
	else
	{
		rect->y=0;
	}
}

bool GetSimpleTextDialog(const char *title,char *enteredText,UINT32 stringBytes,bool *cancel)
// Open a string input dialog with title,
// If there is a problem, SetError, return false
// Otherwise, if cancel is true, the user canceled
// otherwise, enteredText will be set to the text that the user entered in the box
// If enteredText is set to something on entry, it will be placed in the box
{
	DIALOG
		*dialog;
	EDITOR_BUFFER
		*buffer;
	CHUNK_HEADER
		*chunk;
	UINT32
		offset;
	TEXT_BOX_DESCRIPTOR
		textBoxDescriptor;
	STATIC_TEXT_DESCRIPTOR
		textDescriptor;
	BUTTON_DESCRIPTOR
		buttonDescriptor;
	bool
		result;

	result=false;
	*cancel=false;
	if((buffer=EditorNewBuffer("simple text")))		// create a buffer in which to collect the text
	{
		if(InsertUniverseText(buffer->textUniverse,buffer->textUniverse->totalBytes,(UINT8 *)enteredText,strlen(enteredText)))
		{
			if(EditorSelectAll(buffer))				// select it all
			{
				if((dialog=CreateCenteredDialog(600,300," ",BACKGROUND_COLOR)))			// make new empty dialog
				{
					textDescriptor.resizeFunction=GetSimpleTextDialogStatTextResize;
					textDescriptor.backgroundColor=BACKGROUND_COLOR;
					textDescriptor.foregroundColor=BLACK;
					textDescriptor.string=title;
					textDescriptor.fontName=defaultDialogTextFont;

					if(NewDialogItem(dialog,CreateStaticTextItem,&textDescriptor))
					{
						textBoxDescriptor.buffer=buffer;
						textBoxDescriptor.resizeFunction=GetSimpleTextDialogTextBoxResize;
						textBoxDescriptor.topLine=0;
						textBoxDescriptor.leftPixel=0;
						textBoxDescriptor.tabSize=4;
						textBoxDescriptor.focusBackgroundColor=GRAY_3;
						textBoxDescriptor.focusForegroundColor=BLACK;
						textBoxDescriptor.nofocusBackgroundColor=GRAY_3;
						textBoxDescriptor.nofocusForegroundColor=GRAY_1;
						textBoxDescriptor.fontName=defaultViewFont;

						if(NewDialogItem(dialog,CreateTextBoxItem,&textBoxDescriptor))
						{
							buttonDescriptor.resizeFunction=GetSimpleTextDialogButtonOkResize;
							buttonDescriptor.buttonName="Ok";
							buttonDescriptor.fontName=defaultDialogButtonFont;
							buttonDescriptor.hasKey=true;
							buttonDescriptor.keySym=XK_Return;
							buttonDescriptor.pressedProc=ExitPress;
							buttonDescriptor.pressedProcParameters=dialog;
							buttonDescriptor.pressedState=NULL;

							if(NewDialogItem(dialog,CreateButtonItem,&buttonDescriptor))
							{
								buttonDescriptor.resizeFunction=GetSimpleTextDialogButtonCancelResize;
								buttonDescriptor.buttonName="Cancel";
								buttonDescriptor.fontName=defaultDialogButtonFont;
								buttonDescriptor.hasKey=true;
								buttonDescriptor.keySym=XK_Escape;
								buttonDescriptor.pressedProc=ExitPress;
								buttonDescriptor.pressedProcParameters=dialog;
								buttonDescriptor.pressedState=cancel;

								if(NewDialogItem(dialog,CreateButtonItem,&buttonDescriptor))
								{
									HandleDialog(dialog);					// deal with this dialog
									result=true;
								}
							}
						}
					}
					DisposeDialog(dialog);
				}
			}
		}
		if(result&&!(*cancel))
		{
			if(ExtractUniverseText(buffer->textUniverse,buffer->textUniverse->firstChunkHeader,0,(UINT8 *)enteredText,stringBytes,&chunk,&offset))
			{
				if(buffer->textUniverse->totalBytes<stringBytes)
				{
					enteredText[buffer->textUniverse->totalBytes]='\0';
				}
				else
				{
					if(stringBytes)
					{
						enteredText[stringBytes-1]='\0';
					}
				}
			}
			else
			{
				result=false;
			}
		}
		EditorCloseBuffer(buffer);
	}
	return(result);
}

// ----------------------------------------------------------------------------
static void SearchReplaceDialogFindStatTextResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the static text area
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->x=10;
	rect->y=10;
	if(attributes.width>20)
	{
		rect->w=attributes.width-20;
	}
	else
	{
		rect->w=0;
	}
	rect->h=20;
}

static void SearchReplaceDialogReplaceStatTextResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the static text area
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->x=10;

	if(attributes.height>150)
	{
		rect->y=(attributes.height-150)/2+10;
	}
	else
	{
		rect->y=0;
	}

	if(attributes.width>20)
	{
		rect->w=attributes.width-20;
	}
	else
	{
		rect->w=0;
	}
	rect->h=20;
}

static void SearchReplaceDialogTextBoxFindResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the find text box
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;
	INT32
		yStart;
	UINT32
		height;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->x=10;
	if(attributes.width>20)
	{
		rect->w=attributes.width-20;
	}
	else
	{
		rect->w=0;
	}

	if(attributes.height>150)
	{
		yStart=0;
		height=(UINT32)(attributes.height-150)/2;
		if(height>30)			// leave 10 for space above label, 20 for text label
		{
			rect->y=yStart+30;	// move just below the text
			rect->h=height-30;	// take up remaining space
		}
		else
		{
			rect->y=0;
			rect->h=0;
		}
	}
	else
	{
		rect->y=0;
		rect->h=0;
	}
}

static void SearchReplaceDialogTextBoxReplaceResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the find text box
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;
	INT32
		yStart;
	UINT32
		height;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->x=10;

	if(attributes.width>20)
	{
		rect->w=attributes.width-20;
	}
	else
	{
		rect->w=0;
	}

	// get Y location where replace label and text window go
	if(attributes.height>150)
	{
		yStart=(attributes.height-150)/2;
		height=(UINT32)(attributes.height-150)/2;
		if(height>30)			// leave 10 for space above label, 20 for text label
		{
			rect->y=yStart+30;	// move just below the text
			rect->h=height-30;	// take up remaining space
		}
		else
		{
			rect->y=0;
			rect->h=0;
		}
	}
	else
	{
		rect->y=0;
		rect->h=0;
	}
}

static void SearchReplaceDialogCheckBoxBackwardsResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the search backwards check box
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->x=10;

	if(attributes.height>140)
	{
		rect->y=attributes.height-140;
	}
	else
	{
		rect->y=0;
	}

	if(attributes.width/2>20)
	{
		rect->w=attributes.width/2-20;
	}
	else
	{
		rect->w=0;
	}
	rect->h=20;
}

static void SearchReplaceDialogCheckBoxWrapResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the search wrap around check box
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->x=10;

	if(attributes.height>115)
	{
		rect->y=attributes.height-115;
	}
	else
	{
		rect->y=0;
	}

	if(attributes.width/2>20)
	{
		rect->w=attributes.width/2-20;
	}
	else
	{
		rect->w=0;
	}
	rect->h=20;
}

static void SearchReplaceDialogCheckBoxExpressionResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the search expression check box
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->x=attributes.width/2+10;

	if(attributes.height>140)
	{
		rect->y=attributes.height-140;
	}
	else
	{
		rect->y=0;
	}

	if(attributes.width/2>20)
	{
		rect->w=attributes.width/2-20;
	}
	else
	{
		rect->w=0;
	}
	rect->h=20;
}

static void SearchReplaceDialogCheckBoxIgnoreCaseResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the search ignore case check box
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->x=attributes.width/2+10;

	if(attributes.height>115)
	{
		rect->y=attributes.height-115;
	}
	else
	{
		rect->y=0;
	}

	if(attributes.width/2>20)
	{
		rect->w=attributes.width/2-20;
	}
	else
	{
		rect->w=0;
	}
	rect->h=20;
}

static void SearchReplaceDialogCheckBoxLimitScopeResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the search limit scope check box
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->x=10;

	if(attributes.height>90)
	{
		rect->y=attributes.height-90;
	}
	else
	{
		rect->y=0;
	}

	if(attributes.width>20)
	{
		rect->w=attributes.width-20;
	}
	else
	{
		rect->w=0;
	}
	rect->h=20;
}

static void SearchReplaceDialogCheckBoxTclResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the search Tcl replacement check box
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->x=10;

	if(attributes.height>65)
	{
		rect->y=attributes.height-65;
	}
	else
	{
		rect->y=0;
	}

	if(attributes.width>20)
	{
		rect->w=attributes.width-20;
	}
	else
	{
		rect->w=0;
	}
	rect->h=20;
}

static INT32 GetFindButtonX(UINT32 index,UINT32 windowWidth)
// Work out the X index for a button in the find window
// Index is passed as the index of the button (0 through 5)
// There is usually no button at index 4.
// All the buttons are assumed to be 100 pixels wide.
// This is a bit of a hack -- later make all of this work
// better.
{
	if(windowWidth>=(20+100))
	{
		return(10+(windowWidth-(20+100))*index/5);
	}
	return(10);
}

static void SearchReplaceDialogButtonFindResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the find button
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->w=100;
	rect->h=25;

	rect->x=GetFindButtonX(0,attributes.width);

	if(attributes.height>35)
	{
		rect->y=attributes.height-35;
	}
	else
	{
		rect->y=0;
	}
}

static void SearchReplaceDialogButtonFindAllResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the find all button
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->w=100;
	rect->h=25;

	rect->x=GetFindButtonX(1,attributes.width);

	if(attributes.height>35)
	{
		rect->y=attributes.height-35;
	}
	else
	{
		rect->y=0;
	}
}

static void SearchReplaceDialogButtonReplaceResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the replace button
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->w=100;
	rect->h=25;

	rect->x=GetFindButtonX(2,attributes.width);

	if(attributes.height>35)
	{
		rect->y=attributes.height-35;
	}
	else
	{
		rect->y=0;
	}
}

static void SearchReplaceDialogButtonReplaceAllResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the replace all button
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->w=100;
	rect->h=25;

	rect->x=GetFindButtonX(3,attributes.width);

	if(attributes.height>35)
	{
		rect->y=attributes.height-35;
	}
	else
	{
		rect->y=0;
	}
}

static void SearchReplaceDialogButtonCancelResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the cancel button
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->w=100;
	rect->h=25;

	rect->x=GetFindButtonX(5,attributes.width);

	if(attributes.height>35)
	{
		rect->y=attributes.height-35;
	}
	else
	{
		rect->y=0;
	}
}

bool SearchReplaceDialog(EDITOR_BUFFER *searchBuffer,EDITOR_BUFFER *replaceBuffer,bool *backwards,bool *wrapAround,bool *selectionExpr,bool *ignoreCase,bool *limitScope,bool *replaceProc,UINT16 *searchType,bool *cancel)
// Open the search/replace dialog
// If there is a problem, SetError, return false
// Otherwise, if cancel is true, the user canceled
// Otherwise, return the enumerated mode that the user chose, as well as
// the states of backwards, wrapAround, selectionExpr, ignoreCase,
// limitScope, replaceProc
{
	DIALOG
		*dialog;
	TEXT_BOX_DESCRIPTOR
		textBoxDescriptor;
	STATIC_TEXT_DESCRIPTOR
		textDescriptor;
	BUTTON_DESCRIPTOR
		buttonDescriptor;
	CHECK_BOX_DESCRIPTOR
		checkDescriptor;
	bool
		result;
	bool
		hadFind,
		hadFindAll,
		hadReplace,
		hadReplaceAll;

	result=false;
	*cancel=false;
	hadFind=false;
	hadFindAll=false;
	hadReplace=false;
	hadReplaceAll=false;

	if(EditorSelectAll(searchBuffer))							// select it all
	{
		if(EditorSelectAll(replaceBuffer))						// select it all
		{
			if((dialog=CreateCenteredDialog(620,430," ",BACKGROUND_COLOR)))	// make new empty dialog
			{
				textDescriptor.resizeFunction=SearchReplaceDialogFindStatTextResize;
				textDescriptor.backgroundColor=BACKGROUND_COLOR;
				textDescriptor.foregroundColor=BLACK;
				textDescriptor.string="Find:";
				textDescriptor.fontName=defaultDialogTextFont;

				if(NewDialogItem(dialog,CreateStaticTextItem,&textDescriptor))
				{
					textBoxDescriptor.buffer=searchBuffer;
					textBoxDescriptor.resizeFunction=SearchReplaceDialogTextBoxFindResize;
					textBoxDescriptor.topLine=0;
					textBoxDescriptor.leftPixel=0;
					textBoxDescriptor.tabSize=4;
					textBoxDescriptor.focusBackgroundColor=GRAY_3;
					textBoxDescriptor.focusForegroundColor=BLACK;
					textBoxDescriptor.nofocusBackgroundColor=GRAY_3;
					textBoxDescriptor.nofocusForegroundColor=GRAY_1;
					textBoxDescriptor.fontName=defaultViewFont;

					if(NewDialogItem(dialog,CreateTextBoxItem,&textBoxDescriptor))
					{
						textDescriptor.resizeFunction=SearchReplaceDialogReplaceStatTextResize;
						textDescriptor.backgroundColor=BACKGROUND_COLOR;
						textDescriptor.foregroundColor=BLACK;
						textDescriptor.string="Replace with:";
						textDescriptor.fontName=defaultDialogTextFont;

						if(NewDialogItem(dialog,CreateStaticTextItem,&textDescriptor))
						{
							textBoxDescriptor.buffer=replaceBuffer;
							textBoxDescriptor.resizeFunction=SearchReplaceDialogTextBoxReplaceResize;
							textBoxDescriptor.topLine=0;
							textBoxDescriptor.leftPixel=0;
							textBoxDescriptor.tabSize=4;
							textBoxDescriptor.focusBackgroundColor=GRAY_3;
							textBoxDescriptor.focusForegroundColor=BLACK;
							textBoxDescriptor.nofocusBackgroundColor=GRAY_3;
							textBoxDescriptor.nofocusForegroundColor=GRAY_1;
							textBoxDescriptor.fontName=defaultViewFont;

							if(NewDialogItem(dialog,CreateTextBoxItem,&textBoxDescriptor))
							{
								// check boxes
								checkDescriptor.resizeFunction=SearchReplaceDialogCheckBoxBackwardsResize;
								checkDescriptor.checkBoxName="Search Backwards";
								checkDescriptor.fontName=defaultDialogTextFont;
								checkDescriptor.hasKey=false;
								checkDescriptor.pressedProc=NULL;
								checkDescriptor.pressedProcParameters=NULL;
								checkDescriptor.pressedState=backwards;

								if(NewDialogItem(dialog,CreateCheckBoxItem,&checkDescriptor))
								{
									checkDescriptor.resizeFunction=SearchReplaceDialogCheckBoxWrapResize;
									checkDescriptor.checkBoxName="Wrap Around";
									checkDescriptor.fontName=defaultDialogTextFont;
									checkDescriptor.hasKey=false;
									checkDescriptor.pressedProc=NULL;
									checkDescriptor.pressedProcParameters=NULL;
									checkDescriptor.pressedState=wrapAround;

									if(NewDialogItem(dialog,CreateCheckBoxItem,&checkDescriptor))
									{
										checkDescriptor.resizeFunction=SearchReplaceDialogCheckBoxExpressionResize;
										checkDescriptor.checkBoxName="Selection Expression";
										checkDescriptor.fontName=defaultDialogTextFont;
										checkDescriptor.hasKey=false;
										checkDescriptor.pressedProc=NULL;
										checkDescriptor.pressedProcParameters=NULL;
										checkDescriptor.pressedState=selectionExpr;

										if(NewDialogItem(dialog,CreateCheckBoxItem,&checkDescriptor))
										{
											checkDescriptor.resizeFunction=SearchReplaceDialogCheckBoxIgnoreCaseResize;
											checkDescriptor.checkBoxName="Ignore Case";
											checkDescriptor.fontName=defaultDialogTextFont;
											checkDescriptor.hasKey=false;
											checkDescriptor.pressedProc=NULL;
											checkDescriptor.pressedProcParameters=NULL;
											checkDescriptor.pressedState=ignoreCase;

											if(NewDialogItem(dialog,CreateCheckBoxItem,&checkDescriptor))
											{
												checkDescriptor.resizeFunction=SearchReplaceDialogCheckBoxLimitScopeResize;
												checkDescriptor.checkBoxName="Limit scope of 'Find All'/'Replace All' to current selection";
												checkDescriptor.fontName=defaultDialogTextFont;
												checkDescriptor.hasKey=false;
												checkDescriptor.pressedProc=NULL;
												checkDescriptor.pressedProcParameters=NULL;
												checkDescriptor.pressedState=limitScope;

												if(NewDialogItem(dialog,CreateCheckBoxItem,&checkDescriptor))
												{
													checkDescriptor.resizeFunction=SearchReplaceDialogCheckBoxTclResize;
													checkDescriptor.checkBoxName="Treat replace text as Tcl procedure -- substitute results";
													checkDescriptor.fontName=defaultDialogTextFont;
													checkDescriptor.hasKey=false;
													checkDescriptor.pressedProc=NULL;
													checkDescriptor.pressedProcParameters=NULL;
													checkDescriptor.pressedState=replaceProc;

													if(NewDialogItem(dialog,CreateCheckBoxItem,&checkDescriptor))
													{
														buttonDescriptor.resizeFunction=SearchReplaceDialogButtonFindResize;
														buttonDescriptor.buttonName="Find";
														buttonDescriptor.fontName=defaultDialogButtonFont;
														buttonDescriptor.hasKey=true;
														buttonDescriptor.keySym=XK_Return;
														buttonDescriptor.pressedProc=ExitPress;
														buttonDescriptor.pressedProcParameters=dialog;
														buttonDescriptor.pressedState=&hadFind;

														if(NewDialogItem(dialog,CreateButtonItem,&buttonDescriptor))
														{
															buttonDescriptor.resizeFunction=SearchReplaceDialogButtonFindAllResize;
															buttonDescriptor.buttonName="Find All";
															buttonDescriptor.fontName=defaultDialogButtonFont;
															buttonDescriptor.hasKey=false;
															buttonDescriptor.pressedProc=ExitPress;
															buttonDescriptor.pressedProcParameters=dialog;
															buttonDescriptor.pressedState=&hadFindAll;

															if(NewDialogItem(dialog,CreateButtonItem,&buttonDescriptor))
															{
															buttonDescriptor.resizeFunction=SearchReplaceDialogButtonReplaceResize;
																buttonDescriptor.buttonName="Replace";
																buttonDescriptor.fontName=defaultDialogButtonFont;
																buttonDescriptor.hasKey=false;
																buttonDescriptor.pressedProc=ExitPress;
																buttonDescriptor.pressedProcParameters=dialog;
																buttonDescriptor.pressedState=&hadReplace;

																if(NewDialogItem(dialog,CreateButtonItem,&buttonDescriptor))
																{
																	buttonDescriptor.resizeFunction=SearchReplaceDialogButtonReplaceAllResize;
																	buttonDescriptor.buttonName="Replace All";
																	buttonDescriptor.fontName=defaultDialogButtonFont;
																	buttonDescriptor.hasKey=false;
																	buttonDescriptor.pressedProc=ExitPress;
																	buttonDescriptor.pressedProcParameters=dialog;
																	buttonDescriptor.pressedState=&hadReplaceAll;

																	if(NewDialogItem(dialog,CreateButtonItem,&buttonDescriptor))
																	{
																		buttonDescriptor.resizeFunction=SearchReplaceDialogButtonCancelResize;
																		buttonDescriptor.buttonName="Cancel";
																		buttonDescriptor.fontName=defaultDialogButtonFont;
																		buttonDescriptor.hasKey=true;
																		buttonDescriptor.keySym=XK_Escape;
																		buttonDescriptor.pressedProc=ExitPress;
																		buttonDescriptor.pressedProcParameters=dialog;
																		buttonDescriptor.pressedState=cancel;

																		if(NewDialogItem(dialog,CreateButtonItem,&buttonDescriptor))
																		{
																			HandleDialog(dialog);					// deal with this dialog
																			result=true;
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
							}
						}
					}
				}
				DisposeDialog(dialog);
			}
		}
	}
	if(result)
	{
		if(hadFind)
		{
			(*searchType)=ST_FIND;
		}
		if(hadFindAll)
		{
			(*searchType)=ST_FINDALL;
		}
		if(hadReplace)
		{
			(*searchType)=ST_REPLACE;
		}
		if(hadReplaceAll)
		{
			(*searchType)=ST_REPLACEALL;
		}
	}
	return(result);
}

// ----------------------------------------------------------------------------
static void SimpleListBoxDialogStatTextResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the static text area
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->x=10;
	rect->y=10;
	if(attributes.width>20)
	{
		rect->w=attributes.width-20;
	}
	else
	{
		rect->w=0;
	}
	rect->h=20;
}

static void SimpleListBoxDialogListBoxResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the list box area
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->x=10;
	rect->y=30;
	if(attributes.width>20)
	{
		rect->w=attributes.width-20;
	}
	else
	{
		rect->w=0;
	}
	if(attributes.height>80)
	{
		rect->h=attributes.height-80;
	}
	else
	{
		rect->h=0;
	}
}

static void SimpleListDialogButtonOkResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the ok button
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->w=50;
	rect->h=25;

	rect->x=attributes.width/5-rect->w/2;

	if(attributes.height>35)
	{
		rect->y=attributes.height-35;
	}
	else
	{
		rect->y=0;
	}
}

static void SimpleListDialogButtonCancelResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the cancel button
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->w=70;
	rect->h=25;

	rect->x=4*attributes.width/5-rect->w/2;

	if(attributes.height>35)
	{
		rect->y=attributes.height-35;
	}
	else
	{
		rect->y=0;
	}
}

static void ListExitPress(void *parameters)
// When the simple list is double clicked in, it is just like
// pressing the Ok button, so we do that!
{
	DIALOG
		*dialog;

	dialog=(DIALOG *)parameters;
	PressButtonItem((DIALOG_ITEM *)dialog->dialogLocal);
}

bool SimpleListBoxDialog(const char *title,UINT32 numElements,const char **listElements,bool *selectedElements,bool *cancel)
// Open a list selection dialog box with two buttons (ok, cancel)
// The list dialog will be used to modify the selectionListArray, to tell
// which items were selected, and which were not
// If there is a problem, SetError, return false
// Otherwise, if cancel is true, the user canceled
// Otherwise, selectedElements is modified to reflect which items
// have been selected
{
	DIALOG
		*dialog;
	LIST_BOX_DESCRIPTOR
		listBoxDescriptor;
	STATIC_TEXT_DESCRIPTOR
		textDescriptor;
	BUTTON_DESCRIPTOR
		buttonDescriptor;
	bool
		done,
		result;
	UINT32
		index;

	result=false;
	*cancel=false;

	if((dialog=CreateCenteredDialog(600,300," ",BACKGROUND_COLOR)))			// make new empty dialog
	{

		textDescriptor.resizeFunction=SimpleListBoxDialogStatTextResize;
		textDescriptor.backgroundColor=BACKGROUND_COLOR;
		textDescriptor.foregroundColor=BLACK;
		textDescriptor.string=title;
		textDescriptor.fontName=defaultDialogTextFont;

		if(NewDialogItem(dialog,CreateStaticTextItem,&textDescriptor))
		{
			listBoxDescriptor.resizeFunction=SimpleListBoxDialogListBoxResize;
			listBoxDescriptor.topLine=0;						// if no selection found, take top as 0
			done=false;
			for(index=0;!done&&index<numElements;index++)		// make top line the first selection if possible
			{
				if(selectedElements[index])
				{
					listBoxDescriptor.topLine=index;
					done=true;
				}
			}
			listBoxDescriptor.focusBackgroundColor=GRAY_3;
			listBoxDescriptor.focusForegroundColor=BLACK;
			listBoxDescriptor.nofocusBackgroundColor=GRAY_3;
			listBoxDescriptor.nofocusForegroundColor=BLACK;
			listBoxDescriptor.numElements=numElements;
			listBoxDescriptor.listElements=listElements;
			listBoxDescriptor.selectedElements=selectedElements;
			listBoxDescriptor.fontName=defaultViewFont;
			listBoxDescriptor.pressedProc=ListExitPress;
			listBoxDescriptor.pressedProcParameters=dialog;

			if(NewDialogItem(dialog,CreateListBoxItem,&listBoxDescriptor))
			{
				buttonDescriptor.resizeFunction=SimpleListDialogButtonOkResize;
				buttonDescriptor.buttonName="Ok";
				buttonDescriptor.fontName=defaultDialogButtonFont;
				buttonDescriptor.hasKey=true;
				buttonDescriptor.keySym=XK_Return;
				buttonDescriptor.pressedProc=ExitPress;
				buttonDescriptor.pressedProcParameters=dialog;
				buttonDescriptor.pressedState=NULL;

				if((dialog->dialogLocal=NewDialogItem(dialog,CreateButtonItem,&buttonDescriptor)))	// remember this button
				{
					buttonDescriptor.resizeFunction=SimpleListDialogButtonCancelResize;
					buttonDescriptor.buttonName="Cancel";
					buttonDescriptor.fontName=defaultDialogButtonFont;
					buttonDescriptor.hasKey=true;
					buttonDescriptor.keySym=XK_Escape;
					buttonDescriptor.pressedProc=ExitPress;
					buttonDescriptor.pressedProcParameters=dialog;
					buttonDescriptor.pressedState=cancel;

					if(NewDialogItem(dialog,CreateButtonItem,&buttonDescriptor))
					{
						HandleDialog(dialog);					// deal with this dialog
						result=true;
					}
				}
			}
		}
		DisposeDialog(dialog);
	}
	return(result);
}

// ----------------------------------------------------------------------------
// Stuff dealing with open/save dialogs

// This structure is hung off of the open/save dialogs, it allows the dialog item
// procedures access to some things they need

typedef struct
{
	EDITOR_BUFFER
		*buffer;
	char
		pathString[MAXPATHLEN+1];
	char
		**list;
	bool
		*selectionList;
	UINT32
		numElements;
	char
		**outputList;
	UINT32
		outputElements;
	bool
		justOne,
		justPaths;
	DIALOG_ITEM
		*pathText,
		*listBox,
		*textBox,
		*OkButton;
} FILE_DIALOG_PARAMETERS;

static void FileDialogStatTextTitleResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the static text area
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->x=10;
	rect->y=10;
	if(attributes.width>20)
	{
		rect->w=attributes.width-20;
	}
	else
	{
		rect->w=0;
	}
	rect->h=20;
}

static void FileDialogStatTextPathResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the static text area
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->x=10;
	rect->y=30;
	if(attributes.width>20)
	{
		rect->w=attributes.width-20;
	}
	else
	{
		rect->w=0;
	}
	rect->h=40;
}

static void FileDialogTextBoxResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the file dialog path entry box
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->x=10;
	rect->y=66;
	if(attributes.width>20)
	{
		rect->w=attributes.width-20;
	}
	else
	{
		rect->w=0;
	}
	rect->h=30;
}

static void FileDialogListBoxResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the file dialog list box
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->x=10;
	rect->y=110;
	if(attributes.width>20)
	{
		rect->w=attributes.width-20;
	}
	else
	{
		rect->w=0;
	}

	if(attributes.height>160)
	{
		rect->h=attributes.height-160;
	}
	else
	{
		rect->h=0;
	}
}

static void FileDialogButtonOkResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the ok button
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->w=50;
	rect->h=25;

	rect->x=attributes.width/5-rect->w/2;

	if(attributes.height>35)
	{
		rect->y=attributes.height-35;
	}
	else
	{
		rect->y=0;
	}
}

static void FileDialogButtonCancelResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the cancel button
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->w=70;
	rect->h=25;

	rect->x=4*attributes.width/5-rect->w/2;

	if(attributes.height>35)
	{
		rect->y=attributes.height-35;
	}
	else
	{
		rect->y=0;
	}
}

static void FileDialogButtonAcceptPathResize(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Resize the accept path button
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		attributes;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	XGetWindowAttributes(xDisplay,windowElement->xWindow,&attributes);

	rect->w=100;
	rect->h=25;

	rect->x=attributes.width/2-rect->w/2;

	if(attributes.height>35)
	{
		rect->y=attributes.height-35;
	}
	else
	{
		rect->y=0;
	}
}

static char **ProcessDirectoryList(const char *pathString,char **list,UINT32 inElements,UINT32 *outElements,bool justPaths)
// Given a path string, and a sorted raw list of files at the end of the path
// create a new list which contains the processed versions of each entry
// (place PATH_SEP after directories, eliminate files if justPaths is true)
// If there is a problem, SetError, return NULL
// NOTE: this does not dispose of the incoming list
// the output list must be disposed of by calling FreeStringList
{
	char
		**outList;
	char
		tempPath[MAXPATHLEN+1];
	int
		length;
	UINT32
		inElement;
	struct stat
		statOut;

	*outElements=0;
	if((outList=NewStringList()))
	{
		for(inElement=0;inElement<inElements;inElement++)
		{
			ConcatPathAndFile(pathString,list[inElement],tempPath);
			if(stat(tempPath,&statOut)!=-1)				// if this fails, then do not copy the name
			{
				if(S_ISDIR(statOut.st_mode))
				{
					length=strlen(list[inElement]);
					if(length<MAXPATHLEN)
					{
						strcpy(tempPath,list[inElement]);
						tempPath[length++]=PATH_SEP;	// drop PATH_SEP onto the string
						tempPath[length]='\0';
						AddStringToList(tempPath,&outList,outElements);	// ignore errors here
					}
				}
				else
				{
					if(!justPaths)						// see if file names should be added
					{
						AddStringToList(list[inElement],&outList,outElements);	// ignore errors here
					}
				}
			}
		}
	}
	return(outList);
}

static void DisposeSelectionFiles(char **list,bool *selectionList)
// Dispose of the two lists returned by GetPathAndSelectionFiles
{
	MDisposePtr(selectionList);
	FreeStringList(list);
}

static char **GetSelectionFiles(const char *pathString,UINT32 *numElements,bool **selectionList,bool justPaths)
// Given pathString, create a list of the files/directories at the end of the path
// DisposeSelectionFiles must be called on list, and selectionList to dispose of them
// If there is a problem, SetError, return NULL
{
	char
		**rawList,
		**list;
	UINT32
		rawElements;

	if((rawList=GlobAll(pathString,&rawElements)))
	{
		list=ProcessDirectoryList(pathString,rawList,rawElements,numElements,justPaths);
		FreeStringList(rawList);					// get rid of the raw list
		if(list)
		{
			if((*selectionList=(bool *)MNewPtrClr((*numElements)*sizeof(bool))))
			{
				return(list);
			}
			FreeStringList(list);
		}
	}

	*selectionList=NULL;
	*numElements=0;

	return(NULL);
}

static void FileListPress(void *parameters)
// When a selection has been made from the file list, this
// will decide what to do
{
	DIALOG
		*dialog;
	FILE_DIALOG_PARAMETERS
		*dialogParameters;

	dialog=(DIALOG *)parameters;
	dialogParameters=(FILE_DIALOG_PARAMETERS *)dialog->dialogLocal;
	PressButtonItem(dialogParameters->OkButton);		// behave as if Ok was pressed
}

static bool AddPathAndFileToList(const char *path,const char *file,char ***list,UINT32 *numElements)
// Like AddStringToList, but takes a path, and a filename which are
// concatenated into a complete path, and added to the list
// If there is a problem, SetError, and return false
{
	char
		fullPath[MAXPATHLEN+1];

	ConcatPathAndFile(path,file,fullPath);
	return(AddStringToList(fullPath,list,numElements));
}

static bool FileOkInText(DIALOG *dialog,FILE_DIALOG_PARAMETERS *dialogParameters)
// Ok was pressed while the text was active
{
	CHUNK_HEADER
		*chunk;
	UINT32
		offset;
	char
		globPath[MAXPATHLEN+1];
	char
		**globList;
	UINT32
		currentElement,
		globElements;
	bool
		fail;

	fail=false;

	if((dialogParameters->outputList=NewStringList()))
	{
		dialogParameters->outputElements=0;
		if(ExtractUniverseText(dialogParameters->buffer->textUniverse,dialogParameters->buffer->textUniverse->firstChunkHeader,0,(UINT8 *)globPath,MAXPATHLEN,&chunk,&offset))
		{
			if(dialogParameters->buffer->textUniverse->totalBytes<MAXPATHLEN)
			{
				globPath[dialogParameters->buffer->textUniverse->totalBytes]='\0';
			}
			else
			{
				globPath[MAXPATHLEN]='\0';
			}
			if((globList=Glob(&(dialogParameters->pathString[0]),globPath,false,true,&globElements)))
			{
				for(currentElement=0;!fail&&currentElement<globElements;currentElement++)
				{
					fail=!AddPathAndFileToList(&(dialogParameters->pathString[0]),globList[currentElement],&dialogParameters->outputList,&dialogParameters->outputElements);
				}
				FreeStringList(globList);
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
		if(fail)
		{
			FreeStringList(dialogParameters->outputList);
			dialogParameters->outputElements=0;
		}
	}
	return(!fail);
}

static bool FileOkInList(DIALOG *dialog,FILE_DIALOG_PARAMETERS *dialogParameters)
// Ok was pressed while the list was active
// Create an output list which contains all the selected items
// If there is a problem, SetError, return false
// NOTE: to make the interface seem more intuitive, if there
// are no selected items in the list, this will revert to making
// a list from the text box
{
	UINT32
		currentElement;
	bool
		noneSelected,
		fail;

	fail=false;
	noneSelected=true;
	if((dialogParameters->outputList=NewStringList()))
	{
		dialogParameters->outputElements=0;
		for(currentElement=0;!fail&&currentElement<dialogParameters->numElements;currentElement++)
		{
			if(dialogParameters->selectionList[currentElement])	// is this one selected?
			{
				noneSelected=false;
				fail=!AddPathAndFileToList(&(dialogParameters->pathString[0]),dialogParameters->list[currentElement],&dialogParameters->outputList,&dialogParameters->outputElements);
			}
		}
		if(fail||noneSelected)
		{
			FreeStringList(dialogParameters->outputList);
			dialogParameters->outputElements=0;
		}
	}
	if(!fail)
	{
		if(noneSelected)									// if not failure, but nothing selected, use dialog text
		{
			return(FileOkInText(dialog,dialogParameters));
		}
		else
		{
			return(true);
		}
	}
	return(false);
}

static void FileOkPress(void *parameters)
// The user pressed "Ok" in the dialog (or it was pressed for him)
// take appropriate action
{
	DIR
		*directory;
	DIALOG
		*dialog;
	FILE_DIALOG_PARAMETERS
		*dialogParameters;
	bool
		fail;

	ClearErrorTrace();
	dialog=(DIALOG *)parameters;
	dialogParameters=(FILE_DIALOG_PARAMETERS *)dialog->dialogLocal;
	fail=false;
	if(dialog->focusItem==dialogParameters->listBox)	// if list has focus, then build a list out of what's selected
	{
		fail=!FileOkInList(dialog,dialogParameters);
	}
	else
	{
		fail=!FileOkInText(dialog,dialogParameters);
	}

	if(!fail)
	{
		if(dialogParameters->outputElements==1)			// if only one element, then see if directory
		{
			if((directory=opendir(dialogParameters->outputList[0])))	// see if it is a directory, and we are allowed to open it
			{
				closedir(directory);					// did not want it open, just wanted to see if we could
				if(!(dialog->focusItem==dialogParameters->listBox))
				{
					ClearBuffer(dialogParameters->buffer);	// get rid of any text in the view
				}
				if(dialogParameters->list)
				{
					DisposeSelectionFiles(dialogParameters->list,dialogParameters->selectionList);
				}
				realpath2(dialogParameters->outputList[0],&(dialogParameters->pathString[0]));	// this is the new path
				dialogParameters->list=GetSelectionFiles(&(dialogParameters->pathString[0]),&dialogParameters->numElements,&dialogParameters->selectionList,dialogParameters->justPaths);
				ResetStaticTextItemText(dialogParameters->pathText,&(dialogParameters->pathString[0]));
				ResetListBoxItemLists(dialogParameters->listBox,dialogParameters->numElements,(const char **)dialogParameters->list,dialogParameters->selectionList);
				ResetMultipleClicks();					// reset multiple click timers
			}
			else
			{
				if(errno==ENOTDIR||errno==ENOENT)
				{
					if(!dialogParameters->justPaths)
					{
						dialog->dialogComplete=true;		// if not a directory, or known file, return it
					}
					else
					{
						ReportMessage("%s\nNot a directory\n",dialogParameters->outputList[0]);	// let user know what's wrong
					}
				}
				else
				{
					SetStdCLibError();						// report whatever error we got
					ReportMessage("%s\n",GetErrorTrace());	// let user know what's wrong
				}
			}
		}
		else
		{
			if(dialogParameters->justOne||dialogParameters->justPaths)
			{
				ReportMessage("Path is ambiguous\n");		// got more elements than asked for, so complain, and keep going
			}
			else
			{
				dialog->dialogComplete=true;				// if anything other that 1 element, pass it back as the result
			}
		}
		if(!dialog->dialogComplete)
		{
			FreeStringList(dialogParameters->outputList);	// not leaving, so kill the list
		}
	}
	else
	{
		ReportMessage("%s\n",GetErrorTrace());				// let user know what's wrong
	}
}

static bool FileDialog(const char *title,char *fullPath,UINT32 stringBytes,char ***list,bool justOne,bool justPaths,bool *cancel)
// Bring up a file open/save dialog
// If there is a problem, SetError, return false
// Otherwise, cancel is true if the user canceled the dialog
//
// Otherwise, if justPaths is true, return the final path chosen in the
// dialog in fullPath, truncating to stringBytes in length
//
// Otherwise, listElements contains the pointer to the start of an array of
// pointers to char which are the selections.
// It must be disposed of at some later time with a call to FreeFileDialogPaths
// NOTE: the array is terminated with a pointer to NULL
{
	DIALOG
		*dialog;
	TEXT_BOX_DESCRIPTOR
		textBoxDescriptor;
	LIST_BOX_DESCRIPTOR
		listBoxDescriptor;
	STATIC_TEXT_DESCRIPTOR
		textDescriptor;
	BUTTON_DESCRIPTOR
		buttonDescriptor;
	FILE_DIALOG_PARAMETERS
		dialogParameters;
	bool
		result;
	char
		filePart[MAXPATHLEN+1];

	result=false;
	*cancel=false;

	dialogParameters.justOne=justOne;
	dialogParameters.justPaths=justPaths;

	if((dialogParameters.buffer=EditorNewBuffer("file dialog")))			// create a buffer in which to collect the file name
	{
		SplitPathAndFile(fullPath,dialogParameters.pathString,filePart);	// split the incoming name into file and path components
		if(InsertUniverseText(dialogParameters.buffer->textUniverse,dialogParameters.buffer->textUniverse->totalBytes,(UINT8 *)filePart,strlen(filePart)))
		{
			if(EditorSelectAll(dialogParameters.buffer))					// select it all
			{
				if((dialog=CreateCenteredDialog(440,470," ",BACKGROUND_COLOR)))		// make new empty dialog
				{
					dialog->dialogLocal=&dialogParameters;				// hang parameter list off of the dialog

					dialogParameters.list=GetSelectionFiles(&(dialogParameters.pathString[0]),&dialogParameters.numElements,&dialogParameters.selectionList,justPaths);	// ignore failure, since list box will handle a NULL list

					textDescriptor.resizeFunction=FileDialogStatTextTitleResize;
					textDescriptor.backgroundColor=BACKGROUND_COLOR;
					textDescriptor.foregroundColor=BLACK;
					textDescriptor.string=title;
					textDescriptor.fontName=defaultDialogTextFont;

					if(NewDialogItem(dialog,CreateStaticTextItem,&textDescriptor))
					{
						textDescriptor.resizeFunction=FileDialogStatTextPathResize;
						textDescriptor.backgroundColor=BACKGROUND_COLOR;
						textDescriptor.foregroundColor=BLACK;
						textDescriptor.string=&(dialogParameters.pathString[0]);
						textDescriptor.fontName=defaultViewFont;

						if((dialogParameters.pathText=NewDialogItem(dialog,CreateStaticTextItem,&textDescriptor)))
						{

							textBoxDescriptor.buffer=dialogParameters.buffer;
							textBoxDescriptor.resizeFunction=FileDialogTextBoxResize;
							textBoxDescriptor.topLine=0;
							textBoxDescriptor.leftPixel=0;
							textBoxDescriptor.tabSize=4;
							textBoxDescriptor.focusBackgroundColor=GRAY_3;
							textBoxDescriptor.focusForegroundColor=BLACK;
							textBoxDescriptor.nofocusBackgroundColor=GRAY_3;
							textBoxDescriptor.nofocusForegroundColor=GRAY_1;
							textBoxDescriptor.fontName=defaultViewFont;

							if((dialogParameters.textBox=NewDialogItem(dialog,CreateTextBoxItem,&textBoxDescriptor)))
							{
								listBoxDescriptor.resizeFunction=FileDialogListBoxResize;
								listBoxDescriptor.topLine=0;
								listBoxDescriptor.focusBackgroundColor=GRAY_3;
								listBoxDescriptor.focusForegroundColor=BLACK;
								listBoxDescriptor.nofocusBackgroundColor=GRAY_3;
								listBoxDescriptor.nofocusForegroundColor=GRAY_1;
								listBoxDescriptor.numElements=dialogParameters.numElements;
								listBoxDescriptor.listElements=(const char **)dialogParameters.list;
								listBoxDescriptor.selectedElements=dialogParameters.selectionList;
								listBoxDescriptor.fontName=defaultViewFont;
								listBoxDescriptor.pressedProc=FileListPress;
								listBoxDescriptor.pressedProcParameters=dialog;

								if((dialogParameters.listBox=NewDialogItem(dialog,CreateListBoxItem,&listBoxDescriptor)))
								{
									buttonDescriptor.resizeFunction=FileDialogButtonOkResize;
									buttonDescriptor.buttonName="Ok";
									buttonDescriptor.fontName=defaultDialogButtonFont;
									buttonDescriptor.hasKey=true;
									buttonDescriptor.keySym=XK_Return;
									buttonDescriptor.pressedProc=FileOkPress;
									buttonDescriptor.pressedProcParameters=dialog;
									buttonDescriptor.pressedState=NULL;

									if((dialogParameters.OkButton=NewDialogItem(dialog,CreateButtonItem,&buttonDescriptor)))	// remember this button
									{
										buttonDescriptor.resizeFunction=FileDialogButtonCancelResize;
										buttonDescriptor.buttonName="Cancel";
										buttonDescriptor.fontName=defaultDialogButtonFont;
										buttonDescriptor.hasKey=true;
										buttonDescriptor.keySym=XK_Escape;
										buttonDescriptor.pressedProc=ExitPress;
										buttonDescriptor.pressedProcParameters=dialog;
										buttonDescriptor.pressedState=cancel;

										if(NewDialogItem(dialog,CreateButtonItem,&buttonDescriptor))
										{
											if(justPaths)			// add another button for paths (use enter to select)
											{
												buttonDescriptor.resizeFunction=FileDialogButtonAcceptPathResize;
												buttonDescriptor.buttonName="Accept Path";
												buttonDescriptor.fontName=defaultDialogButtonFont;
												buttonDescriptor.hasKey=true;
												buttonDescriptor.keySym=XK_KP_Enter;
												buttonDescriptor.pressedProc=ExitPress;
												buttonDescriptor.pressedProcParameters=dialog;
												buttonDescriptor.pressedState=NULL;

												if(NewDialogItem(dialog,CreateButtonItem,&buttonDescriptor))
												{
													HandleDialog(dialog);	// deal with this dialog
													result=true;
												}
											}
											else
											{
												HandleDialog(dialog);	// deal with this dialog
												result=true;
											}
										}
									}
								}
							}
						}
					}
					if(dialogParameters.list)							// delete any lists that are hanging around
					{
						DisposeSelectionFiles(dialogParameters.list,dialogParameters.selectionList);
					}
					DisposeDialog(dialog);
				}
			}
		}
		EditorCloseBuffer(dialogParameters.buffer);
	}
	if(justPaths)
	{
		if(stringBytes)
		{
			strncpy(fullPath,dialogParameters.pathString,stringBytes);	// return the path
			fullPath[stringBytes-1]='\0';
		}
	}
	if(result&&!(*cancel)&&(!justPaths))
	{
		*list=dialogParameters.outputList;
	}
	return(result);
}

static void FreeFileDialogPaths(char **paths)
// Free an array of paths that was returned from FileDialog
{
	FreeStringList(paths);
}

bool OpenFileDialog(const char *title,char *fullPath,UINT32 stringBytes,char ***listElements,bool *cancel)
// Bring up a file open dialog
// If there is a problem, SetError, return false
// Otherwise, cancel is true if the user canceled the dialog
//
// Otherwise, listElements contains the pointer to the start of an array of
// pointers to char which are the selections
// It must be disposed of at some later time with a call to FreeOpenFileDialogPaths
// NOTE: the array is terminated with a pointer to NULL
//
// NOTE: this is NOT allowed to change the current directory of the editor
{
	return(FileDialog(title,fullPath,stringBytes,listElements,false,false,cancel));
}

void FreeOpenFileDialogPaths(char **paths)
// free an array of paths that was returned from OpenFileDialog
{
	FreeFileDialogPaths(paths);
}

bool SaveFileDialog(const char *title,char *fullPath,UINT32 stringBytes,bool *cancel)
// Bring up a file save dialog
// If there is a problem, SetError, return false
// Otherwise, if cancel is true, the user canceled
// Otherwise, return the full path name of the chosen file
// NOTE: if the fullPath (including terminator) would be larger than stringBytes,
// then it will be truncated to fit
// NOTE ALSO: fullPath may be sent to this routine with a pathname, and the save
// dialog will open to that path (as best as it can)
//
// NOTE: this is NOT allowed to change the current directory of the editor
{
	char
		**paths;
	bool
		result;

	result=false;
	if(FileDialog(title,fullPath,stringBytes,&paths,true,false,cancel))
	{
		if(!(*cancel))
		{
			if(stringBytes)
			{
				strncpy(fullPath,paths[0],stringBytes);
				fullPath[stringBytes-1]='\0';
			}
			FreeFileDialogPaths(paths);
		}
		result=true;
	}
	return(result);
}

bool ChoosePathDialog(const char *title,char *fullPath,UINT32 stringBytes,bool *cancel)
// Bring up a path choose dialog
// If there is a problem, SetError, return false
// Otherwise, if cancel is true, the user canceled
// Otherwise, return the full path name of the chosen path
// NOTE: if the fullPath (including terminator) would be larger than stringBytes,
// then it will be truncated to fit
// NOTE ALSO: fullPath may be sent to this routine with a pathname, and the
// dialog will open to that path (as best as it can)
//
// NOTE: this is NOT allowed to change the current directory of the editor
{
	return(FileDialog(title,fullPath,stringBytes,NULL,true,true,cancel));
}

static char **GetSortedFontList(UINT32 *numElements)
// Return a sorted list of fonts, or SetError, and return NULL if there is a problem
// If a list is returned, it must be freed with FreeStringList
{
	int
		actualCount;
	char
		**list,
		**newList;
	UINT32
		index;
	bool
		fail;

	newList=NULL;
	if((list=XListFonts(xDisplay,"*",32767,&actualCount)))
	{
		if((newList=NewStringList()))							// copy the returned list, so it can be sorted (since it is not nice to sort the passed list, because we do not really own it, and should not be able to write to it)
		{
			fail=false;
			*numElements=0;
			for(index=0;!fail&&(int)index<actualCount;index++)
			{
				fail=!AddStringToList(list[index],&newList,numElements);
			}
			if(!fail)
			{
				SortStringList(newList,*numElements);
			}
			else
			{
				FreeStringList(newList);
				newList=NULL;
			}
		}
		XFreeFontNames(list);
	}
	else
	{
		SetError("No Fonts Located");
	}
	return(newList);
}

static UINT32 MatchLength(const char *stringA,const char *stringB)
// Return the number of matching characters between stringA, and stringB
// (the check is case insensitive)
{
	UINT32
		matchIndex;

	matchIndex=0;
	while(stringA[matchIndex]&&(toupper(stringA[matchIndex])==toupper(stringB[matchIndex])))
	{
		matchIndex++;
	}
	return(matchIndex);
}

static UINT32 GetBestMatchIndex(const char **list,UINT32 numElements,const char *font)
// Try to match font against list, return the index of the
// element of list which matches best
{
	UINT32
		currentMatch,
		bestMatch,
		bestMatchPosition;
	UINT32
		index;

	bestMatch=0;
	bestMatchPosition=0;
	for(index=0;index<numElements;index++)
	{
		if((currentMatch=MatchLength(font,list[index]))>bestMatch)
		{
			bestMatch=currentMatch;
			bestMatchPosition=index;
		}
	}
	return(bestMatchPosition);
}

// ----------------------------------------------------------------------------
bool ChooseFontDialog(const char *title,char *font,UINT32 stringBytes,bool *cancel)
// Bring up a font choose dialog
// If there is a problem, SetError, return false
// Otherwise, if cancel is true, the user canceled
// Otherwise, return the name of the chosen font in font
// NOTE: if the font (including terminator) would be larger than stringBytes,
// then it will be truncated to fit
// NOTE ALSO: font may be sent to this routine with a font name, and the
// dialog will open to that font (as best as it can)
{
	char
		**list;
	UINT32
		actualCount;
	bool
		*selectionList,
		done,
		result;
	int
		index;

	if((list=GetSortedFontList(&actualCount)))
	{
		result=false;
		if((selectionList=(bool *)MNewPtrClr(actualCount*sizeof(bool))))
		{
			selectionList[GetBestMatchIndex((const char **)list,actualCount,font)]=true;	// select the item which is closest to the current font
			if((result=SimpleListBoxDialog(title,actualCount,(const char **)list,selectionList,cancel)))
			{
				if(!*cancel)
				{
					done=false;
					for(index=0;!done&&index<(int)actualCount;index++)
					{
						if(selectionList[index])	// find first selected item
						{
							if(stringBytes)
							{
								strncpy(font,list[index],stringBytes);
								font[stringBytes-1]='\0';
							}
							done=true;
						}
					}
					*cancel=!done;				// if none highlighted, take it as a cancel request
				}
			}
			MDisposePtr(selectionList);
		}
		FreeStringList(list);
		return(result);
	}
	return(false);
}
