// Document window handling
// Copyright (C) 2000 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

typedef struct
{
	EDITOR_VIEW
		*view;
	bool
		verticalScrollBarOnLeft;				// gives the preference for vertical scroll bar placement
	SCROLL_BAR
		*verticalScrollBar,
		*horizontalScrollBar;
	EDITOR_FONT
		*statusBarFont;							// info about currently selected status bar font
	EDITOR_COLOR
		backgroundColor,						// color to use when drawing background
		foregroundColor;						// color to use when drawing foreground
} GUI_DOCUMENT_WINDOW_INFO;

static void SetDocumentForegroundColor(EDITOR_WINDOW *window,EDITOR_COLOR foreground)
// set the foreground color used to draw window
{
	GUI_DOCUMENT_WINDOW_INFO
		*windowInfo;

	windowInfo=(GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo;
	windowInfo->foregroundColor=foreground;
}

static void SetDocumentBackgroundColor(EDITOR_WINDOW *window,EDITOR_COLOR background)
// set the background color used to draw window
{
	GUI_DOCUMENT_WINDOW_INFO
		*windowInfo;

	windowInfo=(GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo;
	windowInfo->backgroundColor=background;
}

static void GetDocumentWindowRects(EDITOR_WINDOW *window,EDITOR_RECT *contentRect,EDITOR_RECT *viewRect,EDITOR_RECT *verticalRect,EDITOR_RECT *horizontalRect,EDITOR_RECT *statusRect)
// get the rectangles for all the areas of a document window
{
	XWindowAttributes
		attributes;
	EDITOR_FONT
		*statusBarFont;							// info about currently selected status bar font
	UINT32
		statusHeight;
	GUI_DOCUMENT_WINDOW_INFO
		*windowInfo;

	windowInfo=(GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo;
	XGetWindowAttributes(xDisplay,((WINDOW_LIST_ELEMENT *)window->userData)->xWindow,&attributes);

	contentRect->x=attributes.border_width;
	contentRect->y=attributes.border_width;
	contentRect->w=attributes.width;
	contentRect->h=attributes.height;

	statusRect->x=attributes.border_width;
	statusRect->y=attributes.border_width;
	statusRect->w=attributes.width;

	statusBarFont=((GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo)->statusBarFont;
	statusHeight=statusBarFont->ascent+statusBarFont->descent;

	statusRect->h=statusHeight+STATUS_BAR_LINE_HEIGHT;

	viewRect->x=attributes.border_width;					// move over for the left border
	viewRect->y=statusRect->y+statusRect->h;				// leave room for status bar

	if(windowInfo->verticalScrollBarOnLeft)
	{
		viewRect->x+=SCROLL_BAR_WIDTH;						// if scroll bar on left, then step over it to begin the view
	}

	if(attributes.width>SCROLL_BAR_WIDTH)
	{
		viewRect->w=attributes.width-SCROLL_BAR_WIDTH;		// pull back for scroll bar, and right free area
	}
	else
	{
		viewRect->w=0;
	}
	if((UINT32)attributes.height>(statusHeight+STATUS_BAR_LINE_HEIGHT+SCROLL_BAR_WIDTH))
	{
		viewRect->h=attributes.height-statusHeight-STATUS_BAR_LINE_HEIGHT-SCROLL_BAR_WIDTH;		// pull back for bottom scroll bar, and bottom free area
	}
	else
	{
		viewRect->h=0;										// view cant fit!
	}

	if(windowInfo->verticalScrollBarOnLeft)
	{
		verticalRect->x=attributes.border_width;			// if on left, then place it just after the border
	}
	else
	{
		verticalRect->x=viewRect->x+viewRect->w;			// if on right, place just after the view
	}

	verticalRect->y=viewRect->y;
	verticalRect->w=SCROLL_BAR_WIDTH;
	verticalRect->h=viewRect->h;

	horizontalRect->x=viewRect->x;
	horizontalRect->y=viewRect->y+viewRect->h;
	horizontalRect->w=viewRect->w;
	horizontalRect->h=SCROLL_BAR_WIDTH;
}

static void CreateStatusString(EDITOR_WINDOW *window,char *string)
// create status string for window
// NOTE: string must be large enough to hold the status
{
	EDITOR_BUFFER
		*buffer;
	UINT32
		startPosition,						// selection information
		endPosition,
		startLine,
		endLine,
		startLinePosition,
		endLinePosition,
		totalSegments,
		totalSpan;

	buffer=window->buffer;

	EditorGetSelectionInfo(buffer,buffer->selectionUniverse,&startPosition,&endPosition,&startLine,&endLine,&startLinePosition,&endLinePosition,&totalSegments,&totalSpan);

	sprintf(string,"%s%sLines:%d Chars:%d",
			AtUndoCleanPoint(buffer)?"  ":"* ",
			window->buffer->task?"<Task running> ":"",
			(int)buffer->textUniverse->totalLines,
			(int)buffer->textUniverse->totalBytes
			);

	if(!totalSegments)
	{
		sprintf(&string[strlen(string)]," Cursor(%d:%d %d)",(int)startLine+1,(int)startLinePosition,(int)startPosition);
	}
	else
	{
		sprintf(&string[strlen(string)]," Selections: Number:%d TotalChars:%d Start(%d:%d %d) End(%d:%d %d)",
				(int)totalSegments,
				(int)totalSpan,
				(int)startLine+1,(int)startLinePosition,(int)startPosition,
				(int)endLine+1,(int)endLinePosition,(int)endPosition
				);
	}
}

static void DrawStatusBar(EDITOR_WINDOW *window,Pixmap imageBuffer,Region updateRegion)
// draw the status bar for window
{
	int
		linePosition;
	XRectangle
		regionBox;
	Region
		invalidRegion;
	EDITOR_RECT
		contentRect,
		viewRect,
		verticalRect,
		horizontalRect,
		statusRect;
	EDITOR_FONT
		*statusBarFont;
	int
		stringWidth;
	char
		statusString[256];										// the status string

	invalidRegion=XCreateRegion();

	GetDocumentWindowRects(window,&contentRect,&viewRect,&verticalRect,&horizontalRect,&statusRect);

	regionBox.x=statusRect.x;
	regionBox.y=statusRect.y;
	regionBox.width=statusRect.w;
	regionBox.height=statusRect.h;
	XUnionRectWithRegion(&regionBox,invalidRegion,invalidRegion);	// create a region which is the rectangle of the entire status bar
	XIntersectRegion(invalidRegion,updateRegion,invalidRegion);		// intersect with what is invalid in this window

	if(!XEmptyRegion(invalidRegion))								// see if it has an invalid area
	{
		XSetRegion(xDisplay,xGraphicsContext,invalidRegion);			// set clipping to what's invalid
		XSetFillStyle(xDisplay,xGraphicsContext,FillSolid);
		XClipBox(invalidRegion,&regionBox);							// get bounds of invalid region

		CreateStatusString(window,&statusString[0]);
		statusBarFont=((GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo)->statusBarFont;
		XSetFont(xDisplay,xGraphicsContext,statusBarFont->xFont->fid);	// point to the font

		stringWidth=XTextWidth(statusBarFont->xFont,&statusString[0],strlen(statusString));

		XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(statusBackgroundColor));
		XFillRectangle(xDisplay,imageBuffer,xGraphicsContext,regionBox.x,regionBox.y,regionBox.width,regionBox.height);	// fill with white
		XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(statusForegroundColor));
		XDrawString(xDisplay,imageBuffer,xGraphicsContext,statusRect.x+VIEW_LEFT_MARGIN,statusRect.y+statusBarFont->ascent,&statusString[0],strlen(statusString));

		linePosition=statusRect.y+statusRect.h-STATUS_BAR_LINE_HEIGHT;
		XDrawLine(xDisplay,imageBuffer,xGraphicsContext,statusRect.x,linePosition,statusRect.w,linePosition);
		linePosition+=STATUS_BAR_LINE_HEIGHT-1;
		XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(WHITE));
		XDrawLine(xDisplay,imageBuffer,xGraphicsContext,statusRect.x,linePosition,statusRect.w,linePosition);
		XSetClipMask(xDisplay,xGraphicsContext,None);				// get rid of clip mask
	}
	XDestroyRegion(invalidRegion);									// get rid of invalid region for the view
}

static void DrawGrowArea(EDITOR_WINDOW *window,Pixmap imageBuffer,Region updateRegion)
// draw the grow box area of the window
{
	XRectangle
		regionBox;
	EDITOR_RECT
		contentRect,
		viewRect,
		verticalRect,
		horizontalRect,
		statusRect;

	GetDocumentWindowRects(window,&contentRect,&viewRect,&verticalRect,&horizontalRect,&statusRect);

	regionBox.x=verticalRect.x;
	regionBox.y=horizontalRect.y;
	regionBox.width=verticalRect.w;
	regionBox.height=horizontalRect.h;

	XSetFillStyle(xDisplay,xGraphicsContext,FillSolid);
	XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(0x868693));
	XFillRectangle(xDisplay,imageBuffer,xGraphicsContext,regionBox.x,regionBox.y,regionBox.width,regionBox.height);
}

void AdjustDocumentWindowStatus(EDITOR_WINDOW *window)
// adjust the status of the document window
// this includes updating the status bar, icon, and anything else status related
{
	EDITOR_RECT
		contentRect,
		viewRect,
		verticalRect,
		horizontalRect,
		statusRect;
	XWMHints
		hints;

	GetDocumentWindowRects(window,&contentRect,&viewRect,&verticalRect,&horizontalRect,&statusRect);
	statusRect.h-=STATUS_BAR_LINE_HEIGHT;							// just invalidate the status area
	InvalidateWindowRect(window,&statusRect);

	// adjust the pixmap for the icon based on the window's "clean" status
	hints.flags=IconPixmapHint;
	if(AtUndoCleanPoint(window->buffer))
	{
		hints.icon_pixmap=documentIconPixmap;						// point at document icon
	}
	else
	{
		hints.icon_pixmap=documentModifiedIconPixmap;
	}
	if(hints.icon_pixmap!=((WINDOW_LIST_ELEMENT *)window->userData)->currentIcon)
	{
		((WINDOW_LIST_ELEMENT *)window->userData)->currentIcon=hints.icon_pixmap;	// remember this so we can tell if it changed
		XSetWMHints(xDisplay,((WINDOW_LIST_ELEMENT *)window->userData)->xWindow,&hints);
	}
}

static void AdjustDocumentWindowViewStatus(EDITOR_VIEW *view)
// adjust the status bar of the document window with the given view
{
	AdjustDocumentWindowStatus(view->parentWindow);
}

static void AdjustDocumentWindowScrollThumbs(EDITOR_VIEW *view)
// look at the data of the passed view, and
// adjust the thumbs of its scroll bars so that
// they are correct
{
	UINT32
		scrollBarLeftPixel;

	AdjustScrollBar(((GUI_DOCUMENT_WINDOW_INFO *)view->parentWindow->windowInfo)->verticalScrollBar,view->parentBuffer->textUniverse->totalLines,((GUI_VIEW_INFO *)view->viewInfo)->topLine);
	// NOTE: since views allow for having a left pixel <0, but the scroll bars do not, peg the bar at 0
	if(((GUI_VIEW_INFO *)view->viewInfo)->leftPixel>=0)
	{
		scrollBarLeftPixel=(UINT32)((GUI_VIEW_INFO *)view->viewInfo)->leftPixel;
	}
	else
	{
		scrollBarLeftPixel=0;
	}
	// since the view is allowed to scroll beyond the length of the horizontal scroll bar, pin the bar at the limit
	if(((GUI_VIEW_INFO *)view->viewInfo)->leftPixel>H_SCROLL_MAX_WIDTH)
	{
		scrollBarLeftPixel=H_SCROLL_MAX_WIDTH;
	}
	AdjustScrollBar(((GUI_DOCUMENT_WINDOW_INFO *)view->parentWindow->windowInfo)->horizontalScrollBar,H_SCROLL_MAX_WIDTH,scrollBarLeftPixel);
}

static void AdjustDocumentWindowScrollThumbsAndStatus(EDITOR_VIEW *view)
// adjust the parts of the window that change when the view changes
{
	AdjustDocumentWindowScrollThumbs(view);
	AdjustDocumentWindowViewStatus(view);
}

static void DrawDocumentWindow(EDITOR_WINDOW *window)
// a document window needs to be updated, this
// will draw it
// NOTE: the window's invalidRegion is set to the area which is in need of update
{
	EDITOR_VIEW
		*view;
	WINDOW_LIST_ELEMENT
		*windowElement;
	XWindowAttributes
		xWindowAttributes;
	Window
		xWindow;
	Pixmap
		imageBuffer;
	bool
		bufferedUpdate;			// tells if update is buffered

	view=((GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo)->view;
	windowElement=(WINDOW_LIST_ELEMENT *)view->parentWindow->userData;	// get the x window information associated with this editor window
	xWindow=windowElement->xWindow;

	XGetWindowAttributes(xDisplay,xWindow,&xWindowAttributes);			// get size needed for update pixmap

	imageBuffer=GetWindowUpdatePixmap(xWindowAttributes.width,xWindowAttributes.height);	// go get a pixmap
	bufferedUpdate=true;
	if(imageBuffer==None)		// if we failed to get an offscreen buffer, just update directly
	{
		imageBuffer=xWindow;
		bufferedUpdate=false;
	}

	DrawStatusBar(window,imageBuffer,windowElement->invalidRegion);
	DrawScrollBar(((GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo)->verticalScrollBar,imageBuffer,windowElement->invalidRegion);	// draw the scroll bars
	DrawScrollBar(((GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo)->horizontalScrollBar,imageBuffer,windowElement->invalidRegion);
	DrawGrowArea(window,imageBuffer,windowElement->invalidRegion);
	DrawView(view,imageBuffer,windowElement->invalidRegion);			// draw this view

	if(bufferedUpdate)			// if the update was buffered, the move the data to the screen
	{
		XSetRegion(xDisplay,xGraphicsContext,windowElement->invalidRegion);	// only copy what's needed
		XCopyArea(xDisplay,imageBuffer,xWindow,xGraphicsContext,0,0,xWindowAttributes.width,xWindowAttributes.height,0,0);	// copy the updates
		XSetClipMask(xDisplay,xGraphicsContext,None);						// get rid of clip mask
	}
}

void UpdateDocumentWindow(EDITOR_WINDOW *window)
// make sure the window is up to date
{
	WINDOW_LIST_ELEMENT
		*windowElement;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;				// get the x window information associated with this document window
	if(!XEmptyRegion(windowElement->invalidRegion))						// see if it has an invalid area
	{
		DrawDocumentWindow(window);
		XDestroyRegion(windowElement->invalidRegion);					// clear out the invalid region now that the window has been drawn
		windowElement->invalidRegion=XCreateRegion();
	}
}

void AdjustDocumentWindowForNewSize(EDITOR_WINDOW *window)
// if the window has changed size, then the
// controls will be needing update, so adjust them for the new
// size
{
	EDITOR_RECT
		contentRect,
		viewRect,
		verticalRect,
		horizontalRect,
		statusRect;

	GetDocumentWindowRects(window,&contentRect,&viewRect,&verticalRect,&horizontalRect,&statusRect);

	SetViewBounds(((GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo)->view,&viewRect);		// change the bounds of the view

	RepositionScrollBar(((GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo)->verticalScrollBar,verticalRect.x,verticalRect.y,verticalRect.h,false);
	RepositionScrollBar(((GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo)->horizontalScrollBar,horizontalRect.x,horizontalRect.y,horizontalRect.w,true);
}

void DocumentWindowButtonPress(EDITOR_WINDOW *window,XEvent *event)
// handle button press events for document windows (see what the user is clicking
// on, and attempt to deal with it)
{
	WINDOW_LIST_ELEMENT
		*windowElement;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	if(!TrackView(((GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo)->view,event))
	{
		if(!TrackScrollBar(((GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo)->verticalScrollBar,event))
		{
			TrackScrollBar(((GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo)->horizontalScrollBar,event);
		}
	}
}

void DocumentPeriodicProc(EDITOR_WINDOW *window)
// this is called roughly twice a second to flash cursors, or do whatever in the document
// when the document has focus
{
	ViewCursorTask(((GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo)->view);			// flash the cursor in the view
}

void DocumentWindowEvent(EDITOR_WINDOW *window,XEvent *event)
// handle an event arriving for a document window
{
	WINDOW_LIST_ELEMENT
		*windowElement;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	switch(event->type)
	{
		case ButtonPress:
			XSetInputFocus(xDisplay,windowElement->xWindow,RevertToPointerRoot,event->xbutton.time);	// give it the focus
			DocumentWindowButtonPress(window,event);									// go handle the button press
			break;
		case KeyPress:
			HandleViewKeyEvent(((GUI_DOCUMENT_WINDOW_INFO *)windowElement->editorWindow->windowInfo)->view,event);
			break;
		case ClientMessage:
			if(event->xclient.send_event)
			{
				if(event->xclient.data.l[0]==(int)deleteWindowAtom)
				{
					HandleShellCommand("close",1,&(window->buffer->contentName));		// tell shell to close this
				}
			}
			break;
		default:
			break;
	}
}

void ScrollEditViewEvent(EDITOR_VIEW *view,UINT32 eventCode,bool repeating)
// send a scroll event to the view
{
	VIEW_EVENT
		record;
	EDITOR_KEY
		keyData;

	if(!repeating)
	{
		record.eventType=VET_KEYDOWN;
	}
	else
	{
		record.eventType=VET_KEYREPEAT;
	}
	record.view=view;
	keyData.isVirtual=true;
	keyData.keyCode=eventCode;		// this is the key
	keyData.modifiers=0;			// no modifiers
	record.eventData=&keyData;
	HandleViewEvent(&record);
}

static void ScrollEditViewUpEvent(SCROLL_BAR *scrollBar,bool repeating,void *parameters)
// tell the editor that the view in the window with scrollBar is to be scrolled up
{
	ScrollEditViewEvent((EDITOR_VIEW *)parameters,VVK_SCROLLUP,repeating);
}

static void ScrollEditViewDownEvent(SCROLL_BAR *scrollBar,bool repeating,void *parameters)
// tell the editor that the view in the window with scrollBar is to be scrolled down
{
	ScrollEditViewEvent((EDITOR_VIEW *)parameters,VVK_SCROLLDOWN,repeating);
}

static void ScrollEditViewLeftEvent(SCROLL_BAR *scrollBar,bool repeating,void *parameters)
// tell the editor that the view in the window with scrollBar is to be scrolled left
{
	ScrollEditViewEvent((EDITOR_VIEW *)parameters,VVK_SCROLLLEFT,repeating);
}

static void ScrollEditViewRightEvent(SCROLL_BAR *scrollBar,bool repeating,void *parameters)
// tell the editor that the view in the window with scrollBar is to be scrolled right
{
	ScrollEditViewEvent((EDITOR_VIEW *)parameters,VVK_SCROLLRIGHT,repeating);
}

static void ScrollEditViewPageUpEvent(SCROLL_BAR *scrollBar,bool repeating,void *parameters)
// tell the editor that the view in the window with scrollBar is to be paged up
{
	ScrollEditViewEvent((EDITOR_VIEW *)parameters,VVK_DOCUMENTPAGEUP,repeating);
}

static void ScrollEditViewPageDownEvent(SCROLL_BAR *scrollBar,bool repeating,void *parameters)
// tell the editor that the view in the window with scrollBar is to be paged down
{
	ScrollEditViewEvent((EDITOR_VIEW *)parameters,VVK_DOCUMENTPAGEDOWN,repeating);
}

static void ScrollEditViewPageLeftEvent(SCROLL_BAR *scrollBar,bool repeating,void *parameters)
// tell the editor that the view in the window with scrollBar is to be paged left
{
	ScrollEditViewEvent((EDITOR_VIEW *)parameters,VVK_DOCUMENTPAGELEFT,repeating);
}

static void ScrollEditViewPageRightEvent(SCROLL_BAR *scrollBar,bool repeating,void *parameters)
// tell the editor that the view in the window with scrollBar is to be paged right
{
	ScrollEditViewEvent((EDITOR_VIEW *)parameters,VVK_DOCUMENTPAGERIGHT,repeating);
}

static void ScrollEditViewThumbVerticalEvent(SCROLL_BAR *scrollBar,UINT32 newValue,void *parameters)
// tell the editor that the view in the window with scrollBar is to be positioned vertically
{
	VIEW_EVENT
		record;
	VIEW_POS_EVENT_DATA
		posData;

	record.eventType=VET_POSITIONVERTICAL;
	record.view=(EDITOR_VIEW *)parameters;
	posData.modifiers=0;
	posData.position=newValue;
	record.eventData=&posData;
	HandleViewEvent(&record);
}

static void ScrollEditViewThumbHorizontalEvent(SCROLL_BAR *scrollBar,UINT32 newValue,void *parameters)
// tell the editor that the view in the window with scrollBar is to be positioned horizontally
{
	VIEW_EVENT
		record;
	VIEW_POS_EVENT_DATA
		posData;

	record.eventType=VET_POSITIONHORIZONTAL;
	record.view=(EDITOR_VIEW *)parameters;
	posData.modifiers=0;
	posData.position=newValue;
	record.eventData=&posData;
	HandleViewEvent(&record);
}

static void DisposeDocumentWindow(EDITOR_WINDOW *window)
// close a given window, kill off the view
{
	WINDOW_LIST_ELEMENT
		*windowElement;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;		// get the x window associated with this document window
	DisposeScrollBar(((GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo)->horizontalScrollBar);
	DisposeScrollBar(((GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo)->verticalScrollBar);
	DisposeEditorView(((GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo)->view);	// get rid of the view, unlink it from everywhere
	XDestroyRegion(windowElement->invalidRegion);				// destroy invalid region
	FocusAwayFrom(window);										// move focus off this window
	XDestroyWindow(xDisplay,windowElement->xWindow);			// tell x to make the window go away
	WaitUntilDestroyed(window);
	UnlinkWindowElement(windowElement);							// unlink it from our list
	MDisposePtr(windowElement);
	FreeFont(((GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo)->statusBarFont);
	MDisposePtr(window->windowInfo);
	MDisposePtr(window);
}

static EDITOR_WINDOW *CreateDocumentWindow(EDITOR_BUFFER *buffer,EDITOR_RECT *rect,char *title,char *fontName,UINT32 tabSize,EDITOR_COLOR foreground,EDITOR_COLOR background)
// open a new document window with one EDITOR_VIEW onto buffer
// set the view(s) to use the given defaults
{
	EDITOR_RECT
		contentRect,
		viewRect,
		verticalRect,
		horizontalRect,
		statusRect;
	EDITOR_WINDOW
		*window;
	GUI_DOCUMENT_WINDOW_INFO
		*documentWindowInfo;
	WINDOW_LIST_ELEMENT
		*windowElement;
	SCROLL_BAR_DESCRIPTOR
		scrollDescriptor;
	EDITOR_VIEW_DESCRIPTOR
		viewDescriptor;
	UINT32
		valueMask;
	Atom
		protocols[2];														// number of protocols
	XWMHints
		hints;																// hints to tell window manager how to treat these windows
	XSizeHints
		sizeHints;															// window size info (used to REALLY convince X that I know where I want my own window)
	INT32
		xOffset,
		yOffset;
	UINT32
		screenWidth,
		screenHeight;

	if((window=(EDITOR_WINDOW *)MNewPtr(sizeof(EDITOR_WINDOW))))			// create editor window
	{
		window->buffer=buffer;

		if((documentWindowInfo=(GUI_DOCUMENT_WINDOW_INFO *)MNewPtr(sizeof(GUI_DOCUMENT_WINDOW_INFO))))	// make structure to contain document window info
		{
			if((documentWindowInfo->statusBarFont=LoadFont(defaultStatusBarFont)))			// load in the font
			{
				if((windowElement=(WINDOW_LIST_ELEMENT *)MNewPtr(sizeof(WINDOW_LIST_ELEMENT))))	// create a window list element
				{
					LinkWindowElement(windowElement);						// link it to the global window list that we maintain
					windowElement->editorWindow=window;						// point to our high level concept of window
					window->windowType=EWT_DOCUMENT;						// set the editor window type to "document"
					window->userData=windowElement;							// point back to the window element
					window->windowInfo=documentWindowInfo;					// point to the edit info

					SetDocumentForegroundColor(window,foreground);			// set the colors for this window
					SetDocumentBackgroundColor(window,background);

					valueMask=0;

					GetEditorScreenOffsetAndDimensions(&xOffset,&yOffset,&screenWidth,&screenHeight);
					// Due to X's nasty error handling, we assume this create does not fail!!!
					windowElement->realTopWindow=windowElement->xWindow=XCreateWindow(xDisplay,RootWindow(xDisplay,xScreenNum),rect->x+xOffset,rect->y+yOffset,rect->w,rect->h,XWINDOW_BORDER_WIDTH,CopyFromParent,InputOutput,DefaultVisual(xDisplay,xScreenNum),valueMask,NULL);

					sizeHints.flags=USSize|USPosition|PMinSize;
					sizeHints.x=rect->x;
					sizeHints.y=rect->y;
					sizeHints.width=rect->w;
					sizeHints.height=rect->h;
					sizeHints.min_width=1;
					sizeHints.min_height=1;
					XSetWMNormalHints(xDisplay,windowElement->xWindow,&sizeHints);

					XStoreName(xDisplay,windowElement->xWindow,title);
					XSetIconName(xDisplay,windowElement->xWindow,title);
					XSelectInput(xDisplay,windowElement->xWindow,ExposureMask|ButtonPressMask|ButtonReleaseMask|FocusChangeMask|KeyPressMask|StructureNotifyMask);	// want to see these events

					protocols[0]=takeFocusAtom;
					protocols[1]=deleteWindowAtom;
					XSetWMProtocols(xDisplay,windowElement->xWindow,&(protocols[0]),2);	// tell X we would like to participate in these protocols

					hints.flags=InputHint|StateHint|WindowGroupHint;
					hints.input=True;
					hints.initial_state=NormalState;
					hints.window_group=((WINDOW_LIST_ELEMENT *)rootMenuWindow->userData)->xWindow;	// all documents belong to the group headed by the main menu window
					XSetWMHints(xDisplay,windowElement->xWindow,&hints);

					SetClassHints(window);

					windowElement->invalidRegion=XCreateRegion();				// create new empty invalid region

					windowElement->currentIcon=(Pixmap)NULL;					// no icon associated with this yet

					documentWindowInfo->verticalScrollBarOnLeft=verticalScrollBarOnLeft;	// copy global preference
					GetDocumentWindowRects(window,&contentRect,&viewRect,&verticalRect,&horizontalRect,&statusRect);	// get sizes of areas within the window

					viewDescriptor.buffer=buffer;
					viewDescriptor.rect=viewRect;								// place where the view is going
					viewDescriptor.topLine=0;
					viewDescriptor.leftPixel=0;
					viewDescriptor.tabSize=tabSize;

					viewDescriptor.backgroundColor=background;
					viewDescriptor.foregroundColor=foreground;

					viewDescriptor.fontName=fontName;
					viewDescriptor.active=true;												// create the view active
					viewDescriptor.viewTextChangedVector=AdjustDocumentWindowScrollThumbsAndStatus;		// fix the thumbs, and status when the view text changes
					viewDescriptor.viewPositionChangedVector=AdjustDocumentWindowScrollThumbs;			// fix the thumbs when the view position changes
					viewDescriptor.viewSelectionChangedVector=AdjustDocumentWindowViewStatus;			// fix the status when the selection changes
					if((documentWindowInfo->view=CreateEditorView(window,&viewDescriptor)))	// create a view for this area, set up its font stuff
					{
						scrollDescriptor.x=verticalRect.x;									// set up description of the scroll bar
						scrollDescriptor.y=verticalRect.y;
						scrollDescriptor.length=verticalRect.h;
						scrollDescriptor.horizontal=false;
						scrollDescriptor.thumbMax=0;
						scrollDescriptor.thumbPosition=0;
						scrollDescriptor.stepLowerProc=ScrollEditViewUpEvent;
						scrollDescriptor.stepHigherProc=ScrollEditViewDownEvent;
						scrollDescriptor.pageLowerProc=ScrollEditViewPageUpEvent;
						scrollDescriptor.pageHigherProc=ScrollEditViewPageDownEvent;
						scrollDescriptor.thumbProc=ScrollEditViewThumbVerticalEvent;
						scrollDescriptor.procParameters=documentWindowInfo->view;

						if((documentWindowInfo->verticalScrollBar=CreateScrollBar(window,&scrollDescriptor)))
						{
							scrollDescriptor.x=horizontalRect.x;
							scrollDescriptor.y=horizontalRect.y;
							scrollDescriptor.length=horizontalRect.w;
							scrollDescriptor.horizontal=true;
							scrollDescriptor.thumbMax=0;
							scrollDescriptor.thumbPosition=0;
							scrollDescriptor.stepLowerProc=ScrollEditViewLeftEvent;
							scrollDescriptor.stepHigherProc=ScrollEditViewRightEvent;
							scrollDescriptor.pageLowerProc=ScrollEditViewPageLeftEvent;
							scrollDescriptor.pageHigherProc=ScrollEditViewPageRightEvent;
							scrollDescriptor.thumbProc=ScrollEditViewThumbHorizontalEvent;
							scrollDescriptor.procParameters=documentWindowInfo->view;

							if((documentWindowInfo->horizontalScrollBar=CreateScrollBar(window,&scrollDescriptor)))
							{
								AdjustDocumentWindowScrollThumbsAndStatus(documentWindowInfo->view);	// force the thumbs and status to be up to date
								XMapRaised(xDisplay,windowElement->xWindow);							// map window to the display
								WaitUntilMapped(window);
								return(window);
							}
							DisposeScrollBar(documentWindowInfo->verticalScrollBar);
						}
						DisposeEditorView(documentWindowInfo->view);
					}
					XDestroyRegion(windowElement->invalidRegion);					// destroy invalid region
					XDestroyWindow(xDisplay,windowElement->xWindow);				// get rid of the window
					WaitUntilDestroyed(window);

					UnlinkWindowElement(windowElement);
					MDisposePtr(windowElement);
				}
				FreeFont(documentWindowInfo->statusBarFont);
			}
			MDisposePtr(documentWindowInfo);
		}
		MDisposePtr(window);
	}
	return((EDITOR_WINDOW *)NULL);
}

// routines callable by the editor

void MinimizeDocumentWindow(EDITOR_WINDOW *window)
// minimize the passed window
{
	MinimizeEditorWindow(window);
}

void UnminimizeDocumentWindow(EDITOR_WINDOW *window)
// unminimize the passed window
{
	UnminimizeEditorWindow(window);
}

void SetTopDocumentWindow(EDITOR_WINDOW *window)
// set the passed document window to the top of the display
{
	SetTopEditorWindow(window);
}

EDITOR_WINDOW *GetActiveDocumentWindow()
// return the document window that has the user's attention, or NULL if there is none
// depending on the system, this may or may not return the top-most document window
// (For instance under Motif, the active document window may be the one with the
// cursor over it)
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	Window
		xWindow;
	int
		focusRevert;

	XGetInputFocus(xDisplay,&xWindow,&focusRevert);			// find out who has the focus
	if((windowElement=LocateWindowElement(xWindow)))		// see if it is one of our windows
	{
		if(windowElement->editorWindow->windowType==EWT_DOCUMENT)
		{
			return(windowElement->editorWindow);			// if the window with focus is a document window, then return it
		}
	}
	return(GetTopEditorWindowType(EWT_DOCUMENT));			// otherwise, the active one is the top one
}

EDITOR_WINDOW *GetTopDocumentWindow()
// return the top-most document window, or NULL if there is none
// this indirectly uses the realTopWindow fields of the window elements
{
	return(GetTopEditorWindowType(EWT_DOCUMENT));
}

bool GetSortedDocumentWindowList(UINT32 *numElements,EDITOR_WINDOW ***list)
// return a list of all document windows -- sorted top to bottom
// if there is a problem, SetError, and return false
// otherwise return true. If true is returned, list must be
// MDisposePtr'ed by the caller (even if numElements is 0)
{
	return(GetSortedEditorWindowTypeList(EWT_DOCUMENT,numElements,list));
}

EDITOR_VIEW *GetDocumentWindowCurrentView(EDITOR_WINDOW *window)
// passed a document window, return the current view of the window
{
	return(((GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo)->view);			// currently, document windows contain one view only
}

bool GetDocumentWindowTitle(EDITOR_WINDOW *window,char *title,UINT32 stringBytes)
// copy window's title into title
// if window's title is longer than stringBytes, return false
{
	char
		*name;

	if(XFetchName(xDisplay,((WINDOW_LIST_ELEMENT *)window->userData)->xWindow,&name))
	{
		strncpy(title,name,stringBytes);
		XFree(name);
	}
	else
	{
		strncpy(title,"no name",stringBytes);
	}
	if(stringBytes)
	{
		title[stringBytes-1]='\0';
	}
	return(true);
}

bool SetDocumentWindowTitle(EDITOR_WINDOW *window,const char *title)
// replace title of window
// if there is a problem, return false
{
	XStoreName(xDisplay,((WINDOW_LIST_ELEMENT *)window->userData)->xWindow,title);
	XSetIconName(xDisplay,((WINDOW_LIST_ELEMENT *)window->userData)->xWindow,title);
	return(true);
}

bool GetEditorDocumentWindowForegroundColor(EDITOR_WINDOW *window,EDITOR_COLOR *foregroundColor)
// get the foreground color in use by window
// if there is a problem, SetError, and return false
{
	GUI_DOCUMENT_WINDOW_INFO
		*windowInfo;

	windowInfo=(GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo;
	*foregroundColor=windowInfo->foregroundColor;
	return(true);
}

bool SetEditorDocumentWindowForegroundColor(EDITOR_WINDOW *window,EDITOR_COLOR foregroundColor)
// set the foreground color to be used in window
// if there is a problem, leave the color unchanged, set the error
// and return false
{
	SetDocumentForegroundColor(window,foregroundColor);
	XSetWindowBorder(xDisplay,((WINDOW_LIST_ELEMENT *)window->userData)->xWindow,((GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo)->foregroundColor);
	return(true);
}

bool GetEditorDocumentWindowBackgroundColor(EDITOR_WINDOW *window,EDITOR_COLOR *backgroundColor)
// get the background color in use by window
// if there is a problem, SetError, and return false
{
	GUI_DOCUMENT_WINDOW_INFO
		*windowInfo;

	windowInfo=(GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo;
	*backgroundColor=windowInfo->backgroundColor;
	return(true);
}

bool SetEditorDocumentWindowBackgroundColor(EDITOR_WINDOW *window,EDITOR_COLOR backgroundColor)
// set the background color to be used in window
// if there is a problem, leave the color unchanged, set the error
// and return false
{
	EDITOR_RECT
		contentRect,
		viewRect,
		verticalRect,
		horizontalRect,
		statusRect;

	SetDocumentBackgroundColor(window,backgroundColor);
	GetDocumentWindowRects(window,&contentRect,&viewRect,&verticalRect,&horizontalRect,&statusRect);
	InvalidateWindowRect(window,&contentRect);			// invalidate the whole thing, because the color has changed
	return(true);
}

bool GetEditorDocumentWindowRect(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// get the rectangle of window
// if there is a problem, SetError, and return false
// NOTE: once again, X windows makes this much more of a pain than it needs
// to be....
{
	XWindowAttributes
		attributes;
	Window
		childWindow;
	INT32
		xOffset,
		yOffset;
	UINT32
		screenWidth,
		screenHeight;

	XGetWindowAttributes(xDisplay,((WINDOW_LIST_ELEMENT *)window->userData)->xWindow,&attributes);
	rect->w=attributes.width;
	rect->h=attributes.height;

	// Convert our window coordinates to screen coordinates usable by XMoveResizeWindow()
	XTranslateCoordinates(xDisplay,((WINDOW_LIST_ELEMENT *)window->userData)->xWindow,RootWindow(xDisplay,xScreenNum),attributes.x,attributes.y,&rect->x,&rect->y,&childWindow);

	GetEditorScreenOffsetAndDimensions(&xOffset,&yOffset,&screenWidth,&screenHeight);
	rect->x-=xOffset;
	rect->y-=yOffset;

	return(true);
}

bool SetEditorDocumentWindowRect(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// set window to rect
// if there is a problem, leave rect unchanged, set the error
// and return false
{
	INT32
		xOffset,
		yOffset;
	UINT32
		screenWidth,
		screenHeight;

	GetEditorScreenOffsetAndDimensions(&xOffset,&yOffset,&screenWidth,&screenHeight);
	XMoveResizeWindow(xDisplay,((WINDOW_LIST_ELEMENT *)window->userData)->xWindow,rect->x+xOffset,rect->y+yOffset,rect->w,rect->h);
	return(true);
}

EDITOR_WINDOW *OpenDocumentWindow(EDITOR_BUFFER *buffer,EDITOR_RECT *rect,char *title,char *fontName,UINT32 tabSize,EDITOR_COLOR foreground,EDITOR_COLOR background)
// open a document window onto buffer
// if there is a problem, undo anything that was done,
// set the error, then return NULL
{
	return(CreateDocumentWindow(buffer,rect,title,fontName,tabSize,foreground,background));
}

void CloseDocumentWindow(EDITOR_WINDOW *window)
// close an open document window
{
	DisposeDocumentWindow(window);
}
