// Font handling
// Copyright (C) 2000 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

// fallback fonts, (if we dont get the requested font, try these in succession)
#define	FALLBACK_FONT_0				"-*-lucidatypewriter-medium-r-normal-sans-12-*-*-*-*-*-*-*"
#define	FALLBACK_FONT_1				"fixed"
#define	FALLBACK_FONT_2				"-adobe-courier-medium-r-normal--12-120-75-75-*-70-*-*"
#define	FALLBACK_FONT_3				"7x13"

static EDITOR_FONT
	*topFont;			// holds top font in linked list of gui fonts

static EDITOR_FONT *LocateLoadedFont(const char *name)
// see if an already loaded font can be located which has name
// if so, return a pointer to it, otherwise return NULL
// ### this should use a hash table
{
	EDITOR_FONT
		*testFont;

	testFont=topFont;
	while(testFont&&(strcasecmp(testFont->name,name)!=0))
	{
		testFont=testFont->next;
	}
	return(testFont);
}

static XFontStruct *GetClosestFont(const char *name)
// return the X font which is the best we could find
{
	XFontStruct
		*font;

	if((font=XLoadQueryFont(xDisplay,name)))
	{
		return(font);
	}
	if((font=XLoadQueryFont(xDisplay,FALLBACK_FONT_0)))
	{
		return(font);
	}
	if((font=XLoadQueryFont(xDisplay,FALLBACK_FONT_1)))
	{
		return(font);
	}
	if((font=XLoadQueryFont(xDisplay,FALLBACK_FONT_2)))
	{
		return(font);
	}
	if((font=XLoadQueryFont(xDisplay,FALLBACK_FONT_3)))
	{
		return(font);
	}
	SetError("Could not locate font '%s'",name);
	return(NULL);
}

void GetFontName(EDITOR_FONT *font,char *nameBuffer,UINT32 bufferLength)
// Copy the name of the passed font into nameBuffer which is
// bufferLength bytes long (including the 0 termination)
{
	if(bufferLength)
	{
		strncpy(nameBuffer,font->name,bufferLength);
		nameBuffer[bufferLength-1]='\0';
	}
}

void FreeFont(EDITOR_FONT *font)
// Free a font allocated by LoadFont
{
	if(!(--font->usageCount))
	{
		XFreeFont(xDisplay,font->xFont);
		if(font->previous)
		{
			font->previous->next=font->next;
		}
		else
		{
			topFont=font->next;
		}
		if(font->next)
		{
			font->next->previous=font->previous;
		}
		MDisposePtr(font);
	}
}

EDITOR_FONT *LoadFont(const char *name)
// attempt to load the font given by name
// if there is a problem, SetError, return NULL
{
	EDITOR_FONT
		*result;
	int
		overhang;
	int
		i,j;

	if((result=LocateLoadedFont(name)))		// see if we can find this one already loaded
	{
		result->usageCount++;				// add another user to this font structure
		return(result);
	}
	else
	{
		if((result=(EDITOR_FONT *)MNewPtr(sizeof(EDITOR_FONT))))	// allocate one
		{
			if((result->xFont=GetClosestFont(name)))
			{
				strncpy(result->name,name,sizeof(result->name));
				result->name[sizeof(result->name)-1]='\0';	// terminate string if copy was too long

				result->usageCount=1;

				result->previous=NULL;		// link to font list
				if((result->next=topFont))
				{
					topFont->previous=result;
				}
				topFont=result;

				result->ascent=result->xFont->ascent;
				result->descent=result->xFont->descent;

				result->maxLeftOverhang=0;	// initialize the overhang amounts
				result->maxRightOverhang=0;

				for(i=0;i<256;i++)
				{
					result->charWidths[i]=0;	// clear widths for non-existent characters
				}
				if(result->xFont->per_char)		// see if there is information for individual characters
				{
					j=0;
					for(i=result->xFont->min_char_or_byte2;i<=(int)result->xFont->max_char_or_byte2;i++)
					{
						// handle left overhang
						overhang=-result->xFont->per_char[j].lbearing;
						if(overhang>(INT32)result->maxLeftOverhang)
						{
							result->maxLeftOverhang=(UINT32)overhang;
						}

						// handle right overhang
						overhang=result->xFont->per_char[j].rbearing-result->xFont->per_char[j].width;
						if(overhang>(INT32)result->maxRightOverhang)
						{
							result->maxRightOverhang=(UINT32)overhang;
						}

						result->charWidths[i]=result->xFont->per_char[j].width;	// set widths for existing characters
						j++;
					}
				}
				else
				{
					for(i=result->xFont->min_char_or_byte2;i<=(int)result->xFont->max_char_or_byte2;i++)
					{
						result->charWidths[i]=result->xFont->max_bounds.width;	// set widths for existing characters
					}
				}
				return(result);
			}
			MDisposePtr(result);
		}
	}
	return(NULL);
}

void UnInitFonts()
// Undo what InitFonts did
{
}

bool InitFonts()
// set up font handling stuff
{
	topFont=NULL;					// no top font loaded in yet
	memset(defaultViewFont,0,sizeof(defaultViewFont));
	memset(defaultMenuFont,0,sizeof(defaultMenuFont));
	memset(defaultDialogButtonFont,0,sizeof(defaultDialogButtonFont));
	memset(defaultDialogTextFont,0,sizeof(defaultDialogTextFont));
	memset(defaultStatusBarFont,0,sizeof(defaultStatusBarFont));
	return(true);
}
