// List box handling
// Copyright (C) 2000 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

#define	LIST_BOX_TOP_OFFSET		4			// pixel offsets within the list box that keep the text of the items from drawing too close to the edge
#define	LIST_BOX_LEFT_OFFSET	4
#define	LIST_BOX_BOTTOM_OFFSET	4

#define	ITEM_SPACE_TOP			2			// pixels to add to each line of text
#define	ITEM_SPACE_BOTTOM		2

#define	SHADOW_COLOR_UL			0x314F5E	// shadow colors (upper left)
#define	SHADOW_COLOR_LR			0x91AFBE	// (lower right)

#define	MAX_LIST_CHAR_BUFFER	64			// max characters in the character buffer
#define	LIST_CHAR_TIMEOUT		40			// max number of ticks between consecutive typed characters
#define	LIST_BOX_BORDER_WIDTH	3			// width of border that surrounds the list box

typedef struct listBox
{
	EDITOR_WINDOW
		*parentWindow;					// the window that contains this list
	listBoxResizeFunction
		*resizeFunction;				// function which is used to resize for this window
	EDITOR_RECT
		rect;							// position within the parent
	EDITOR_COLOR
		focusBackgroundColor,			// color to use when drawing background during focus
		focusForegroundColor,			// color to use when drawing foreground during focus
		nofocusBackgroundColor,			// color to use when drawing background during non focus
		nofocusForegroundColor;			// color to use when drawing foreground during non focus
	bool
		verticalScrollBarOnLeft;		// gives the preference for vertical scroll bar placement
	SCROLL_BAR
		*scrollBar;						// pointer to the scrollbar attached to this list
	UINT32
		numElements;					// tells how many elements are in the list
	const char
		**listElements;					// pointer to NULL terminated array of pointers to items in the list
	bool
		*selectedElements;				// pointer to array of boolS which tell which items are selected
	UINT32
		currentElement;					// this is the element last operated on
	UINT32
		numLines;						// tells how many lines in the current font, and list box size can be displayed
	UINT32
		topLine;						// tells which line is displayed at the top of the list box
	UINT32
		lineHeight;						// height of a line of text
	UINT32
		charBufferLength;				// number of characters in the charBuffer
	char
		charBuffer[MAX_LIST_CHAR_BUFFER];	// keeps track of the characters typed
	UINT32
		charTimer;						// keeps track of the time between characters when hunting for an item (if the time is too long between characters, charBuffer is flushed)
	EDITOR_FONT
		*font;							// info about currently selected font
	void
		(*pressedProc)(void *parameters);	// routine called when the list items have been chosen (by double clicking, or hitting return in the list)
	void
		*pressedProcParameters;			// data passed to pressedProc routine
} LIST_BOX;


static void InvalidateListBox(LIST_BOX *listBox)
// invalidate the list box within its parent window
{
	InvalidateWindowRect(listBox->parentWindow,&listBox->rect);
}

static void GetListBoxControlRects(LIST_BOX *listBox,EDITOR_RECT *borderRect,EDITOR_RECT *listContentRect,EDITOR_RECT *scrollBarRect)
// return the rectangles that enclose the various pieces of listBox
// if the list box starts getting too small, certain rects may be returned with 0 width or
// height
{
	EDITOR_RECT
		insetRect;

	insetRect=(*borderRect)=listBox->rect;				// border rect is the whole thing

	insetRect.x+=LIST_BOX_BORDER_WIDTH;					// inset from the border to the content area of the list box
	insetRect.y+=LIST_BOX_BORDER_WIDTH;
	if(insetRect.w>2*LIST_BOX_BORDER_WIDTH)
	{
		insetRect.w-=2*LIST_BOX_BORDER_WIDTH;
	}
	else
	{
		insetRect.w=0;
	}
	if(insetRect.h>2*LIST_BOX_BORDER_WIDTH)
	{
		insetRect.h-=2*LIST_BOX_BORDER_WIDTH;
	}
	else
	{
		insetRect.h=0;
	}
	listContentRect->x=insetRect.x;
	listContentRect->y=insetRect.y;
	if(insetRect.w>SCROLL_BAR_WIDTH)
	{
		listContentRect->w=insetRect.w-SCROLL_BAR_WIDTH;
		if(listBox->verticalScrollBarOnLeft)
		{
			scrollBarRect->x=insetRect.x;
			listContentRect->x+=SCROLL_BAR_WIDTH;
		}
		else
		{
			scrollBarRect->x=insetRect.x+insetRect.w-SCROLL_BAR_WIDTH;
		}
		scrollBarRect->w=SCROLL_BAR_WIDTH;
	}
	else
	{
		listContentRect->w=0;
		scrollBarRect->x=insetRect.x;
		scrollBarRect->w=insetRect.w;
	}
	listContentRect->h=insetRect.h;

	scrollBarRect->y=insetRect.y;
	scrollBarRect->h=insetRect.h;
}

static void InvalidateListContents(LIST_BOX *listBox)
// invalidate the list area of listBox
{
	EDITOR_RECT
		borderRect,
		listContentRect,
		scrollBarRect;

	GetListBoxControlRects(listBox,&borderRect,&listContentRect,&scrollBarRect);
	InvalidateWindowRect(listBox->parentWindow,&listContentRect);
}

static void GetListBoxElementRect(LIST_BOX *listBox,UINT32 element,EDITOR_RECT *rect)
// get back the rectangle for element in listBox
// NOTE: if element is invalid, or element is not displayed,
// an empty rectangle is returned
{
	EDITOR_RECT
		borderRect,
		listContentRect,
		scrollBarRect;

	if(element>=listBox->topLine&&element<listBox->numElements&&element<listBox->topLine+listBox->numLines)
	{
		GetListBoxControlRects(listBox,&borderRect,&listContentRect,&scrollBarRect);

		rect->x=listContentRect.x;
		rect->y=listContentRect.y+LIST_BOX_TOP_OFFSET+(listBox->lineHeight)*(element-listBox->topLine);
		rect->w=listContentRect.w;
		rect->h=listBox->lineHeight;
	}
	else
	{
		rect->x=rect->y=0;
		rect->w=rect->h=0;
	}
}

static void RecalculateListBoxElementInfo(LIST_BOX *listBox)
// calculate some often needed constants
{
	EDITOR_RECT
		borderRect,
		listContentRect,
		scrollBarRect;

	GetListBoxControlRects(listBox,&borderRect,&listContentRect,&scrollBarRect);

	listBox->lineHeight=ITEM_SPACE_TOP+listBox->font->ascent+listBox->font->descent+ITEM_SPACE_BOTTOM;

// find out how many lines can be displayed in the box given the size, and font
	if(listContentRect.h>(LIST_BOX_TOP_OFFSET+LIST_BOX_BOTTOM_OFFSET))
	{
		listBox->numLines=(listContentRect.h-LIST_BOX_TOP_OFFSET-LIST_BOX_BOTTOM_OFFSET)/listBox->lineHeight;
	}
	else
	{
		listBox->numLines=0;
	}
}

static void AdjustListBoxTopLine(LIST_BOX *listBox,UINT32 newTopLine)
// Set the top line of listBox to the given value (or the closest value that
// is legal)
// update the scroll bar, and all that...
{
	UINT32
		maxLines;

	if(newTopLine+listBox->numLines<listBox->numElements)
	{
		maxLines=listBox->numElements-listBox->numLines;
	}
	else
	{
		if(listBox->numElements>listBox->numLines)
		{
			newTopLine=listBox->numElements-listBox->numLines;
			maxLines=listBox->numElements-listBox->numLines;
		}
		else
		{
			newTopLine=0;
			maxLines=0;
		}
	}
	listBox->topLine=newTopLine;
	AdjustScrollBar(listBox->scrollBar,maxLines,newTopLine);
}

static void SetListBoxRectangle(LIST_BOX *listBox)
// Work out the rectangle given the current window
{
	EDITOR_RECT
		borderRect,
		listContentRect,
		scrollBarRect;

	if(listBox->resizeFunction)	// see if non-null resize function specified
	{
		listBox->resizeFunction(listBox->parentWindow,&listBox->rect);	// get new rectangle
		RecalculateListBoxElementInfo(listBox);							// recalculate some stuff
		GetListBoxControlRects(listBox,&borderRect,&listContentRect,&scrollBarRect);
		RepositionScrollBar(listBox->scrollBar,scrollBarRect.x,scrollBarRect.y,scrollBarRect.h,false);	// move scroll bar
		AdjustListBoxTopLine(listBox,listBox->topLine);
		InvalidateListContents(listBox);				// invalidate contents
	}
}

static void SetListBoxTopLine(LIST_BOX *listBox,UINT32 newTopLine)
// Set the top line of listBox to the given value (or the closest value that
// is legal)
// update the scroll bar, and all that...
{
	AdjustListBoxTopLine(listBox,newTopLine);
	InvalidateListContents(listBox);				// invalidate contents
	UpdateEditorWindows();							// redraw anything that is needed
}

static void HomeListBoxToCurrentElement(LIST_BOX *listBox)
// make sure the current element is visible within listBox
{
	if(listBox->currentElement<listBox->topLine)
	{
		SetListBoxTopLine(listBox,listBox->currentElement);
	}
	else
	{
		if(listBox->currentElement>=listBox->topLine+listBox->numLines)
		{
			SetListBoxTopLine(listBox,listBox->currentElement-listBox->numLines+1);
		}
	}
}

static void CenterListBoxToCurrentElement(LIST_BOX *listBox)
// Place current element at the center of the list box
{
	UINT32
		centerElement;

	centerElement=0;
	if(listBox->currentElement>listBox->numLines/2)
	{
		centerElement=listBox->currentElement-listBox->numLines/2;
		if(centerElement>listBox->numElements-listBox->numLines+1)
		{
			centerElement=listBox->numElements-listBox->numLines+1;
		}
	}
	SetListBoxTopLine(listBox,centerElement);
}

static void ClearListBoxSelections(LIST_BOX *listBox)
// run through the elements of listBox, clearing the
// selections, invalidate any items that were changed
{
	UINT32
		currentElement;
	EDITOR_RECT
		itemRect;

	for(currentElement=0;currentElement<listBox->numElements;currentElement++)
	{
		if(listBox->selectedElements[currentElement])	// unselect any that were selected
		{
			listBox->selectedElements[currentElement]=false;
			GetListBoxElementRect(listBox,currentElement,&itemRect);
			InvalidateWindowRect(listBox->parentWindow,&itemRect);
		}
	}
}

static void SetListBoxSelections(LIST_BOX *listBox)
// run through the elements of listBox, setting the
// selections, invalidate any items that were changed
{
	UINT32
		currentElement;
	EDITOR_RECT
		itemRect;

	for(currentElement=0;currentElement<listBox->numElements;currentElement++)
	{
		if(!listBox->selectedElements[currentElement])	// select any that were unselected
		{
			listBox->selectedElements[currentElement]=true;
			GetListBoxElementRect(listBox,currentElement,&itemRect);
			InvalidateWindowRect(listBox->parentWindow,&itemRect);
		}
	}
}

static bool GetPointedToElement(LIST_BOX *listBox,INT32 x,INT32 y,UINT32 *element,bool strict)
// return element as the index to the item of listBox that
// falls under x,y in the parent window
// NOTE: if strict is true, and no item falls under the point, return false
// if strict is false, and no item falls under the point, return the item
// nearest the point (if the point is below all visible items, return
// the item just below the last visible item, if the point is above
// all visible items, return the item just above the first visible
// item.) The non-strict mode is used while scrolling.
{
	EDITOR_RECT
		borderRect,
		listContentRect,
		scrollBarRect;
	UINT32
		localElement;

	if(listBox->numElements)
	{
		GetListBoxControlRects(listBox,&borderRect,&listContentRect,&scrollBarRect);
		if(PointInRECT(x,y,&listContentRect)||((!strict)&&(y>=listContentRect.y)&&(y<listContentRect.y+(int)listContentRect.h)))
		{
			if(y>=listContentRect.y+LIST_BOX_TOP_OFFSET)
			{
				localElement=(y-listContentRect.y-LIST_BOX_TOP_OFFSET)/listBox->lineHeight;
			}
			else
			{
				localElement=0;
			}
			if(localElement>=listBox->numLines)
			{
				if(listBox->numLines)
				{
					localElement=listBox->numLines-1;
				}
				else
				{
					return(false);		// no lines displayed in list!
				}
			}
			if(localElement+listBox->topLine<listBox->numElements)
			{
				*element=localElement+listBox->topLine;
			}
			else
			{
				*element=listBox->numElements-1;
			}
			return(true);
		}
		else
		{
			if(!strict)
			{
				if(y<listContentRect.y)
				{
					if(listBox->topLine)
					{
						*element=listBox->topLine-1;
					}
					else
					{
						*element=0;
					}
				}
				else
				{
					if(listBox->topLine+listBox->numLines<listBox->numElements)
					{
						*element=listBox->topLine+listBox->numLines;
					}
					else
					{
						*element=listBox->numElements-1;
					}
				}
				return(true);
			}
		}
	}
	return(false);
}

static bool PointInListBox(LIST_BOX *listBox,INT32 x,INT32 y)
// return true if parent window relative x/y are in listBox
// false if not
{
	return(PointInRECT(x,y,&listBox->rect));
}

static void DrawListBox(LIST_BOX *listBox,Pixmap imageBuffer,Region updateRegion,bool hasFocus)
// redraw listBox in its parent window, obey the invalid area of the parent's window
{
	EDITOR_COLOR
		foreground,
		background;
	XRectangle
		regionBox;
	Region
		invalidRegion;
	EDITOR_RECT
		borderRect,
		listContentRect,
		scrollBarRect,
		itemRect;
	UINT32
		activeHeight,
		currentElement;

	GetListBoxControlRects(listBox,&borderRect,&listContentRect,&scrollBarRect);

	invalidRegion=XCreateRegion();

	regionBox.x=borderRect.x;
	regionBox.y=borderRect.y;
	regionBox.width=borderRect.w;
	regionBox.height=borderRect.h;
	XUnionRectWithRegion(&regionBox,invalidRegion,invalidRegion);			// create a region which is the rectangle of the list box
	XIntersectRegion(invalidRegion,updateRegion,invalidRegion);				// intersect with what is invalid in this window

	if(!XEmptyRegion(invalidRegion))										// see if it has an invalid area
	{
		XSetRegion(xDisplay,xGraphicsContext,invalidRegion);				// set clipping to what's invalid

		if(!hasFocus)
		{
			OutlineShadowRectangle(imageBuffer,&borderRect,SHADOW_COLOR_UL,SHADOW_COLOR_LR,LIST_BOX_BORDER_WIDTH);	// drop in border
		}
		else
		{
			OutlineShadowRectangle(imageBuffer,&borderRect,BLACK,GRAY_1,LIST_BOX_BORDER_WIDTH);	// drop in border
		}

		XSubtractRegion(invalidRegion,invalidRegion,invalidRegion);			// clear the region
		regionBox.x=listContentRect.x;
		regionBox.y=listContentRect.y;
		regionBox.width=listContentRect.w;
		regionBox.height=listContentRect.h;
		XUnionRectWithRegion(&regionBox,invalidRegion,invalidRegion);		// create a region which is the rectangle of the list box
		XIntersectRegion(invalidRegion,updateRegion,invalidRegion);			// intersect with what is invalid in this window

		if(!XEmptyRegion(invalidRegion))									// see if it has an invalid area
		{
			XSetRegion(xDisplay,xGraphicsContext,invalidRegion);			// set clipping to what's invalid

			if(hasFocus)
			{
				foreground=listBox->focusForegroundColor;
				background=listBox->focusBackgroundColor;
			}
			else
			{
				foreground=listBox->nofocusForegroundColor;
				background=listBox->nofocusBackgroundColor;
			}

			XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(background));

			// fill space at top and bottom where no text draws
			XFillRectangle(xDisplay,imageBuffer,xGraphicsContext,listContentRect.x,listContentRect.y,listContentRect.w,LIST_BOX_TOP_OFFSET);
			if((listBox->numElements-listBox->topLine)>listBox->numLines)
			{
				activeHeight=LIST_BOX_TOP_OFFSET+(listBox->numLines*listBox->lineHeight);
			}
			else
			{
				activeHeight=LIST_BOX_TOP_OFFSET+((listBox->numElements-listBox->topLine)*listBox->lineHeight);
			}
			XFillRectangle(xDisplay,imageBuffer,xGraphicsContext,listContentRect.x,listContentRect.y+activeHeight,listContentRect.w,listContentRect.h-activeHeight);

			XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(foreground));
			XSetFont(xDisplay,xGraphicsContext,listBox->font->xFont->fid);		// point to the current font for this text

			for(currentElement=listBox->topLine;(currentElement<listBox->numElements)&&((currentElement-listBox->topLine)<listBox->numLines);currentElement++)
			{
				GetListBoxElementRect(listBox,currentElement,&itemRect);
				if(!listBox->selectedElements[currentElement])					// see if this one is selected, if so, draw it as such
				{
					XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(background));
					XFillRectangle(xDisplay,imageBuffer,xGraphicsContext,itemRect.x,itemRect.y,itemRect.w,itemRect.h);
					XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(foreground));
				}
				else
				{
					XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(foreground));
					XFillRectangle(xDisplay,imageBuffer,xGraphicsContext,itemRect.x,itemRect.y,itemRect.w,itemRect.h);
					XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(background));
				}
				XDrawString(xDisplay,imageBuffer,xGraphicsContext,itemRect.x+LIST_BOX_LEFT_OFFSET,itemRect.y+ITEM_SPACE_TOP+listBox->font->ascent,listBox->listElements[currentElement],strlen(listBox->listElements[currentElement]));
			}
		}
		XSetClipMask(xDisplay,xGraphicsContext,None);							// get rid of clip mask
	}
	XDestroyRegion(invalidRegion);												// get rid of invalid region
	DrawScrollBar(listBox->scrollBar,imageBuffer,updateRegion);
}

static void TrackNormalListElements(LIST_BOX *listBox,XEvent *event,UINT32 element)
// track list elements normally
{
	EDITOR_RECT
		itemRect;
	Window
		root,
		child;
	int
		rootX,
		rootY,
		windowX,
		windowY;
	unsigned int
		state;

	ClearListBoxSelections(listBox);
	listBox->selectedElements[element]=true;
	listBox->currentElement=element;
	GetListBoxElementRect(listBox,element,&itemRect);
	InvalidateWindowRect(listBox->parentWindow,&itemRect);
	UpdateEditorWindows();
	while(StillDown(event->xbutton.button,1))
	{
		if(XQueryPointer(xDisplay,((WINDOW_LIST_ELEMENT *)listBox->parentWindow->userData)->xWindow,&root,&child,&rootX,&rootY,&windowX,&windowY,&state))
		{
			if(GetPointedToElement(listBox,windowX,windowY,&element,false))
			{
				if(element!=listBox->currentElement)
				{
					listBox->selectedElements[listBox->currentElement]=false;
					GetListBoxElementRect(listBox,listBox->currentElement,&itemRect);
					InvalidateWindowRect(listBox->parentWindow,&itemRect);

					listBox->selectedElements[element]=true;
					GetListBoxElementRect(listBox,element,&itemRect);
					InvalidateWindowRect(listBox->parentWindow,&itemRect);

					listBox->currentElement=element;
					HomeListBoxToCurrentElement(listBox);	// this will scroll the window if needed
					UpdateEditorWindows();
				}
			}
		}
	}
}

static void TrackShiftListElements(LIST_BOX *listBox,XEvent *event,UINT32 element)
// track list elements in a mode which selects/deselects
{
	UINT32
		tempElement;
	EDITOR_RECT
		itemRect;
	Window
		root,
		child;
	int
		rootX,
		rootY,
		windowX,
		windowY;
	unsigned int
		state;
	bool
		fillMode;

	fillMode=listBox->selectedElements[element]=listBox->selectedElements[element]?false:true;
	listBox->currentElement=element;
	GetListBoxElementRect(listBox,element,&itemRect);
	InvalidateWindowRect(listBox->parentWindow,&itemRect);
	UpdateEditorWindows();
	while(StillDown(event->xbutton.button,1))
	{
		if(XQueryPointer(xDisplay,((WINDOW_LIST_ELEMENT *)listBox->parentWindow->userData)->xWindow,&root,&child,&rootX,&rootY,&windowX,&windowY,&state))
		{
			if(GetPointedToElement(listBox,windowX,windowY,&element,false))
			{
				if(element>listBox->currentElement)	// fill from current to new
				{
					for(tempElement=element;tempElement>listBox->currentElement;tempElement--)
					{
						listBox->selectedElements[tempElement]=fillMode;
						GetListBoxElementRect(listBox,tempElement,&itemRect);
						InvalidateWindowRect(listBox->parentWindow,&itemRect);
					}
					listBox->currentElement=element;
					HomeListBoxToCurrentElement(listBox); // this will scroll the window if needed
					UpdateEditorWindows();
				}
				else
				{
					if(element<listBox->currentElement)
					{
						for(tempElement=element;tempElement<listBox->currentElement;tempElement++)
						{
							listBox->selectedElements[tempElement]=fillMode;
							GetListBoxElementRect(listBox,tempElement,&itemRect);
							InvalidateWindowRect(listBox->parentWindow,&itemRect);
						}
						listBox->currentElement=element;
						HomeListBoxToCurrentElement(listBox); // this will scroll the window if needed
						UpdateEditorWindows();
					}
				}
			}
		}
	}
}

static void TrackHandListElements(LIST_BOX *listBox,XEvent *event,UINT32 element)
// move around in the list box by tracking the mouse
{
	int
		initialY;
	int
		lineOffset;
	UINT32
		initialTopLine;
	UINT32
		newTopLine;
	Window
		root,
		child;
	int
		rootX,
		rootY,
		windowX,
		windowY;
	unsigned int
		state;

	initialY=event->xbutton.y;
	initialTopLine=listBox->topLine;

	while(StillDown(event->xbutton.button,1))
	{
		if(XQueryPointer(xDisplay,((WINDOW_LIST_ELEMENT *)listBox->parentWindow->userData)->xWindow,&root,&child,&rootX,&rootY,&windowX,&windowY,&state))
		{
			lineOffset=((initialY-windowY)/(int)listBox->lineHeight);
			if(lineOffset<0)
			{
				if(initialTopLine>=(UINT32)-(lineOffset))
				{
					newTopLine=initialTopLine+lineOffset;
				}
				else
				{
					newTopLine=0;
				}
			}
			else
			{
				newTopLine=initialTopLine+lineOffset;
			}
			SetListBoxTopLine(listBox,newTopLine);
		}
	}
}

static bool TrackListElements(LIST_BOX *listBox,XEvent *event)
// Track the elements of the list
{
	UINT32
		element;
	EDITOR_RECT
		borderRect,
		listContentRect,
		scrollBarRect;
	UINT32
		numClicks,
		modifiers;

	GetListBoxControlRects(listBox,&borderRect,&listContentRect,&scrollBarRect);
	if(PointInRECT(event->xbutton.x,event->xbutton.y,&listContentRect))
	{
		if(GetPointedToElement(listBox,event->xbutton.x,event->xbutton.y,&element,true))
		{
			numClicks=HandleMultipleClicks(event);
			XStateToEditorModifiers(event->xbutton.state,numClicks,&modifiers);
			switch(event->xbutton.button)
			{
				case Button1:
					if(!(modifiers&EEM_SHIFT))
					{
						TrackNormalListElements(listBox,event,element);
					}
					else
					{
						TrackShiftListElements(listBox,event,element);
					}
					if(numClicks&&listBox->pressedProc)
					{
						listBox->pressedProc(listBox->pressedProcParameters);
					}
					break;
				case Button2:
					TrackHandListElements(listBox,event,element);
					break;
				case Button3:
					TrackShiftListElements(listBox,event,element);
					break;
			}
			return(true);
		}
	}
	return(false);
}

static bool TrackListBox(LIST_BOX *listBox,XEvent *event)
// a button was pressed, see if it was in listBox, if so, track it, and return true
// if it was not within the list box, return false
{
	if(!TrackScrollBar(listBox->scrollBar,event))
	{
		if(!TrackListElements(listBox,event))
		{
			return(false);
		}
	}
	return(true);
}

static void ResetTypedCharacters(LIST_BOX *listBox)
// reset the typed character search buffer for listBox
{
	listBox->charBufferLength=0;						// number of characters in the charBuffer
	listBox->charTimer=0;								// clear the typed character timer
}

static bool BetterMatch(LIST_BOX *listBox,UINT32 element,UINT32 previousElement)
// return true if element is a better match for the contents
// of currentBuffer, than previousElement
// NOTE: if there is a tie, previousElement wins
{
	int
		result;

	result=strncmp(listBox->charBuffer,listBox->listElements[previousElement],listBox->charBufferLength);	// see if char buffer so far is an exact match for the current best element
	if(result==0)
	{
		return(false);				// new element cannot be better
	}

	result=strcmp(listBox->listElements[element],listBox->listElements[previousElement]);	// see which of the two elements is larger
	if(result>0)					// new element is larger
	{
		return(strcmp(listBox->charBuffer,listBox->listElements[previousElement])>0);	// if char buffer is larger than previous, move to new
	}
	else if(result<0)				// old element is larger
	{
		return(strcmp(listBox->charBuffer,listBox->listElements[element])<=0);			// if char buffer is smaller than new, move to new
	}
	return(false);					// elements match, previous wins
}

static UINT32 SearchElements(LIST_BOX *listBox)
// search through the elements of listBox, looking for
// the first, best match to charBuffer
// return the element located
// NOTE: if numElements is 0, this will return 0
{
	UINT32
		currentElement,
		bestMatchElement;

	bestMatchElement=0;
	for(currentElement=1;currentElement<listBox->numElements;currentElement++)
	{
		if(BetterMatch(listBox,currentElement,bestMatchElement))
		{
			bestMatchElement=currentElement;
		}
	}
	return(bestMatchElement);
}

static void CheckTypedChar(LIST_BOX *listBox,char character)
// See where char should be added to the search buffer,
// add it, reset timers as needed, and hunt for the contents
// of the search buffer in listBox
// take the first, best match, set the current element to it,
// clear all other selections, and finally home the list to the current element
// This allows the user to type characters while a list box is open,
// locating the element of the list that is the best match to what
// has been typed.
// NOTE: this does not depend on the list to be sorted in any way,
// but the results are better if it is
{
	EDITOR_RECT
		itemRect;

	if(listBox->charTimer+LIST_CHAR_TIMEOUT<timer)
	{
		listBox->charBufferLength=0;								// reset index to character buffer if timeout has expired
	}
	if(listBox->charBufferLength+1<MAX_LIST_CHAR_BUFFER)
	{
		listBox->charBuffer[listBox->charBufferLength++]=character;	// if room, add the character to the buffer
		listBox->charBuffer[listBox->charBufferLength]='\0';		// terminate the string
	}
	if(listBox->numElements)
	{
		ClearListBoxSelections(listBox);
		listBox->currentElement=SearchElements(listBox);
		listBox->selectedElements[listBox->currentElement]=true;
		GetListBoxElementRect(listBox,listBox->currentElement,&itemRect);
		InvalidateWindowRect(listBox->parentWindow,&itemRect);
		CenterListBoxToCurrentElement(listBox); 					// center the selected element
//		HomeListBoxToCurrentElement(listBox);
		UpdateEditorWindows();
	}
	listBox->charTimer=timer;										// reset timer to current time
}

static void ScrollListBoxLower(SCROLL_BAR *scrollBar,bool repeating,void *parameters)
// scroll bar for this box is reporting a click
{
	LIST_BOX
		*listBox;

	listBox=(LIST_BOX *)parameters;
	if(listBox->topLine>0)
	{
		SetListBoxTopLine(listBox,listBox->topLine-1);
	}
}

static void ScrollListBoxHigher(SCROLL_BAR *scrollBar,bool repeating,void *parameters)
// scroll bar for this box is reporting a click
{
	LIST_BOX
		*listBox;

	listBox=(LIST_BOX *)parameters;
	SetListBoxTopLine(listBox,listBox->topLine+1);
}

static void PageListBoxLower(SCROLL_BAR *scrollBar,bool repeating,void *parameters)
// scroll bar for this box is reporting a click
{
	LIST_BOX
		*listBox;

	listBox=(LIST_BOX *)parameters;
	if(listBox->topLine>listBox->numLines)
	{
		SetListBoxTopLine(listBox,listBox->topLine-listBox->numLines);
	}
	else
	{
		SetListBoxTopLine(listBox,0);
	}
}

static void PageListBoxHigher(SCROLL_BAR *scrollBar,bool repeating,void *parameters)
// scroll bar for this box is reporting a click
{
	LIST_BOX
		*listBox;

	listBox=(LIST_BOX *)parameters;
	SetListBoxTopLine(listBox,listBox->topLine+listBox->numLines);
}

static void ListBoxThumbProc(SCROLL_BAR *scrollBar,UINT32 newValue,void *parameters)
// scroll bar for this box is reporting a click in the thumb
{
	LIST_BOX
		*listBox;

	listBox=(LIST_BOX *)parameters;
	SetListBoxTopLine(listBox,newValue);
}

static void HandleListBoxKeyEvent(LIST_BOX *listBox,XEvent *event)
// A keyboard event is arriving for the list box, handle it here
{
	EDITOR_RECT
		itemRect;
	XKeyEvent
		localKeyEvent;									// make a local copy so we can strip annoying bits
	EDITOR_KEY
		key;

	if(listBox->numElements)
	{
		localKeyEvent=*(XKeyEvent *)event;				// copy the event
		localKeyEvent.state&=(~LockMask);					// remove this from consideration
		if(KeyEventToEditorKey((XEvent *)&localKeyEvent,&key))
		{
			if(key.isVirtual)							// see if virtual key or not
			{
				ResetTypedCharacters(listBox);
				switch(key.keyCode)
				{
					case VVK_UPARROW:
						if(listBox->selectedElements[listBox->currentElement])
						{
							if(listBox->currentElement)
							{
								ClearListBoxSelections(listBox);
								listBox->currentElement--;
								listBox->selectedElements[listBox->currentElement]=true;
								GetListBoxElementRect(listBox,listBox->currentElement,&itemRect);
								InvalidateWindowRect(listBox->parentWindow,&itemRect);
							}
						}
						else
						{
							ClearListBoxSelections(listBox);
							listBox->selectedElements[listBox->currentElement]=true;
							GetListBoxElementRect(listBox,listBox->currentElement,&itemRect);
							InvalidateWindowRect(listBox->parentWindow,&itemRect);
						}
						HomeListBoxToCurrentElement(listBox);
						break;
					case VVK_DOWNARROW:
						if(listBox->selectedElements[listBox->currentElement])
						{
							if(listBox->currentElement+1<listBox->numElements)
							{
								ClearListBoxSelections(listBox);
								listBox->currentElement++;
								listBox->selectedElements[listBox->currentElement]=true;
								GetListBoxElementRect(listBox,listBox->currentElement,&itemRect);
								InvalidateWindowRect(listBox->parentWindow,&itemRect);
							}
						}
						else
						{
							ClearListBoxSelections(listBox);
							listBox->selectedElements[listBox->currentElement]=true;
							GetListBoxElementRect(listBox,listBox->currentElement,&itemRect);
							InvalidateWindowRect(listBox->parentWindow,&itemRect);
						}
						HomeListBoxToCurrentElement(listBox);
						break;
					case VVK_PAGEUP:
						if(listBox->selectedElements[listBox->currentElement])
						{
							if(listBox->currentElement)
							{
								if(listBox->currentElement>listBox->numLines)
								{
									listBox->currentElement-=listBox->numLines;
								}
								else
								{
									listBox->currentElement=0;
								}
								ClearListBoxSelections(listBox);
								listBox->selectedElements[listBox->currentElement]=true;
								GetListBoxElementRect(listBox,listBox->currentElement,&itemRect);
								InvalidateWindowRect(listBox->parentWindow,&itemRect);
							}
						}
						else
						{
							ClearListBoxSelections(listBox);
							listBox->selectedElements[listBox->currentElement]=true;
							GetListBoxElementRect(listBox,listBox->currentElement,&itemRect);
							InvalidateWindowRect(listBox->parentWindow,&itemRect);
						}
						HomeListBoxToCurrentElement(listBox);
						break;
					case VVK_PAGEDOWN:
						if(listBox->selectedElements[listBox->currentElement])
						{
							if(listBox->currentElement+1<listBox->numElements)
							{
								if(listBox->currentElement+listBox->numLines<listBox->numElements)
								{
									listBox->currentElement+=listBox->numLines;
								}
								else
								{
									listBox->currentElement=listBox->numElements-1;
								}
								ClearListBoxSelections(listBox);
								listBox->selectedElements[listBox->currentElement]=true;
								GetListBoxElementRect(listBox,listBox->currentElement,&itemRect);
								InvalidateWindowRect(listBox->parentWindow,&itemRect);
							}
						}
						else
						{
							ClearListBoxSelections(listBox);
							listBox->selectedElements[listBox->currentElement]=true;
							GetListBoxElementRect(listBox,listBox->currentElement,&itemRect);
							InvalidateWindowRect(listBox->parentWindow,&itemRect);
						}
						HomeListBoxToCurrentElement(listBox);
						break;
					case VVK_HOME:
						ClearListBoxSelections(listBox);
						listBox->currentElement=0;
						listBox->selectedElements[listBox->currentElement]=true;
						GetListBoxElementRect(listBox,listBox->currentElement,&itemRect);
						InvalidateWindowRect(listBox->parentWindow,&itemRect);
						HomeListBoxToCurrentElement(listBox);
						break;
					case VVK_END:
						ClearListBoxSelections(listBox);
						if(listBox->numElements)
						{
							listBox->currentElement=listBox->numElements-1;
							listBox->selectedElements[listBox->currentElement]=true;
						}
						else
						{
							listBox->currentElement=0;
						}
						GetListBoxElementRect(listBox,listBox->currentElement,&itemRect);
						InvalidateWindowRect(listBox->parentWindow,&itemRect);
						HomeListBoxToCurrentElement(listBox);
						break;
				}
			}
			else
			{
				if(key.keyCode=='a'&&(key.modifiers&EEM_MOD0))	// select all
				{
					ResetTypedCharacters(listBox);
					SetListBoxSelections(listBox);
				}
				else
				{
					// asc to search on
					CheckTypedChar(listBox,key.keyCode);
				}
			}
		}
	}
}

static void ResetListBoxLists(LIST_BOX *listBox,UINT32 numElements,const char **list,bool *selectedElements)
// reset the list in listBox to list with selectedElements
{
	UINT32
		maxLines;

	listBox->numElements=numElements;
	listBox->listElements=list;
	listBox->selectedElements=selectedElements;

	ResetTypedCharacters(listBox);
	listBox->topLine=listBox->currentElement=0;

	if(listBox->numLines<listBox->numElements)
	{
		maxLines=listBox->numElements-listBox->numLines;
	}
	else
	{
		if(listBox->numElements>listBox->numLines)
		{
			maxLines=listBox->numElements-listBox->numLines;
		}
		else
		{
			maxLines=0;
		}
	}
	AdjustScrollBar(listBox->scrollBar,maxLines,0);
	InvalidateListContents(listBox);				// invalidate contents
}

static LIST_BOX *CreateListBox(EDITOR_WINDOW *window,LIST_BOX_DESCRIPTOR *description)
// create a list box in window, invalidate the area of window
// where the list box is to be placed
// return a pointer to it if successful
// SetError, return NULL if not
{
	LIST_BOX
		*listBox;
	SCROLL_BAR_DESCRIPTOR
		scrollDescriptor;

	if((listBox=(LIST_BOX *)MNewPtr(sizeof(LIST_BOX))))	// create list box data structure
	{
		listBox->parentWindow=window;

		listBox->resizeFunction=description->resizeFunction;

		listBox->focusForegroundColor=description->focusForegroundColor;
		listBox->focusBackgroundColor=description->focusBackgroundColor;
		listBox->nofocusForegroundColor=description->nofocusForegroundColor;
		listBox->nofocusBackgroundColor=description->nofocusBackgroundColor;

		listBox->verticalScrollBarOnLeft=verticalScrollBarOnLeft;	// copy global preference
		listBox->numElements=description->numElements;
		listBox->listElements=description->listElements;
		listBox->selectedElements=description->selectedElements;

		ResetTypedCharacters(listBox);								// reset keyboard input delays

		listBox->pressedProc=description->pressedProc;
		listBox->pressedProcParameters=description->pressedProcParameters;

		if((listBox->font=LoadFont(description->fontName)))
		{
			listBox->currentElement=description->topLine;
			listBox->topLine=description->topLine;

			scrollDescriptor.x=0;									// place scroll bar at 0 initially, resize later
			scrollDescriptor.y=0;
			scrollDescriptor.length=0;
			scrollDescriptor.horizontal=false;
			scrollDescriptor.thumbMax=listBox->numElements;
			scrollDescriptor.thumbPosition=listBox->topLine;
			scrollDescriptor.stepLowerProc=ScrollListBoxLower;
			scrollDescriptor.stepHigherProc=ScrollListBoxHigher;
			scrollDescriptor.pageLowerProc=PageListBoxLower;
			scrollDescriptor.pageHigherProc=PageListBoxHigher;
			scrollDescriptor.thumbProc=ListBoxThumbProc;
			scrollDescriptor.procParameters=(void *)listBox;

			if((listBox->scrollBar=CreateScrollBar(window,&scrollDescriptor)))
			{
				SetListBoxRectangle(listBox);
				return(listBox);
			}
			FreeFont(listBox->font);					// free the font
		}
		MDisposePtr(listBox);
	}
	return(NULL);
}

static void DisposeListBox(LIST_BOX *listBox)
// unlink listBox from its parent window, and delete it
// invalidate the area of the window where the list box was
{
	InvalidateListBox(listBox);							// invalidate where it is now

	DisposeScrollBar(listBox->scrollBar);
	FreeFont(listBox->font);							// free the font

	MDisposePtr(listBox);
}

// dialog interface routines

static void DrawListBoxItem(DIALOG_ITEM *item,Pixmap imageBuffer,Region updateRegion)
// draw the list box pointed to by item
{
	LIST_BOX
		*listBox;

	listBox=(LIST_BOX *)item->itemStruct;
	DrawListBox(listBox,imageBuffer,updateRegion,(item==item->dialog->focusItem));
}

static void ResizeListBoxItem(DIALOG_ITEM *item)
// When the dialog window size changes, call this to update the item itself
// This will invalidate the old and new item areas
{
	LIST_BOX
		*listBox;

	listBox=(LIST_BOX *)item->itemStruct;
	InvalidateListBox(listBox);				// invalidate where we're at currently
	SetListBoxRectangle(listBox);
}

static bool HandleListBoxMouseWheelEvents(LIST_BOX *listBox,XEvent *event)
// Added 2/12/2004 FEA
// Call this to see if buttons 4 or 5 have been clicked. X sends button press events for buttons 4 and 5 when you
// scroll the mouse wheel. Returns true if it was a wheel event, else false
{
	UINT32
		delta;
	UINT32
		i;
	XEvent
		xevent;

	if(event->xbutton.button==Button4||event->xbutton.button==Button5)
	{
		delta=2;
		//
		// check for other events of the same kind in the event queue and pull them off now
		// the more we have, the more we should scroll
		//
		while(XCheckTypedWindowEvent(xDisplay,((WINDOW_LIST_ELEMENT *)listBox->parentWindow->userData)->xWindow,ButtonPress,&xevent))
		{
			if(xevent.xbutton.button!=event->xbutton.button)
			{
				XPutBackEvent(xDisplay,&xevent);
				break;
			}
			delta++;
		}

		if(delta>1)
		{
			delta*=3;
		}
		if(event->xbutton.button==Button4)
		{
			for(i=0;(listBox->topLine)&&(i<delta);i++)
			{
				SetListBoxTopLine(listBox,listBox->topLine-1);
			}
		}
		else
		{
			for(i=0;i<delta;i++)
			{
				SetListBoxTopLine(listBox,listBox->topLine+1);
			}
		}
		return(true);
	}
	return(false);
}

static bool TrackListBoxItem(DIALOG_ITEM *item,XEvent *event)
// attempt to track the list box pointed to by item
{
	if(PointInListBox((LIST_BOX *)item->itemStruct,event->xbutton.x,event->xbutton.y))
	{
		if(!HandleListBoxMouseWheelEvents((LIST_BOX *)item->itemStruct,event))
		{
			DialogTakeFocus(item);				// set the focus
			return(TrackListBox((LIST_BOX *)item->itemStruct,event));
		}
		return(true);
	}
	return(false);
}

static void DisposeListBoxItem(DIALOG_ITEM *item)
// dispose of the local list box structure pointed to by item
{
	DisposeListBox((LIST_BOX *)item->itemStruct);
}

static void ListBoxItemFocusChange(DIALOG_ITEM *item)
// focus status has changed for list box
// handle it here
{
	InvalidateListBox((LIST_BOX *)item->itemStruct);
}

static void ListBoxItemFocusKey(DIALOG_ITEM *item,XEvent *event)
// give the key to the list box
{
	HandleListBoxKeyEvent((LIST_BOX *)item->itemStruct,event);
}

void ResetListBoxItemLists(DIALOG_ITEM *item,UINT32 numElements,const char **list,bool *selectedElements)
// place a new list into item
{
	ResetListBoxLists((LIST_BOX *)item->itemStruct,numElements,list,selectedElements);
}

bool CreateListBoxItem(DIALOG_ITEM *item,void *descriptor)
// create the list box, and fill in item
// if there is a problem, SetError, return false
{
	if((item->itemStruct=(void *)CreateListBox(item->dialog->parentWindow,(LIST_BOX_DESCRIPTOR *)descriptor)))
	{
		item->disposeProc=DisposeListBoxItem;
		item->drawProc=DrawListBoxItem;
		item->resizeProc=ResizeListBoxItem;
		item->trackProc=TrackListBoxItem;
		item->earlyKeyProc=NULL;
		item->wantFocus=true;
		item->focusChangeProc=ListBoxItemFocusChange;
		item->focusKeyProc=ListBoxItemFocusKey;
		item->focusPeriodicProc=NULL;
		return(true);
	}
	return(false);
}
