// Menu handling
// Copyright (C) 2000 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

#define BACKGROUND_COLOR			GRAY_2		// color of the background of the menus
#define	SHADOW_COLOR_UL				GRAY_3		// shadow colors (upper left)
#define	SHADOW_COLOR_LR				GRAY_1		// (lower right)

#define MENU_BORDER_SIZE			2			// number of pixels of border around menus (not items, but whole menus)

#define	MENU_MIN_ITEM_COMMAND_SPACING	7		// minimum number of pixels to leave between menu item, and right justified command string
#define	MENU_ADDITIONAL_WIDTH			6		// additional pixel width added to each menu item to make it look better when drawn
#define	MENU_ADDITIONAL_HEIGHT			6		// additional pixel height added to each menu item to make it look better when drawn

typedef struct
{
	EDITOR_MENU
		*ownerMenu;								// points to the menu that owns this window (if NULL, root menu owns window)
} GUI_MENU_WINDOW_INFO;

static EDITOR_MENU
	*firstMenu;

enum
{
	BAD_RELATIONSHIP,
	BAD_PATH
};

static const char *errorDescriptions[]=
{
	"Invalid menu relationship",
	"Invalid menu path"
};

EDITOR_MENU *GetRelatedEditorMenu(EDITOR_MENU *menu,int menuRelationship)
// return the menu with menuRelationship to menu
// return NULL if there is no menu with menuRelationship
{
	EDITOR_MENU
		*outputMenu;

	outputMenu=NULL;
	switch(menuRelationship)
	{
		case MR_NEXTSIBLING:
			if(menu->parent)
			{
				if(menu->nextSibling!=menu->parent->firstChild)
				{
					outputMenu=menu->nextSibling;
				}
			}
			else
			{
				if(menu->nextSibling!=firstMenu)
				{
					outputMenu=menu->nextSibling;
				}
			}
			break;
		case MR_PREVIOUSSIBLING:
			if(menu->parent)
			{
				if(menu!=menu->parent->firstChild)
				{
					outputMenu=menu->previousSibling;
				}
			}
			else
			{
				if(menu!=firstMenu)
				{
					if(menu->previousSibling!=firstMenu)
					{
						outputMenu=menu->previousSibling;
					}
				}
				else
				{
					outputMenu=firstMenu;
				}
			}
			break;
		case MR_FIRSTCHILD:
			if(menu)
			{
				outputMenu=menu->firstChild;
			}
			else
			{
				outputMenu=firstMenu;
			}
			break;
		case MR_LASTCHILD:
			if(menu)
			{
				if(menu->firstChild)
				{
					outputMenu=menu->firstChild->previousSibling;
				}
			}
			else
			{
				if(firstMenu)
				{
					outputMenu=firstMenu->previousSibling;
				}
			}
			break;
		default:
			break;
	}
	return(outputMenu);
}

static EDITOR_WINDOW *GetMenuWindow(EDITOR_MENU *menu)
// return the window that belongs to menu if it has one, or NULL
// if it does not
{
	if(menu)						// see if menu, or root
	{
		return(menu->window);
	}
	return(rootMenuWindow);
}

static void CalculateMenuItemSize(EDITOR_MENU *menu,UINT32 *width,UINT32 *height)
// get the ideal width and height for menu
{
	if(menu->separator)
	{
		(*height)=2;												// use exact height for separators
		(*width)=MENU_ADDITIONAL_WIDTH;
	}
	else
	{
		(*height)=menu->font->ascent+menu->font->descent+MENU_ADDITIONAL_HEIGHT;	// get the height of the menu item (use total height of the font, plus some slop to make it look better)
		(*width)=XTextWidth(menu->font->xFont,menu->itemText,strlen(menu->itemText))+MENU_ADDITIONAL_WIDTH;

		if(menu->firstChild)										// if this menu has children, we need to draw the hierarchical right arrow, otherwise, we can use the command string
		{
			(*width)+=MENU_MIN_ITEM_COMMAND_SPACING;				// add spacing between item and command string
			(*width)+=SUB_MENU_ARROW_WIDTH;
			if((*height)<SUB_MENU_ARROW_HEIGHT)						// if the font is tiny, then adjust height
			{
				(*height)=SUB_MENU_ARROW_HEIGHT;
			}
		}
		else
		{
			if(menu->commandString)
			{
				(*width)+=MENU_MIN_ITEM_COMMAND_SPACING;			// add spacing between item and command string
				(*width)+=XTextWidth(menu->font->xFont,menu->commandString,strlen(menu->commandString));
			}
		}
	}
}

static void PinMenuRect(EDITOR_RECT *rect)
// Make sure rect does not go off the display (as best as possible).
{
	UINT32
		displayWidth,
		displayHeight;

	GetEditorDisplayDimensions(&displayWidth,&displayHeight);
	if(rect->x+rect->w>displayWidth)
	{
		rect->x=displayWidth-rect->w;
	}
	if(rect->y+rect->h>displayHeight)
	{
		rect->y=displayHeight-rect->h;
	}
	if(rect->x<0)
	{
		rect->x=0;
	}
	if(rect->y<0)
	{
		rect->y=0;
	}
}

static void CalculateMenuItemRect(EDITOR_MENU *menu,EDITOR_RECT *rect)
// get the rect of item as it will be drawn into its parent menu
{
	EDITOR_MENU
		*currentMenu,
		*parentMenu;
	UINT32
		tempWidth,
		tempHeight;
	INT32
		currentY;

	rect->x=rect->y=MENU_BORDER_SIZE;									// offset by border
	rect->w=rect->h=0;
	currentY=0;
	parentMenu=menu->parent;
	currentMenu=GetRelatedEditorMenu(parentMenu,MR_FIRSTCHILD);
	while(currentMenu)
	{
		CalculateMenuItemSize(currentMenu,&tempWidth,&tempHeight);
		if(tempWidth>rect->w)
		{
			rect->w=tempWidth;											// find the maximum width of all the items
		}
		if(currentMenu==menu)											// when we get to our menu remember its Y position
		{
			rect->y+=currentY;
			rect->h=tempHeight;
		}
		currentY+=tempHeight;
		currentMenu=GetRelatedEditorMenu(currentMenu,MR_NEXTSIBLING);
	}
}

static void CalculateMenuWindowSize(EDITOR_MENU *menu,UINT32 *width,UINT32 *height)
// run though the list of children of menu, determining the size of each, use these
// sizes to get the composite size for menu
{
	EDITOR_MENU
		*currentMenu;
	UINT32
		tempWidth,
		tempHeight;

	(*width)=0;
	(*height)=0;
	currentMenu=GetRelatedEditorMenu(menu,MR_FIRSTCHILD);
	while(currentMenu)
	{
		CalculateMenuItemSize(currentMenu,&tempWidth,&tempHeight);
		if(tempWidth>(*width))
		{
			(*width)=tempWidth;											// find the maximum width of all the items
		}
		(*height)+=tempHeight;											// and the total height of all items
		currentMenu=GetRelatedEditorMenu(currentMenu,MR_NEXTSIBLING);
	}
	*width+=2*MENU_BORDER_SIZE;											// fancy border drawn around entire menu
	*height+=2*MENU_BORDER_SIZE;
}

static void DrawMenuItem(Window xWindow,EDITOR_MENU *menu)
// use this to draw menu's item into xWindow
{
	EDITOR_RECT
		itemRect;
	int
		stringLength,
		stringWidth;
	int
		itemX,
		itemY;
	INT32
		xLeft,
		xRight,
		yTop,
		yBottom;

	CalculateMenuItemRect(menu,&itemRect);								// get the rectangle for this item
	if(itemRect.w&&itemRect.h)
	{
		XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(BACKGROUND_COLOR));
		XFillRectangle(xDisplay,xWindow,xGraphicsContext,itemRect.x,itemRect.y,itemRect.w,itemRect.h);

		if(menu->separator)
		{
			xLeft=itemRect.x;
			xRight=itemRect.x+itemRect.w-1;
			yTop=itemRect.y;
			yBottom=yTop+1;

			XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(BLACK));
			XDrawLine(xDisplay,xWindow,xGraphicsContext,xLeft,yTop,xRight,yTop);
			XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(WHITE));
			XDrawLine(xDisplay,xWindow,xGraphicsContext,xLeft,yBottom,xRight,yBottom);
		}
		else
		{
			stringLength=strlen(menu->itemText);
			stringWidth=XTextWidth(menu->font->xFont,menu->itemText,stringLength);
			itemX=itemRect.x+(MENU_ADDITIONAL_WIDTH/2);
			itemY=itemRect.y+(itemRect.h/2)-(menu->font->ascent+menu->font->descent)/2+menu->font->ascent;

			XSetFont(xDisplay,xGraphicsContext,menu->font->xFont->fid);	// point to the current font for this item

			XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(BLACK));
			XDrawString(xDisplay,xWindow,xGraphicsContext,itemX,itemY,menu->itemText,stringLength);

			if(menu->firstChild)
			{
				itemX=itemRect.x+(itemRect.w-SUB_MENU_ARROW_WIDTH-(MENU_ADDITIONAL_WIDTH/2));
				itemY=itemRect.y+(itemRect.h-SUB_MENU_ARROW_HEIGHT)/2;
				XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(GRAY_3));
				XSetBackground(xDisplay,xGraphicsContext,EditorColorToXPixel(BACKGROUND_COLOR));
				XCopyPlane(xDisplay,subMenuPixmap,xWindow,xGraphicsContext,0,0,SUB_MENU_ARROW_WIDTH,SUB_MENU_ARROW_HEIGHT,itemX,itemY,1);
			}
			else
			{
				if(menu->commandString)
				{
					stringLength=strlen(menu->commandString);
					stringWidth=XTextWidth(menu->font->xFont,menu->commandString,stringLength);
					itemX=itemRect.x+(itemRect.w-stringWidth-(MENU_ADDITIONAL_WIDTH/2));
					itemY=itemRect.y+(itemRect.h/2)-(menu->font->ascent+menu->font->descent)/2+menu->font->ascent;

					XSetFont(xDisplay,xGraphicsContext,menu->font->xFont->fid);	// point to the current font for this item

					XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(GRAY_0));
					XDrawString(xDisplay,xWindow,xGraphicsContext,itemX+1,itemY+1,menu->commandString,stringLength);

					XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(WHITE));
					XDrawString(xDisplay,xWindow,xGraphicsContext,itemX,itemY,menu->commandString,stringLength);
				}
			}
			if(menu->highlight)
			{
				OutlineShadowRectangle(xWindow,&itemRect,WHITE,GRAY_0,2);
			}
		}
	}
}

static void DrawMenuWindow(EDITOR_WINDOW *window)
// draw all of the buttons into window
// also draw the menu border
{
	EDITOR_MENU
		*owner,							// the edit menu that owns this window
		*currentMenu;
	WINDOW_LIST_ELEMENT
		*windowElement;
	Window
		xWindow;
	XWindowAttributes
		attributes;
	EDITOR_RECT
		windowRect;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;				// get the x window information associated with this window
	xWindow=windowElement->xWindow;										// get x window for this window
	XSetRegion(xDisplay,xGraphicsContext,windowElement->invalidRegion);	// set clipping to what's invalid

	XGetWindowAttributes(xDisplay,xWindow,&attributes);
	windowRect.x=0;
	windowRect.y=0;
	windowRect.w=attributes.width;
	windowRect.h=attributes.height;
	OutlineShadowRectangle(xWindow,&windowRect,SHADOW_COLOR_UL,SHADOW_COLOR_LR,MENU_BORDER_SIZE);

	owner=((GUI_MENU_WINDOW_INFO *)window->windowInfo)->ownerMenu;
	currentMenu=GetRelatedEditorMenu(owner,MR_FIRSTCHILD);
	while(currentMenu)
	{
		DrawMenuItem(xWindow,currentMenu);								// draw the items into the menu
		currentMenu=GetRelatedEditorMenu(currentMenu,MR_NEXTSIBLING);
	}
	XSetClipMask(xDisplay,xGraphicsContext,None);						// get rid of clip mask
}

void UpdateMenuWindow(EDITOR_WINDOW *window)
// make sure window is up to date
{
	WINDOW_LIST_ELEMENT
		*windowElement;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;				// get the x window information associated with this menu window
	if(!XEmptyRegion(windowElement->invalidRegion))						// see if it has an invalid area
	{
		DrawMenuWindow(window);
		XDestroyRegion(windowElement->invalidRegion);					// clear out the invalid region now that the window has been drawn
		windowElement->invalidRegion=XCreateRegion();
	}
}

static void ResizeMenuWindow(EDITOR_MENU *menu)
// if menu has a window, call this to size and invalidate the window when a child of menu
// has changed in some way, or when the menu is created
// if NULL is passed for menu, it indicates the root menu window
{
	EDITOR_WINDOW
		*window;
	WINDOW_LIST_ELEMENT
		*windowElement;
	EDITOR_RECT
		rect;
	XSizeHints
		sizeHints;

	rect.x=rect.y=0;

	if((window=GetMenuWindow(menu)))
	{
		windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
		CalculateMenuWindowSize(menu,&rect.w,&rect.h);					// get the ideal dimensions of this window (based on its children)
		if(rect.w==0)													// ### this is because X will not allow me to open windows of 0 width or height!
		{
			rect.w=1;
		}
		if(rect.h==0)
		{
			rect.h=1;
		}

		sizeHints.flags=PMinSize|PMaxSize;
		sizeHints.min_width=rect.w;
		sizeHints.min_height=rect.h;
		sizeHints.max_width=rect.w;
		sizeHints.max_height=rect.h;
		XSetWMNormalHints(xDisplay,windowElement->xWindow,&sizeHints);

		XResizeWindow(xDisplay,windowElement->xWindow,rect.w,rect.h);	// set the window to the new size, now that child has changed

		InvalidateWindowRect(window,&rect);								// invalidate the entire contents of the window so it will be redrawn
	}
}

static EDITOR_MENU *LocateMenuWithKey(EDITOR_MENU *topLevel,KeySym keySym)
// attempt to locate an active menu below topLevel, that has keySym defined
// if one can be located, return it, else return NULL
{
	EDITOR_MENU
		*currentMenu,
		*locatedMenu;

	currentMenu=GetRelatedEditorMenu(topLevel,MR_FIRSTCHILD);
	locatedMenu=NULL;
	while(currentMenu&&!locatedMenu)
	{
		if(!(locatedMenu=LocateMenuWithKey(currentMenu,keySym)))		// recurse to try to find child that has this key
		{
			if(currentMenu->hasKey&&currentMenu->active)
			{
				if(currentMenu->keySym==keySym)							// see if this item has the key
				{
					locatedMenu=currentMenu;
				}
			}
		}
		currentMenu=GetRelatedEditorMenu(currentMenu,MR_NEXTSIBLING);
	}
	return(locatedMenu);
}

static EDITOR_MENU *LocateKeyPressMenu(XEvent *event)
// return the menu item associated with event, or NULL
// if none
{
	KeySym
		keySym;
	EDITOR_MENU
		*menu;
	XKeyEvent
		ourEvent;

	if(event->xkey.state&Mod1Mask)
	{
		ourEvent=*((XKeyEvent *)event);
		ourEvent.state&=(~(Mod1Mask|LockMask));						// remove these from consideration
		ComposeXLookupString(&ourEvent,NULL,0,&keySym);
		if(!IsModifierKey(keySym))									// do not pass lone modifiers
		{
			if((menu=LocateMenuWithKey(NULL,keySym)))
			{
				if(!menu->firstChild)								// must be leaf on the tree
				{
					return(menu);
				}
			}
		}
	}
	return(NULL);
}

static void UnAssignFont(EDITOR_MENU *menu)
// Get rid of font assigned to a menu item
{
	if(menu->font)
	{
		FreeFont(menu->font);
		menu->font=NULL;				// use the default font
	}
}

static void AssignFont(EDITOR_MENU *menu,const char *fontName)
// given a font name, apply it to menu
// if the menu already has a font assigned to it, get rid of it
// if there is a problem getting fontName, use the defaultMenuFont
{
	if(menu->font)
	{
		FreeFont(menu->font);
	}
	if(!(menu->font=LoadFont(fontName)))
	{
		if(!(menu->font=LoadFont(defaultMenuFont)))
		{
			SetError("Failed to AssignFont");
		}
	}
}

static void UnAssignKeyEquivalent(EDITOR_MENU *menu)
// remove key code assigned to menu
{
	if(menu->hasKey)
	{
		if(menu->commandString)
		{
			MDisposePtr(menu->commandString);
			menu->commandString=NULL;
		}
		menu->hasKey=false;
	}
}

static void AssignKeyEquivalent(EDITOR_MENU *menu,const char *keyName)
// given a key name, assign that key to menu
// if the another menu already has the key assigned to it, get rid of it
// if keyName does not map to a valid key, then the menu will have no key equivalent
{
	KeySym
		keySym;
	EDITOR_MENU
		*oldMenu;

	UnAssignKeyEquivalent(menu);
	if((keySym=XStringToKeysym(keyName))!=NoSymbol)
	{
		if((oldMenu=LocateMenuWithKey(NULL,keySym)))					// see if another menu item had this key, if so, take it
		{
			UnAssignKeyEquivalent(oldMenu);
			ResizeMenuWindow(oldMenu->parent);							// update the old menu
		}
		menu->keySym=keySym;
		if((menu->commandString=(char *)MNewPtr(strlen(keyName)+3)))	// make room for <,>, and 0 terminator
		{
			sprintf(menu->commandString,"<%s>",keyName);				// copy in the command key string
		}
		menu->hasKey=true;
	}
}

static void ExtractAttributeParameter(const char *parameterInput,UINT32 *parameterIndex,char *parameterBuffer,UINT32 parameterBufferSize)
// extract the next parameter from parameterInput[parameterIndex] into parameter buffer
// if the parameter would be too long, just truncate it
// leave paramterIndex updated to point to the end of the given parameter, no matter
// how long it is
{
	UINT32
		outIndex;
	char
		character;
	bool
		done;

	outIndex=0;
	done=false;
	while((character=parameterInput[*parameterIndex])&&!done)	// make sure not end of string, or end of parameter
	{
		if(character!='\\')
		{
			(*parameterIndex)++;
			if(outIndex<parameterBufferSize)
			{
				parameterBuffer[outIndex++]=character;
			}
		}
		else
		{
			if(parameterInput[*parameterIndex+1]=='\\')			// see if quoted separator (not a separator at all then)
			{
				(*parameterIndex)+=2;
				if(outIndex<parameterBufferSize)
				{
					parameterBuffer[outIndex++]='\\';			// place separator into buffer
				}
			}
			else
			{
				done=true;										// ran into end of parameter list
			}
		}
	}
	if(outIndex<parameterBufferSize)
	{
		parameterBuffer[outIndex]='\0';							// terminate buffer
	}
	else
	{
		if(parameterBufferSize)
		{
			parameterBuffer[outIndex-1]='\0';					// terminate buffer by writing over last character
		}
	}
}

static void UnRealizeAttributes(EDITOR_MENU *menu)
// undo anything that RealizeAttributes did
{
	UnAssignFont(menu);
	UnAssignKeyEquivalent(menu);
}

static bool RealizeAttributes(EDITOR_MENU *menu)
// Attempt to fill the parts of menu's record that are described in its string
// Currently supports:
// \Ffontname		specify the font to use when drawing the item
// \S				specify that this item is a menu separation line (text is not drawn)
// \Kkey			specify that mod1 <key> could be used to access the given menu
// if there is a problem, return false
{
	char
		parameterBuffer[256];						// max size for parameter string, anything larger gets truncated!
	UINT32
		attributeIndex;
	char
		character;

	menu->separator=false;
	menu->hasKey=false;
	menu->keySym=NoSymbol;
	menu->commandString=NULL;						// no command string yet
	menu->font=NULL;								// no font yet

	attributeIndex=0;
	while((character=menu->attributes[attributeIndex]))
	{
		if(character!='\\')							// eat separators
		{
			ExtractAttributeParameter(menu->attributes,&attributeIndex,&(parameterBuffer[0]),256);
			switch(parameterBuffer[0])				// check command part
			{
				case 'F':
					AssignFont(menu,&(parameterBuffer[1]));
					break;
				case 'S':
					menu->separator=true;
					break;
				case 'K':
					AssignKeyEquivalent(menu,&(parameterBuffer[1]));
					break;
				default:
					break;
			}
		}
		else
		{
			attributeIndex++;
		}
	}

	if(!menu->font)							// if not font specified, use the default
	{
		AssignFont(menu,defaultMenuFont);
	}
	return(true);
}

static void SetMenuEnableState(EDITOR_MENU *menu)
// set the active state for menu
{
}

static EDITOR_WINDOW *CreateSubMenuWindow(EDITOR_MENU *owner,EDITOR_RECT *rect,const char *title)
// this is called to create a sub menu window for owner
// sub menu windows do not have normal window decoration
// NOTE: this checks to see if the submenu will go off the screen, and pins it
// into range
{
	EDITOR_WINDOW
		*window;
	WINDOW_LIST_ELEMENT
		*windowElement;
	GUI_MENU_WINDOW_INFO
		*menuInfo;
	XSetWindowAttributes
		attributes;
	UINT32
		valueMask;
	Atom
		protocols[2];															// number of protocols

	if((window=(EDITOR_WINDOW *)MNewPtr(sizeof(EDITOR_WINDOW))))				// create editor window
	{
		if((menuInfo=(GUI_MENU_WINDOW_INFO *)MNewPtr(sizeof(GUI_MENU_WINDOW_INFO))))		// make structure to contain menu window info
		{
			if((windowElement=(WINDOW_LIST_ELEMENT *)MNewPtr(sizeof(WINDOW_LIST_ELEMENT))))	// create a window list element
			{
				LinkWindowElement(windowElement);								// link it to the global window list that we maintain
				windowElement->editorWindow=window;								// point to our high level concept of window
				window->windowType=EWT_MENU;									// set the editor window type to "menu"
				window->userData=windowElement;									// point back to the window element
				window->windowInfo=menuInfo;									// menu window info pointer
				menuInfo->ownerMenu=owner;										// point at the owner of this window

				attributes.save_under=True;
				attributes.background_pixmap=None;
				attributes.background_pixel=EditorColorToXPixel(GRAY_2);
				attributes.border_pixmap=None;
				attributes.border_pixel=EditorColorToXPixel(BLACK);
				attributes.override_redirect=True;
				valueMask=CWSaveUnder|CWBackPixmap|CWBackPixel|CWBorderPixmap|CWBorderPixel|CWOverrideRedirect;

				CalculateMenuWindowSize(owner,&rect->w,&rect->h);

				if(rect->w==0)													// ### this is because X will not allow me to open windows of 0 width or height!
				{
					rect->w=1;
				}
				if(rect->h==0)													// ### this is because X will not allow me to open windows of 0 width or height!
				{
					rect->h=1;
				}
				PinMenuRect(rect);												// pin the rectangle to the screen

				// Due to X's nasty error handling, we assume this create does not fail!!!
				windowElement->realTopWindow=windowElement->xWindow=XCreateWindow(xDisplay,RootWindow(xDisplay,xScreenNum),rect->x,rect->y,rect->w,rect->h,0,CopyFromParent,InputOutput,DefaultVisual(xDisplay,xScreenNum),valueMask,&attributes);

				XStoreName(xDisplay,windowElement->xWindow,title);
				XSetIconName(xDisplay,windowElement->xWindow,title);
				XSelectInput(xDisplay,windowElement->xWindow,ExposureMask|ButtonPressMask|ButtonReleaseMask|KeyPressMask|StructureNotifyMask);	// want to see these events

				protocols[0]=deleteWindowAtom;
				XSetWMProtocols(xDisplay,windowElement->xWindow,&(protocols[0]),1);	// tell X we would like to participate in these protocols

				windowElement->invalidRegion=XCreateRegion();					// create new empty invalid region

				windowElement->currentIcon=(Pixmap)NULL;						// no icon for sub-menus
				owner->window=window;											// point this menu at its window
				XDefineCursor(xDisplay,windowElement->xWindow,arrowCursor);		// this is the cursor to use in the submenu
				XMapRaised(xDisplay,windowElement->xWindow);					// map window to the top of the display
				WaitUntilMappedNoFocus(window);									// wait until this window is actually mapped before continuing (olwm likes to mess with the temporal ordering of our windows, this attempts to prevent its inane meddling
				return(window);
			}
			MDisposePtr(menuInfo);
		}
		MDisposePtr(window);
	}
	SetError("Failed to CreateSubMenuWindow");
	return((EDITOR_WINDOW *)NULL);
}

static void DisposeSubMenuWindow(EDITOR_MENU *menu)
// remove the window associated with menu
{
	WINDOW_LIST_ELEMENT
		*windowElement;

	if(menu->window)
	{
		windowElement=(WINDOW_LIST_ELEMENT *)menu->window->userData;
		XDestroyRegion(windowElement->invalidRegion);					// destroy invalid region
		XDestroyWindow(xDisplay,windowElement->xWindow);				// tell x to make the window go away
		WaitUntilDestroyed(menu->window);								// wait for it to die
		UnlinkWindowElement(windowElement);								// unlink it from our list
		MDisposePtr(windowElement);
		MDisposePtr(menu->window->windowInfo);							// get rid of any information hanging on this window
		MDisposePtr(menu->window);
		menu->window=NULL;												// point to no window
	}
}

static EDITOR_WINDOW *CreateRootMenuWindow(EDITOR_RECT *rect,const char *title)
// create a window which will be used as the root of a menu tree
// if there is a problem, SetError, return NULL
{
	EDITOR_WINDOW
		*window;
	WINDOW_LIST_ELEMENT
		*windowElement;
	GUI_MENU_WINDOW_INFO
		*menuInfo;
	XSetWindowAttributes
		attributes;
	UINT32
		valueMask;
	Atom
		protocols[2];													// number of protocols
	XWMHints
		hints;															// hints to tell window manager how to treat these windows
	XSizeHints
		sizeHints;														// window size info (used to REALLY convince X that I know where I want my own window)
	INT32
		xOffset,
		yOffset;
	UINT32
		screenWidth,
		screenHeight;

	if((window=(EDITOR_WINDOW *)MNewPtr(sizeof(EDITOR_WINDOW))))		// create editor window
	{
		if((menuInfo=(GUI_MENU_WINDOW_INFO *)MNewPtr(sizeof(GUI_MENU_WINDOW_INFO))))		// make structure to contain menu window info
		{
			if((windowElement=(WINDOW_LIST_ELEMENT *)MNewPtr(sizeof(WINDOW_LIST_ELEMENT))))	// create a window list element
			{
				LinkWindowElement(windowElement);						// link it to the global window list that we maintain
				windowElement->editorWindow=window;						// point to our high level concept of window
				window->windowType=EWT_MENU;							// set the editor window type to "menu"
				window->userData=windowElement;							// point back to the window element
				window->windowInfo=menuInfo;							// point to the menu info
				menuInfo->ownerMenu=NULL;								// this window is the root, so no menu owns it

				attributes.background_pixmap=None;
				attributes.background_pixel=EditorColorToXPixel(GRAY_2);
				attributes.border_pixmap=None;
				attributes.border_pixel=EditorColorToXPixel(BLACK);
				valueMask=CWBackPixmap|CWBackPixel|CWBorderPixmap|CWBorderPixel;

				CalculateMenuWindowSize(NULL,&rect->w,&rect->h);
				if(rect->w==0)											// ### this is because X will not allow me to open windows of 0 width or height!
				{
					rect->w=1;
				}
				if(rect->h==0)											// ### this is because X will not allow me to open windows of 0 width or height!
				{
					rect->h=1;
				}
				// Due to X's nasty error handling, we assume this create does not fail!!!

				GetEditorScreenOffsetAndDimensions(&xOffset,&yOffset,&screenWidth,&screenHeight);
				windowElement->realTopWindow=windowElement->xWindow=XCreateWindow(xDisplay,RootWindow(xDisplay,xScreenNum),rect->x+xOffset,rect->y+yOffset,rect->w,rect->h,0,CopyFromParent,InputOutput,DefaultVisual(xDisplay,xScreenNum),valueMask,&attributes);

				sizeHints.flags=USSize|USPosition|PMinSize|PMaxSize;	// ### LIE to X, tell it the user requested the size and position so it will honor them
				sizeHints.x=rect->x;
				sizeHints.y=rect->y;
				sizeHints.width=rect->w;
				sizeHints.height=rect->h;
				sizeHints.min_width=1;
				sizeHints.min_height=1;
				sizeHints.max_width=1;
				sizeHints.max_height=1;
				XSetWMNormalHints(xDisplay,windowElement->xWindow,&sizeHints);

				XStoreName(xDisplay,windowElement->xWindow,title);
				XSetIconName(xDisplay,windowElement->xWindow,title);
				XSelectInput(xDisplay,windowElement->xWindow,ExposureMask|ButtonPressMask|ButtonReleaseMask|FocusChangeMask|KeyPressMask|StructureNotifyMask|PropertyChangeMask);	// want to see these events

				protocols[0]=takeFocusAtom;
				protocols[1]=deleteWindowAtom;
				XSetWMProtocols(xDisplay,windowElement->xWindow,&(protocols[0]),2);	// tell X we would like to participate in these protocols

				hints.flags=InputHint|StateHint|WindowGroupHint|IconPixmapHint;
				hints.input=True;
				hints.initial_state=NormalState;
				hints.window_group=windowElement->xWindow;
				hints.icon_pixmap=editorIconPixmap;						// point at application icon
				XSetWMHints(xDisplay,windowElement->xWindow,&hints);

				windowElement->currentIcon=hints.icon_pixmap;			// remember the icon in use

				windowElement->invalidRegion=XCreateRegion();			// create new empty invalid region

				SetClassHints(window);

				XMapRaised(xDisplay,windowElement->xWindow);			// map window to the top of the display
				WaitUntilMapped(window);								// wait until this window is actually mapped before continuing (olwm likes to mess with the temporal ordering of our windows, this attempts to prevent its inane meddling
				return(window);
			}
			MDisposePtr(menuInfo);
		}
		MDisposePtr(window);
	}
	SetError("Failed to CreateRootMenuWindow");
	return((EDITOR_WINDOW *)NULL);
}

static void DisposeRootMenuWindow(EDITOR_WINDOW *window)
// close a given root menu window
{
	WINDOW_LIST_ELEMENT
		*windowElement;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;			// get the x window associated with this document window
	XDestroyRegion(windowElement->invalidRegion);					// destroy invalid region
	FocusAwayFrom(window);
	XDestroyWindow(xDisplay,windowElement->xWindow);				// tell x to make the window go away
	WaitUntilDestroyed(window);										// wait for it to die
	UnlinkWindowElement(windowElement);								// unlink it from our list
	MDisposePtr(windowElement);
	MDisposePtr(window->windowInfo);
	MDisposePtr(window);
}

void DeactivateEditorMenu(EDITOR_MENU *menu)
// make menu inactive
// if menu contains sub-Menus, they should be inactive because
// their parent is now inactive, but they should not have their active
// state changed. This way, when the parent becomes active again, the children
// will resume their old states
{
	menu->active=false;
	SetMenuEnableState(menu);
}

void ActivateEditorMenu(EDITOR_MENU *menu)
// make menu active
{
	menu->active=true;
	SetMenuEnableState(menu);
}

char *GetEditorMenuName(EDITOR_MENU *menu)
// return a pointer to the name of menu
{
	return(menu->itemText);
}

bool SetEditorMenuName(EDITOR_MENU *menu,const char *name)
// set the name for menu
// if this fails, leave the old name, set the error, and return false
{
	char
		*newName;
	if((newName=(char *)MNewPtr(strlen(name)+1)))
	{
		strcpy(newName,name);
		MDisposePtr(menu->itemText);
		menu->itemText=newName;
		return(true);
	}
	return(false);
}

char *GetEditorMenuAttributes(EDITOR_MENU *menu)
// return a pointer to the attributes string for menu
{
	return(menu->attributes);
}

bool SetEditorMenuAttributes(EDITOR_MENU *menu,const char *attributes)
// set the attributes for menu
// if this fails, set the error, and return false
{
	char
		*newAttributes;

	if((newAttributes=(char *)MNewPtr(strlen(attributes)+1)))
	{
		strcpy(newAttributes,attributes);
		MDisposePtr(menu->attributes);
		menu->attributes=newAttributes;
		return(true);
	}
	return(false);
}

char *GetEditorMenuDataText(EDITOR_MENU *menu)
// return a pointer to the data text string for menu
{
	return(menu->dataText);
}

bool SetEditorMenuDataText(EDITOR_MENU *menu,const char *dataText)
// set the data text for menu
// if this fails, leave the data text, set the error, and return false
{
	char
		*newDataText;

	if((newDataText=(char *)MNewPtr(strlen(dataText)+1)))
	{
		strcpy(newDataText,dataText);
		MDisposePtr(menu->dataText);
		menu->dataText=newDataText;
		return(true);
	}
	return(false);
}

bool GetEditorMenu(int argc,const char *argv[],EDITOR_MENU **menu)
// traverse the elements of the menu path given by argc,argv
// and attempt to locate the menu that they point to
// if the path is invalid, return false
// otherwise return true with menu set to the located item
// NOTE: on entry, menu is used as the starting point for
// the search, if it is passed as NULL, it indicates that the search
// should start at the top
{
	EDITOR_MENU
		*currentMenu,
		*testMenu;
	int
		elementIndex;
	bool
		matching;

	currentMenu=testMenu=*menu;
	matching=true;
	for(elementIndex=0;(elementIndex<argc)&&matching;elementIndex++)
	{
		if(currentMenu)							// locate root of current menu
		{
			testMenu=currentMenu->firstChild;
		}
		else
		{
			testMenu=firstMenu;
		}
		matching=false;
		while(testMenu&&!matching)
		{
			if(strcmp(argv[elementIndex],testMenu->itemText)==0)
			{
				matching=true;					// located a match, this is the menu we want
			}
			else
			{
				testMenu=GetRelatedEditorMenu(testMenu,MR_NEXTSIBLING);
			}
		}
		currentMenu=testMenu;
	}
	*menu=testMenu;
	return(matching);
}

static bool LinkEditorMenuElement(EDITOR_MENU *menu,EDITOR_MENU *baseMenu,int menuRelationship)
// link menu to baseMenu with menuRelationship
// if baseMenu is NULL, then menuRelationship must be MR_FIRSTCHILD, or MR_LASTCHILD as it is not possible to
// link menus as siblings of the root
// if there is a problem, Set the error, and return false
{
	bool
		fail;

	fail=false;
	menu->firstChild=NULL;
	switch(menuRelationship)
	{
		case MR_NEXTSIBLING:
			if(baseMenu)
			{
				menu->nextSibling=baseMenu->nextSibling;
				menu->nextSibling->previousSibling=menu;
				menu->previousSibling=baseMenu;
				baseMenu->nextSibling=menu;
				menu->parent=baseMenu->parent;
			}
			else
			{
				SetError(errorDescriptions[BAD_RELATIONSHIP]);
				fail=true;
			}
			break;
		case MR_PREVIOUSSIBLING:
			if(baseMenu)
			{
				menu->previousSibling=baseMenu->previousSibling;
				menu->previousSibling->nextSibling=menu;
				menu->nextSibling=baseMenu;
				baseMenu->previousSibling=menu;
				menu->parent=baseMenu->parent;
				if(menu->parent)
				{
					if(menu->parent->firstChild==baseMenu)
					{
						menu->parent->firstChild=menu;		// make us the first child if linked before old first child
					}
				}
				else
				{
					if(firstMenu==baseMenu)
					{
						firstMenu=menu;
					}
				}
			}
			else
			{
				SetError(errorDescriptions[BAD_RELATIONSHIP]);
				fail=true;
			}
			break;
		case MR_FIRSTCHILD:
			menu->parent=baseMenu;								// set parent
			if(baseMenu)										// see if linking as a child of a current menu
			{
				if((menu->nextSibling=baseMenu->firstChild))
				{
					menu->previousSibling=menu->nextSibling->previousSibling;
					menu->nextSibling->previousSibling=menu;	// link me in as previous sibling of old first child if there was one
					menu->previousSibling->nextSibling=menu;	// link me in as next sibling
				}
				else
				{
					menu->previousSibling=menu;					// loop previous sibling to myself
					menu->nextSibling=menu;						// loop next sibling to myself
				}
				baseMenu->firstChild=menu;						// this is now the first child
			}
			else												// creating a child of the universe
			{
				if(firstMenu)
				{
					menu->nextSibling=firstMenu;
					menu->previousSibling=firstMenu->previousSibling;
					menu->previousSibling->nextSibling=menu;
					firstMenu->previousSibling=menu;
				}
				else
				{
					menu->previousSibling=menu;
					menu->nextSibling=menu;
				}
				firstMenu=menu;
			}
			break;
		case MR_LASTCHILD:
			menu->parent=baseMenu;								// set parent
			if(baseMenu)										// see if linking as a child of a current menu
			{
				if((menu->nextSibling=baseMenu->firstChild))
				{
					menu->previousSibling=menu->nextSibling->previousSibling;
					menu->nextSibling->previousSibling=menu;	// link me in as previous sibling of old first child if there was one
					menu->previousSibling->nextSibling=menu;	// link me in as next sibling
				}
				else
				{
					menu->previousSibling=menu;					// loop previous sibling to myself
					menu->nextSibling=menu;						// loop next sibling to myself
					baseMenu->firstChild=menu;
				}
			}
			else												// creating a child of the universe
			{
				if(firstMenu)
				{
					menu->nextSibling=firstMenu;
					menu->previousSibling=firstMenu->previousSibling;

					menu->previousSibling->nextSibling=menu;

					firstMenu->previousSibling=menu;
				}
				else
				{
					menu->previousSibling=menu;
					menu->nextSibling=menu;
					firstMenu=menu;
				}
			}
			break;
		default:
			SetError(errorDescriptions[BAD_RELATIONSHIP]);
			fail=true;
			break;
	}
	return(!fail);
}

static void UnlinkEditorMenuElement(EDITOR_MENU *menu)
// unlink menu from the menu tree
{
	if(menu->parent)
	{
		if(menu->parent->firstChild==menu)
		{
			if(menu->nextSibling!=menu)
			{
				menu->parent->firstChild=menu->nextSibling;
			}
			else
			{
				menu->parent->firstChild=NULL;
			}
		}
	}
	else
	{
		if(firstMenu==menu)
		{
			if(menu->nextSibling!=menu)
			{
				firstMenu=menu->nextSibling;
			}
			else
			{
				firstMenu=NULL;
			}
		}
	}
	menu->nextSibling->previousSibling=menu->previousSibling;
	menu->previousSibling->nextSibling=menu->nextSibling;
}

static EDITOR_MENU *CreateEditorMenuElement(const char *name,const char *attributes,const char *dataText)
// create an editor menu structure (not linked to anything, but fill in
// name, and attributes)
// if there is a problem, set the error, and return NULL
{
	EDITOR_MENU
		*menu;

	if((menu=(EDITOR_MENU *)MNewPtr(sizeof(EDITOR_MENU))))				// create editor menu element
	{
		if((menu->itemText=(char *)MNewPtr(strlen(name)+1)))
		{
			strcpy(menu->itemText,name);								// copy the name
			if((menu->attributes=(char *)MNewPtr(strlen(attributes)+1)))
			{
				strcpy(menu->attributes,attributes);					// copy the attributes
				if((menu->dataText=(char *)MNewPtr(strlen(dataText)+1)))
				{
					strcpy(menu->dataText,dataText);					// copy the data text for this menu
					menu->active=menu->highlight=false;					// item is not active, and not highlighted by default
					menu->window=NULL;									// this menu currently has no window
					if(RealizeAttributes(menu))							// parse through the attribute string, set up font, and command string
					{
						return(menu);
					}
					MDisposePtr(menu->dataText);
				}
				MDisposePtr(menu->attributes);
			}
			MDisposePtr(menu->itemText);
		}
		MDisposePtr(menu);
	}
	return(NULL);
}

static void DisposeEditorMenuElement(EDITOR_MENU *menu)
// dispose of an editor menu structure
{
	DisposeSubMenuWindow(menu);
	UnRealizeAttributes(menu);
	MDisposePtr(menu->dataText);
	MDisposePtr(menu->attributes);
	MDisposePtr(menu->itemText);
	MDisposePtr(menu);
}

static EDITOR_MENU *CreateEditorMenuAtBase(EDITOR_MENU *baseMenu,int menuRelationship,const char *name,const char *attributes,const char *dataText,bool active)
// create a new menu with name, as menuRelationship to baseMenu
// and return a pointer to it
// if baseMenu is NULL it indicates the top of the hierarchy
// therefore, menuRelationship must be MR_FIRSTCHILD, or MR_LASTCHILD as it is not possible to
// create menus which are siblings of the root
// attributes is a machine specific string that describes machine specific attributes of the menu
// if active is false, the menu should be created inactive
// if there is a problem, set the error, and return NULL
{
	EDITOR_MENU
		*menu;

	if((menu=CreateEditorMenuElement(name,attributes,dataText)))
	{
		menu->active=active;										// set the active state
		if(LinkEditorMenuElement(menu,baseMenu,menuRelationship))
		{
			ResizeMenuWindow(menu->parent);							// if the parent of menu has a window, then update the window so that this child item appears
			if(menu->parent)
			{
				if((menu->previousSibling==menu)&&(menu->nextSibling==menu))	// if this is the only child of the parent, then the parent needs to be updated to show he has children
				{
					ResizeMenuWindow(menu->parent->parent);			// creating this item has forced my parent to change (to show he has children)
				}
			}
			return(menu);
		}
		DisposeEditorMenuElement(menu);
	}
	return(NULL);
}

static void InvalidateMenuRect(EDITOR_MENU *menu,EDITOR_RECT *rect)
// invalidate the given rectangle of menu's window if menu has
// a window
// NULL can be passed as menu, and indicates the root menu
{
	EDITOR_WINDOW
		*window;

	if((window=GetMenuWindow(menu)))
	{
		InvalidateWindowRect(window,rect);		// invalidate the rectangle
	}
}

static void InvalidateMenuItem(EDITOR_MENU *menu)
// if menu is within a parent window, invalidate the rectangle of
// the parent that corresponds to menu
// it is incorrect to call this with menu==NULL
{
	EDITOR_RECT
		rect;

	CalculateMenuItemRect(menu,&rect);		// find the rectangle of this item within its parent
	InvalidateMenuRect(menu->parent,&rect);	// invalidate this rectangle of the parent's window (if he has one)
}

static void CloseSubMenus(EDITOR_MENU *menu)
// close all of the windows for all inferiors of menu
// the close is done depth first
{
	EDITOR_MENU
		*currentMenu;

	currentMenu=GetRelatedEditorMenu(menu,MR_FIRSTCHILD);
	while(currentMenu)
	{
		if(currentMenu->window)
		{
			CloseSubMenus(currentMenu);					// close down any this menu has first
			DisposeSubMenuWindow(currentMenu);
		}
		if(currentMenu->highlight)
		{
			currentMenu->highlight=false;				// unhighlight this menu
			InvalidateMenuItem(currentMenu);
		}
		currentMenu=GetRelatedEditorMenu(currentMenu,MR_NEXTSIBLING);
	}
}

static void OpenSubMenu(EDITOR_MENU *menu)
// if menu is hierarchical, call this to open its submenu
// an attempt is made to open the submenu to the right and down of menu
// if it does not fit on the screen, some other method should be implemented
{
	EDITOR_RECT
		rect;
	EDITOR_WINDOW
		*window;

	menu->highlight=true;								// highlight this menu
	InvalidateMenuItem(menu);
	if((window=GetMenuWindow(menu->parent))&&menu->firstChild)	// find parent's window, make sure this one has children
	{
		CalculateMenuItemRect(menu,&rect);				// get my rectangle within it
		LocalToGlobal(((WINDOW_LIST_ELEMENT *)window->userData)->xWindow,rect.x+rect.w,rect.y,&rect.x,&rect.y);	// get global location for new window
		rect.w=rect.h=0;								// initialize to zero
		CreateSubMenuWindow(menu,&rect,"submenu");
	}
}

void RealizeEditorMenu(EDITOR_MENU* menu)
// Realize menu attributes (and any children it may have)
// if NULL is passed, realize ALL menus
// NOTE: this calls itself recursively
{
	EDITOR_MENU
		*subMenu;

	subMenu=GetRelatedEditorMenu(menu,MR_FIRSTCHILD);	// see if this one has any children
	while(subMenu)
	{
		RealizeEditorMenu(subMenu);
		subMenu=GetRelatedEditorMenu(subMenu,MR_NEXTSIBLING);
	}
	if(menu)
	{
		UnRealizeAttributes(menu);
		RealizeAttributes(menu);
	}
	ResizeMenuWindow(menu);
}

void DisposeEditorMenu(EDITOR_MENU *menu)
// remove menu (and any children it may have)
// if NULL is passed, dispose of ALL menus
// NOTE: this calls itself recursively
{
	EDITOR_MENU
		*parent,
		*subMenu;

	while((subMenu=GetRelatedEditorMenu(menu,MR_LASTCHILD)))	// see if this one has any children
	{
		DisposeEditorMenu(subMenu);
	}
	if(menu)
	{
		parent=menu->parent;								// remember the parent, so after this is disposed of, he can be updated
		UnlinkEditorMenuElement(menu);						// unlink this item
		DisposeEditorMenuElement(menu);						// get rid of it
		if(parent)											// see if parent is not root
		{
			if(parent->window&&!parent->firstChild)			// if the parent has a window, but now has no children, kill the window
			{
				DisposeSubMenuWindow(parent);				// get rid of the window if it contained the last child
			}
			if(!parent->firstChild)
			{
				ResizeMenuWindow(parent->parent);			// tell the parent's parent to resize (to show that there are no children)
			}
		}
		ResizeMenuWindow(parent);							// if the parent of menu has a window, then update the window so that this child item disappears
	}
}

EDITOR_MENU *CreateEditorMenu(int argc,const char *argv[],int menuRelationship,const char *name,const char *attributes,const char *dataText,bool active)
// create a new menu with name, as menuRelationship to the passed argc,argv path
// the path is absolute within the menu hierarchy
// attributes is a machine specific string that describes machine specific attributes of the menu
// if active is false, the menu should be created inactive
// if there is a problem, set the error, and return false
{
	EDITOR_MENU
		*parentMenu,
		*testMenu;
	bool
		fail;

	parentMenu=NULL;									// start at the top
	if(GetEditorMenu(argc,argv,&parentMenu))			// locate item to link this one to
	{
		testMenu=parentMenu;
		fail=false;
		if(menuRelationship==MR_NEXTSIBLING||menuRelationship==MR_PREVIOUSSIBLING)	// see if hunting for sibling
		{
			if(parentMenu)								// if he wants sibling, better not be at top
			{
				testMenu=testMenu->parent;				// back up one
				if(GetEditorMenu(1,&name,&testMenu))	// see if an item already exists with this name in the place we are putting it
				{
					if(testMenu!=parentMenu)			// cannot delete the one we want to add to!!
					{
						DisposeEditorMenu(testMenu);	// kill it before adding new item
					}
					else
					{
						SetError(errorDescriptions[BAD_RELATIONSHIP]);
						fail=true;
					}
				}
			}
			else
			{
				SetError(errorDescriptions[BAD_RELATIONSHIP]);
				fail=true;
			}
		}
		else
		{
			if(GetEditorMenu(1,&name,&testMenu))			// see if an item already exists with this name in the place we are putting it
			{
				DisposeEditorMenu(testMenu);				// kill it before adding new item
			}
		}
		if(!fail)
		{
			return(CreateEditorMenuAtBase(parentMenu,menuRelationship,name,attributes,dataText,active));
		}
	}
	else
	{
		SetError(errorDescriptions[BAD_PATH]);
	}
	return(NULL);
}

static void PressMenuItem(EDITOR_MENU *menu)
// when a menu has been located, do what is needed to highlight it, and also
// if it has a submenu, open it (closing any submenus of any other menus at this level)
{
	if(!menu->window)							// if this one does not have a window, close any other windows at this level
	{
		CloseSubMenus(menu->parent);			// close any open submenus at this level
		OpenSubMenu(menu);						// open, or at least highlight this item
		UpdateEditorWindows();					// force it to update
	}
	else
	{
		if(!menu->highlight)					// if it has a window, it may still be unhighlighted, so highlight it
		{
			menu->highlight=true;				// highlight this menu
			InvalidateMenuItem(menu);
			UpdateEditorWindows();				// force it to update
		}
	}
}

static void UnpressMenuItem(EDITOR_MENU *menu)
// when tracking has left a menu item, call this to unhighlight it
// if an item has submenus open, it is not unpressed
{
	if((!menu->window)&&menu->highlight)
	{
		menu->highlight=false;					// unhighlight this menu
		InvalidateMenuItem(menu);
		UpdateEditorWindows();					// force it to update
	}
}

static EDITOR_MENU *LocateMenuItem(EDITOR_MENU *parent,INT32 x,INT32 y)
// locate the item (if any) that X and Y are over in parent
// X and Y are relative to parent
{
	EDITOR_MENU
		*currentMenu;
	EDITOR_RECT
		rect;
	bool
		found;

	currentMenu=GetRelatedEditorMenu(parent,MR_FIRSTCHILD);
	found=false;
	while(currentMenu&&!found)
	{
		CalculateMenuItemRect(currentMenu,&rect);
		if(currentMenu->active&&!currentMenu->separator&&PointInRECT(x,y,&rect))
		{
			found=true;
		}
		else
		{
			currentMenu=GetRelatedEditorMenu(currentMenu,MR_NEXTSIBLING);
		}
	}
	return(currentMenu);
}

static bool TrackLocateMenu(EDITOR_MENU *topLevel,INT32 rootX,INT32 rootY,EDITOR_MENU **menu)
// given an X, and Y that are relative to the X root, run through the open
// menu window list, and return the menu item that the X and Y are over.
// if X,Y are not over any item, return menu=NULL
// if false is returned, the given point did not fall in topLevel, or any of its sub windows
{
	EDITOR_MENU
		*currentMenu;
	INT32
		localX,
		localY;
	bool
		foundHigher;
	EDITOR_WINDOW
		*window;
	XWindowAttributes
		attributes;

	(*menu)=NULL;
	foundHigher=false;
	currentMenu=GetRelatedEditorMenu(topLevel,MR_FIRSTCHILD);
	while(currentMenu&&!foundHigher)
	{
		if(currentMenu->window)
		{
			foundHigher=TrackLocateMenu(currentMenu,rootX,rootY,menu);
		}
		currentMenu=GetRelatedEditorMenu(currentMenu,MR_NEXTSIBLING);
	}
	if(!foundHigher)
	{
		if((window=GetMenuWindow(topLevel)))
		{
			XGetWindowAttributes(xDisplay,((WINDOW_LIST_ELEMENT *)window->userData)->xWindow,&attributes);
			GlobalToLocal(((WINDOW_LIST_ELEMENT *)window->userData)->xWindow,rootX,rootY,&localX,&localY);	// convert global coordinates to those local to this window
			if(localX>=0&&localY>=0&&localX<attributes.width&&localY<attributes.height)
			{
				(*menu)=LocateMenuItem(topLevel,localX,localY);	// see if we can find an item here (either way, we are done because we saw the point in our window)
				foundHigher=true;
			}
		}
	}
	return(foundHigher);
}

static EDITOR_MENU *MenuWindowButtonPress(EDITOR_WINDOW *window,XEvent *event)
// handle button press events for menu windows (see what the user is clicking
// on, and return it)
// window is passed in as the root menu window
// NOTE: this does not close submenus when it exits, and it leaves the selected
// menu pressed
{
	WINDOW_LIST_ELEMENT
		*windowElement;
	EDITOR_MENU
		*menu,
		*lastMenu;
	Window
		root,
		child;
	int
		rootX,
		rootY,
		windowX,
		windowY,
		lastX,
		lastY;
	unsigned int
		state;
	int
		button;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	button=event->xbutton.button;
	TrackLocateMenu(NULL,event->xbutton.x_root,event->xbutton.y_root,&menu);	// find the menu item (if any) that was clicked on
	lastMenu=menu;
	if(menu)
	{
		PressMenuItem(menu);
	}
	lastX=event->xbutton.x_root;
	lastY=event->xbutton.y_root;
	while(StillDown(button,1))
	{
		if(XQueryPointer(xDisplay,windowElement->xWindow,&root,&child,&rootX,&rootY,&windowX,&windowY,&state))
		{
			if(lastX!=rootX||lastY!=rootY)					// make sure something has changed
			{
				TrackLocateMenu(NULL,rootX,rootY,&menu);
				if(menu!=lastMenu)
				{
					if(lastMenu)
					{
						UnpressMenuItem(lastMenu);			// unpress the item that we are leaving
					}
					if(menu)
					{
						PressMenuItem(menu);				// if a new item was located, press it
					}
					lastMenu=menu;
				}
				lastX=rootX;
				lastY=rootY;
			}
		}
		else
		{
			if(lastMenu)
			{
				UnpressMenuItem(lastMenu);					// left the screen, so unpress last
				lastMenu=menu=NULL;
			}
		}
	}
	return(menu);
}

bool AttemptMenuKeyPress(XEvent *event)
// see if the keypress event can be handled by one of the menu items
// if so, handle it, and return true
// if no menu item wants to handle the key, return false
{
	EDITOR_MENU
		*menu;

	if((menu=LocateKeyPressMenu(event)))
	{
		HandleMenuEvent(menu);								// call the shell to handle this menu
		return(true);
	}
	return(false);
}

static Bool MenuEventsPredicate(Display *display,XEvent *event,char *arg)
// match events we want while waiting for menus to be selected
{
	if((event->type==ButtonPress)||(event->type==ButtonRelease)||(event->type==KeyPress))
	{
		return(True);
	}
	return(False);
}

static EDITOR_MENU *GetFinalMenu(Time time)
// user has selected a menu item that has a sub-menu, so, collect and
// process events, attempting to complete the menu selection process
// NOTE: this will not return unless the user cancels (by clicking outside
// the menus), or selects an item which does not have a submenu
{
	EDITOR_MENU
		*menu;
	XEvent
		event;
	bool
		done;

	menu=NULL;											// cannot grab pointer, then cancel menu choose
	if(XGrabPointer(xDisplay,RootWindow(xDisplay,xScreenNum),True,ButtonPressMask|ButtonReleaseMask,GrabModeAsync,GrabModeAsync,None,None,time)==GrabSuccess)
	{
		done=false;
		while(!done)
		{
			if(XCheckIfEvent(xDisplay,&event,MenuEventsPredicate,NULL))
			{
				switch(event.type)
				{
					case KeyPress:
						if((menu=LocateKeyPressMenu(&event)))	// if there is a menu for this key, then get it
						{
							time=event.xkey.time;
							done=(!menu->firstChild);
						}
						break;
					case ButtonPress:
						menu=MenuWindowButtonPress(rootMenuWindow,&event);	// go handle the button press
						if(!menu||!menu->firstChild)
						{
							time=event.xbutton.time;
							done=true;
						}
						break;
				}
			}
			else
			{
				EditorDoBackground();					// handle background stuff while waiting for menu selection
			}
		}
		XUngrabPointer(xDisplay,time);
	}
	return(menu);
}

void MenuWindowEvent(EDITOR_WINDOW *window,XEvent *event)
// handle an event arriving for a menu window
{
	EDITOR_MENU
		*menu;
	WINDOW_LIST_ELEMENT
		*windowElement;

	windowElement=(WINDOW_LIST_ELEMENT *)window->userData;
	switch(event->type)
	{
		case ButtonPress:
			XRaiseWindow(xDisplay,windowElement->xWindow);		// make sure menu is on top
			if((menu=MenuWindowButtonPress(window,event)))		// go handle the button press
			{
				if(menu->firstChild)							// pressed button that got us to a menu, go try to get final menu press
				{
					menu=GetFinalMenu(event->xbutton.time);		// enter event loop waiting for final menu selection, or cancellation
				}
			}
			CloseSubMenus(NULL);								// get rid of sub menus
			if(menu)
			{
				UpdateEditorWindows();							// make windows update before handling menu
				HandleMenuEvent(menu);							// call the shell to handle this menu
			}
			break;
		case KeyPress:
			// all keys for menus are attempted outside, if we get a keypress, it means we dont want it!
			break;
		case ClientMessage:
			if(event->xclient.data.l[0]==(int)deleteWindowAtom)
			{
				HandleShellCommand("quit",0,NULL);
			}
			break;
		case SelectionRequest:
			HandleSelectionRequest(event);
			break;
		default:
			break;
	}
}

void RaiseRootMenu()
// X specific menu function to pull menus to the top of the window pile
{
	XRaiseWindow(xDisplay,((WINDOW_LIST_ELEMENT *)(rootMenuWindow->userData))->xWindow);	// raise the root menu window to the top
}

void UnInitEditorMenus()
// uninitialize the editor menuing system
{
	DisposeEditorMenu(NULL);									// get rid of any menus that were created
	DisposeRootMenuWindow(rootMenuWindow);						// get rid of the root of the menu windows
}

bool InitEditorMenus()
// initialize the editor menuing system
// if there is a problem, SetError, return false
{
	EDITOR_RECT
		rect;

	firstMenu=NULL;												// no menus exist yet
	rect.x=10;													// this is the desired place for the main menu window
	rect.y=20;
	rect.w=1;													// make tiny window, because we have no menus
	rect.h=1;
	if((rootMenuWindow=CreateRootMenuWindow(&rect,programName)))	// attempt to create the root of the menu windows
	{
		return(true);
	}
	SetError("Failed to InitEditorMenus");
	return(false);
}
