// Separator handling (draw separation lines for dialogs)
// Copyright (C) 2000 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

typedef struct separator
{
	EDITOR_WINDOW
		*parentWindow;
	separatorResizeFunction
		*resizeFunction;
	bool
		horizontal;												// true if separator is horizontal
	INT32
		x,
		y;
	UINT32
		length;
} SEPARATOR;

static void InvalidateSeparator(SEPARATOR *separator)
// invalidate separator within its parent window
{
	EDITOR_RECT
		rect;

	rect.x=separator->x;
	rect.y=separator->y;

	if(separator->horizontal)
	{
		rect.w=separator->length;
		rect.h=2;
	}
	else
	{
		rect.w=2;
		rect.h=separator->length;
	}
	InvalidateWindowRect(separator->parentWindow,&rect);
}

static void SetSeparatorSize(SEPARATOR *separator)
// Work out the size given the current window
{
	if(separator->resizeFunction)	// see if non-null resize function specified
	{
		separator->resizeFunction(separator->parentWindow,&separator->x,&separator->y,&separator->length);
		InvalidateSeparator(separator);								// invalidate it
	}
}

static void DrawSeparator(SEPARATOR *separator,Pixmap imageBuffer,Region updateRegion)
// redraw separator in its parent window, obey the invalid area of the parent's window
{
	XSetRegion(xDisplay,xGraphicsContext,updateRegion);							// clip to area we wish to draw in

	if(separator->horizontal)
	{
		XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(BLACK));
		XDrawLine(xDisplay,imageBuffer,xGraphicsContext,separator->x,separator->y,separator->x+separator->length,separator->y);
		XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(WHITE));
		XDrawLine(xDisplay,imageBuffer,xGraphicsContext,separator->x,separator->y+1,separator->x+separator->length,separator->y+1);
	}
	else
	{
		XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(BLACK));
		XDrawLine(xDisplay,imageBuffer,xGraphicsContext,separator->x,separator->y,separator->x,separator->y+separator->length);
		XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(WHITE));
		XDrawLine(xDisplay,imageBuffer,xGraphicsContext,separator->x+1,separator->y,separator->x+1,separator->y+separator->length);
	}
	XSetClipMask(xDisplay,xGraphicsContext,None);								// get rid of clip mask
}

static SEPARATOR *CreateSeparator(EDITOR_WINDOW *window,SEPARATOR_DESCRIPTOR *description)
// create a separator item in window, invalidate the area of window
// where the separator is to be placed
// return a pointer to it if successful
// SetError, and return NULL if not
{
	SEPARATOR
		*separator;

	if((separator=(SEPARATOR *)MNewPtr(sizeof(SEPARATOR))))			// create data structure
	{
		separator->parentWindow=window;
		separator->resizeFunction=description->resizeFunction;
		separator->horizontal=description->horizontal;

		SetSeparatorSize(separator);
		return(separator);
	}
	return(NULL);
}

static void DisposeSeparator(SEPARATOR *separator)
// unlink separator from its parent window, and delete it
// invalidate the area of the window where separator was
{
	InvalidateSeparator(separator);									// invalidate where it is now
	MDisposePtr(separator);
}

// dialog interface routines

static void DrawSeparatorItem(DIALOG_ITEM *item,Pixmap imageBuffer,Region updateRegion)
// draw the separator pointed to by item
{
	DrawSeparator((SEPARATOR *)item->itemStruct,imageBuffer,updateRegion);
}

static void ResizeSeparatorItem(DIALOG_ITEM *item)
// When the dialog window size changes, call this to update the item itself
// This will invalidate the old and new item areas
{
	SEPARATOR
		*separator;

	separator=(SEPARATOR *)item->itemStruct;
	InvalidateSeparator(separator);				// invalidate where we're at currently
	SetSeparatorSize(separator);
}

static void DisposeSeparatorItem(DIALOG_ITEM *item)
// dispose of the local separator structure pointed to by item
{
	DisposeSeparator((SEPARATOR *)item->itemStruct);
}

bool CreateSeparatorItem(DIALOG_ITEM *item,void *description)
// create the separator, and fill in item
// if there is a problem, SetError, return false
{
	if((item->itemStruct=(void *)CreateSeparator(item->dialog->parentWindow,(SEPARATOR_DESCRIPTOR *)description)))
	{
		item->disposeProc=DisposeSeparatorItem;
		item->drawProc=DrawSeparatorItem;
		item->resizeProc=ResizeSeparatorItem;
		item->trackProc=NULL;
		item->earlyKeyProc=NULL;
		item->wantFocus=false;
		item->focusChangeProc=NULL;
		item->focusKeyProc=NULL;
		item->focusPeriodicProc=NULL;
		return(true);
	}
	return(false);
}
