// Static text handler for dialogs
// Copyright (C) 2000 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

#define	MAXSTATIC_TEXTSTRING	512

typedef struct staticText
{
	EDITOR_WINDOW
		*parentWindow;											// the window that contains this text
	EDITOR_COLOR
		backgroundColor,										// color to use when drawing background
		foregroundColor;										// color to use when drawing foreground
	staticTextResizeFunction
		*resizeFunction;										// function which is used to resize for this window
	EDITOR_RECT
		rect;
	EDITOR_FONT
		*font;													// info about currently selected font
	char
		string[MAXSTATIC_TEXTSTRING];
} STATIC_TEXT;

static void InvalidateStaticText(STATIC_TEXT *text)
// invalidate text within its parent window
{
	InvalidateWindowRect(text->parentWindow,&text->rect);
}

static void SetStaticTextRectangle(STATIC_TEXT *text)
// Work out the rectangle given the current window
{
	if(text->resizeFunction)	// see if non-null resize function specified
	{
		text->resizeFunction(text->parentWindow,&text->rect);	// get new rectangle
		InvalidateStaticText(text);								// invalidate it
	}
}

static void DrawStaticText(STATIC_TEXT *text,Pixmap imageBuffer,Region updateRegion)
// redraw text in its parent window, obey the invalid area of the parent's window
{
	XRectangle
		textRect;
	Region
		textRegion;
	int
		stringLength,
		currentIndex,
		lastIndex;
	int
		stringX,
		stringY;

	textRect.x=text->rect.x;
	textRect.y=text->rect.y;
	textRect.width=text->rect.w;
	textRect.height=text->rect.h;
	textRegion=XCreateRegion();
	XUnionRectWithRegion(&textRect,textRegion,textRegion);						// make region for the text we are about to draw
	XIntersectRegion(textRegion,updateRegion,textRegion);						// see if it intersects with the invalid region

	XSetRegion(xDisplay,xGraphicsContext,textRegion);							// clip to area we wish to draw text in

	XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(text->backgroundColor));
	XFillRectangle(xDisplay,imageBuffer,xGraphicsContext,text->rect.x,text->rect.y,text->rect.w,text->rect.h);

	XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(text->foregroundColor));
	XSetFont(xDisplay,xGraphicsContext,text->font->xFont->fid);					// point to the current font for this text

	stringX=text->rect.x;
	stringY=text->rect.y+text->font->ascent;

	stringLength=strlen(&(text->string[0]));
	currentIndex=0;
	while(currentIndex<stringLength&&(stringY<text->rect.y+(int)text->rect.h))	// stop drawing when we get below the bottom edge
	{
		lastIndex=currentIndex;
		while(currentIndex<stringLength&&text->string[currentIndex]!='\n'&&(XTextWidth(text->font->xFont,&(text->string[lastIndex]),(currentIndex+1)-lastIndex)<(int)text->rect.w))	// race up to the next new line, or until we have to wrap
		{
			currentIndex++;
		}
		XDrawString(xDisplay,imageBuffer,xGraphicsContext,stringX,stringY,&(text->string[lastIndex]),currentIndex-lastIndex);
		stringY+=text->font->ascent+text->font->descent;
		if(text->string[currentIndex]=='\n')
		{
			currentIndex++;														// skip over the new line
		}
	}
	XSetClipMask(xDisplay,xGraphicsContext,None);								// get rid of clip mask
	XDestroyRegion(textRegion);
}

static void ResetStaticText(STATIC_TEXT *text,char *newString)
// place a new static text string into text
{
	strncpy(text->string,newString,MAXSTATIC_TEXTSTRING-1);
	text->string[MAXSTATIC_TEXTSTRING-1]='\0';
	InvalidateStaticText(text);
}

static STATIC_TEXT *CreateStaticText(EDITOR_WINDOW *window,STATIC_TEXT_DESCRIPTOR *description)
// create a static text item in window, invalidate the area of window
// where the text is to be placed
// return a pointer to it if successful
// SetError, and return NULL if not
{
	STATIC_TEXT
		*text;

	if((text=(STATIC_TEXT *)MNewPtr(sizeof(STATIC_TEXT))))			// create static text data structure
	{
		text->parentWindow=window;

		text->foregroundColor=description->foregroundColor;
		text->backgroundColor=description->backgroundColor;

		text->resizeFunction=description->resizeFunction;

		strncpy(text->string,description->string,MAXSTATIC_TEXTSTRING-1);
		text->string[MAXSTATIC_TEXTSTRING-1]='\0';
		if((text->font=LoadFont(description->fontName)))
		{
			SetStaticTextRectangle(text);
			return(text);
		}
		MDisposePtr(text);
	}
	return(NULL);
}

static void DisposeStaticText(STATIC_TEXT *text)
// unlink text from its parent window, and delete it
// invalidate the area of the window where text was
{
	InvalidateStaticText(text);				// invalidate where it is now
	FreeFont(text->font);					// free the font
	MDisposePtr(text);
}

// dialog interface routines

static void DrawStaticTextItem(DIALOG_ITEM *item,Pixmap imageBuffer,Region updateRegion)
// draw the static text pointed to by item
{
	DrawStaticText((STATIC_TEXT *)item->itemStruct,imageBuffer,updateRegion);
}

static void ResizeStaticTextItem(DIALOG_ITEM *item)
// When the dialog window size changes, call this to update the item itself
// This will invalidate the old and new item areas
{
	STATIC_TEXT
		*text;

	text=(STATIC_TEXT *)item->itemStruct;
	InvalidateStaticText(text);				// invalidate where we're at currently
	SetStaticTextRectangle(text);
}

static void DisposeStaticTextItem(DIALOG_ITEM *item)
// dispose of the local static text structure pointed to by item
{
	DisposeStaticText((STATIC_TEXT *)item->itemStruct);
}

void ResetStaticTextItemText(DIALOG_ITEM *item,char *newString)
// reset the text of item to newString, invalidate the item
{
	ResetStaticText((STATIC_TEXT *)item->itemStruct,newString);
}

bool CreateStaticTextItem(DIALOG_ITEM *item,void *description)
// create the static text, and fill in item
// if there is a problem, SetError, return false
{
	if((item->itemStruct=(void *)CreateStaticText(item->dialog->parentWindow,(STATIC_TEXT_DESCRIPTOR *)description)))
	{
		item->disposeProc=DisposeStaticTextItem;
		item->drawProc=DrawStaticTextItem;
		item->resizeProc=ResizeStaticTextItem;
		item->trackProc=NULL;
		item->earlyKeyProc=NULL;
		item->wantFocus=false;
		item->focusChangeProc=NULL;
		item->focusKeyProc=NULL;
		item->focusPeriodicProc=NULL;
		return(true);
	}
	return(false);
}
