// Text box handling for dialogs
// Copyright (C) 2000 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

#define	BORDER_WIDTH			3
#define	SHADOW_COLOR_UL			0x314F5E	// shadow colors (upper left)
#define	SHADOW_COLOR_LR			0x91AFBE	// (lower right)

typedef struct textBox
{
	EDITOR_WINDOW
		*parentWindow;					// the window that contains this text box
	textBoxResizeFunction
		*resizeFunction;				// function which is used to resize for this window
	EDITOR_RECT
		rect;							// position and size within parent window
	EDITOR_COLOR
		focusBackgroundColor,			// color to use when drawing background when we have the focus
		focusForegroundColor,			// color to use when drawing foreground when we have the focus
		nofocusBackgroundColor,			// color to use when drawing background when we do not have focus
		nofocusForegroundColor;			// color to use when drawing foreground when we do not have focus
	EDITOR_VIEW
		*view;							// the view that this text box item made
} TEXT_BOX;

static void InvalidateTextBox(TEXT_BOX *textBox)
// invalidate the text box within its parent window
{
	InvalidateWindowRect(textBox->parentWindow,&textBox->rect);
}

static void GetTextBoxControlRects(TEXT_BOX *textBox,EDITOR_RECT *borderRect,EDITOR_RECT *textContentRect)
// return the rectangles that enclose the various pieces of textBox
// if the text box starts getting too small, certain rects may be returned with 0 width or
// height
{
	EDITOR_RECT
		insetRect;

	insetRect=*borderRect=textBox->rect;					// border rect is the whole thing

	insetRect.x+=BORDER_WIDTH;								// inset from the border to the content area of the text box
	insetRect.y+=BORDER_WIDTH;
	if(insetRect.w>2*BORDER_WIDTH)
	{
		insetRect.w-=2*BORDER_WIDTH;
	}
	else
	{
		insetRect.w=0;
	}
	if(insetRect.h>2*BORDER_WIDTH)
	{
		insetRect.h-=2*BORDER_WIDTH;
	}
	else
	{
		insetRect.h=0;
	}
	*textContentRect=insetRect;
}

static void SetTextBoxRectangle(TEXT_BOX *textBox)
// Work out the rectangle given the current window
{
	EDITOR_RECT
		borderRect,
		textContentRect;

	if(textBox->resizeFunction)			// see if non-null resize function specified
	{
		textBox->resizeFunction(textBox->parentWindow,&textBox->rect);	// get new rectangle
		GetTextBoxControlRects(textBox,&borderRect,&textContentRect);
		SetViewBounds(textBox->view,&textContentRect);		// move the view over
		InvalidateTextBox(textBox);							// invalidate it
	}
}

static bool PointInTextBox(TEXT_BOX *textBox,INT32 x,INT32 y)
// return true if parent window relative x/y are in textBox
// false if not
{
	return(PointInRECT(x,y,&textBox->rect));
}

static void DrawTextBox(TEXT_BOX *textBox,Pixmap imageBuffer,Region updateRegion,bool hasFocus)
// redraw textBox in its parent window, obey the invalid area of the parent's window
{
	XRectangle
		regionBox;
	Region
		invalidRegion;
	EDITOR_RECT
		borderRect,
		textContentRect;

	GetTextBoxControlRects(textBox,&borderRect,&textContentRect);

	invalidRegion=XCreateRegion();

	regionBox.x=borderRect.x;
	regionBox.y=borderRect.y;
	regionBox.width=borderRect.w;
	regionBox.height=borderRect.h;
	XUnionRectWithRegion(&regionBox,invalidRegion,invalidRegion);					// create a region which is the rectangle of the text box
	XIntersectRegion(invalidRegion,updateRegion,invalidRegion);						// intersect with what is invalid in this window

	if(!XEmptyRegion(invalidRegion))												// see if it has an invalid area
	{
		XSetRegion(xDisplay,xGraphicsContext,invalidRegion);						// set clipping to what's invalid
		if(!hasFocus)
		{
			OutlineShadowRectangle(imageBuffer,&borderRect,SHADOW_COLOR_UL,SHADOW_COLOR_LR,BORDER_WIDTH);	// drop in border
		}
		else
		{
			OutlineShadowRectangle(imageBuffer,&borderRect,BLACK,GRAY_1,BORDER_WIDTH);	// drop in border
		}
		XSetClipMask(xDisplay,xGraphicsContext,None);								// get rid of clip mask
	}
	XDestroyRegion(invalidRegion);													// get rid of invalid region
	DrawView(textBox->view,imageBuffer,updateRegion);
}

static bool TrackTextBox(TEXT_BOX *textBox,XEvent *event)
// a button was pressed, see if it was in textBox, if so, track it, and return true
// if it was not within the text box, return false
{
	return(TrackView(textBox->view,event));	// only thing tracking in the text box is the view
}

static void HandleTextBoxKeyEvent(TEXT_BOX *textBox,XEvent *event)
// A keyboard event is arriving for the text box, handle it here
{
	HandleViewKeyEvent(textBox->view,event);
}

static void ActivateTextBox(TEXT_BOX *textBox)
// when the text box gets focus, call this to make it active
{
	SetEditorViewStyleForegroundColor(textBox->view,0,textBox->focusForegroundColor);
	SetEditorViewStyleBackgroundColor(textBox->view,0,textBox->focusBackgroundColor);
	EditorActivateView(textBox->view);
	ResetEditorViewCursorBlink(textBox->view);
	InvalidateTextBox(textBox);						// invalidate where it is now
}

static void DeactivateTextBox(TEXT_BOX *textBox)
// when the text box loses focus, call this to make it inactive
{
	SetEditorViewStyleForegroundColor(textBox->view,0,textBox->nofocusForegroundColor);
	SetEditorViewStyleBackgroundColor(textBox->view,0,textBox->nofocusBackgroundColor);
	EditorDeactivateView(textBox->view);
	InvalidateTextBox(textBox);						// invalidate where it is now
}

static void HandleTextBoxPeriodicProc(TEXT_BOX *textBox)
// call this to handle periodic things a text box must do
{
	ViewCursorTask(textBox->view);
}

static TEXT_BOX *CreateTextBox(EDITOR_WINDOW *window,TEXT_BOX_DESCRIPTOR *description)
// create a text box in window, invalidate the area of window
// where the text box is to be placed
// return a pointer to it if successful
// SetError, return NULL if not
{
	TEXT_BOX
		*textBox;
	EDITOR_VIEW_DESCRIPTOR
		viewDescriptor;

	if((textBox=(TEXT_BOX *)MNewPtr(sizeof(TEXT_BOX))))			// create text box data structure
	{
		textBox->parentWindow=window;

		textBox->resizeFunction=description->resizeFunction;

		textBox->focusBackgroundColor=description->focusBackgroundColor;	// copy over the colors
		textBox->focusForegroundColor=description->focusForegroundColor;
		textBox->nofocusBackgroundColor=description->nofocusBackgroundColor;
		textBox->nofocusForegroundColor=description->nofocusForegroundColor;

		viewDescriptor.buffer=description->buffer;				// the buffer on which to view

		viewDescriptor.rect.x=0;								// initally make empty view
		viewDescriptor.rect.y=0;
		viewDescriptor.rect.w=0;
		viewDescriptor.rect.h=0;
		viewDescriptor.topLine=description->topLine;
		viewDescriptor.leftPixel=description->leftPixel;
		viewDescriptor.tabSize=description->tabSize;
		viewDescriptor.backgroundColor=textBox->nofocusBackgroundColor;
		viewDescriptor.foregroundColor=textBox->nofocusForegroundColor;
		viewDescriptor.fontName=description->fontName;			// initial font to use
		viewDescriptor.active=false;							// tells if the view is created active or not
		viewDescriptor.viewTextChangedVector=NULL;
		viewDescriptor.viewSelectionChangedVector=NULL;
		viewDescriptor.viewPositionChangedVector=NULL;

		if((textBox->view=CreateEditorView(textBox->parentWindow,&viewDescriptor)))
		{
			SetTextBoxRectangle(textBox);
			return(textBox);
		}
		MDisposePtr(textBox);
	}
	return(NULL);
}

static void DisposeTextBox(TEXT_BOX *textBox)
// unlink textBox from its parent window, and delete it
// invalidate the area of the window where the text box was
{
	InvalidateTextBox(textBox);					// invalidate where it is now
	DisposeEditorView(textBox->view);
	MDisposePtr(textBox);
}

// dialog interface routines

static void DrawTextBoxItem(DIALOG_ITEM *item,Pixmap imageBuffer,Region updateRegion)
// draw the view pointed to by item
{
	DrawTextBox((TEXT_BOX *)item->itemStruct,imageBuffer,updateRegion,(item==item->dialog->focusItem));	// draw the text box
}

static void ResizeTextBoxItem(DIALOG_ITEM *item)
// When the dialog window size changes, call this to update the item itself
// This will invalidate the old and new item areas
{
	TEXT_BOX
		*textBox;

	textBox=(TEXT_BOX *)item->itemStruct;
	InvalidateTextBox(textBox);					// invalidate where we're at currently
	SetTextBoxRectangle(textBox);
}

static bool TrackTextBoxItem(DIALOG_ITEM *item,XEvent *event)
// attempt to track the text box pointed to by item
{
	if(PointInTextBox((TEXT_BOX *)item->itemStruct,event->xbutton.x,event->xbutton.y))
	{
		DialogTakeFocus(item);
		return(TrackTextBox((TEXT_BOX *)item->itemStruct,event));
	}
	return(false);
}

static void DisposeTextBoxItem(DIALOG_ITEM *item)
// dispose of the local text box structure pointed to by item
{
	DisposeTextBox((TEXT_BOX *)item->itemStruct);
}

static void TextBoxItemFocusChange(DIALOG_ITEM *item)
// the focus is changing into, or out of this view item, deal with it
{
	if(item==item->dialog->focusItem)
	{
		ActivateTextBox((TEXT_BOX *)item->itemStruct);
	}
	else
	{
		DeactivateTextBox((TEXT_BOX *)item->itemStruct);
	}
}

static void TextBoxItemFocusKey(DIALOG_ITEM *item,XEvent *event)
// give the key to the text box
{
	HandleTextBoxKeyEvent((TEXT_BOX *)item->itemStruct,event);
}

static void TextBoxItemPeriodicProc(DIALOG_ITEM *item)
// this is used to flash the cursor of the view
{
	HandleTextBoxPeriodicProc((TEXT_BOX *)item->itemStruct);
}

bool CreateTextBoxItem(DIALOG_ITEM *item,void *descriptor)
// create the textBox, and fill in item
// if there is a problem, SetError, return false
{
	if((item->itemStruct=(void *)CreateTextBox(item->dialog->parentWindow,(TEXT_BOX_DESCRIPTOR *)descriptor)))
	{
		item->disposeProc=DisposeTextBoxItem;
		item->drawProc=DrawTextBoxItem;
		item->resizeProc=ResizeTextBoxItem;
		item->trackProc=TrackTextBoxItem;
		item->earlyKeyProc=NULL;
		item->wantFocus=true;
		item->focusChangeProc=TextBoxItemFocusChange;
		item->focusKeyProc=TextBoxItemFocusKey;
		item->focusPeriodicProc=TextBoxItemPeriodicProc;		// set up cursor flash task
		return(true);
	}
	return(false);
}
