// Low level stuff needed to manage views
// Copyright (C) 2000 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

#define	STAGING_BUFFER_SIZE		8192										// number of bytes used in the staging buffer (must be larger than STAGING_MAX_ADD)
#define	STAGING_MAX_ADD			256											// maximum number of bytes that will ever be added to the staging buffer in one call to GetCharacterEquiv

static const char hexConv[]="0123456789ABCDEF";								// used for speedy hex conversions (for display of non-ascii codes)

// when text is being built, or widths are being calculated, these globals
// are used, they allow for much more efficient creation of text because things
// are not being passed all over for each character
static UINT8
	gceStagingBuffer[STAGING_BUFFER_SIZE];									// this is where get character equivalent builds its output
static UINT32
	gceBufferIndex;															// offset into the staging buffer
static UINT32
	gceColumnNumber;														// running column number
static UINT16
	*gceWidthTable;															// width table to use while calculating positions
static UINT32
	gceTabSize;																// number of spaces to use per tab
static INT32
	gceRunningWidth;														// running width

static EDITOR_FONT *GetFontForStyle(VIEW_STYLE *styles,UINT32 desiredStyle)
// Return the font that should be used for the given style
// if there is none specified, then use the default
{
	if(styles[desiredStyle].font)									// if there is a font loaded for this style, return it
	{
		return(styles[desiredStyle].font);
	}
	return(styles[0].font);											// otherwise, return the default font
}

static EDITOR_COLOR GetForegroundForStyle(VIEW_STYLE *styles,UINT32 desiredStyle)
// Return the foreground color that should be used for the given style
// if there is none specified, then use the default
{
	if(styles[desiredStyle].haveForegroundColor)					// if there is a color loaded for this style, return it
	{
		return(styles[desiredStyle].foregroundColor);
	}
	return(styles[0].foregroundColor);								// otherwise, return the default color
}

static EDITOR_COLOR GetBackgroundForStyle(VIEW_STYLE *styles,UINT32 desiredStyle)
// Return the background color that should be used for the given style
// if there is none specified, then use the default
{
	if(styles[desiredStyle].haveBackgroundColor)					// if there is a color loaded for this style, return it
	{
		return(styles[desiredStyle].backgroundColor);
	}
	return(styles[0].backgroundColor);								// otherwise, return the default color
}

static bool PointInView(EDITOR_VIEW *view,INT32 x,INT32 y)
// Return true if parent window relative x/y are in view
// false if not
{
	return(PointInRECT(x,y,&(view->viewInfo->bounds)));
}

static void LocalClickToViewClick(EDITOR_VIEW *view,INT32 clickX,INT32 clickY,INT32 *xPosition,INT32 *lineNumber)
// Given a point local to the window that contains view,
// convert it into a view position
{
	GUI_VIEW_INFO
		*viewInfo;
	INT32
		viewY;

	viewInfo=view->viewInfo;									// point at the information record for this view
	(*xPosition)=clickX-viewInfo->textBounds.x;
	viewY=clickY-viewInfo->textBounds.y;
	if(viewY<0)
	{
		viewY-=viewInfo->lineHeight;							// account for cases about 0
	}
	(*lineNumber)=viewY/((INT32)(viewInfo->lineHeight));
}

void HandleViewKeyEvent(EDITOR_VIEW *view,XEvent *event)
// Handle an incoming key event for view
{
	VIEW_EVENT
		viewEvent;
	EDITOR_KEY
		keyData;

	viewEvent.view=view;									// point at the view
	viewEvent.eventType=VET_KEYDOWN;						// let it know what type
	if(KeyEventToEditorKey(event,&keyData))					// fill in the key record
	{
		viewEvent.eventData=&keyData;
		HandleViewEvent(&viewEvent);						// pass this off to high level handler
	}
}

static bool GetCharacterEquiv(UINT8 character)
// Put the correct characters into the gceStagingBuffer at gceBufferIndex
// update gceColumnNumber, gceBufferIndex, and gceRunningWidth
// if an end of line character is seen, return false, else return true
// NOTE: this is allowed to add at most STAGING_MAX_ADD characters to the staging buffer
// When called, there must be at least that many bytes available, or
// there will be trouble
{
	UINT32
		i,
		numTabSpaces;

	if(character>=' '&&gceWidthTable[character])				// make sure the font can display this, if not, display it as hex
	{
		gceStagingBuffer[gceBufferIndex++]=character;
		gceRunningWidth+=gceWidthTable[character];
		gceColumnNumber++;
	}
	else
	{
		switch(character)
		{
			case '\n':
				return(false);
				break;
			case '\t':
				if(gceTabSize)
				{
					numTabSpaces=gceTabSize-(gceColumnNumber%gceTabSize);
					for(i=0;i<numTabSpaces;i++)
					{
						gceStagingBuffer[gceBufferIndex++]=' ';			// space over to tab stop
					}
					gceRunningWidth+=gceWidthTable[' ']*numTabSpaces;	// add in space to running width
					gceColumnNumber+=numTabSpaces;						// push to correct column
				}
				else
				{
					gceRunningWidth+=gceWidthTable[gceStagingBuffer[gceBufferIndex++]='['];
					gceRunningWidth+=gceWidthTable[gceStagingBuffer[gceBufferIndex++]=hexConv[character>>4]];
					gceRunningWidth+=gceWidthTable[gceStagingBuffer[gceBufferIndex++]=hexConv[character&0x0F]];
					gceRunningWidth+=gceWidthTable[gceStagingBuffer[gceBufferIndex++]=']'];
					gceColumnNumber+=4;
				}
				break;
			default:
				// this is a little scary looking, but it is fast
				gceRunningWidth+=gceWidthTable[gceStagingBuffer[gceBufferIndex++]='['];
				gceRunningWidth+=gceWidthTable[gceStagingBuffer[gceBufferIndex++]=hexConv[character>>4]];
				gceRunningWidth+=gceWidthTable[gceStagingBuffer[gceBufferIndex++]=hexConv[character&0x0F]];
				gceRunningWidth+=gceWidthTable[gceStagingBuffer[gceBufferIndex++]=']'];
				gceColumnNumber+=4;
				break;
		}
	}
	return(true);
}

static bool PositionAtPixel(EDITOR_VIEW *view,UINT32 linePosition,INT32 xPosition,UINT32 *charPosition,UINT32 *charColumn,INT32 *charStartPixel,UINT32 *charWidth)
// Given linePosition which points to the character at the start of a line, and xPosition
// as a "pixel" position along the line, determine the offset to the character
// at xPosition, the pixel where that character begins, and the width of that character.
// if linePosition is past EOF, charPosition will be returned as linePosition,
// charColumn will be 0, charStartPixel will be VIEW_LEFT_MARGIN, and charWidth will be 0
// if the xPosition is past the end of the line, or end of the data, this will return false
{
	GUI_VIEW_INFO
		*viewInfo;
	CHUNK_HEADER
		*chunk;
	UINT32
		offset;
	bool
		haveStyle,
		inLine,
		loopDone;
	INT32
		lastRunningWidth;
	UINT32
		lastColumnNumber;
	UINT32
		stylePosition,
		numStyleElements,
		currentStyle;

	viewInfo=view->viewInfo;								// point at the information record for this view

	PositionToChunkPosition(view->parentBuffer->textUniverse,linePosition,&chunk,&offset);

	if((haveStyle=GetStyleRange(view->parentBuffer->styleUniverse,linePosition,&stylePosition,&numStyleElements,&currentStyle)))
	{
		ForceStyleIntoRange(currentStyle);
		stylePosition+=numStyleElements;					// style will change at this position
	}

	loopDone=false;
	inLine=true;

	gceTabSize=view->tabSize;
	gceRunningWidth=lastRunningWidth=VIEW_LEFT_MARGIN;
	gceColumnNumber=lastColumnNumber=0;
	gceWidthTable=&(GetFontForStyle(view->viewStyles,currentStyle)->charWidths[0]);

	while(!loopDone&&inLine&&chunk)							// run out to the position where the text begins
	{
		if(haveStyle&&(linePosition>=stylePosition))		// time to update style?
		{
			if((haveStyle=GetStyleRange(view->parentBuffer->styleUniverse,linePosition,&stylePosition,&numStyleElements,&currentStyle)))
			{
				ForceStyleIntoRange(currentStyle);
				stylePosition+=numStyleElements;			// style will change at this position
			}
			gceWidthTable=&(GetFontForStyle(view->viewStyles,currentStyle)->charWidths[0]);
		}
		gceBufferIndex=0;
		if((inLine=GetCharacterEquiv(chunk->data[offset])))
		{
			if(gceRunningWidth>xPosition)
			{
				loopDone=true;
			}
			else
			{
				lastColumnNumber=gceColumnNumber;			// remember these
				lastRunningWidth=gceRunningWidth;
				linePosition++;
				offset++;
				if(offset>=chunk->totalBytes)				// see if time for a new chunk
				{
					chunk=chunk->nextHeader;
					offset=0;
				}
			}
		}
	}
	*charPosition=linePosition;								// return the results
	*charColumn=lastColumnNumber;
	*charStartPixel=lastRunningWidth;
	*charWidth=gceRunningWidth-lastRunningWidth;

	return(chunk&&inLine);
}

static void PixelAtPosition(EDITOR_VIEW *view,UINT32 position,INT32 *xPosition,bool limitMax)
// Given position within view, determine the "pixel" offset to the start of the character
// of the line position is in.
// if position is passed in past the end of the buffer, this will return 0
// NOTE: if limitMax is true, this will abort as soon as the returned position
// grows larger than that passed in
{
	GUI_VIEW_INFO
		*viewInfo;
	CHUNK_HEADER
		*chunk;
	UINT32
		offset;
	UINT32
		linePosition;
	UINT32
		line,
		lineOffset;
	bool
		haveStyle;
	UINT32
		stylePosition,
		numStyleElements,
		currentStyle;

	viewInfo=view->viewInfo;								// point at the information record for this view

	PositionToLinePosition(view->parentBuffer->textUniverse,position,&line,&lineOffset,&chunk,&offset);

	linePosition=position-lineOffset;

	if((haveStyle=GetStyleRange(view->parentBuffer->styleUniverse,linePosition,&stylePosition,&numStyleElements,&currentStyle)))
	{
		ForceStyleIntoRange(currentStyle);
		stylePosition+=numStyleElements;					// style will change at this position
	}

	gceTabSize=view->tabSize;
	gceRunningWidth=VIEW_LEFT_MARGIN;
	gceColumnNumber=0;
	gceWidthTable=&(GetFontForStyle(view->viewStyles,currentStyle)->charWidths[0]);

	while((linePosition<position)&&chunk&&(!limitMax||(gceRunningWidth<*xPosition)))	// run out to the given position
	{
		if(haveStyle&&(linePosition>=stylePosition))		// time to update style?
		{
			if((haveStyle=GetStyleRange(view->parentBuffer->styleUniverse,linePosition,&stylePosition,&numStyleElements,&currentStyle)))
			{
				ForceStyleIntoRange(currentStyle);
				stylePosition+=numStyleElements;			// style will change at this position
			}
			gceWidthTable=&(GetFontForStyle(view->viewStyles,currentStyle)->charWidths[0]);
		}
		gceBufferIndex=0;
		GetCharacterEquiv(chunk->data[offset]);
		linePosition++;
		offset++;
		if(offset>=chunk->totalBytes)						// see if time for a new chunk
		{
			chunk=chunk->nextHeader;
			offset=0;
		}
	}
	(*xPosition)=gceRunningWidth;
}

void GetEditorViewGraphicToTextPosition(EDITOR_VIEW *view,UINT32 linePosition,INT32 xPosition,UINT32 *betweenOffset,UINT32 *charOffset)
// Given linePosition which points to the character at the start of a line, and xPosition
// as a "pixel" position along the line, determine two offsets:
// betweenOffset is the offset of the character boundary closest to the xPosition,
// charOffset is the offset of the character closest to the xPosition
// if linePosition is past EOF, both offsets will be returned as 0
{
	UINT32
		charPosition;
	UINT32
		charColumn;
	INT32
		charStartPixel;
	UINT32
		charWidth;
	bool
		inLine;

	inLine=PositionAtPixel(view,linePosition,xPosition,&charPosition,&charColumn,&charStartPixel,&charWidth);
	*betweenOffset=*charOffset=charPosition-linePosition;
	if(inLine&&(xPosition>(charStartPixel+((INT32)charWidth/2))))
	{
		(*betweenOffset)++;
	}
}

void GetEditorViewTextToGraphicPosition(EDITOR_VIEW *view,UINT32 position,INT32 *xPosition,bool limitMax,UINT32 *slopLeft,UINT32 *slopRight)
// Given position within view, determine the "pixel" offset to the start of the character
// of the line position is in.
// if position is passed in past the end of the buffer, this will return the width of the last line
// NOTE: if limitMax is true, xPosition arrives at this function with a maximum
// pixel position. This routine is then allowed to stop as soon as it
// sees the return result-slopLeft as greater than the passed value.
// NOTE: slopLeft and slopRight are used to allow fonts which have overhang
// into the previous or next character to be dealt with correctly.
// slopLeft tells how far the character at position draws to the left of
// its given position.
// slopRight tells how far the character 1 to the left of position draws
// to the right (past its character width).
{
	GUI_VIEW_INFO
		*viewInfo;

	viewInfo=view->viewInfo;
	*slopLeft=viewInfo->maxLeftOverhang;
	*slopRight=viewInfo->maxRightOverhang;
	if(limitMax)					// if limiting to pixel position, move out by max overhang
	{
		*xPosition+=*slopLeft;
	}
	PixelAtPosition(view,position,xPosition,limitMax);
}

void GetEditorViewTextInfo(EDITOR_VIEW *view,UINT32 *topLine,UINT32 *numLines,INT32 *leftPixel,UINT32 *numPixels)
// Get information about the given view
{
	GUI_VIEW_INFO
		*viewInfo;

	viewInfo=view->viewInfo;
	*topLine=viewInfo->topLine;
	*numLines=viewInfo->numLines;
	*leftPixel=viewInfo->leftPixel;
	*numPixels=viewInfo->textBounds.w;
}

static void DrawViewCursor(EDITOR_VIEW *view,Pixmap imageBuffer)
// Draw the cursor wherever it happens to be in view into
// the passed pixmap constrained by the passed graphics context.
{
	GUI_VIEW_INFO
		*viewInfo;
	CHUNK_HEADER
		*chunk;
	UINT32
		offset;
	UINT32
		line,
		lineOffset;
	INT32
		xPosition,
		maxXPosition;
	UINT32
		slopLeft,
		slopRight;
	EDITOR_RECT
		cursorRect;
	UINT32
		style;

	viewInfo=view->viewInfo;										// point at the information record for this view

	if((IsSelectionEmpty(view->parentBuffer->selectionUniverse))&&viewInfo->blinkState&&view->active)
	{
		PositionToLinePosition(view->parentBuffer->textUniverse,GetSelectionCursorPosition(view->parentBuffer->selectionUniverse),&line,&lineOffset,&chunk,&offset);
		if((line>=viewInfo->topLine)&&(line<viewInfo->topLine+viewInfo->numLines))	// make sure cursor is on the view vertically
		{
			maxXPosition=xPosition=viewInfo->leftPixel+viewInfo->textBounds.w;
			GetEditorViewTextToGraphicPosition(view,GetSelectionCursorPosition(view->parentBuffer->selectionUniverse),&xPosition,true,&slopLeft,&slopRight);

			style=0;														// always paint the cursor in style 0

			if((xPosition>=viewInfo->leftPixel)&&(xPosition<maxXPosition))	// see if on view horizontally
			{
				cursorRect.x=viewInfo->textBounds.x+(xPosition-viewInfo->leftPixel);
				cursorRect.w=1;
				cursorRect.y=viewInfo->textBounds.y+viewInfo->lineHeight*(line-viewInfo->topLine);
				cursorRect.h=viewInfo->lineHeight;

				XSetFunction(xDisplay,xGraphicsContext,GXxor);				// prepare to fill in xor format
				XSetFillStyle(xDisplay,xGraphicsContext,FillSolid);			// use solid fill

				XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(GetForegroundForStyle(view->viewStyles,style)^GetBackgroundForStyle(view->viewStyles,style)));	// this is the value to xor with

				XFillRectangle(xDisplay,imageBuffer,xGraphicsContext,cursorRect.x,cursorRect.y,cursorRect.w,cursorRect.h);
				XSetFunction(xDisplay,xGraphicsContext,GXcopy);				// put function back
			}
		}
	}
}

static void AddSelectionLineToRegion(EDITOR_VIEW *view,SELECTION_UNIVERSE *selection,Region region,UINT32 currentPosition,INT32 yWindowOffset)
// Read bytes from the line of text at chunk, offset
// Add pieces where the passed selection is active to region
// NOTE: this must work as an exact mirror to DrawViewLine
// NOTE: this adds extra space to the left and right of the region
// as it creates it, to accommodate the overhang of fonts
{
	GUI_VIEW_INFO
		*viewInfo;
	CHUNK_HEADER
		*chunk;
	UINT32
		offset;
	INT32
		xWindowOffset,
		xWindowEndOffset;
	INT32
		skippedWidth;
	bool
		currentSelection,
		nextSelection;
	UINT32
		currentStyle,
		nextStyle;
	INT32
		scrollX;					// pixel offset in text where we need to start drawing
	INT32
		textXOffset;				// window relative pixel offset where text is being drawn
	INT32
		distanceToStart;
	XRectangle
		selectionRect;
	UINT32
		charWidth;

	viewInfo=view->viewInfo;								// point at the information record for this view

	xWindowOffset=viewInfo->textBounds.x;
	xWindowEndOffset=viewInfo->textBounds.x+viewInfo->textBounds.w;

	selectionRect.y=yWindowOffset;
	selectionRect.height=viewInfo->lineHeight;

	scrollX=viewInfo->leftPixel;							// get pixel offset in text where we need to start drawing

	PositionAtPixel(view,currentPosition,scrollX,&currentPosition,&gceColumnNumber,&skippedWidth,&charWidth);

	PositionToChunkPosition(view->parentBuffer->textUniverse,currentPosition,&chunk,&offset);

	gceBufferIndex=0;										// begin at start of buffer
	gceRunningWidth=0;

	distanceToStart=skippedWidth-scrollX;

	if(distanceToStart>(xWindowEndOffset-xWindowOffset))	// do not go past the end of the draw area
	{
		distanceToStart=xWindowEndOffset-xWindowOffset;
	}

	textXOffset=xWindowOffset+distanceToStart;				// move the offset slightly back to beginning of first character which enters the view (or forward if view is scrolled negatively)

	currentSelection=GetSelectionAtPosition(selection,currentPosition);
	currentStyle=GetStyleAtPosition(view->parentBuffer->styleUniverse,currentPosition);
	ForceStyleIntoRange(currentStyle);

	gceWidthTable=&(GetFontForStyle(view->viewStyles,currentStyle)->charWidths[0]);

	while(chunk&&((textXOffset+gceRunningWidth)<xWindowEndOffset)&&GetCharacterEquiv(chunk->data[offset]))		// run through the text
	{
		nextSelection=GetSelectionAtPosition(selection,currentPosition+1);

		nextStyle=GetStyleAtPosition(view->parentBuffer->styleUniverse,currentPosition+1);
		ForceStyleIntoRange(nextStyle);

		if((currentStyle!=nextStyle)||(currentSelection!=nextSelection)||(gceBufferIndex>STAGING_BUFFER_SIZE-STAGING_MAX_ADD))	// is something about to change? (if so, output what we have so far) (this will also output if the staging buffer is getting full)
		{
			if(currentSelection)
			{
				selectionRect.x=textXOffset-viewInfo->maxLeftOverhang;
				selectionRect.width=gceRunningWidth+viewInfo->maxLeftOverhang+viewInfo->maxRightOverhang;
				if(selectionRect.x<xWindowOffset)			// pin to bounds
				{
					selectionRect.width-=(xWindowOffset-selectionRect.x);
					selectionRect.x=xWindowOffset;
				}
				if(selectionRect.x+selectionRect.width>xWindowEndOffset)
				{
					selectionRect.width=xWindowEndOffset-selectionRect.x;
				}
				XUnionRectWithRegion(&selectionRect,region,region);
			}

			textXOffset+=gceRunningWidth;

			gceBufferIndex=0;								// begin at start of buffer
			gceRunningWidth=0;

			currentSelection=nextSelection;
			currentStyle=nextStyle;
			gceWidthTable=&(GetFontForStyle(view->viewStyles,currentStyle)->charWidths[0]);
		}

		currentPosition++;
		offset++;
		if(offset>=chunk->totalBytes)						// push through end of chunk
		{
			chunk=chunk->nextHeader;
			offset=0;
		}
	}
	// ran off the end of the line, or the end of the universe
	if(gceBufferIndex)
	{
		if(currentSelection)
		{
			selectionRect.x=textXOffset-viewInfo->maxLeftOverhang;
			selectionRect.width=gceRunningWidth+viewInfo->maxLeftOverhang+viewInfo->maxRightOverhang;
			if(selectionRect.x<xWindowOffset)					// pin to bounds
			{
				selectionRect.width-=(xWindowOffset-selectionRect.x);
				selectionRect.x=xWindowOffset;
			}
			if(selectionRect.x+selectionRect.width>xWindowEndOffset)
			{
				selectionRect.width=xWindowEndOffset-selectionRect.x;
			}
			XUnionRectWithRegion(&selectionRect,region,region);
		}
		textXOffset+=gceRunningWidth;
	}

	if(textXOffset<xWindowEndOffset)							// do we need to go out to the end of the line?
	{
		if(textXOffset<xWindowOffset)							// did this end before the draw area began?
		{
			textXOffset=xWindowOffset;
		}
		currentSelection=GetSelectionAtPosition(selection,currentPosition);		// get selection information for newline (or off of end)

		if(currentSelection)
		{
			selectionRect.x=textXOffset;
			selectionRect.width=xWindowEndOffset-textXOffset;
			if(selectionRect.x<xWindowOffset)					// pin to bounds
			{
				selectionRect.width-=(xWindowOffset-selectionRect.x);
				selectionRect.x=xWindowOffset;
			}
			if(selectionRect.x+selectionRect.width>xWindowEndOffset)
			{
				selectionRect.width=xWindowEndOffset-selectionRect.x;
			}
			XUnionRectWithRegion(&selectionRect,region,region);
		}
	}
}

static EDITOR_COLOR SumColors(EDITOR_COLOR colorA,EDITOR_COLOR colorB)
// When marks overlap, this is called to blend the colors of the overlapping
// marks together
{
	unsigned int
		red,
		green,
		blue;

	red=((colorA>>16)&0xFF)+((colorB>>16)&0xFF);
	green=((colorA>>8)&0xFF)+((colorB>>8)&0xFF);
	blue=(colorA&0xFF)+(colorB&0xFF);
	if(red>255)
	{
		red=255;
	}
	if(green>255)
	{
		green=255;
	}
	if(blue>255)
	{
		blue=255;
	}
	return((red<<16)|(green<<8)|blue);
}

static bool DetermineMarksBackgroundColorAtPosition(EDITOR_VIEW *view,UINT32 position,EDITOR_COLOR *color)
// Work out what the mark background color contribution is for the given position
// This looks at all visible marks, and sums their individual background color
// components. If there are no visible marks at the position, false is returned
{
	bool
		sawMark;
	MARK_VISIBILITY
		*visibleMark;

	sawMark=false;
	visibleMark=view->headVisibility;					// get head of visible marks
	while(visibleMark)
	{
		if(GetSelectionAtPosition(visibleMark->parentMark->selectionUniverse,position))	// mark active?
		{
			if(!sawMark)								// if first mark, just copy the color
			{
				*color=visibleMark->backgroundColor;
				sawMark=true;
			}
			else										// sum in other colors
			{
				*color=SumColors(*color,visibleMark->backgroundColor);
			}
		}
		visibleMark=visibleMark->nextMark;				// move to next mark hanging off of this view
	}
	return(sawMark);
}

static void DrawViewBackgroundPiece(EDITOR_VIEW *view,Pixmap imageBuffer,bool selected,bool haveMarkColor,EDITOR_COLOR markColor,UINT32 style,INT32 xWindowOffset,UINT32 width,INT32 yWindowOffset)
// Draw the background into the view in the given style and selection
{
	GUI_VIEW_INFO
		*viewInfo;
	EDITOR_COLOR
		background;

	viewInfo=view->viewInfo;							// point at the information record for this view

	if(selected)
	{
		if(!view->haveSelectionForegroundColor&&!view->haveSelectionBackgroundColor)		// if neither specified, invert
		{
			background=GetForegroundForStyle(view->viewStyles,style);
		}
		else	// otherwise, do not invert
		{
			if(view->haveSelectionBackgroundColor)
			{
				background=view->selectionBackgroundColor;
			}
			else
			{
				background=GetBackgroundForStyle(view->viewStyles,style);
			}
		}
	}
	else
	{
		if(!haveMarkColor)
		{
			background=GetBackgroundForStyle(view->viewStyles,style);
		}
		else
		{
			background=markColor;
		}
	}

	XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(background));	// set fill color

	XFillRectangle(xDisplay,imageBuffer,xGraphicsContext,xWindowOffset,yWindowOffset,width,viewInfo->lineHeight);
}

static void DrawViewTextPiece(EDITOR_VIEW *view,Pixmap imageBuffer,bool selected,UINT32 style,INT32 xWindowOffset,UINT32 width,INT32 yWindowOffset,INT32 clipOffset)
// Draw the piece of text in the staging buffer into the view in the given style and selection
{
	GUI_VIEW_INFO
		*viewInfo;
	EDITOR_COLOR
		foreground;

	viewInfo=view->viewInfo;						// point at the information record for this view

	XSetFont(xDisplay,xGraphicsContext,GetFontForStyle(view->viewStyles,style)->xFont->fid);	// point to the font needed for this draw
	if(selected)
	{
		if(!view->haveSelectionForegroundColor&&!view->haveSelectionBackgroundColor)		// if neither specified, invert
		{
			foreground=GetBackgroundForStyle(view->viewStyles,style);
		}
		else	// otherwise, do not invert
		{
			if(view->haveSelectionForegroundColor)
			{
				foreground=view->selectionForegroundColor;
			}
			else
			{
				foreground=GetForegroundForStyle(view->viewStyles,style);
			}
		}
	}
	else
	{
		foreground=GetForegroundForStyle(view->viewStyles,style);
	}

	XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(foreground));

	XDrawString(xDisplay,imageBuffer,xGraphicsContext,xWindowOffset,yWindowOffset+viewInfo->lineAscent,(char *)gceStagingBuffer,gceBufferIndex);
}

static void DrawViewLine(EDITOR_VIEW *view,Pixmap imageBuffer,UINT32 currentPosition,INT32 xWindowOffset,INT32 xWindowEndOffset,INT32 yWindowOffset)
// Read bytes from the line of text at currentPosition
// Draw them onto the view
// This is responsible for drawing ALL of the pixels in the view for the line
// on which it was called (even if that line contains no text)
// NOTE: this performs the draw in 2 passes.
// The first pass fills the background for the line, and the second
// pass draws the text over the background.
// The draw is done this way so that fonts which have pixels which
// overhang their character cells will draw without being clipped off
{
	CHUNK_HEADER
		*chunk;
	UINT32
		offset;
	GUI_VIEW_INFO
		*viewInfo;
	INT32
		skippedWidth;
	UINT32
		currentStyle,
		nextStyle;
	bool
		currentSelection,
		nextSelection,
		currentMark,
		nextMark;
	EDITOR_COLOR
		currentMarkColor,
		nextMarkColor;
	INT32
		scrollX;					// pixel offset in text where we need to start drawing
	INT32
		textXOffset;				// window relative pixel offset where text is being drawn
	INT32
		distanceToStart;
	UINT32
		charWidth;

	UINT32
		beginningPosition;
	CHUNK_HEADER
		*beginningChunk;
	UINT32
		beginningOffset;
	UINT32
		beginningColumn;

	viewInfo=view->viewInfo;	// point at the information record for this view

	scrollX=viewInfo->leftPixel+(xWindowOffset-viewInfo->textBounds.x);	// get pixel offset in text where we need to start drawing

	PositionAtPixel(view,currentPosition,scrollX,&currentPosition,&gceColumnNumber,&skippedWidth,&charWidth);

	PositionToChunkPosition(view->parentBuffer->textUniverse,currentPosition,&chunk,&offset);

	beginningPosition=currentPosition;							// remember these for later
	beginningChunk=chunk;
	beginningOffset=offset;
	beginningColumn=gceColumnNumber;

	gceBufferIndex=0;											// begin at start of buffer
	gceRunningWidth=0;

	distanceToStart=skippedWidth-scrollX;

	if(distanceToStart>0)										// blank space before text starts?
	{
		if(distanceToStart>(xWindowEndOffset-xWindowOffset))	// do not go past the end of the draw area
		{
			distanceToStart=xWindowEndOffset-xWindowOffset;
		}
		XSetFillStyle(xDisplay,xGraphicsContext,FillSolid);
		XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(view->viewStyles[0].backgroundColor));
		XFillRectangle(xDisplay,imageBuffer,xGraphicsContext,xWindowOffset,yWindowOffset,distanceToStart,viewInfo->lineHeight);	// fill with background color
	}

	// first draw all background

	textXOffset=xWindowOffset+distanceToStart;		// move the offset slightly back to beginning of first character which enters the view (or forward if view is scrolled negatively)

	currentSelection=GetSelectionAtPosition(view->parentBuffer->selectionUniverse,currentPosition);
	currentMark=DetermineMarksBackgroundColorAtPosition(view,currentPosition,&currentMarkColor);
	currentStyle=GetStyleAtPosition(view->parentBuffer->styleUniverse,currentPosition);
	ForceStyleIntoRange(currentStyle);

	gceWidthTable=&(GetFontForStyle(view->viewStyles,currentStyle)->charWidths[0]);

	while(chunk&&((textXOffset+gceRunningWidth)<xWindowEndOffset)&&GetCharacterEquiv(chunk->data[offset]))		// fill in all of the background needed
	{
		nextSelection=GetSelectionAtPosition(view->parentBuffer->selectionUniverse,currentPosition+1);
		nextMark=DetermineMarksBackgroundColorAtPosition(view,currentPosition+1,&nextMarkColor);
		nextStyle=GetStyleAtPosition(view->parentBuffer->styleUniverse,currentPosition+1);
		ForceStyleIntoRange(nextStyle);

		if((currentStyle!=nextStyle)||(currentSelection!=nextSelection)||(currentMark!=nextMark)||(currentMark&&(currentMarkColor!=nextMarkColor))||(gceBufferIndex>STAGING_BUFFER_SIZE-STAGING_MAX_ADD))	// is something about to change? (if so, output background so far) (this will also output if the staging buffer is getting full)
		{
			DrawViewBackgroundPiece(view,imageBuffer,currentSelection,currentMark,currentMarkColor,currentStyle,textXOffset,gceRunningWidth,yWindowOffset);

			textXOffset+=gceRunningWidth;

			gceBufferIndex=0;						// begin at start of buffer
			gceRunningWidth=0;

			currentSelection=nextSelection;
			currentMark=nextMark;
			currentMarkColor=nextMarkColor;
			currentStyle=nextStyle;
			gceWidthTable=&(GetFontForStyle(view->viewStyles,currentStyle)->charWidths[0]);
		}

		currentPosition++;
		offset++;
		if(offset>=chunk->totalBytes)				// push through end of chunk
		{
			chunk=chunk->nextHeader;
			offset=0;
		}
	}

	// ran off the end of the line, or the end of the universe
	if(gceBufferIndex)
	{
		DrawViewBackgroundPiece(view,imageBuffer,currentSelection,currentMark,currentMarkColor,currentStyle,textXOffset,gceRunningWidth,yWindowOffset);
		textXOffset+=gceRunningWidth;
	}

	if(textXOffset<xWindowEndOffset)				// do we need to draw out to the end of the line?
	{
		if(textXOffset<xWindowOffset)				// did this end before the draw area began?
		{
			textXOffset=xWindowOffset;
		}
		currentSelection=GetSelectionAtPosition(view->parentBuffer->selectionUniverse,currentPosition);
		currentMark=DetermineMarksBackgroundColorAtPosition(view,currentPosition,&currentMarkColor);
		currentStyle=GetStyleAtPosition(view->parentBuffer->styleUniverse,currentPosition);	// get style and selection information for newline (or off of end)
		ForceStyleIntoRange(currentStyle);

		DrawViewBackgroundPiece(view,imageBuffer,currentSelection,currentMark,currentMarkColor,currentStyle,textXOffset,xWindowEndOffset-textXOffset,yWindowOffset);
	}


	// next draw all the text on top of the background

	currentPosition=beginningPosition;				// restore these from above
	chunk=beginningChunk;
	offset=beginningOffset;
	gceColumnNumber=beginningColumn;

	gceBufferIndex=0;								// begin at start of buffer
	gceRunningWidth=0;

	textXOffset=xWindowOffset+distanceToStart;		// move the offset slightly back to beginning of first character which enters the view (or forward if view is scrolled negatively)

	currentSelection=GetSelectionAtPosition(view->parentBuffer->selectionUniverse,currentPosition);
	currentStyle=GetStyleAtPosition(view->parentBuffer->styleUniverse,currentPosition);
	ForceStyleIntoRange(currentStyle);
	gceWidthTable=&(GetFontForStyle(view->viewStyles,currentStyle)->charWidths[0]);

	while(chunk&&((textXOffset+gceRunningWidth)<xWindowEndOffset)&&GetCharacterEquiv(chunk->data[offset]))		// fill in all of the text needed
	{
		nextSelection=GetSelectionAtPosition(view->parentBuffer->selectionUniverse,currentPosition+1);
		nextStyle=GetStyleAtPosition(view->parentBuffer->styleUniverse,currentPosition+1);
		ForceStyleIntoRange(nextStyle);

		if((currentStyle!=nextStyle)||(currentSelection!=nextSelection)||(gceBufferIndex>STAGING_BUFFER_SIZE-STAGING_MAX_ADD))	// is something about to change? (if so, output what we have so far) (this will also output if the staging buffer is getting full)
		{
			DrawViewTextPiece(view,imageBuffer,currentSelection,currentStyle,textXOffset,gceRunningWidth,yWindowOffset,xWindowEndOffset);

			textXOffset+=gceRunningWidth;

			gceBufferIndex=0;						// begin at start of buffer
			gceRunningWidth=0;

			currentSelection=nextSelection;
			currentStyle=nextStyle;
			gceWidthTable=&(GetFontForStyle(view->viewStyles,currentStyle)->charWidths[0]);
		}

		currentPosition++;
		offset++;
		if(offset>=chunk->totalBytes)				// push through end of chunk
		{
			chunk=chunk->nextHeader;
			offset=0;
		}
	}
	// ran off the end of the line, or the end of the universe
	if(gceBufferIndex)
	{
		DrawViewTextPiece(view,imageBuffer,currentSelection,currentStyle,textXOffset,gceRunningWidth,yWindowOffset,xWindowEndOffset);
	}
}

static void DrawViewLines(EDITOR_VIEW *view,Pixmap imageBuffer)
// Draw the lines into view, obey the invalid region of the view
{
	GUI_VIEW_INFO
		*viewInfo;
	XRectangle
		lineRect,
		regionBox;
	Region
		invalidTextRegion,
		lineRegion;
	UINT32
		topLineToDraw,
		bottomLineToDraw;
	INT32
		xWindowOffset;
	INT32
		xWindowEndOffset;
	CHUNK_HEADER
		*chunk;
	UINT32
		offset,
		currentPosition;
	UINT32
		i;

	viewInfo=view->viewInfo;			// point at the information record for this view

	gceTabSize=view->tabSize;

	regionBox.x=lineRect.x=viewInfo->textBounds.x;
	regionBox.width=lineRect.width=viewInfo->textBounds.w;
	lineRect.height=viewInfo->lineHeight;
	regionBox.y=viewInfo->textBounds.y;
	regionBox.height=viewInfo->textBounds.h;

	invalidTextRegion=XCreateRegion();
	XUnionRectWithRegion(&regionBox,invalidTextRegion,invalidTextRegion);		// make rectangular region of entire text space
	XIntersectRegion(invalidTextRegion,((WINDOW_LIST_ELEMENT *)view->parentWindow->userData)->invalidRegion,invalidTextRegion);	// intersect with invalid region of window

	XClipBox(invalidTextRegion,&regionBox);							// get bounds for actual invalid text area

	topLineToDraw=(regionBox.y-viewInfo->textBounds.y)/viewInfo->lineHeight;	// screen relative top line to draw
	bottomLineToDraw=(regionBox.y-viewInfo->textBounds.y+regionBox.height-1)/viewInfo->lineHeight;	// get last line to draw

	for(i=topLineToDraw;i<=bottomLineToDraw;i++)
	{
		lineRect.y=viewInfo->textBounds.y+i*viewInfo->lineHeight;
		lineRegion=XCreateRegion();
		XUnionRectWithRegion(&lineRect,lineRegion,lineRegion);		// make region for the line we are about to draw
		XIntersectRegion(lineRegion,invalidTextRegion,lineRegion);	// see if it intersects with the invalid region
		if(!XEmptyRegion(lineRegion))
		{
			XClipBox(lineRegion,&regionBox);						// get bounds for this part
			LineToChunkPosition(view->parentBuffer->textUniverse,viewInfo->topLine+i,&chunk,&offset,&currentPosition);	// point to data for this line

			xWindowOffset=regionBox.x;
			xWindowEndOffset=regionBox.x+regionBox.width;

			// extend the draw area slightly to account for left and right text overhang

			xWindowOffset-=viewInfo->maxRightOverhang;				// move back far enough to let any character which extends to the right to draw what it needs
			xWindowEndOffset+=viewInfo->maxLeftOverhang;			// do same for right hand side

			DrawViewLine(view,imageBuffer,currentPosition,xWindowOffset,xWindowEndOffset,lineRect.y);
		}
		XDestroyRegion(lineRegion);
	}
	XDestroyRegion(invalidTextRegion);
}

void DrawView(EDITOR_VIEW *view,Pixmap imageBuffer,Region updateRegion)
// Draw the given view in its parent window
// also handle highlighting of text
// NOTE: the invalidRegion of the view's parent
// window should be set to the region that needs update
{
	XRectangle
		regionBox;
	GUI_VIEW_INFO
		*viewInfo;
	Region
		viewInvalidRegion;

	viewInfo=view->viewInfo;											// point at the information record for this view

	viewInvalidRegion=XCreateRegion();
	regionBox.x=viewInfo->bounds.x;
	regionBox.y=viewInfo->bounds.y;
	regionBox.width=viewInfo->bounds.w;
	regionBox.height=viewInfo->bounds.h;
	XUnionRectWithRegion(&regionBox,viewInvalidRegion,viewInvalidRegion);	// create a region which is the rectangle of the entire view
	XIntersectRegion(viewInvalidRegion,updateRegion,viewInvalidRegion);	// intersect with what needs update

	if(!XEmptyRegion(viewInvalidRegion))								// see if it has an invalid area
	{
		XClipBox(viewInvalidRegion,&regionBox);							// get bounds of invalid region

		XSetRegion(xDisplay,xGraphicsContext,viewInvalidRegion);			// set clipping to what's invalid in the view

		XSetFillStyle(xDisplay,xGraphicsContext,FillSolid);
		XSetForeground(xDisplay,xGraphicsContext,EditorColorToXPixel(view->viewStyles[0].backgroundColor));

		if(regionBox.x<viewInfo->textBounds.x)
		{
			XFillRectangle(xDisplay,imageBuffer,xGraphicsContext,viewInfo->bounds.x,viewInfo->bounds.y,viewInfo->textBounds.x-viewInfo->bounds.x,viewInfo->bounds.h);
		}

		if(regionBox.y<viewInfo->textBounds.y)
		{
			XFillRectangle(xDisplay,imageBuffer,xGraphicsContext,viewInfo->bounds.x,viewInfo->bounds.y,viewInfo->bounds.w,viewInfo->textBounds.y-viewInfo->bounds.y);
		}

		if((regionBox.y+(int)regionBox.height)>(viewInfo->textBounds.y+(int)viewInfo->textBounds.h))
		{
			XFillRectangle(xDisplay,imageBuffer,xGraphicsContext,viewInfo->bounds.x,viewInfo->textBounds.y+viewInfo->textBounds.h,viewInfo->bounds.w,(viewInfo->bounds.y+viewInfo->bounds.h)-(viewInfo->textBounds.y-viewInfo->textBounds.h));
		}

		DrawViewLines(view,imageBuffer);
		DrawViewCursor(view,imageBuffer);
		XSetClipMask(xDisplay,xGraphicsContext,None);					// get rid of clip mask
	}
	XDestroyRegion(viewInvalidRegion);									// get rid of invalid region for the view
}

static void RecalculateViewFontInfo(EDITOR_VIEW *view)
// Recalculate the cached information for view
// from its current size and set of styles
{
	INT32
		viewHeight;
	UINT32
		maxDescent;
	UINT32
		i;

	maxDescent=0;
	view->viewInfo->lineAscent=0;
	view->viewInfo->maxLeftOverhang=0;
	view->viewInfo->maxRightOverhang=0;
	for(i=0;i<VIEW_MAX_STYLES;i++)
	{
		if(view->viewStyles[i].font)						// see if a font is assigned for this style
		{
			if(view->viewInfo->lineAscent<(view->viewStyles[i].font->ascent))
			{
				view->viewInfo->lineAscent=view->viewStyles[i].font->ascent;
			}
			if(maxDescent<(view->viewStyles[i].font->descent))
			{
				maxDescent=view->viewStyles[i].font->descent;
			}
			if(view->viewInfo->maxLeftOverhang<view->viewStyles[i].font->maxLeftOverhang)
			{
				view->viewInfo->maxLeftOverhang=view->viewStyles[i].font->maxLeftOverhang;
			}
			if(view->viewInfo->maxRightOverhang<view->viewStyles[i].font->maxRightOverhang)
			{
				view->viewInfo->maxRightOverhang=view->viewStyles[i].font->maxRightOverhang;
			}
		}
	}
	view->viewInfo->lineHeight=view->viewInfo->lineAscent+maxDescent;

	if(view->viewInfo->lineHeight<=0)
	{
		view->viewInfo->lineHeight=1;								// lines are minimally one pixel tall (so we do not get an infinite number of lines on the view)
	}

	view->viewInfo->textBounds=view->viewInfo->bounds;				// start with copy of the view bounds
	view->viewInfo->textBounds.y+=VIEW_TOP_MARGIN;					// leave a small margin at the top

	if(view->viewInfo->bounds.h>VIEW_TOP_MARGIN+VIEW_BOTTOM_MARGIN)
	{
		viewHeight=view->viewInfo->bounds.h-VIEW_TOP_MARGIN-VIEW_BOTTOM_MARGIN;	// find out how much can be used for the text
		view->viewInfo->numLines=viewHeight/view->viewInfo->lineHeight;	// get integer number of lines in this view
		view->viewInfo->textBounds.h=view->viewInfo->numLines*view->viewInfo->lineHeight;	// alter the textBounds to be the exact rectangle to hold the text
	}
	else
	{
		view->viewInfo->numLines=0;
		view->viewInfo->textBounds.h=0;								// no room to draw text
	}
}

void SetViewBounds(EDITOR_VIEW *view,EDITOR_RECT *newBounds)
// Set the bounds of view to newBounds
// invalidate area where the view used to be but now is not
// and invalidate new areas of the view
{
	GUI_VIEW_INFO
		*viewInfo;
	EDITOR_RECT
		*viewBounds;
	unsigned int
		width,
		height;

	viewInfo=view->viewInfo;									// point at the information record for this view

	viewBounds=&(viewInfo->bounds);								// point at the view bounds rect

	if((newBounds->x!=viewBounds->x)||(newBounds->y!=viewBounds->y)||(newBounds->w!=viewBounds->w)||(newBounds->h!=viewBounds->h))
	{
		InvalidateWindowRect(view->parentWindow,viewBounds);	// invalidate old area
		InvalidateWindowRect(view->parentWindow,newBounds);		// invalidate new area

		(*viewBounds)=*newBounds;								// copy new rect into place
		RecalculateViewFontInfo(view);							// update additional information

		if(!(width=viewInfo->textBounds.w))
		{
			width=1;
		}
		if(!(height=viewInfo->textBounds.h))
		{
			height=1;
		}
		XMoveResizeWindow(xDisplay,viewInfo->cursorWindow,viewInfo->textBounds.x,viewInfo->textBounds.y,width,height);	// adjust the cursor change window
	}
}

void InvalidateViewPortion(EDITOR_VIEW *view,UINT32 startLine,UINT32 endLine,UINT32 startPixel,UINT32 endPixel)
// Invalidate the area of the view between startLine, endLine, and startPixel, endPixel
// NOTE: startLine, endLine, startPixel, and endPixel are all relative to the displayed view
// so that startLine of 0 always means the top line, and startPixel of 0 always
// means the left-most pixel
// NOTE: startLine must be <= endLine, startPixel must be <= endPixel
// the invalidated area is exclusive of endLine, and endPixel
{
	UINT32
		totalPixels;
	GUI_VIEW_INFO
		*viewInfo;
	EDITOR_RECT
		invalidRect;

	viewInfo=view->viewInfo;						// point at the information record for this view

	totalPixels=viewInfo->textBounds.w;
	if(viewInfo->numLines&&totalPixels)				// make sure we can actually invalidate
	{
		if(startLine>viewInfo->numLines)
		{
			startLine=viewInfo->numLines;
		}
		if(endLine>viewInfo->numLines)
		{
			endLine=viewInfo->numLines;
		}

		if(startPixel>totalPixels)
		{
			startPixel=totalPixels;
		}
		if(endPixel>totalPixels)
		{
			endPixel=totalPixels;
		}
		invalidRect.x=viewInfo->textBounds.x+startPixel;
		invalidRect.y=viewInfo->textBounds.y+startLine*viewInfo->lineHeight;
		invalidRect.w=endPixel-startPixel;
		invalidRect.h=(endLine-startLine)*viewInfo->lineHeight;
		InvalidateWindowRect(view->parentWindow,&invalidRect);
	}
}

void InvalidateViewSelection(EDITOR_VIEW *view,SELECTION_UNIVERSE *selection)
// Passed a view, and a selection, invalidate the visible area of the view
// covered by selection
{
	GUI_VIEW_INFO
		*viewInfo;
	Region
		newSelectionRegion;
	CHUNK_HEADER
		*chunk;
	UINT32
		offset,
		currentPosition;
	UINT32
		i;

	viewInfo=(GUI_VIEW_INFO *)view->viewInfo;				// point at the information record for this view

	newSelectionRegion=XCreateRegion();						// create a region to hold what is now in the selection

	gceTabSize=view->tabSize;
	for(i=0;i<viewInfo->numLines;i++)
	{
		LineToChunkPosition(view->parentBuffer->textUniverse,viewInfo->topLine+i,&chunk,&offset,&currentPosition);	// point to data for line
		AddSelectionLineToRegion(view,selection,newSelectionRegion,currentPosition,viewInfo->textBounds.y+i*viewInfo->lineHeight);
	}
	InvalidateWindowRegion(view->parentWindow,newSelectionRegion);

	XDestroyRegion(newSelectionRegion);						// get rid of the region we just created
}

static void InvalidateViewCursor(EDITOR_VIEW *view)
// Passed a view, invalidate the visible area of the view
// covered by the cursor (assuming the cursor currently
// is active).
{
	GUI_VIEW_INFO
		*viewInfo;
	CHUNK_HEADER
		*chunk;
	UINT32
		offset;
	UINT32
		line,
		lineOffset;
	INT32
		xPosition,
		maxXPosition;
	UINT32
		slopLeft,
		slopRight;
	EDITOR_RECT
		cursorRect;

	if(IsSelectionEmpty(view->parentBuffer->selectionUniverse))
	{
		viewInfo=view->viewInfo;													// point at the information record for this view

		PositionToLinePosition(view->parentBuffer->textUniverse,GetSelectionCursorPosition(view->parentBuffer->selectionUniverse),&line,&lineOffset,&chunk,&offset);
		if((line>=viewInfo->topLine)&&(line<viewInfo->topLine+viewInfo->numLines))	// make sure cursor is on the view vertically
		{
			maxXPosition=xPosition=viewInfo->leftPixel+viewInfo->textBounds.w;
			GetEditorViewTextToGraphicPosition(view,GetSelectionCursorPosition(view->parentBuffer->selectionUniverse),&xPosition,true,&slopLeft,&slopRight);

			if((xPosition>=viewInfo->leftPixel)&&(xPosition<maxXPosition))			// see if on view horizontally
			{
				cursorRect.x=viewInfo->textBounds.x+(xPosition-viewInfo->leftPixel);
				cursorRect.w=1;
				cursorRect.y=viewInfo->textBounds.y+viewInfo->lineHeight*(line-viewInfo->topLine);
				cursorRect.h=viewInfo->lineHeight;

				InvalidateWindowRect(view->parentWindow,&cursorRect);
			}
		}
	}
}

void InvalidateView(EDITOR_VIEW *view)
// invalidate the entire area of the view
{
	InvalidateWindowRect(view->parentWindow,&view->viewInfo->bounds);
}

void ScrollViewPortion(EDITOR_VIEW *view,UINT32 startLine,UINT32 endLine,INT32 numLines,UINT32 startPixel,UINT32 endPixel,INT32 numPixels)
// Scroll the area of the view between startLine, endLine, and startPixel, endPixel
// by numLines, and numPixels
// If numLines > 0, the data is scrolled down, else it is scrolled up
// If numPixels > 0, the data is scrolled right, else it is scrolled left
// any area left open by the scroll will be left erased, and invalidated
// NOTE: startLine, endLine, startPixel, and endPixel are all relative to the displayed view
// so that startLine of 0 always means the top line, and startPixel of 0 always
// means the left-most pixel
// NOTE: startLine must be <= endLine, startPixel must be <= endPixel
// the scrolled area is exclusive of endLine, and endPixel
{
	Window
		xWindow;
	UINT32
		totalPixels;
	INT32
		copySrcX,					// source x and y for copy
		copySrcY,
		copyDestX,
		copyDestY,
		invalidX,
		invalidY;
	UINT32
		copyWidth,
		copyHeight,
		invalidWidth,
		invalidHeight;
	EDITOR_RECT
		invalidRect;
	GUI_VIEW_INFO
		*viewInfo;

	viewInfo=view->viewInfo;										// point at the information record for this view
	xWindow=((WINDOW_LIST_ELEMENT *)view->parentWindow->userData)->xWindow;					// get x window for this view

	totalPixels=viewInfo->textBounds.w;
	if(viewInfo->numLines&&totalPixels)								// make sure we can actually scroll
	{
		if(startLine>viewInfo->numLines)
		{
			startLine=viewInfo->numLines;
		}
		if(endLine>viewInfo->numLines)
		{
			endLine=viewInfo->numLines;
		}

		if(startPixel>totalPixels)
		{
			startPixel=totalPixels;
		}
		if(endPixel>totalPixels)
		{
			endPixel=totalPixels;
		}

		if(numLines>((INT32)(endLine-startLine)))						// allow a scroll that erases everything
		{
			numLines=endLine-startLine;
		}
		else
		{
			if(numLines<((INT32)startLine-(INT32)endLine))
			{
				numLines=(INT32)startLine-(INT32)endLine;
			}
		}

		if(numPixels>((INT32)(endPixel-startPixel)))
		{
			numPixels=endPixel-startPixel;
		}
		else
		{
			if(numPixels<(INT32)startPixel-(INT32)endPixel)
			{
				numPixels=(INT32)startPixel-(INT32)endPixel;
			}
		}
		if(numLines||numPixels)
		{
			if(numLines>=0)												// see if scrolling down, or not scrolling
			{
				invalidY=copySrcY=startLine;
				copyDestY=startLine+numLines;
				copyHeight=(endLine-startLine)-numLines;
				invalidHeight=numLines;
			}
			else														// scrolling up
			{
				copySrcY=startLine-numLines;
				copyDestY=startLine;
				copyHeight=endLine-copySrcY;
				invalidY=endLine+numLines;
				invalidHeight=-numLines;
			}
			if(numPixels>=0)											// scrolling right, or not scrolling horizontally
			{
				invalidX=copySrcX=startPixel;
				copyDestX=startPixel+numPixels;
				copyWidth=(endPixel-startPixel)-numPixels;
				invalidWidth=numPixels;
			}
			else														// left
			{
				copySrcX=startPixel-numPixels;
				copyDestX=startPixel;
				copyWidth=endPixel-copySrcX;
				invalidX=endPixel+numPixels;
				invalidWidth=-numPixels;
			}
			if(copyWidth&&copyHeight)
			{
				XSetGraphicsExposures(xDisplay,xGraphicsContext,True);	// tell x to give events when part of scroll is revealed
				XGrabServer(xDisplay);									// keep anyone else from invalidating while the scroll is in progress (so they do not invalidate BEFORE the scroll, causing invalid data to be scrolled and never updated)
				XSync(xDisplay,False);									// get all invalidations
				UpdateEditorWindows();									// redraw everything which is invalid before scrolling
				XCopyArea(xDisplay,xWindow,xWindow,xGraphicsContext,viewInfo->textBounds.x+copySrcX,viewInfo->textBounds.y+copySrcY*viewInfo->lineHeight,copyWidth,copyHeight*viewInfo->lineHeight,viewInfo->textBounds.x+copyDestX,viewInfo->textBounds.y+copyDestY*viewInfo->lineHeight);
				XUngrabServer(xDisplay);								// let go after the scroll
				XSetGraphicsExposures(xDisplay,xGraphicsContext,False);
			}
			if(invalidWidth)
			{
				invalidRect.x=viewInfo->textBounds.x+invalidX;
				invalidRect.y=viewInfo->textBounds.y+startLine*viewInfo->lineHeight;
				invalidRect.w=invalidWidth;
				invalidRect.h=(endLine-startLine)*viewInfo->lineHeight;
				InvalidateWindowRect(view->parentWindow,&invalidRect);
			}
			if(invalidHeight)
			{
				invalidRect.x=viewInfo->textBounds.x+startPixel;
				invalidRect.y=viewInfo->textBounds.y+invalidY*viewInfo->lineHeight;
				invalidRect.w=endPixel-startPixel;
				invalidRect.h=invalidHeight*viewInfo->lineHeight;
				InvalidateWindowRect(view->parentWindow,&invalidRect);
			}
		}
	}
}

void SetViewTopLeft(EDITOR_VIEW *view,UINT32 newTopLine,INT32 newLeftPixel)
// Set the upper left corner of view
// NOTE: newTop, and newLeft will be forced into bounds by this routine
{
	UINT32
		oldTopLine;
	INT32
		oldLeftPixel;
	UINT32
		numLines;
	GUI_VIEW_INFO
		*viewInfo;

	viewInfo=view->viewInfo;										// point at the information record for this view
	numLines=view->parentBuffer->textUniverse->totalLines;
	oldTopLine=viewInfo->topLine;
	oldLeftPixel=viewInfo->leftPixel;

	if(newTopLine>numLines)
	{
		newTopLine=numLines;
	}
	if(newLeftPixel>EDIT_MAX_WIDTH)
	{
		newLeftPixel=EDIT_MAX_WIDTH;
	}
	if(newLeftPixel<-EDIT_MAX_WIDTH)
	{
		newLeftPixel=-EDIT_MAX_WIDTH;
	}
	ScrollViewPortion(view,0,viewInfo->numLines,oldTopLine-newTopLine,0,viewInfo->textBounds.w,oldLeftPixel-newLeftPixel);
	viewInfo->topLine=newTopLine;
	viewInfo->leftPixel=newLeftPixel;
	if(view->viewPositionChangedVector)								// see if someone needs to be called because of my change
	{
		view->viewPositionChangedVector(view);						// call him with my pointer
	}
	UpdateEditorWindows();											// redraw now to make the update more crisp
}

void SetEditorViewTopLine(EDITOR_VIEW *view,UINT32 lineNumber)
// Set the line number of the top line that is
// displayed in the given edit view
// NOTE: this does no invalidation of the view
{
	view->viewInfo->topLine=lineNumber;
}

static void BlinkViewCursor(EDITOR_VIEW *view)
// calling this will cause the cursor in view to change blink states, updating it as needed
{
	GUI_VIEW_INFO
		*viewInfo;

	viewInfo=(GUI_VIEW_INFO *)view->viewInfo;				// point at the information record for this view
	viewInfo->blinkState=!viewInfo->blinkState;				// invert blink state
	InvalidateViewCursor(view);								// invalidate the area of the window where the cursor is currently, so we'll see the changeon the next redraw
}

void ResetEditorViewCursorBlink(EDITOR_VIEW *view)
// When this is called, it will reset the cursor blinking, force the
// cursor to the blink-on state, and reset the blink timeout
{
	GUI_VIEW_INFO
		*viewInfo;

	viewInfo=(GUI_VIEW_INFO *)view->viewInfo;				// point at the information record for this view
	if(!viewInfo->blinkState)								// if it is currently off, blink it to on
	{
		BlinkViewCursor(view);
	}
	ResetCursorBlinkTime();
}

void ViewCursorTask(EDITOR_VIEW *view)
// Handle blinking the cursor for view
{
	BlinkViewCursor(view);
}

static bool HandleViewMouseWheelEvents(EDITOR_VIEW *view,XEvent *event)
// Call this to see if mounse wheel has been rolled.
// X sends button press events 4 ( for UP) and 5 (for DOWN) when you scroll the mouse wheel.
// Returns true if the event was a mouse wheel event.
{
	if((event->xbutton.button==Button4)||(event->xbutton.button==Button5))
	{
		ScrollEditViewEvent(view,(event->xbutton.button==Button4)?VVK_SCROLLUP:VVK_SCROLLDOWN,1);
		UpdateEditorWindows();
		return(true);
	}
	return(false);
}

bool TrackView(EDITOR_VIEW *view,XEvent *event)
// call this to see if the click was in the view, and if so, track it
// if the click was not in the view, return false
{
	VIEW_EVENT
		record;
	VIEW_CLICK_EVENT_DATA
		clickData;
	GUI_VIEW_INFO
		*viewInfo;
	Window
		root,
		child;
	int
		rootX,
		rootY,
		windowX,
		windowY;
	unsigned int
		state;
	UINT32
		numClicks,
		keyCode,
		modifiers;

	if(PointInView(view,event->xbutton.x,event->xbutton.y))
	{
		if(!HandleViewMouseWheelEvents(view,event))
		{
			numClicks=HandleMultipleClicks(event);
			XStateToEditorModifiers(event->xbutton.state,numClicks,&modifiers);
			switch(event->xbutton.button)
			{
				case Button1:
					keyCode=0;
					break;
				case Button2:
					keyCode=1;
					break;
				case Button3:
					keyCode=2;
					break;
				default:
					keyCode=0;
					break;
			}
			clickData.keyCode=keyCode;
			clickData.modifiers=modifiers;

			record.eventType=VET_CLICK;
			record.view=view;
			record.eventData=&clickData;
			viewInfo=(GUI_VIEW_INFO *)(view->viewInfo);
			LocalClickToViewClick(record.view,event->xbutton.x,event->xbutton.y,&clickData.xClick,&clickData.yClick);
			if((clickData.xClick>=0)&&(clickData.xClick<(int)viewInfo->textBounds.w))	// make sure initial click is in the text part of the view
			{
				if((clickData.yClick>=0)&&(clickData.yClick<(int)viewInfo->numLines))	// make sure initial click is in the view
				{
					HandleViewEvent(&record);
					while(StillDown(event->xbutton.button,1))							// send repeat events for the view click
					{
						if(XQueryPointer(xDisplay,((WINDOW_LIST_ELEMENT *)view->parentWindow->userData)->xWindow,&root,&child,&rootX,&rootY,&windowX,&windowY,&state))
						{
							record.view=view;
							record.eventType=VET_CLICKHOLD;
							record.eventData=&clickData;
							clickData.modifiers=keyCode;
							clickData.modifiers=modifiers;
							LocalClickToViewClick(record.view,windowX,windowY,&clickData.xClick,&clickData.yClick);
							HandleViewEvent(&record);
						}
						UpdateEditorWindows();
					}
				}
			}
		}
		return(true);
	}
	return(false);
}

void ViewStyleFontsChanged(EDITOR_VIEW *view)
// When the fonts have been modified in styles for the passed view,
// this is called so that it may update the line height, and anything else
// that may be necessary.
{
	RecalculateViewFontInfo(view);
}

void ViewEndChange(EDITOR_VIEW *view)
// The passed view has finished being changed. Bring the view back
// up to date
{
	InvalidateViewCursor(view);
}

void ViewStartChange(EDITOR_VIEW *view)
// The passed view is about to be changed in some way, do whatever is needed
// so that when ViewEndChange is called, the correct parts of the view are
// invalidated.
{
	InvalidateViewCursor(view);
}

void DisposeViewInfo(EDITOR_VIEW *view)
// unlink view from everywhere, and dispose of it
{
	XDestroyWindow(xDisplay,view->viewInfo->cursorWindow);	// tell x to make the cursor window go away
	MDisposePtr(view->viewInfo);
}

bool CreateViewInfo(EDITOR_VIEW *view,EDITOR_VIEW_DESCRIPTOR *descriptor)
// Called after all editor specific fields of a new view have been filled in
// This creates and stuffs local GUI specific information, and sets up the view in the GUI
// if there is a problem, SetError, and return false
{
	unsigned int
		width,
		height;

	if((view->viewInfo=(GUI_VIEW_INFO *)MNewPtr(sizeof(GUI_VIEW_INFO))))
	{
		view->viewInfo->topLine=descriptor->topLine;
		view->viewInfo->leftPixel=descriptor->leftPixel;		// initial pixel to display
		view->viewInfo->bounds=descriptor->rect;				// copy bounds rectangle

		view->viewInfo->blinkState=false;						// next time it blinks, it will blink to on
		view->viewInfo->nextBlinkTime=0;						// next time it checks, it will blink

		RecalculateViewFontInfo(view);							// update view information based on fonts and stuff

		if(!(width=view->viewInfo->textBounds.w))
		{
			width=1;
		}
		if(!(height=view->viewInfo->textBounds.h))
		{
			height=1;
		}

		// create a window which has the sole purpose of changing the cursor to a caret when it enters the view
		view->viewInfo->cursorWindow=XCreateWindow(xDisplay,((WINDOW_LIST_ELEMENT *)view->parentWindow->userData)->xWindow,view->viewInfo->textBounds.x,view->viewInfo->textBounds.y,width,height,0,0,InputOnly,DefaultVisual(xDisplay,xScreenNum),0,NULL);
		XDefineCursor(xDisplay,view->viewInfo->cursorWindow,caretCursor);	// this is the cursor to use in the view
		XMapRaised(xDisplay,view->viewInfo->cursorWindow);		// map window to the top of the pile

		return(true);
	}
	return(false);
}

